﻿using System;
using System.Threading;

using wclCommon;
using wclBluetooth;

namespace Beacons
{
    class Program
    {
        private static Guid BEACON_UUID = new Guid("{09039835-4A80-443B-87AA-DC565D09EA61}");

        private static wclBluetoothRadio Radio = null;
        private static Int32 InitResult = wclErrors.WCL_E_SUCCESS;
        private static AutoResetEvent InitEvent = null;
        private static AutoResetEvent Event = null;
        private static Thread RunThread = null;
        private static wclBluetoothLeAdvertiser Advertiser = new wclBluetoothLeAdvertiser();

        #region Beacons Watcher
        // Beacons monitoring thread.
        private static void MonitorBeacons()
        {
            Console.Clear();
            Console.WriteLine("MONITORING BLUETOOTH LE BEACONS");
            Console.WriteLine("===============================");

            // Create beacon watcher and set event handlers.
            wclBluetoothLeBeaconWatcher Watcher = new wclBluetoothLeBeaconWatcher();
            Watcher.OnStarted += Watcher_OnStarted;
            Watcher.OnStopped += Watcher_OnStopped;
            Watcher.OnAdvertisementRawFrame += Watcher_OnAdvertisementRawFrame;
            Watcher.OnEddystoneTlmFrame += Watcher_OnEddystoneTlmFrame;
            Watcher.OnEddystoneUidFrame += Watcher_OnEddystoneUidFrame;
            Watcher.OnEddystoneUrlFrame += Watcher_OnEddystoneUrlFrame;
            Watcher.OnManufacturerRawFrame += Watcher_OnManufacturerRawFrame;
            Watcher.OnProximityBeaconFrame += Watcher_OnProximityBeaconFrame;
            Watcher.OnAltBeaconFrame += Watcher_OnAltBeaconFrame;
            Watcher.OnAdvertisementFrameInformation += Watcher_OnAdvertisementFrameInformation;
            Watcher.OnAdvertisementUuidFrame += Watcher_OnAdvertisementUuidFrame;
            Watcher.OnAdvertisementService128DataFrame += Watcher_OnAdvertisementService128DataFrame;
            Watcher.OnAdvertisementService16DataFrame += Watcher_OnAdvertisementService16DataFrame;
            Watcher.OnAdvertisementService32DataFrame += Watcher_OnAdvertisementService32DataFrame;
            Watcher.OnAdvertisementServiceSol128Frame += Watcher_OnAdvertisementServiceSol128Frame;
            Watcher.OnAdvertisementServiceSol16Frame += Watcher_OnAdvertisementServiceSol16Frame;
            Watcher.OnAdvertisementServiceSol32Frame += Watcher_OnAdvertisementServiceSol32Frame;

            // Try to start watcher.
            InitResult = Watcher.Start(Radio);
            if (InitResult != wclErrors.WCL_E_SUCCESS)
                Console.WriteLine("Failed to start monitoring. Error: 0x" + InitResult.ToString("X8"));

            // Initialization completed.
            InitEvent.Set();

            if (InitResult == wclErrors.WCL_E_SUCCESS)
            {
                wclMessageBroadcaster.Wait(Event);
                Watcher.Stop();
            }

            Watcher = null;
        }

        #region Beacons Watcher event handlers.
        private static void Watcher_OnAdvertisementServiceSol32Frame(Object Sender, Int64 Address,
            Int64 Timestamp, SByte Rssi, UInt32 Uuid)
        {
            Console.WriteLine("32 bit SOL service data");
            Console.WriteLine("  Address: " + Address.ToString("X12"));
            Console.WriteLine("  Timestamp: " + Timestamp.ToString("X16"));
            Console.WriteLine("  RSSI: " + Rssi.ToString());
            Console.WriteLine("  UUID: " + Uuid.ToString("X8"));
            Console.WriteLine("");
        }

        private static void Watcher_OnAdvertisementServiceSol16Frame(Object Sender, Int64 Address,
            Int64 Timestamp, SByte Rssi, UInt16 Uuid)
        {
            Console.WriteLine("16 bit SOL service data");
            Console.WriteLine("  Address: " + Address.ToString("X12"));
            Console.WriteLine("  Timestamp: " + Timestamp.ToString("X16"));
            Console.WriteLine("  RSSI: " + Rssi.ToString());
            Console.WriteLine("  UUID: " + Uuid.ToString("X4"));
            Console.WriteLine("");
        }

        private static void Watcher_OnAdvertisementServiceSol128Frame(Object Sender, Int64 Address,
            Int64 Timestamp, SByte Rssi, Guid Uuid)
        {
            Console.WriteLine("128 bit SOL service data");
            Console.WriteLine("  Address: " + Address.ToString("X12"));
            Console.WriteLine("  Timestamp: " + Timestamp.ToString("X16"));
            Console.WriteLine("  RSSI: " + Rssi.ToString());
            Console.WriteLine("  UUID: " + Uuid.ToString());
            Console.WriteLine("");
        }

        private static void Watcher_OnAdvertisementService32DataFrame(Object Sender, Int64 Address,
            Int64 Timestamp, SByte Rssi, UInt32 Uuid, Byte[] Data)
        {
            Console.WriteLine("32 bit service data");
            Console.WriteLine("  Address: " + Address.ToString("X12"));
            Console.WriteLine("  Timestamp: " + Timestamp.ToString("X16"));
            Console.WriteLine("  RSSI: " + Rssi.ToString());
            Console.WriteLine("  UUID: " + Uuid.ToString("X8"));
            Console.WriteLine("");
        }

        private static void Watcher_OnAdvertisementService16DataFrame(Object Sender, Int64 Address,
            Int64 Timestamp, SByte Rssi, UInt16 Uuid, Byte[] Data)
        {
            Console.WriteLine("16 bit service data");
            Console.WriteLine("  Address: " + Address.ToString("X12"));
            Console.WriteLine("  Timestamp: " + Timestamp.ToString("X16"));
            Console.WriteLine("  RSSI: " + Rssi.ToString());
            Console.WriteLine("  UUID: " + Uuid.ToString("X4"));
            Console.WriteLine("");
        }

        private static void Watcher_OnAdvertisementService128DataFrame(Object Sender, Int64 Address,
            Int64 Timestamp, SByte Rssi, Guid Uuid, Byte[] Data)
        {
            Console.WriteLine("128 bit service data");
            Console.WriteLine("  Address: " + Address.ToString("X12"));
            Console.WriteLine("  Timestamp: " + Timestamp.ToString("X16"));
            Console.WriteLine("  RSSI: " + Rssi.ToString());
            Console.WriteLine("  UUID: " + Uuid.ToString());
            Console.WriteLine("");
        }

        private static void Watcher_OnAdvertisementUuidFrame(Object Sender, Int64 Address, Int64 Timestamp,
            SByte Rssi, Guid Uuid)
        {
            Console.WriteLine("UUID FRAME");
            Console.WriteLine("  Address: " + Address.ToString("X12"));
            Console.WriteLine("  Timestamp: " + Timestamp.ToString("X16"));
            Console.WriteLine("  RSSI: " + Rssi.ToString());
            Console.WriteLine("  UUID: " + Uuid.ToString());
            Console.WriteLine("");
        }

        private static void Watcher_OnAdvertisementFrameInformation(Object Sender, Int64 Address, Int64 Timestamp,
            SByte Rssi, String Name, wclBluetoothLeAdvertisementType PacketType, wclBluetoothLeAdvertisementFlag Flags)
        {
            Console.WriteLine("FRAME INFORMATION");
            Console.WriteLine("  Address: " + Address.ToString("X12"));
            Console.WriteLine("  Timestamp: " + Timestamp.ToString("X16"));
            Console.WriteLine("  RSSI: " + Rssi.ToString());
            Console.WriteLine("  Name: " + Name);

            String Str;

            switch (PacketType)
            {
                case wclBluetoothLeAdvertisementType.atConnectableUndirected:
                    Str = "Connectable Undirected";
                    break;
                case wclBluetoothLeAdvertisementType.atConnectableDirected:
                    Str = "Connectable Directed";
                    break;
                case wclBluetoothLeAdvertisementType.atScannableUndirected:
                    Str = "Scannable Undirecte";
                    break;
                case wclBluetoothLeAdvertisementType.atNonConnectableUndirected:
                    Str = "Non Connectable Undirected";
                    break;
                case wclBluetoothLeAdvertisementType.atScanResponse:
                    Str = "Scan Response";
                    break;
                default:
                    Str = "Unknown";
                    break;
            }
            Console.WriteLine("  Frame type: " + Str);

            Str = "[";
            if ((Flags & wclBluetoothLeAdvertisementFlag.afLimitedDiscoverableMode) != 0)
                Str = Str + " afLimitedDiscoverableMode";
            if ((Flags & wclBluetoothLeAdvertisementFlag.afGeneralDiscoverableMode) != 0)
                Str = Str + " afGeneralDiscoverableMode";
            if ((Flags & wclBluetoothLeAdvertisementFlag.afClassicNotSupported) != 0)
                Str = Str + " afClassicNotSupported";
            if ((Flags & wclBluetoothLeAdvertisementFlag.afDualModeControllerCapable) != 0)
                Str = Str + " afDualModeControllerCapable";
            if ((Flags & wclBluetoothLeAdvertisementFlag.afDualModeHostCapable) != 0)
                Str = Str + " afDualModeHostCapable";
            Str = Str + " ]";
            Console.WriteLine("  Flags: " + Str);

            Console.WriteLine("");
        }

        private static void Watcher_OnProximityBeaconFrame(Object Sender, Int64 Address, Int64 Timestamp,
            SByte Rssi, UInt16 CompanyId, UInt16 Major, UInt16 Minor, Guid Uuid, SByte TxRssi, Byte[] Data)
        {
            Console.WriteLine("PROXIMITY BEACON FRAME");
            Console.WriteLine("  Address: " + Address.ToString("X12"));
            Console.WriteLine("  Timestamp: " + Timestamp.ToString("X16"));
            Console.WriteLine("  RSSI: " + Rssi.ToString());
            Console.WriteLine("  TX RSSI: " + TxRssi.ToString());
            Console.WriteLine("  Company ID: " + CompanyId.ToString("X4"));
            Console.WriteLine("  Minor: " + Minor.ToString("X4"));
            Console.WriteLine("  Major: " + Major.ToString("X4"));
            Console.WriteLine("  UUID: " + Uuid.ToString());
            Console.WriteLine("");
        }

        private static void Watcher_OnAltBeaconFrame(Object Sender, Int64 Address, Int64 Timestamp, SByte Rssi,
            UInt16 CompanyId, UInt16 Major, UInt16 Minor, Guid Uuid, SByte TxRssi, Byte Reserved, Byte[] Data)
        {
            Console.WriteLine("ALT BEACON FRAME");
            Console.WriteLine("  Address: " + Address.ToString("X12"));
            Console.WriteLine("  Timestamp: " + Timestamp.ToString("X16"));
            Console.WriteLine("  RSSI: " + Rssi.ToString());
            Console.WriteLine("  TX RSSI: " + TxRssi.ToString());
            Console.WriteLine("  Company ID: " + CompanyId.ToString("X4"));
            Console.WriteLine("  Minor: " + Minor.ToString("X4"));
            Console.WriteLine("  Major: " + Major.ToString("X4"));
            Console.WriteLine("  Reserved: " + Reserved.ToString("X2"));
            Console.WriteLine("  UUID: " + Uuid.ToString());
            Console.WriteLine("");
        }

        private static void Watcher_OnManufacturerRawFrame(Object Sender, Int64 Address, Int64 Timestamp,
            SByte Rssi, UInt16 CompanyId, Byte[] Data)
        {
            Console.WriteLine("MANUFACTURER RAW FRAME");
            Console.WriteLine("  Address: " + Address.ToString("X12"));
            Console.WriteLine("  Timestamp: " + Timestamp.ToString("X16"));
            Console.WriteLine("  RSSI: " + Rssi.ToString());
            Console.WriteLine("  Company ID: " + CompanyId.ToString("X4"));
            if (Data != null)
                Console.WriteLine("  Data length: " + Data.Length.ToString());
            else
                Console.WriteLine("  Data length: 0");
            Console.WriteLine("");
        }

        private static void Watcher_OnEddystoneUrlFrame(Object Sender, Int64 Address, Int64 Timestamp, SByte Rssi,
            SByte TxRssi, String Url)
        {
            Console.WriteLine("EDDYSTONE URL FRAME");
            Console.WriteLine("  Address: " + Address.ToString("X12"));
            Console.WriteLine("  Timestamp: " + Timestamp.ToString("X16"));
            Console.WriteLine("  RSSI: " + Rssi.ToString());
            Console.WriteLine("  TX RSSI: " + TxRssi.ToString());
            Console.WriteLine("  URL: " + Url);
            Console.WriteLine("");
        }

        private static void Watcher_OnEddystoneUidFrame(Object Sender, Int64 Address, Int64 Timestamp, SByte Rssi,
            SByte TxRssi, Guid Uuid, Byte[] Data)
        {
            Console.WriteLine("EDDYSTONE UID FRAME");
            Console.WriteLine("  Address: " + Address.ToString("X12"));
            Console.WriteLine("  Timestamp: " + Timestamp.ToString("X16"));
            Console.WriteLine("  RSSI: " + Rssi.ToString());
            Console.WriteLine("  TX RSSI: " + TxRssi.ToString());
            Console.WriteLine("  UUID: " + Uuid.ToString());
            Console.WriteLine("");
        }

        private static void Watcher_OnEddystoneTlmFrame(Object Sender, Int64 Address, Int64 Timestamp, SByte Rssi,
            UInt32 AdvCnt, UInt16 Batt, UInt32 SecCnt, Double Temp, Byte[] Data)
        {
            Console.WriteLine("EDDYSTONE TLM FRAME");
            Console.WriteLine("  Address: " + Address.ToString("X12"));
            Console.WriteLine("  Timestamp: " + Timestamp.ToString("X16"));
            Console.WriteLine("  RSSI: " + Rssi.ToString());
            Console.WriteLine("  Frames sent: " + AdvCnt.ToString());
            Console.WriteLine("  Battery level: " + Batt.ToString());
            Console.WriteLine("  Running time (secs): " + SecCnt.ToString());
            Console.WriteLine("  Temperature: " + Temp.ToString());
            Console.WriteLine("");
        }

        private static void Watcher_OnAdvertisementRawFrame(Object Sender, Int64 Address, Int64 Timestamp,
            SByte Rssi, Byte DataType, Byte[] Data)
        {
            Console.WriteLine("RAW FRAME");
            Console.WriteLine("  Address: " + Address.ToString("X12"));
            Console.WriteLine("  Timestamp: " + Timestamp.ToString("X16"));
            Console.WriteLine("  RSSI: " + Rssi.ToString());
            Console.WriteLine("  Data type: " + DataType.ToString("X2"));
            if (Data != null)
                Console.WriteLine("  Data length: " + Data.Length.ToString());
            else
                Console.WriteLine("  Data length: 0");
            Console.WriteLine("");
        }

        private static void Watcher_OnStopped(Object sender, EventArgs e)
        {
            Console.WriteLine("Monitoring has been stopped");
        }

        private static void Watcher_OnStarted(Object sender, EventArgs e)
        {
            Console.WriteLine("Monitoring has been started");
        }
        #endregion
        #endregion

        #region Beacons.
        #region Common beacons event handlers.
        private static void Advertiser_OnStarted(object sender, EventArgs e)
        {
            Console.WriteLine("Advertiser has been started");
        }

        private static void Advertiser_OnStopped(object sender, EventArgs e)
        {
            Console.WriteLine("Advertiser has been sopped");
            Advertiser.Clear();
        }

        private static void StartAdvertising(wclBluetoothLeAdvertisement Adv)
        {
            Advertiser.Clear();

            InitResult = Advertiser.Add(Adv);
            if (InitResult != wclErrors.WCL_E_SUCCESS)
                Console.WriteLine("  Add advertising failed. Error: 0x" + InitResult.ToString("X8"));
            else
            {
                InitResult = Advertiser.Start(Radio);
                if (InitResult != wclErrors.WCL_E_SUCCESS)
                {
                    Advertiser.Clear();
                    Console.WriteLine("  Start advertiser failed. Error: 0x" + InitResult.ToString("X8"));
                }

                InitEvent.Set();

                if (InitResult == wclErrors.WCL_E_SUCCESS)
                {
                    wclMessageBroadcaster.Wait(Event);
                    Advertiser.Stop();
                }
            }
        }
        #endregion

        // Proximity beacon thread.
        private static void ProximityBeacon()
        {
            Console.Clear();
            Console.WriteLine("BLUETOOTH LE PROXIMITY BEACON");
            Console.WriteLine("=============================");

            StartAdvertising(new wclBluetoothLeProximityBeaconAdvertisement(-5, 0x0101, 0x0202,
                BEACON_UUID, 0xFFFE));
        }

        private static void AltBeacon()
        {
            Console.Clear();
            Console.WriteLine("BLUETOOTH LE ALT BEACON");
            Console.WriteLine("=======================");

            StartAdvertising(new wclBluetoothLeAltBeaconAdvertisement(-5, 0x0101, 0x0202, BEACON_UUID,
                0xFFFE, 0x11));
        }

        private static void EddystoneUidBeacon()
        {
            Console.Clear();
            Console.WriteLine("BLUETOOTH LE EDDYSTONE UID BEACON");
            Console.WriteLine("=================================");

            StartAdvertising(new wclBluetoothLeEddystoneUidBeaconAdvertisement(-5, BEACON_UUID));
        }

        private static void EddystoneUrlBeacon()
        {
            Console.Clear();
            Console.WriteLine("BLUETOOTH LE EDDYSTONE URL BEACON");
            Console.WriteLine("=================================");

            StartAdvertising(new wclBluetoothLeEddystoneUrlBeaconAdvertisement(-5, "https://www.btframework.com"));
        }

        private static void CustomBeacon()
        {
            Console.Clear();
            Console.WriteLine("BLUETOOTH LE CUSTOM BEACON");
            Console.WriteLine("==========================");

            String Str = "WCL Custom Beacon";
            StartAdvertising(new wclBluetoothLeManufacturerAdvertisement(0xFFFE, System.Text.Encoding.ASCII.GetBytes(Str)));
        }
        #endregion

        private static void StartThread(ThreadStart Proc)
        {
            try { RunThread = new Thread(Proc); } catch { RunThread = null; }
            if (RunThread == null)
                Console.WriteLine("Start thread failed");
            else
            {
                RunThread.Start();
                //. Wait for initialization.
                InitEvent.WaitOne();
                if (InitResult == wclErrors.WCL_E_SUCCESS)
                {
                    Console.WriteLine("Press ENTER to terminate thread.");
                    Console.ReadLine();
                    Event.Set();
                }
                RunThread.Join();
                RunThread = null;
            }
        }

        static void Main(string[] args)
        {
            Console.Clear();

            Console.WriteLine("Beacon IoT demo.");
            Console.WriteLine("The demo shows how to use Bluetooth Framework on Windows 10 IoT devices");
            Console.WriteLine("in console mode using APC synchronization.");

            Advertiser.OnStarted += Advertiser_OnStarted;
            Advertiser.OnStopped += Advertiser_OnStopped;

            // Set syncrhonization method
            wclMessageBroadcaster.SetSyncMethod(wclMessageSynchronizationKind.skApc);

            // Create termination event.
            try { Event = new AutoResetEvent(false); } catch { Event = null; }
            if (Event == null)
                Console.WriteLine("Create termination event failed");
            else
            {
                // Try to create initialization event.
                try { InitEvent = new AutoResetEvent(false); } catch { InitEvent = null; }
                if (InitEvent == null)
                    Console.WriteLine("Create initialization event failed");
                else
                {
                    // Create Bluetooth Manager.
                    wclBluetoothManager Manager = new wclBluetoothManager();
                    // Try to open Bluetooth Manager.
                    Int32 Res = Manager.Open();
                    if (Res != wclErrors.WCL_E_SUCCESS)
                        Console.WriteLine("Error opening Bluetooth Manager: 0x" + Res.ToString("X8"));
                    else
                    {
                        if (Manager.Count == 0)
                            Console.WriteLine("Not Bluetooth hardware found.");
                        else
                        {
                            for (int i = 0; i < Manager.Count; i++)
                            {
                                if (Manager[i].Available)
                                {
                                    Radio = Manager[i];
                                    break;
                                }
                            }

                            if (Radio == null)
                                Console.WriteLine("No available Bluetooth hardware were found.");
                            else
                            {
                                Boolean Terminated = false;
                                do
                                {
                                    Console.WriteLine("BLUETOOTH LE BEACONS DEMO");
                                    Console.WriteLine("=========================");
                                    Console.WriteLine("  1 - Monitor Beacons");
                                    Console.WriteLine("  2 - Start Proximity Beacon");
                                    Console.WriteLine("  3 - Start Eddystone UID Beacon");
                                    Console.WriteLine("  4 - Start Eddystone URL Beacon");
                                    Console.WriteLine("  5 - Start Custom Beacon");
                                    Console.WriteLine("  6 - Start Alt Beacon");
                                    Console.WriteLine("  0 - Exit");
                                    Console.Write("Select option: ");
                                    String Opt = Console.ReadLine();

                                    try
                                    {
                                        Byte Item = Convert.ToByte(Opt);
                                        switch (Item)
                                        {
                                            case 0:
                                                Terminated = true;
                                                break;
                                            case 1:
                                                StartThread(MonitorBeacons);
                                                break;
                                            case 2:
                                                StartThread(ProximityBeacon);
                                                break;
                                            case 3:
                                                StartThread(EddystoneUidBeacon);
                                                break;
                                            case 4:
                                                StartThread(EddystoneUrlBeacon);
                                                break;
                                            case 5:
                                                StartThread(CustomBeacon);
                                                break;
                                            case 6:
                                                StartThread(AltBeacon);
                                                break;
                                            default:
                                                Console.WriteLine("Wrong selection. Try again.");
                                                break;
                                        }
                                    }
                                    catch
                                    {
                                        Console.WriteLine("Wrong selection. Try again.");
                                    }
                                } while (!Terminated);

                                Radio = null;
                            }
                        }
                        // Close Bluetooth Manager.
                        Manager.Close();
                    }
                    Manager = null;

                    // Release initialization event.
                    InitEvent.Close();
                    InitEvent = null;
                }
                // Close termination event.
                Event.Close();
                Event = null;
            }
        }
    }
}
