﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;

using wclCommon;
using wclAudio;

namespace AudioSwitcher
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        private wclAudioSwitcher Switcher;

        private class DeviceItem
        {
            public String Id { get; set; }
            public String FriendlyName { get; set; }
            public String Description { get; set; }
            public String Flow { get; set; }
            public String Roles { get; set; }
            public String State { get; set; }
            public String Mac { get; set; }
            public String Service { get; set; }
        };

        public MainWindow()
        {
            InitializeComponent();
        }

        private void fmMain_Loaded(object sender, RoutedEventArgs e)
        {
            Switcher = new wclAudioSwitcher();
            Switcher.OnOpened += new EventHandler(Switcher_OnOpened);
            Switcher.OnClosed += new EventHandler(Switcher_OnClosed);
            Switcher.OnDefaultDeviceChanged += new wclAudioDefaultDeviceChangedEvent(Switcher_OnDefaultDeviceChanged);
            Switcher.OnDeviceAdded += new wclAudioDeviceEvent(Switcher_OnDeviceAdded);
            Switcher.OnDeviceRemoved += new wclAudioDeviceEvent(Switcher_OnDeviceRemoved);
            Switcher.OnStateChanged += new wclAudioDeviceStateChangedEvent(Switcher_OnStateChanged);

            cbActive.IsChecked = true;
            cbRole.SelectedIndex = 0;

            Int32 Res = Switcher.Open();
            if (Res != wclErrors.WCL_E_SUCCESS)
                lbLog.Items.Add("Open failed: 0x" + Res.ToString("X8"));
        }

        private void Switcher_OnStateChanged(Object Sender, String Id, wclAudioDeviceState State)
        {
            String Str = "";
            switch (State)
            {
                case wclAudioDeviceState.asActive:
                    Str = "Active";
                    break;
                case wclAudioDeviceState.asDisabled:
                    Str = "Disabled";
                    break;
                case wclAudioDeviceState.asNotPresent:
                    Str = "NotPresent";
                    break;
                case wclAudioDeviceState.asUnplugged:
                    Str = "Unplugged";
                    break;
            }
            lbLog.Items.Add("Device " + Id + " state changed: " + Str);
        }

        private void Switcher_OnDeviceRemoved(Object Sender, String Id)
        {
            lbLog.Items.Add("Device removed: " + Id);
        }

        private void Switcher_OnDeviceAdded(Object Sender, String Id)
        {
            lbLog.Items.Add("Device added: " + Id);
        }

        private void Switcher_OnDefaultDeviceChanged(Object Sender, String Id, wclAudioDeviceDataFlow Flow,
            wclAudioDeviceRole Role)
        {
            String Str = "Default device changed: " + Id + "; Flow: ";
            if (Flow == wclAudioDeviceDataFlow.dfRender)
                Str += "Render";
            else
                Str += "Capture";
            Str += "; Role: ";
            switch (Role)
            {
                case wclAudioDeviceRole.drConsole:
                    Str += "Console";
                    break;
                case wclAudioDeviceRole.drMultimedia:
                    Str += "Multimedia";
                    break;
                case wclAudioDeviceRole.drCommunications:
                    Str += "Communications";
                    break;
            }
            lbLog.Items.Add(Str);
        }

        private void Switcher_OnClosed(Object sender, EventArgs e)
        {
            lvDevices.Items.Clear();
            lbLog.Items.Add("Audio switcher closed");
        }

        private void Switcher_OnOpened(Object sender, EventArgs e)
        {
            lbLog.Items.Add("Audio switcher opened");
            EnumDevices();
        }

        private void fmMain_Closed(object sender, EventArgs e)
        {
            Switcher.Close();
        }

        private void btEnum_Click(object sender, RoutedEventArgs e)
        {
            EnumDevices();
        }

        private void btDefault_Click(object sender, RoutedEventArgs e)
        {
            if (lvDevices.SelectedItems == null)
                MessageBox.Show("Select device");
            else
            {
                wclAudioDeviceRole Role;
                switch (cbRole.SelectedIndex)
                {
                    case 0:
                        Role = wclAudioDeviceRole.drConsole;
                        break;
                    case 1:
                        Role = wclAudioDeviceRole.drMultimedia;
                        break;
                    default:
                        Role = wclAudioDeviceRole.drCommunications;
                        break;
                }

                DeviceItem Item = (DeviceItem)lvDevices.SelectedItems[0];
                Int32 Res = Switcher.SetDefault(Role, Item.Id);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Unable to set default device: 0x" + Res.ToString("X8"));
            }
        }

        private void EnumDevices()
        {
            wclAudioDeviceState States = 0;
            if (cbActive.IsChecked.Value)
                States |= wclAudioDeviceState.asActive;
            if (cbDisabled.IsChecked.Value)
                States |= wclAudioDeviceState.asDisabled;
            if (cbNotPresent.IsChecked.Value)
                States |= wclAudioDeviceState.asNotPresent;
            if (cbUnplugged.IsChecked.Value)
                States |= wclAudioDeviceState.asUnplugged;
            if (States == 0)
                MessageBox.Show("Select device state.");
            else
            {
                lvDevices.Items.Clear();

                wclAudioDevice[] Devices;
                Int32 Res = Switcher.Enum(States, out Devices);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Enum failed: 0x" + Res.ToString("X8"));
                else
                {
                    if (Devices != null && Devices.Length > 0)
                    {
                        foreach (wclAudioDevice Device in Devices)
                        {
                            int Item = lvDevices.Items.Add(new DeviceItem
                            {
                                Id = "",
                                FriendlyName = "",
                                Description = "",
                                Flow = "",
                                Roles = "",
                                State = "",
                                Mac = "",
                                Service = ""
                            });
                            UpdateDevice(Item, Device);
                        }
                    }
                }
            }
        }

        private void btClear_Click(object sender, RoutedEventArgs e)
        {
            lbLog.Items.Clear();
        }

        private void UpdateDevice(int i, wclAudioDevice Device)
        {
            DeviceItem Item = (DeviceItem)lvDevices.Items[i];

            String FlowStr;
            if (Device.Flow == wclAudioDeviceDataFlow.dfRender)
                FlowStr = "Render";
            else
                FlowStr = "Capture";

            String RolesStr = "[ ";
            if ((Device.Roles & wclAudioDeviceRole.drConsole) != 0)
                RolesStr += "Console ";
            if ((Device.Roles & wclAudioDeviceRole.drMultimedia) != 0)
                RolesStr += "Multimedia ";
            if ((Device.Roles & wclAudioDeviceRole.drCommunications) != 0)
                RolesStr += "Communications ";
            RolesStr += "]";

            String StateStr = "";
            switch (Device.State)
            {
                case wclAudioDeviceState.asActive:
                    StateStr = "Active";
                    break;
                case wclAudioDeviceState.asDisabled:
                    StateStr = "Disabled";
                    break;
                case wclAudioDeviceState.asNotPresent:
                    StateStr = "NotPresent";
                    break;
                case wclAudioDeviceState.asUnplugged:
                    StateStr = "Unplugged";
                    break;
            }

            String MacStr;
            String ServiceStr;
            if (Device.IsBluetooth)
            {
                MacStr = Device.Mac.ToString("X12");
                ServiceStr = Device.Service.ToString();
            }
            else
            {
                MacStr = "";
                ServiceStr = "";
            }

            Item.Id = Device.Id;
            Item.FriendlyName = Device.FriendlyName;
            Item.Description = Device.Description;
            Item.Flow = FlowStr;
            Item.Roles = RolesStr;
            Item.State = StateStr;
            Item.Mac = MacStr;
            Item.Service = ServiceStr;
            lvDevices.Items.Refresh();
        }

        private void btRefresh_Click(object sender, RoutedEventArgs e)
        {
            if (lvDevices.SelectedItem == null)
                MessageBox.Show("Select device");
            else
            {
                wclAudioDevice Device;
                Int32 Res = Switcher.GetDeviceDetails(((DeviceItem)lvDevices.SelectedItem).Id,
                    out Device);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Refresh device error: 0x" + Res.ToString("X8"));
                else
                    UpdateDevice(lvDevices.SelectedIndex, Device);
            }
        }

        private void btConnect_Click(object sender, RoutedEventArgs e)
        {
            if (lvDevices.SelectedItems == null)
                MessageBox.Show("Select device");
            else
            {
                Int32 Res;
                if (cbUseMac.IsChecked.Value)
                {
                    if (((DeviceItem)lvDevices.SelectedItem).Mac == "")
                    {
                        MessageBox.Show("Selected device is not Bluetooth enabled");
                        return;
                    }

                    Res = Switcher.Connect(Convert.ToInt64(((DeviceItem)lvDevices.SelectedItem).Mac, 16));
                }
                else
                    Res = Switcher.Connect(((DeviceItem)lvDevices.SelectedItem).Id);

                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Connect error: 0x" + Res.ToString("X8"));
            }
        }

        private void btDisconect_Click(object sender, RoutedEventArgs e)
        {
            if (lvDevices.SelectedItems == null)
                MessageBox.Show("Select device");
            else
            {
                Int32 Res;
                if (cbUseMac.IsChecked.Value)
                {
                    if (((DeviceItem)lvDevices.SelectedItem).Mac == "")
                    {
                        MessageBox.Show("Selected device is not Bluetooth enabled");
                        return;
                    }

                    Res = Switcher.Disconnect(Convert.ToInt64(((DeviceItem)lvDevices.SelectedItem).Mac, 16));
                }
                else
                    Res = Switcher.Disconnect(((DeviceItem)lvDevices.SelectedItem).Id);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Disconnect error: 0x" + Res.ToString("X8"));
            }
        }

        private void btGetDefault_Click(object sender, RoutedEventArgs e)
        {
            wclAudioDeviceRole Role;
            switch (cbRole.SelectedIndex)
            {
                case 0:
                    Role = wclAudioDeviceRole.drConsole;
                    break;
                case 1:
                    Role = wclAudioDeviceRole.drMultimedia;
                    break;
                default:
                    Role = wclAudioDeviceRole.drCommunications;
                    break;
            }

            wclAudioDeviceDataFlow Flow;
            if (cbFlow.SelectedIndex == 0)
                Flow = wclAudioDeviceDataFlow.dfCapture;
            else
                Flow = wclAudioDeviceDataFlow.dfRender;

            String Id;
            Int32 Res = Switcher.GetDefault(Role, Flow, out Id);
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Unable to get default device: 0x" + Res.ToString("X8"));
            else
                MessageBox.Show("Default device ID: " + Id);
        }
    }
}
