﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;

using wclCommon;
using wclBluetooth;

namespace BleSniffer
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        private wclBleSniffer FSniffer;

        public MainWindow()
        {
            InitializeComponent();
        }

        private void Window_Loaded(Object sender, RoutedEventArgs e)
        {
            cbChannel.SelectedIndex = 0;
            FSniffer = new wclBleSniffer();
            FSniffer.OnRawPacketReceived += new wclBleSnifferRawPacketReceivedEvent(FSniffer_OnRawPacketReceived);
            FSniffer.OnAdvIndReceived += new wclBleSnifferAdvIndReceivedEvent(FSniffer_OnAdvIndReceived);
            FSniffer.OnAdvDirectIndReceived += new wclBleSnifferAdvDirectIndReceivedEvent(FSniffer_OnAdvDirectIndReceived);
            FSniffer.OnAdvNonConnIndReceived += new wclBleSnifferAdvNonConnIndReceivedEvent(FSniffer_OnAdvNonConnIndReceived);
            FSniffer.OnAdvScanIndReceived += new wclBleSnifferAdvScanIndReceivedEvent(FSniffer_OnAdvScanIndReceived);
            FSniffer.OnScanReqReceived += new wclBleSnifferScanReqReceivedEvent(FSniffer_OnScanReqReceived);
            FSniffer.OnScanRspReceived += new wclBleSnifferScanRspReceivedEvent(FSniffer_OnScanRspReceived);
            FSniffer.OnConnectIndReceived += new wclBleSnifferConnectIndReceivedEvent(FSniffer_OnConnectIndReceived);
            FSniffer.OnStarted += FSniffer_OnStarted;
            FSniffer.OnStopped += FSniffer_OnStopped;
        }

        void FSniffer_OnConnectIndReceived(Object Sender, wclBluetoothLeAdvertisingPduHeader PduHeader,
            Int64 InitA, Int64 AdvA, wclBleSnifferLlData LlData)
        {
            lbPackets.Items.Add("CONNECT_IND RECEIVED");
            lbPackets.Items.Add("  InitA: " + InitA.ToString("X12"));
            lbPackets.Items.Add("  AdvA: " + AdvA.ToString("X12"));
            lbPackets.Items.Add("  LLDATA");
            lbPackets.Items.Add("    Aa: " + LlData.Aa.ToString("X8"));
            lbPackets.Items.Add("    CrcInit: " + LlData.CrcInit.ToString("X8"));
            lbPackets.Items.Add("    WinSize: " + LlData.WinSize.ToString());
            lbPackets.Items.Add("    WinOffset: " + LlData.WinOffset.ToString());
            lbPackets.Items.Add("    Interval: " + LlData.Interval.ToString());
            lbPackets.Items.Add("    Latency: " + LlData.Latency.ToString());
            lbPackets.Items.Add("    Timeout: " + LlData.Timeout.ToString());
            lbPackets.Items.Add("    ChM: " + LlData.ChM.ToString("X16"));
            lbPackets.Items.Add("    Hop: " + LlData.Hop.ToString());
            lbPackets.Items.Add("    Sca: " + LlData.Sca.ToString());
            DumpPduHeader(PduHeader);
        }

        void FSniffer_OnAdvIndReceived(Object Sender, wclBluetoothLeAdvertisingPduHeader PduHeader,
            Int64 AdvA, Byte[] AdvData)
        {
            lbPackets.Items.Add("ADV_IND RECEIVED");
            lbPackets.Items.Add("  AdvA: " + AdvA.ToString("X12"));
            DumpPduHeader(PduHeader);
            DumpPayload(AdvData);
        }

        void FSniffer_OnAdvDirectIndReceived(Object Sender, wclBluetoothLeAdvertisingPduHeader PduHeader,
            Int64 AdvA, Int64 TargetA)
        {
            lbPackets.Items.Add("ADV_DIRECT_IND RECEIVED");
            lbPackets.Items.Add("  AdvA: " + AdvA.ToString("X12"));
            lbPackets.Items.Add("  TargetA: " + TargetA.ToString("X12"));
            DumpPduHeader(PduHeader);
        }

        void FSniffer_OnAdvNonConnIndReceived(Object Sender, wclBluetoothLeAdvertisingPduHeader PduHeader,
            Int64 AdvA, Byte[] AdvData)
        {
            lbPackets.Items.Add("ADV_NONCONN_IND RECEIVED");
            lbPackets.Items.Add("  AdvA: " + AdvA.ToString("X12"));
            DumpPduHeader(PduHeader);
            DumpPayload(AdvData);
        }

        void FSniffer_OnAdvScanIndReceived(Object Sender, wclBluetoothLeAdvertisingPduHeader PduHeader,
            Int64 AdvA, Byte[] AdvData)
        {
            lbPackets.Items.Add("ADV_SCAN_IND RECEIVED");
            lbPackets.Items.Add("  AdvA: " + AdvA.ToString("X12"));
            DumpPduHeader(PduHeader);
            DumpPayload(AdvData);
        }

        void FSniffer_OnScanReqReceived(Object Sender, wclBluetoothLeAdvertisingPduHeader PduHeader,
            Int64 ScanA, Int64 AdvA)
        {
            lbPackets.Items.Add("SCAN_REQ RECEIVED");
            lbPackets.Items.Add("  ScanA: " + ScanA.ToString("X12"));
            lbPackets.Items.Add("  AdvA: " + AdvA.ToString("X12"));
            DumpPduHeader(PduHeader);
        }

        void FSniffer_OnScanRspReceived(Object Sender, wclBluetoothLeAdvertisingPduHeader PduHeader,
            Int64 AdvA, Byte[] ScanRspData)
        {
            lbPackets.Items.Add("SCAN_RSP RECEIVED");
            lbPackets.Items.Add("  AdvA: " + AdvA.ToString("X12"));
            DumpPduHeader(PduHeader);
            DumpPayload(ScanRspData);
        }

        private void DumpHeader(wclBluetoothLePacketHeader Header)
        {
            lbPackets.Items.Add("   HEADER");
            lbPackets.Items.Add("  ========");
            lbPackets.Items.Add("    Timestamp: " + Header.Timestamp.ToString("X8"));
            lbPackets.Items.Add("    Access Address: " + Header.AccessAddress.ToString("X8"));
            lbPackets.Items.Add("    Channel: " + Header.Channel.ToString("X2"));
            lbPackets.Items.Add("    Crc: " + Header.Crc.ToString("X8"));
            lbPackets.Items.Add("    Rssi: " + Header.Rssi.ToString());
            if (Header.Valid)
                lbPackets.Items.Add("    PACKET: OK");
            else
                lbPackets.Items.Add("    PACKET: ERROR");
        }

        private void DumpPduHeader(wclBluetoothLeAdvertisingPduHeader PduHeader)
        {
            DumpHeader(PduHeader.Header);
            lbPackets.Items.Add("   PDU HEADER");
            lbPackets.Items.Add("  ============");
            lbPackets.Items.Add("    ChSel: " + PduHeader.ChSel.ToString());
            lbPackets.Items.Add("    TxAdd: " + PduHeader.TxAdd.ToString());
            lbPackets.Items.Add("    RxAdd: " + PduHeader.RxAdd.ToString());
        }

        private void DumpPayload(Byte[] Payload)
        {
            if (Payload != null && Payload.Length > 0)
            {
                String Str = "";
                Byte Ndx = 1;
                for (Int32 i = 0; i < Payload.Length; i++)
                {
                    Str = Str + Payload[i].ToString("X2") + " ";
                    Ndx++;
                    if (Ndx == 16)
                    {
                        lbPackets.Items.Add("    " + Str);
                        Str = "";
                        Ndx = 1;
                    }
                }
                if (Str != "")
                    lbPackets.Items.Add("    " + Str);
            }
            lbPackets.Items.Add("");
        }

        void FSniffer_OnRawPacketReceived(object Sender, wclBluetoothLePacketHeader Header, Byte[] Payload)
        {
            lbPackets.Items.Add("RAW PACKET RECEIVED");
            DumpHeader(Header);
            DumpPayload(Payload);
        }

        private void FSniffer_OnStopped(Object sender, EventArgs e)
        {
            lbPackets.Items.Add("Stopped");
        }

        private void FSniffer_OnStarted(Object sender, EventArgs e)
        {
            lbPackets.Items.Add("Started");
        }

        private void Window_Closed(Object sender, EventArgs e)
        {
            FSniffer.Stop();
            FSniffer = null;
        }

        private void btClear_Click(Object sender, RoutedEventArgs e)
        {
            lbPackets.Items.Clear();
        }

        private void btStart_Click(Object sender, RoutedEventArgs e)
        {
            Int32 Res = FSniffer.Start((Byte)(cbChannel.SelectedIndex + 37));
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Start failed: 0x" + Res.ToString("X8"));
        }

        private void btStop_Click(Object sender, RoutedEventArgs e)
        {
            Int32 Res = FSniffer.Stop();
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Stop failed: 0x" + Res.ToString("X8"));
        }
    }
}
