﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;

using wclCommon;
using wclBluetooth;

namespace BluetoothManager
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        private wclBluetoothManager Manager;
        private wclPowerEventsMonitor FPowerMonitor;

        public MainWindow()
        {
            InitializeComponent();
        }

        private void btClear_Click(object sender, RoutedEventArgs e)
        {
            lvEvents.Items.Clear();
        }

        private class EventItem
        {
            public String Driver { get; set; }
            public String Event { get; set; }
            public String Parameter { get; set; }
            public String Value { get; set; }
        };

        private void TraceEvent(wclBluetoothRadio Radio, String EventStr, String ParameterStr, String ValueStr)
        {
            String Str;
            if (Radio != null)
                Str = Radio.ApiName;
            else
                Str = "";

            lvEvents.Items.Add(new EventItem { Driver = Str, Event = EventStr,
                Parameter = ParameterStr, Value = ValueStr });
        }

        private void fmMain_Loaded(object sender, RoutedEventArgs e)
        {
            Manager = new wclBluetoothManager();

            Manager.AfterOpen += new EventHandler(Manager_AfterOpen);
            Manager.BeforeClose += new EventHandler(Manager_BeforeClose);
            Manager.OnAuthenticationCompleted += new wclBluetoothDeviceResultEvent(Manager_OnAuthenticationCompleted);
            Manager.OnDeviceFound += new wclBluetoothDeviceEvent(Manager_OnDeviceFound);
            Manager.OnDiscoveringCompleted += new wclBluetoothResultEvent(Manager_OnDiscoveringCompleted);
            Manager.OnDiscoveringStarted += new wclBluetoothEvent(Manager_OnDiscoveringStarted);
            Manager.OnNumericComparison += new wclBluetoothNumericComparisonEvent(Manager_OnNumericComparison);
            Manager.OnPasskeyNotification += new wclBluetoothPasskeyNotificationEvent(Manager_OnPasskeyNotification);
            Manager.OnPasskeyRequest += new wclBluetoothPasskeyRequestEvent(Manager_OnPasskeyRequest);
            Manager.OnPinRequest += new wclBluetoothPinRequestEvent(Manager_OnPinRequest);
            Manager.OnStatusChanged +=new wclBluetoothEvent(Manager_OnStatusChanged);
            Manager.OnConfirm += new wclBluetoothConfirmEvent(Manager_OnConfirm);
            Manager.OnIoCapabilityRequest += new wclBluetoothIoCapabilityRequestEvent(Manager_OnIoCapabilityRequest);
            Manager.OnOobDataRequest += new wclBluetoothOobDataRequestEvent(Manager_OnOobDataRequest);
            Manager.OnProtectionLevelRequest += new wclBluetoothProtectionLevelRequestEvent(Manager_OnProtectionLevelRequest);
            Manager.OnClosed += new EventHandler(Manager_OnClosed);

            FPowerMonitor = new wclPowerEventsMonitor();
            FPowerMonitor.OnPowerStateChanged += new wclPowerStateChangedEvent(FPowerMonitor_OnPowerStateChanged);
            FPowerMonitor.Open();
        }

        void Manager_OnClosed(object sender, EventArgs e)
        {
            ClearRadios();

            TraceEvent(null, "Closed", "", "");
        }

        private Boolean WasOpened = false;

        void FPowerMonitor_OnPowerStateChanged(object Sender, wclPowerState State)
        {
            switch (State)
            {
                case wclPowerState.psResumeAutomatic:
                    TraceEvent(null, "Power", "State", "psResumeAutomatic");
                    if (WasOpened)
                    {
                        WasOpened = false;
                        Manager.Open();
                    }
                    break;

                case wclPowerState.psResume:
                    TraceEvent(null, "Power", "State", "psResume");
                    break;

                case wclPowerState.psSuspend:
                    TraceEvent(null, "Power", "State", "psSuspend");
                    if (Manager.Active)
                    {
                        WasOpened = true;
                        Manager.Close();
                    }
                    break;

                case wclPowerState.psUnknown:
                    TraceEvent(null, "Power", "State", "psUnknown");
                    break;
            }
        }

        void Manager_OnProtectionLevelRequest(object Sender, wclBluetoothRadio Radio, long Address, out wclBluetoothLeProtectionLevel Protection)
        {
            switch (cbBleProtection.SelectedIndex)
            {
                case 0:
                    Protection = wclBluetoothLeProtectionLevel.pplDefault;
                    break;
                case 1:
                    Protection = wclBluetoothLeProtectionLevel.pplNone;
                    break;
                case 2:
                    Protection = wclBluetoothLeProtectionLevel.pplEncryption;
                    break;
                case 3:
                    Protection = wclBluetoothLeProtectionLevel.pplEncryptionAndAuthentication;
                    break;
                default:
                    Protection = wclBluetoothLeProtectionLevel.pplDefault;
                    break;
            }
        }

        void Manager_OnOobDataRequest(object Sender, wclBluetoothRadio Radio, long Address, out wclBluetoothOobData OobData)
        {
            // This event fires when a remote device requests OOB data.
            OobData = null;
        }

        void Manager_OnIoCapabilityRequest(object Sender, wclBluetoothRadio Radio, long Address, out wclBluetoothMitmProtection Mitm, out wclBluetoothIoCapability IoCapability, out bool OobPresent)
        {
            switch (cbMitmProtection.SelectedIndex)
            {
                case 0:
                    Mitm = wclBluetoothMitmProtection.mitmProtectionNotRequired;
                    break;
                case 1:
                    Mitm = wclBluetoothMitmProtection.mitmProtectionRequired;
                    break;
                case 2:
                    Mitm = wclBluetoothMitmProtection.mitmProtectionNotRequiredBonding;
                    break;
                case 3:
                    Mitm = wclBluetoothMitmProtection.mitmProtectionRequiredBonding;
                    break;
                case 4:
                    Mitm = wclBluetoothMitmProtection.mitmProtectionNotRequiredGeneralBonding;
                    break;
                case 5:
                    Mitm = wclBluetoothMitmProtection.mitmProtectionRequiredGeneralBonding;
                    break;
                default:
                    Mitm = wclBluetoothMitmProtection.mitmProtectionNotDefined;
                    break;
            }

            switch (cbIoCap.SelectedIndex)
            {
                case 0:
                    IoCapability = wclBluetoothIoCapability.iocapDisplayOnly;
                    break;
                case 1:
                    IoCapability = wclBluetoothIoCapability.iocapDisplayYesNo;
                    break;
                case 2:
                    IoCapability = wclBluetoothIoCapability.iocapKeyboardOnly;
                    break;
                case 3:
                    IoCapability = wclBluetoothIoCapability.iocapNoInputNoOutput;
                    break;
                case 4:
                    IoCapability = wclBluetoothIoCapability.iocapDisplayKeyboard;
                    break;
                default:
                    IoCapability = wclBluetoothIoCapability.iocapNotDefined;
                    break;
            }

            // To accept OOB pairing set this parameter to True.
            OobPresent = false;
        }

        void Manager_OnConfirm(object Sender, wclBluetoothRadio Radio, long Address, out bool Confirm)
        {
            Confirm = true;

            TraceEvent(Radio, "Just Works pairing", "Address", Address.ToString("X12"));
        }

        private class RadioItem
        {
            public String Api { get; set; }
            public String Available { get; set; }
            public String Address { get; set; }
            public String Connectable { get; set; }
            public String Discoverable { get; set; }
            public String HciVersion { get; set; }
            public String LmpVersion { get; set; }
            public String Manufacturer { get; set; }
            public String Name { get; set; }
            public String Cod { get; set; }
            
            internal wclBluetoothRadio Radio { get; set; }
        };

        void Manager_AfterOpen(object sender, EventArgs e)
        {
            TraceEvent(null, "AfterOpen", "Handle pairing", Manager.HandlePairing.ToString());
        }

        void Manager_BeforeClose(object sender, EventArgs e)
        {
            TraceEvent(null, "BeforeClose", "", "");
        }

        void Manager_OnAuthenticationCompleted(object Sender, wclBluetoothRadio Radio, long Address, int Error)
        {
            TraceEvent(Radio, "AuthenticationCompleted", "Address", Address.ToString("X12"));
            TraceEvent(null, "", "Error", "0x" + Error.ToString("X8"));
        }

        void Manager_OnDeviceFound(object Sender, wclBluetoothRadio Radio, long Address)
        {
            TraceEvent(Radio, "DeviceFound", "Address", Address.ToString("X12"));

            AddDevice(Radio, Address);
        }

        void Manager_OnDiscoveringCompleted(object Sender, wclBluetoothRadio Radio, int Error)
        {
            TraceEvent(Radio, "DiscoveringCompleted", "Error", "0x" + Error.ToString("X8"));

            RefreshDevices();
        }

        void Manager_OnDiscoveringStarted(object Sender, wclBluetoothRadio Radio)
        {
            ClearDevices();

            TraceEvent(Radio, "DiscoveringStarted", "", "");
        }

        void Manager_OnNumericComparison(object Sender, wclBluetoothRadio Radio, long Address, uint Number, out bool Confirm)
        {
            Confirm = true;

            TraceEvent(Radio, "NumericComparison", "Address", Address.ToString("X12"));
            TraceEvent(null, "", "Number", Number.ToString());
            TraceEvent(null, "", "Confirm", Confirm.ToString());
        }

        void Manager_OnPasskeyNotification(object Sender, wclBluetoothRadio Radio, long Address, uint Passkey)
        {
            TraceEvent(Radio, "PasskeyNotification", "Address", Address.ToString("X12"));
            TraceEvent(null, "", "Passkey", Passkey.ToString());
        }

        void Manager_OnPasskeyRequest(object Sender, wclBluetoothRadio Radio, long Address, out uint Passkey)
        {
            Passkey = 123456;

            TraceEvent(Radio, "PasskeyRequest", "Address", Address.ToString());
            TraceEvent(null, "", "Passkey", Passkey.ToString());
        }

        void Manager_OnPinRequest(object Sender, wclBluetoothRadio Radio, long Address, out string Pin)
        {
            Pin = edPIN.Text;

            TraceEvent(Radio, "PinRequest", "Address", Address.ToString("X12"));
            TraceEvent(null, "", "PIN", Pin);
        }
        
        void  Manager_OnStatusChanged(object Sender, wclBluetoothRadio Radio)
        {
            String Status;
            if (Radio.Available)
                Status = "Available";
            else
            {
                if (Radio.Plugged)
                    Status = "Not available, Plugged";
                else
                    Status = "Unplugged";
            }
            TraceEvent(Radio, "StatusChanged", "New status", Status);

            Boolean Found = false;
            for (int i = 0; i < lvRadios.Items.Count; i++)
            {
                if (((RadioItem)lvRadios.Items[i]).Radio == Radio)
                {
                    Found = true;
                    RefreshRadio(i);
                    break;
                }
            }

            if (!Found)
            {
                wclBluetoothRadio _Radio = Radio;
                int Item = lvRadios.Items.Add(new RadioItem
                {
                    Api = "",
                    Available = "",
                    Address = "",
                    Connectable = "",
                    Discoverable = "",
                    HciVersion = "",
                    LmpVersion = "",
                    Manufacturer = "",
                    Name = "",
                    Cod = "",
                    Radio = _Radio
                });

                RefreshRadio(Item);
            }
        }

        private void fmMain_Closed(object sender, EventArgs e)
        {
            Manager.Close();
            Manager = null;

            FPowerMonitor.Close();
            FPowerMonitor = null;
        }

        private void ClearRadios()
        {
            lvRadios.Items.Clear();
            ClearDevices();
        }

        private void RefreshRadio(int nItem)
        {
            RadioItem Item = (RadioItem)lvRadios.Items[nItem];
            wclBluetoothRadio Radio = Item.Radio;

            Item.Api = Radio.ApiName;
            Item.Available = Radio.Available.ToString();

            Int64 Address = 0;
            Int32 Res = Radio.GetAddress(out Address);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Item.Address = "Error: 0x" + Res.ToString("X8");
            else
                Item.Address = Address.ToString("X12");

            Boolean BoolVal;
            Res = Radio.GetConnectable(out BoolVal);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Item.Connectable = "Error: 0x" + Res.ToString("X8");
            else
                Item.Connectable = BoolVal.ToString();

            Res = Radio.GetDiscoverable(out BoolVal);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Item.Discoverable = "Error: 0x" + Res.ToString("X8");
            else
                Item.Discoverable = BoolVal.ToString();

            Byte ByteVal;
            UInt16 WordVal;
            Res = Radio.GetHciVersion(out ByteVal, out WordVal);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Item.HciVersion = "Error: 0x" + Res.ToString("X8");
            else
                Item.HciVersion = ByteVal.ToString() + "." + WordVal.ToString();

            Res = Radio.GetLmpVersion(out ByteVal, out WordVal);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Item.LmpVersion = "Error: 0x" + Res.ToString("X8");
            else
                Item.LmpVersion = ByteVal.ToString() + "." + WordVal.ToString();

            Res = Radio.GetManufacturer(out WordVal);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Item.Manufacturer = "Error: 0x" + Res.ToString("X8");
            else
                Item.Manufacturer = WordVal.ToString();

            String StrVal;
            Res = Radio.GetName(out StrVal);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Item.Name = "Error: 0x" + Res.ToString("X8");
            else
                Item.Name = StrVal;

            UInt32 Cod;
            Res = Radio.GetCod(out Cod);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Item.Cod = "Error: 0x" + Res.ToString("X8");
            else
                Item.Cod = "0x" + Cod.ToString("X8");

            lvRadios.Items.Refresh();
        }

        private void btRefreshRadio_Click(object sender, RoutedEventArgs e)
        {
            if (lvRadios.SelectedItems.Count == 0)
                MessageBox.Show("Select radio", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
                RefreshRadio(lvRadios.SelectedIndex);
        }

        private void btOpen_Click(object sender, RoutedEventArgs e)
        {
            Int32 Res = Manager.Open(cbHandlePairing.IsChecked.Value);
            if (Res != wclErrors.WCL_E_SUCCESS)
                ShowError(Res);
        }

        private void btClose_Click(object sender, RoutedEventArgs e)
        {
            Int32 Res = Manager.Close();
            if (Res != wclErrors.WCL_E_SUCCESS)
                ShowError(Res);
        }

        private void ShowError(Int32 Error)
        {
            MessageBox.Show("Error: 0x" + Error.ToString("X8"), "Error", MessageBoxButton.OK, MessageBoxImage.Error);
        }

        private void btSetConnectable_Click(object sender, RoutedEventArgs e)
        {
            if (lvRadios.SelectedItems.Count == 0)
                MessageBox.Show("Select radio", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                MessageBoxResult Mr = MessageBox.Show("[YES] - turn connectable ON, [NO] - turn connectable OFF",
                    "Set Connectable", MessageBoxButton.YesNoCancel, MessageBoxImage.Question);
                if (Mr != MessageBoxResult.Cancel)
                {
                    RadioItem Item = (RadioItem)lvRadios.SelectedItems[0];
                    wclBluetoothRadio Radio = Item.Radio;

                    Int32 Res = Radio.SetConnectable(Mr == MessageBoxResult.Yes);
                    if (Res != wclErrors.WCL_E_SUCCESS)
                        ShowError(Res);

                    RefreshRadio(lvRadios.SelectedIndex);
                }
            }
        }

        private void btSetDiscoverable_Click(object sender, RoutedEventArgs e)
        {
            if (lvRadios.SelectedItems.Count == 0)
                MessageBox.Show("Select radio", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                MessageBoxResult Mr = MessageBox.Show("[YES] - turn discoverable ON, [NO] - turn discoverable OFF",
                  "Set discoverable", MessageBoxButton.YesNoCancel, MessageBoxImage.Question);
                if (Mr != MessageBoxResult.Cancel)
                {
                    RadioItem Item = (RadioItem)lvRadios.SelectedItems[0];
                    wclBluetoothRadio Radio = Item.Radio;

                    Int32 Res = Radio.SetDiscoverable(Mr == MessageBoxResult.Yes);
                    if (Res != wclErrors.WCL_E_SUCCESS)
                        ShowError(Res);

                    RefreshRadio(lvRadios.SelectedIndex);
                }
            }
        }

        private void btSetName_Click(object sender, RoutedEventArgs e)
        {
            if (lvRadios.SelectedItems.Count == 0)
                MessageBox.Show("Select radio", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                RadioItem Item = (RadioItem)lvRadios.SelectedItems[0];
                wclBluetoothRadio Radio = Item.Radio;

                Int32 Res = Radio.SetName(edName.Text);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    ShowError(Res);

                RefreshRadio(lvRadios.SelectedIndex);
            }
        }

        private void btEnumPaired_Click(object sender, RoutedEventArgs e)
        {
            if (lvRadios.SelectedItems.Count == 0)
                MessageBox.Show("Select radio", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                ClearDevices();

                RadioItem Item = (RadioItem)lvRadios.SelectedItems[0];
                wclBluetoothRadio Radio = Item.Radio;

                Int64[] Devices;
                Int32 Res = Radio.EnumPairedDevices(out Devices);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    ShowError(Res);
                else
                {
                    if (Devices != null)
                    {
                        for (Int32 i = 0; i < Devices.Length; i++)
                            AddDevice(Radio, Devices[i]);

                        RefreshDevices();
                    }
                }
            }
        }

        private void ClearDevices()
        {
            lvDevices.Items.Clear();
            ClearServices();
        }

        internal class DeviceItem
        {
            public String Driver { get; set; }
            public String Address { get; set; }
            public String Cod { get; set; }
            public String Type { get; set; }
            public String Name { get; set; }
            public String Paired { get; set; }
            public String Connected { get; set; }
            public String AddressType { get; set; }
            internal wclBluetoothRadio Radio { get; set; }
        };

        private void RefreshDevice(int nItem)
        {
            DeviceItem Item = (DeviceItem)lvDevices.Items[nItem];
            wclBluetoothRadio Radio = Item.Radio;
            Int64 Address = Convert.ToInt64(Item.Address, 16);

            UInt32 Cod;
            Int32 Res = Radio.GetRemoteCod(Address, out Cod);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Item.Cod = "Error: 0x" + Res.ToString("X8");
            else
                Item.Cod = Cod.ToString("X8");

            wclBluetoothDeviceType DevType;
            Res = Radio.GetRemoteDeviceType(Address, out DevType);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Item.Type = "Error: 0x" + Res.ToString("X8");
            else
            {
                switch (DevType)
                {
                    case wclBluetoothDeviceType.dtClassic:
                        Item.Type = "Classic";
                        break;
                    case wclBluetoothDeviceType.dtBle:
                        Item.Type = "BLE";
                        break;
                    case wclBluetoothDeviceType.dtMixed:
                        Item.Type = "Mixed";
                        break;
                    default:
                        Item.Type = "Unknown";
                        break;
                }
            }

            String DevName;
            Res = Radio.GetRemoteName(Address, out DevName);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Item.Name = "Error: 0x" + Res.ToString("X8");
            else
                Item.Name = DevName;

            Boolean Paired;
            Res = Radio.GetRemotePaired(Address, out Paired);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Item.Paired = "Error: 0x" + Res.ToString("X8");
            else
                Item.Paired = Paired.ToString();

            Boolean Connected;
            Res = Radio.GetRemoteConnectedStatus(Address, out Connected);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Item.Connected = "Error: 0x" + Res.ToString("X8");
            else
                Item.Connected = Connected.ToString();

            wclBluetoothAddressType AddrType;
            Res = Radio.GetRemoteAddressType(Address, out AddrType);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Item.AddressType = "Error: 0x" + Res.ToString("X8");
            else
            {
                switch (AddrType)
                {
                    case wclBluetoothAddressType.atClassic:
                        Item.AddressType = "Classic";
                        break;
                    case wclBluetoothAddressType.atPublic:
                        Item.AddressType = "LE Public";
                        break;
                    case wclBluetoothAddressType.atRandom:
                        Item.AddressType = "LE Random";
                        break;
                    case wclBluetoothAddressType.atUnspecified:
                        Item.AddressType = "Unspecified";
                        break;
                    default:
                        Item.AddressType = "Unknown";
                        break;
                }
            }

            lvDevices.Items.Refresh();
        }

        private void btRefresh_Click(object sender, RoutedEventArgs e)
        {
            if (lvDevices.SelectedItems.Count == 0)
                MessageBox.Show("Select device", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
                RefreshDevice(lvDevices.SelectedIndex);
        }

        private void ClearServices()
        {
            lvServices.Items.Clear();
            ClearComPorts();
        }

        internal class ServiceItem
        {
            public String Driver { get; set; }
            public String Address { get; set; }
            public String Handle { get; set; }
            public String Uuid { get; set; }
            public String Channel { get; set; }
            public String Name { get; set; }
            public String Comment { get; set; }
        }

        private void btEnumServices_Click(object sender, RoutedEventArgs e)
        {
            if (lvDevices.SelectedItems.Count == 0)
                MessageBox.Show("Select device", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                ClearServices();

                DeviceItem Item = (DeviceItem)lvDevices.SelectedItems[0];
                wclBluetoothRadio Radio = Item.Radio;
                Int64 Address = Convert.ToInt64(Item.Address, 16);

                wclBluetoothService[] Services;
                Guid g = Guid.Empty;
                Int32 Res = Radio.EnumRemoteServices(Address, g, out Services);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    ShowError(Res);
                else
                {
                    if (Services != null)
                    {
                        foreach (wclBluetoothService Service in Services)
                        {
                            lvServices.Items.Add(new ServiceItem
                            {
                                Driver = Radio.ApiName,
                                Address = Address.ToString("X12"),
                                Handle = Service.Handle.ToString("X8"),
                                Uuid = Service.Uuid.ToString(),
                                Channel = Service.Channel.ToString(),
                                Name = Service.Name,
                                Comment = Service.Comment
                            });
                        }
                    }
                }
            }
        }

        private void btDiscover_Click(object sender, RoutedEventArgs e)
        {
            if (lvRadios.SelectedItems.Count == 0)
                MessageBox.Show("Select radio", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                RadioItem Item = (RadioItem)lvRadios.SelectedItems[0];
                wclBluetoothRadio Radio = Item.Radio;

                wclBluetoothDiscoverKind Kind;
                if (cbDiscoveringMode.SelectedIndex == 0)
                    Kind = wclBluetoothDiscoverKind.dkClassic;
                else
                    Kind = wclBluetoothDiscoverKind.dkBle;

                Int32 Res = Radio.Discover(Convert.ToByte(edTimeout.Text), Kind);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    ShowError(Res);
            }
        }

        private void btTerminate_Click(object sender, RoutedEventArgs e)
        {
            if (lvRadios.SelectedItems.Count == 0)
                MessageBox.Show("Select radio", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                RadioItem Item = (RadioItem)lvRadios.SelectedItems[0];
                wclBluetoothRadio Radio = Item.Radio;

                Int32 Res = Radio.Terminate();
                if (Res != wclErrors.WCL_E_SUCCESS)
                    ShowError(Res);
            }
        }

        private void AddDevice(wclBluetoothRadio Radio, Int64 Address)
        {
            lvDevices.Items.Add(new DeviceItem
            {
                Driver = Radio.ApiName,
                Address = Address.ToString("X12"),
                Cod = "",
                Type = "",
                Name = "",
                Paired = "",
                Connected = "",
                AddressType = "",
                Radio = Radio
            });
        }

        private void RefreshDevices()
        {
            for (int i = 0; i < lvDevices.Items.Count; i++ )
                RefreshDevice(i);
        }

        private void btUnpair_Click(object sender, RoutedEventArgs e)
        {
            if (lvDevices.SelectedItems.Count == 0)
                MessageBox.Show("Select device", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                DeviceItem Item = (DeviceItem)lvDevices.SelectedItems[0];
                wclBluetoothRadio Radio = Item.Radio;
                Int64 Address = Convert.ToInt64(Item.Address, 16);

                wclBluetoothPairingMethod Method;
                switch (cbPairingMethod.SelectedIndex)
                {
                    case 1:
                        Method = wclBluetoothPairingMethod.pmClassic;
                        break;
                    case 2:
                        Method = wclBluetoothPairingMethod.pmLe;
                        break;
                    default:
                        Method = wclBluetoothPairingMethod.pmAuto;
                        break;
                }

                Int32 Res = Radio.RemoteUnpair(Address, Method);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    ShowError(Res);

                RefreshDevice(lvDevices.SelectedIndex);
            }
        }

        private void btPair_Click(object sender, RoutedEventArgs e)
        {
            if (lvDevices.SelectedItems.Count == 0)
                MessageBox.Show("Select device", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                DeviceItem Item = (DeviceItem)lvDevices.SelectedItems[0];
                wclBluetoothRadio Radio = Item.Radio;
                Int64 Address = Convert.ToInt64(Item.Address, 16);

                wclBluetoothPairingMethod Method;
                switch (cbPairingMethod.SelectedIndex)
                {
                    case 1:
                        Method = wclBluetoothPairingMethod.pmClassic;
                        break;
                    case 2:
                        Method = wclBluetoothPairingMethod.pmLe;
                        break;
                    default:
                        Method = wclBluetoothPairingMethod.pmAuto;
                        break;
                }

                Int32 Res = Radio.RemotePair(Address, Method);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    ShowError(Res);
            }
        }

        private void btRssi_Click(object sender, RoutedEventArgs e)
        {
            if (lvDevices.SelectedItems.Count == 0)
                MessageBox.Show("Select device", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                DeviceItem Item = (DeviceItem)lvDevices.SelectedItems[0];
                wclBluetoothRadio Radio = Item.Radio;
                Int64 Address = Convert.ToInt64(Item.Address, 16);

                SByte Rssi;
                Int32 Res = Radio.GetRemoteRssi(Address, out Rssi);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    ShowError(Res);
                else
                    MessageBox.Show("RSSI: " + Rssi.ToString() + "dB", "RSSI", MessageBoxButton.OK, MessageBoxImage.Information);
            }
        }

        private void btTurnOn_Click(object sender, RoutedEventArgs e)
        {
            if (lvRadios.SelectedItems.Count == 0)
                MessageBox.Show("Select radio", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                RadioItem Item = (RadioItem)lvRadios.SelectedItems[0];
                wclBluetoothRadio Radio = Item.Radio;

                Int32 Res = Radio.TurnOn();
                if (Res != wclErrors.WCL_E_SUCCESS)
                    ShowError(Res);

                RefreshRadio(lvRadios.SelectedIndex);
            }
        }

        private void btTurnOff_Click(object sender, RoutedEventArgs e)
        {
            if (lvRadios.SelectedItems.Count == 0)
                MessageBox.Show("Select radio", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                RadioItem Item = (RadioItem)lvRadios.SelectedItems[0];
                wclBluetoothRadio Radio = Item.Radio;

                Int32 Res = Radio.TurnOff();
                if (Res != wclErrors.WCL_E_SUCCESS)
                    ShowError(Res);

                RefreshRadio(lvRadios.SelectedIndex);
            }
        }

        private void ClearComPorts()
        {
            lvComPorts.Items.Clear();
        }

        private void btEnumvCOMs_Click(object sender, RoutedEventArgs e)
        {
            if (lvRadios.SelectedItems.Count == 0)
                MessageBox.Show("Select radio", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
                RefreshVComs();
        }

        internal class ComItem
        {
            public String Driver { get; set; }
            public String Address { get; set; }
            public String Uuid { get; set; }
            public String Number { get; set; }
            internal wclBluetoothRadio Radio { get; set; }
        }

        private void RefreshVComs()
        {
            ClearComPorts();

            RadioItem Item = (RadioItem)lvRadios.SelectedItems[0];
            wclBluetoothRadio radio = Item.Radio;

            Int32 Res;
            wclVirtualComPort[] Ports;
            if (cbEnumPreinstalled.IsChecked == true)
                Res = radio.EnumComPorts(out Ports);
            else
            {
                Ports = radio.ComPorts;
                Res = wclErrors.WCL_E_SUCCESS;
            }

            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Error enumerating vCOMs: 0x" + Res.ToString("X8"));
            else
            {
                if (Ports != null && Ports.Length > 0)
                {
                    for (int i = 0; i < Ports.Length; i++)
                    {
                        wclVirtualComPort Port = Ports[i];

                        lvComPorts.Items.Add(new ComItem
                        {
                            Driver = radio.ApiName,
                            Address = Port.Address.ToString("X12"),
                            Uuid = Port.Service.ToString(),
                            Number = Port.Number.ToString(),
                            Radio = radio
                        });
                    }
                }
            }
        }

        private void btCreatevCOM_Click(object sender, RoutedEventArgs e)
        {
            if (lvDevices.SelectedItems.Count == 0)
                MessageBox.Show("Select device", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                if (lvServices.SelectedItems.Count == 0)
                    MessageBox.Show("Select service", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
                else
                {
                    DeviceItem Item = (DeviceItem)lvDevices.SelectedItems[0];
                    wclBluetoothRadio Radio = Item.Radio;
                    Int64 Address = Convert.ToInt64(Item.Address, 16);
                    Guid Service = new Guid(((ServiceItem)lvServices.SelectedItems[0]).Uuid);

                    UInt16 Number;
                    Int32 Res = Radio.CreateComPort(Address, Service, out Number);
                    if (Res != wclErrors.WCL_E_SUCCESS)
                        ShowError(Res);
                    else
                    {
                        TraceEvent(Radio, "Create vCOM", "Number", Number.ToString());

                        RefreshVComs();
                    }
                }
            }
        }

        private void btDestroyvCOM_Click(object sender, RoutedEventArgs e)
        {
            if (lvComPorts.SelectedItems.Count == 0)
                MessageBox.Show("Select vCOM", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                ComItem Item = (ComItem)lvComPorts.SelectedItems[0];
                wclBluetoothRadio Radio = Item.Radio;
                UInt16 Number = Convert.ToUInt16(Item.Number);

                Int32 Res = Radio.DestroyComPort(Number);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    ShowError(Res);
                else
                {
                    TraceEvent(Radio, "Destroy vCOM", "Number", Number.ToString());

                    RefreshVComs();
                }
            }
        }

        private void btInstallDevice_Click(object sender, RoutedEventArgs e)
        {
            if (lvDevices.SelectedItems.Count == 0)
                MessageBox.Show("Select device", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                if (lvServices.SelectedItems.Count == 0)
                    MessageBox.Show("Select service", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
                else
                {
                    DeviceItem Item = (DeviceItem)lvDevices.SelectedItems[0];
                    wclBluetoothRadio Radio = Item.Radio;
                    Int64 Address = Convert.ToInt64(Item.Address, 16);
                    Guid Service = new Guid(((ServiceItem)lvServices.SelectedItems[0]).Uuid);

                    Int32 Res = Radio.InstallDevice(Address, Service);
                    if (Res == wclBluetoothErrors.WCL_E_BLUETOOTH_DEVICE_ALREADY_INSTALLED)
                    {
                        // Reinstall to reconnect!
                        Radio.UninstallDevice(Address, Service);
                        Res = Radio.InstallDevice(Address, Service);
                    }

                    if (Res != wclErrors.WCL_E_SUCCESS)
                        ShowError(Res);
                    else
                        MessageBox.Show("Device installed");
                }
            }
        }

        private void btUninstallDevice_Click(object sender, RoutedEventArgs e)
        {
            if (lvDevices.SelectedItems.Count == 0)
                MessageBox.Show("Select device", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                if (lvServices.SelectedItems.Count == 0)
                    MessageBox.Show("Select service", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
                else
                {
                    DeviceItem Item = (DeviceItem)lvDevices.SelectedItems[0];
                    wclBluetoothRadio Radio = Item.Radio;
                    Int64 Address = Convert.ToInt64(Item.Address, 16);
                    Guid Service = new Guid(((ServiceItem)lvServices.SelectedItems[0]).Uuid);

                    Int32 Res = Radio.UninstallDevice(Address, Service);
                    if (Res != wclErrors.WCL_E_SUCCESS)
                        ShowError(Res);
                    else
                        MessageBox.Show("Device uninstalled");
                }
            }
        }

        private void btIsInRange_Click(object sender, RoutedEventArgs e)
        {
            if (lvDevices.SelectedItems.Count == 0)
                MessageBox.Show("Select device", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                DeviceItem Item = (DeviceItem)lvDevices.SelectedItems[0];
                wclBluetoothRadio Radio = Item.Radio;
                Int64 Address = Convert.ToInt64(Item.Address, 16);

                Boolean InRange;
                Int32 Res = Radio.IsRemoteDeviceInRange(Address, out InRange);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    ShowError(Res);
                else
                {
                    if (InRange)
                        MessageBox.Show("Device is in range");
                    else
                        MessageBox.Show("Device is NOT in range");
                }
            }
        }

        private void btDisconnect_Click(object sender, RoutedEventArgs e)
        {
            if (lvDevices.SelectedItems.Count == 0)
                MessageBox.Show("Select device", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                DeviceItem Item = (DeviceItem)lvDevices.SelectedItems[0];
                wclBluetoothRadio Radio = Item.Radio;
                Int64 Address = Convert.ToInt64(Item.Address, 16);

                Int32 Res = Radio.RemoteDisconnect(Address);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    ShowError(Res);
            }
        }

        private void btEnumConnected_Click(object sender, RoutedEventArgs e)
        {
            if (lvRadios.SelectedItems.Count == 0)
                MessageBox.Show("Select radio", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                ClearDevices();

                RadioItem Item = (RadioItem)lvRadios.SelectedItems[0];
                wclBluetoothRadio Radio = Item.Radio;

                Int64[] Devices;
                Int32 Res = Radio.EnumConnectedDevices(out Devices);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    ShowError(Res);
                else
                {
                    if (Devices != null)
                    {
                        for (Int32 i = 0; i < Devices.Length; i++)
                            AddDevice(Radio, Devices[i]);

                        RefreshDevices();
                    }
                }
            }
        }

        private void btEnumInstalledServices_Click(object sender, RoutedEventArgs e)
        {
            if (lvDevices.SelectedItems.Count == 0)
                MessageBox.Show("Select device", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                ClearServices();

                DeviceItem Item = (DeviceItem)lvDevices.SelectedItems[0];
                wclBluetoothRadio Radio = Item.Radio;
                Int64 Address = Convert.ToInt64(Item.Address, 16);

                Guid[] Services;
                Int32 Res = Radio.EnumInstalledServices(Address, out Services);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    ShowError(Res);
                else
                {
                    if (Services != null && Services.Length > 0)
                    {
                        foreach (Guid Service in Services)
                        {
                            lvServices.Items.Add(new ServiceItem
                            {
                                Driver = Radio.ApiName,
                                Address = Address.ToString("X12"),
                                Handle = "",
                                Uuid = Service.ToString(),
                                Channel = "",
                                Name = "",
                                Comment = ""
                            });
                        }
                    }
                }
            }
        }
    }
}
