﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;

using wclCommon;
using wclBluetooth;

namespace GattClient
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        private wclBluetoothManager Manager;
        private wclGattClient Client;

        private wclGattCharacteristic[] FCharacteristics;
        private wclGattDescriptor[] FDescriptors;
        private wclGattService[] FServices;

        internal class DeviceItem
        {
            public String Address { get; set; }
            public String Name { get; set; }
            public String Type { get; set; }
            public String Paired { get; set; }
            internal wclBluetoothRadio Radio { get; set; }
        };

        internal class EventItem
        {
            public String Address { get; set; }
            public String Event { get; set; }
            public String Param { get; set; }
            public String Value { get; set; }
        };

        internal class ServiceItem
        {
            public String Uuid { get; set; }
            public String Short { get; set; }
            public String Handle { get; set; }
        };

        internal class CharItem
        {
            public String Uuid { get; set; }
            public String Short { get; set; }
            public String ServiceHandle { get; set; }
            public String Handle { get; set; }
            public String ValueHandle { get; set; }
            public String IsBroadcastable { get; set; }
            public String IsReadable { get; set; }
            public String IsWritable { get; set; }
            public String IsWritableWithoutResponse { get; set; }
            public String IsSignedWritable { get; set; }
            public String IsNotifiable { get; set; }
            public String IsIndicatable { get; set; }
            public String HasExtendedProperties { get; set; }
        };

        internal class DescItem
        {
            public String Uuid { get; set; }
            public String Short { get; set; }
            public String ServiceHandle { get; set; }
            public String CharHandle { get; set; }
            public String Handle { get; set; }
            public String Type { get; set; }
        };

        internal class DescValItem
        {
            public String Name { get; set; }
            public String Value { get; set; }
        };

        public MainWindow()
        {
            InitializeComponent();
        }

        private void fmMain_Loaded(object sender, RoutedEventArgs e)
        {
            Manager = new wclBluetoothManager();
            Client = new wclGattClient();

            Manager.OnNumericComparison += new wclBluetoothNumericComparisonEvent(Manager_OnNumericComparison);
            Manager.OnPasskeyNotification += new wclBluetoothPasskeyNotificationEvent(Manager_OnPasskeyNotification);
            Manager.OnPasskeyRequest += new wclBluetoothPasskeyRequestEvent(Manager_OnPasskeyRequest);
            Manager.OnPinRequest += new wclBluetoothPinRequestEvent(Manager_OnPinRequest);
            Manager.OnAuthenticationCompleted += new wclBluetoothDeviceResultEvent(Manager_OnAuthenticationCompleted);
            Manager.OnConfirm += new wclBluetoothConfirmEvent(Manager_OnConfirm);
            Manager.OnProtectionLevelRequest += new wclBluetoothProtectionLevelRequestEvent(Manager_OnProtectionLevelRequest);
            Manager.OnDeviceFound += new wclBluetoothDeviceEvent(Manager_OnDeviceFound);
            Manager.OnDiscoveringCompleted += new wclBluetoothResultEvent(Manager_OnDiscoveringCompleted);
            Manager.OnDiscoveringStarted += new wclBluetoothEvent(Manager_OnDiscoveringStarted);
            Manager.OnIoCapabilityRequest += new wclBluetoothIoCapabilityRequestEvent(Manager_OnIoCapabilityRequest);

            Client.OnCharacteristicChanged += new wclGattCharacteristicChangedEvent(Client_OnCharacteristicChanged);
            Client.OnConnect += new wclCommunication.wclClientConnectionConnectEvent(Client_OnConnect);
            Client.OnDisconnect += new wclCommunication.wclClientConnectionDisconnectEvent(Client_OnDisconnect);
            Client.OnConnectionParamsChanged += new EventHandler(Client_OnConnectionParamsChanged);
            Client.OnMaxPduSizeChanged += new EventHandler(Client_OnMaxPduSizeChanged);
            Client.OnConnectionPhyChanged += new EventHandler(Client_OnConnectionPhyChanged);

            Manager.Open();

            Cleanup();
        }

        void Manager_OnIoCapabilityRequest(Object Sender, wclBluetoothRadio Radio, Int64 Address,
            out wclBluetoothMitmProtection Mitm, out wclBluetoothIoCapability IoCapability, out Boolean OobPresent)
        {
            switch (cbMitmProtection.SelectedIndex)
            {
                case 0:
                    Mitm = wclBluetoothMitmProtection.mitmProtectionNotRequired;
                    break;
                case 1:
                    Mitm = wclBluetoothMitmProtection.mitmProtectionRequired;
                    break;
                case 2:
                    Mitm = wclBluetoothMitmProtection.mitmProtectionNotRequiredBonding;
                    break;
                case 3:
                    Mitm = wclBluetoothMitmProtection.mitmProtectionRequiredBonding;
                    break;
                case 4:
                    Mitm = wclBluetoothMitmProtection.mitmProtectionNotRequiredGeneralBonding;
                    break;
                case 5:
                    Mitm = wclBluetoothMitmProtection.mitmProtectionRequiredGeneralBonding;
                    break;
                default:
                    Mitm = wclBluetoothMitmProtection.mitmProtectionNotDefined;
                    break;
            }
            
            switch (cbIoCap.SelectedIndex)
            {
                case 0:
                    IoCapability = wclBluetoothIoCapability.iocapDisplayOnly;
                    break;
                case 1:
                    IoCapability = wclBluetoothIoCapability.iocapDisplayYesNo;
                    break;
                case 2:
                    IoCapability = wclBluetoothIoCapability.iocapKeyboardOnly;
                    break;
                case 3:
                    IoCapability = wclBluetoothIoCapability.iocapNoInputNoOutput;
                    break;
                case 4:
                    IoCapability = wclBluetoothIoCapability.iocapDisplayKeyboard;
                    break;
                default:
                    IoCapability = wclBluetoothIoCapability.iocapNotDefined;
                    break;
            }
            
            OobPresent = false;
        }

        private void fmMain_Closed(object sender, EventArgs e)
        {
            Client.Disconnect();
            Client = null;

            Manager.Close();
            Manager = null;

            Cleanup();
        }

        private wclBluetoothRadio GetRadio()
        {
            wclBluetoothRadio Radio;
            Int32 Res = Manager.GetLeRadio(out Radio);
            if (Res != wclErrors.WCL_E_SUCCESS)
            {
                MessageBox.Show("Get working radio failed: 0x" + Res.ToString("X8"));
                return null;
            }
            return Radio;
        }

        private void btDiscover_Click(object sender, RoutedEventArgs e)
        {
            wclBluetoothRadio Radio = GetRadio();
            if (Radio != null)
            {
                Int32 Res = Radio.Discover(10, wclBluetoothDiscoverKind.dkBle);
                if (Res != wclErrors.WCL_E_SUCCESS)
                {
                    MessageBox.Show("Error starting discovering: 0x" + Res.ToString("X8"),
                        "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                }
            }
        }

        private void btClear_Click(object sender, RoutedEventArgs e)
        {
            lvEvents.Items.Clear();
        }

        private void TraceEvent(Int64 address, String ev, String param, String value)
        {
            String s = "";
            if (address != 0)
                s = address.ToString("X12");

            lvEvents.Items.Add(new EventItem
            {
                Address = s,
                Event = ev,
                Param = param,
                Value = value
            });
        }

        private void btConnect_Click(object sender, RoutedEventArgs e)
        {
            if (lvDevices.SelectedItems.Count == 0)
                MessageBox.Show("Select device", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                DeviceItem Item = (DeviceItem)lvDevices.SelectedItems[0];
                try
                {
                    Client.Address = Convert.ToInt64(Item.Address, 16);
                    Client.ConnectOnRead = cbConnectOnRead.IsChecked.Value;
                    Client.ForceNotifications = cbForceIndications.IsChecked.Value;
                    Int32 Res = Client.Connect(Item.Radio);
                    if (Res != wclErrors.WCL_E_SUCCESS)
                        MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.Message, "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                }
            }
        }

        private void btDisconnect_Click(object sender, RoutedEventArgs e)
        {
            Int32 Res = Client.Disconnect();
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButton.OK, MessageBoxImage.Error);
        }

        private void btGetServices_Click(object sender, RoutedEventArgs e)
        {
            lvServices.Items.Clear();
            FServices = null;

            Int32 Res = Client.ReadServices(OpFlag(), out FServices);
            if (Res != wclErrors.WCL_E_SUCCESS)
            {
                MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                return;
            }

            if (FServices == null)
                return;

            foreach (wclGattService Service in FServices)
            {
                String s;
                if (Service.Uuid.IsShortUuid)
                    s = Service.Uuid.ShortUuid.ToString("X4");
                else
                    s = Service.Uuid.LongUuid.ToString();
                lvServices.Items.Add(new ServiceItem
                {
                    Uuid = s,
                    Short = Service.Uuid.IsShortUuid.ToString(),
                    Handle = Service.Handle.ToString("X4")
                });
            }
        }

        private void btGetIncludedServices_Click(object sender, RoutedEventArgs e)
        {
            if (lvServices.SelectedItems.Count == 0)
            {
                MessageBox.Show("Select service", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
                return;
            }

            wclGattService Service = FServices[lvServices.SelectedIndex];

            lvServices.Items.Clear();
            FServices = null;

            Int32 Res = Client.ReadIncludedServices(Service, OpFlag(), out FServices);
            if (Res != wclErrors.WCL_E_SUCCESS)
            {
                MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                return;
            }

            if (FServices == null)
                return;

            foreach (wclGattService IncService in FServices)
            {
                String s;
                if (IncService.Uuid.IsShortUuid)
                    s = IncService.Uuid.ShortUuid.ToString("X4");
                else
                    s = IncService.Uuid.LongUuid.ToString();

                lvServices.Items.Add(new ServiceItem
                {
                    Uuid = s,
                    Short = IncService.Uuid.IsShortUuid.ToString(),
                    Handle = IncService.Handle.ToString("X4")
                });
            }
        }

        private void btGetCharacteristics_Click(object sender, RoutedEventArgs e)
        {
            FCharacteristics = null;
            lvCharacteristics.Items.Clear();

            if (lvServices.SelectedItems.Count == 0)
            {
                MessageBox.Show("Select service", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
                return;
            }

            wclGattService Service = FServices[lvServices.SelectedIndex];
            Int32 Res = Client.ReadCharacteristics(Service, OpFlag(), out FCharacteristics);
            if (Res != wclErrors.WCL_E_SUCCESS)
            {
                MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                return;
            }

            if (FCharacteristics == null)
                return;

            foreach (wclGattCharacteristic Character in FCharacteristics)
            {
                String s;
                if (Character.Uuid.IsShortUuid)
                    s = Character.Uuid.ShortUuid.ToString("X4");
                else
                    s = Character.Uuid.LongUuid.ToString();
                lvCharacteristics.Items.Add(new CharItem
                {
                    Uuid = s,
                    Short = Character.Uuid.IsShortUuid.ToString(),
                    ServiceHandle = Character.ServiceHandle.ToString("X4"),
                    Handle = Character.Handle.ToString("X4"),
                    ValueHandle = Character.ValueHandle.ToString("X4"),
                    IsBroadcastable = Character.IsBroadcastable.ToString(),
                    IsReadable = Character.IsReadable.ToString(),
                    IsWritable = Character.IsWritable.ToString(),
                    IsWritableWithoutResponse = Character.IsWritableWithoutResponse.ToString(),
                    IsSignedWritable = Character.IsSignedWritable.ToString(),
                    IsNotifiable = Character.IsNotifiable.ToString(),
                    IsIndicatable = Character.IsIndicatable.ToString(),
                    HasExtendedProperties = Character.HasExtendedProperties.ToString()
                });
            }
        }

        private void btGetValue_Click(object sender, RoutedEventArgs e)
        {
            if (lvCharacteristics.SelectedItems.Count == 0)
            {
                MessageBox.Show("Select characteristic", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
                return;
            }

            wclGattCharacteristic Characteristic = FCharacteristics[lvCharacteristics.SelectedIndex];
            Byte[] Value;
            Int32 Res = Client.ReadCharacteristicValue(Characteristic, OpFlag(), out Value, Protection());
            if (Res != wclErrors.WCL_E_SUCCESS)
            {
                MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                return;
            }

            if (Value == null)
                return;

            if (Value.Length == 0)
                MessageBox.Show("Value is empty", "Information", MessageBoxButton.OK, MessageBoxImage.Information);
            else
            {
                String Str = "";
                for (Int32 i = 0; i < Value.Length; i++)
                    Str = Str + Value[i].ToString("X2");

                MessageBox.Show("Value: " + Str, "Characterist value", MessageBoxButton.OK, MessageBoxImage.Information);
            }
        }

        private void btGetDesc_Click(object sender, RoutedEventArgs e)
        {
            FDescriptors = null;
            lvDescriptors.Items.Clear();

            if (lvCharacteristics.SelectedItems.Count == 0)
            {
                MessageBox.Show("Select characteristic", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
                return;
            }

            wclGattCharacteristic Characteristic = FCharacteristics[lvCharacteristics.SelectedIndex];
            Int32 Res = Client.ReadDescriptors(Characteristic, OpFlag(), out FDescriptors);
            if (Res != wclErrors.WCL_E_SUCCESS)
            {
                MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                return;
            }

            if (FDescriptors == null)
                return;

            foreach (wclGattDescriptor Descriptor in FDescriptors)
            {
                String s;
                if (Descriptor.Uuid.IsShortUuid)
                    s = Descriptor.Uuid.ShortUuid.ToString("X4");
                else
                    s = Descriptor.Uuid.LongUuid.ToString();

                String t;
                switch (Descriptor.DescriptorType)
                {
                    case wclGattDescriptorType.dtCharacteristicExtendedProperties:
                        t = "dtCharacteristicExtendedProperties";
                        break;
                    case wclGattDescriptorType.dtCharacteristicUserDescription:
                        t = "dtCharacteristicUserDescription";
                        break;
                    case wclGattDescriptorType.dtClientCharacteristicConfiguration:
                        t = "dtClientCharacteristicConfiguration";
                        break;
                    case wclGattDescriptorType.dtServerCharacteristicConfiguration:
                        t = "dtServerCharacteristicConfiguration";
                        break;
                    case wclGattDescriptorType.dtCharacteristicFormat:
                        t = "dtCharacteristicFormat";
                        break;
                    case wclGattDescriptorType.dtCharacteristicAggregateFormat:
                        t = "dtCharacteristicAggregateFormat";
                        break;
                    default:
                        t = "dtCustomDescriptor";
                        break;
                }

                lvDescriptors.Items.Add(new DescItem
                {
                    Uuid = s,
                    Short = Descriptor.Uuid.IsShortUuid.ToString(),
                    ServiceHandle = Descriptor.ServiceHandle.ToString("X4"),
                    CharHandle = Descriptor.CharacteristicHandle.ToString("X4"),
                    Handle = Descriptor.Handle.ToString("X4"),
                    Type = t
                });
            }
        }

        private void Cleanup()
        {
            FCharacteristics = null;
            FDescriptors = null;
            FServices = null;

            lvServices.Items.Clear();
            lvCharacteristics.Items.Clear();
            lvDescriptors.Items.Clear();
            lvDescriptorValue.Items.Clear();
        }

        private void btGetDescvalue_Click(object sender, RoutedEventArgs e)
        {
            lvDescriptorValue.Items.Clear();

            if (lvDescriptors.SelectedItems.Count == 0)
            {
                MessageBox.Show("Select descriptor", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
                return;
            }

            wclGattDescriptor Descriptor = FDescriptors[lvDescriptors.SelectedIndex];
            wclGattDescriptorValue Value;
            Int32 Res = Client.ReadDescriptorValue(Descriptor, OpFlag(), out Value, Protection());
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButton.OK, MessageBoxImage.Error);
            else
            {
                switch (Value.AType)
                {
                    case wclGattDescriptorType.dtCharacteristicExtendedProperties:
                        ListProp("Type", "dtCharacteristicExtendedProperties");
                        ListProp("IsReliableWriteEnabled", Value.CharacteristicExtendedProperties.IsReliableWriteEnabled.ToString());
                        ListProp("IsAuxiliariesWritable", Value.CharacteristicExtendedProperties.IsAuxiliariesWritable.ToString());
                        break;

                    case wclGattDescriptorType.dtCharacteristicUserDescription:
                        ListProp("Type", "dtCharacteristicUserDescription");
                        ListProp("Descriptipon", Value.UserDescription.Description);
                        break;

                    case wclGattDescriptorType.dtClientCharacteristicConfiguration:
                        ListProp("Type", "dtClientCharacteristicConfiguration");
                        ListProp("IsSubscribeToNotification", Value.ClientCharacteristicConfiguration.IsSubscribeToNotification.ToString());
                        ListProp("IsSubscribeToIndication", Value.ClientCharacteristicConfiguration.IsSubscribeToIndication.ToString());
                        break;

                    case wclGattDescriptorType.dtServerCharacteristicConfiguration:
                        ListProp("Type", "dtServerCharacteristicConfiguration");
                        ListProp("IsBroadcast", Value.ServerCharacteristicConfiguration.IsBroadcast.ToString());
                        break;

                    case wclGattDescriptorType.dtCharacteristicFormat:
                        ListProp("Type", "dtCharacteristicFormat");
                        ListProp("Format", Value.CharacteristicFormat.Format.ToString());
                        ListProp("Exponent", Value.CharacteristicFormat.Exponent.ToString("X2"));
                        ListProp("Unit", Value.CharacteristicFormat.AUnit.ToString("X4"));
                        ListProp("NameSpace", Value.CharacteristicFormat.NameSpace.ToString("X2"));
                        ListProp("Description", Value.CharacteristicFormat.Description.ToString("X4"));
                        break;

                    case wclGattDescriptorType.dtCharacteristicAggregateFormat:
                        ListProp("Type", "dtCharacteristicAggregateFormat");
                        if (Value.CharacteristicAggregateFormat.Handles != null && Value.CharacteristicAggregateFormat.Handles.Length > 0)
                        {
                            for (Int32 i = 0; i < Value.CharacteristicAggregateFormat.Handles.Length; i++)
                            {
                                ListProp("Handle [" + i.ToString() + "]", Value.CharacteristicAggregateFormat.Handles[i].ToString("X4"));
                            }
                        }
                        break;

                    default:
                        ListProp("Type", "UNKNOWN");
                        break;
                }

                if (Value.Data != null)
                {
                    String Str = "";
                    for (Int32 i = 0; i < Value.Data.Length; i++)
                        Str = Str + Value.Data[i].ToString("X2");
                    ListProp("Data", Str);
                }
            }
        }

        private void ListProp(String name, String value)
        {
            lvDescriptorValue.Items.Add(new DescValItem
            {
                Name = name,
                Value = value
            });
        }

        private void btSetValue_Click(object sender, RoutedEventArgs e)
        {
            if (lvCharacteristics.SelectedItems.Count == 0)
            {
                MessageBox.Show("Select characteristic", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
                return;
            }

            wclGattCharacteristic Characteristic = FCharacteristics[lvCharacteristics.SelectedIndex];

            String Str = edCharVal.Text;
            if (Str.Length % 2 != 0)
                Str = "0" + Str;

            Byte[] Val = new Byte[Str.Length / 2];
            for (Int32 i = 0; i < Val.Length; i++)
            {
                String b = Str.Substring(i * 2, 2);
                Val[i] = Convert.ToByte(b, 16);
            }

            Int32 Res;
            wclGattWriteKind WriteKind;
            switch (cbWriteKind.SelectedIndex)
            {
                case 0:
                    WriteKind = wclGattWriteKind.wkWithResponse;
                    break;
                case 1:
                    WriteKind = wclGattWriteKind.wkWithoutResponse;
                    break;
                default:
                    WriteKind = wclGattWriteKind.wkAuto;
                    break;
            }

            Res = Client.WriteCharacteristicValue(Characteristic, Val, Protection(), WriteKind);
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButton.OK, MessageBoxImage.Error);
        }

        private void btSubscribe_Click(object sender, RoutedEventArgs e)
        {
            if (lvCharacteristics.SelectedItems.Count == 0)
            {
                MessageBox.Show("Select characteristic", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
                return;
            }

            wclGattCharacteristic Characteristic = FCharacteristics[lvCharacteristics.SelectedIndex];

            wclGattSubscribeKind SubscribeKind;
            switch (cbSubscribeKind.SelectedIndex)
            {
                case 0:
                    SubscribeKind = wclGattSubscribeKind.skNotification;
                    break;
                case 1:
                    SubscribeKind = wclGattSubscribeKind.skIndication;
                    break;
                default:
                    SubscribeKind = wclGattSubscribeKind.skManual;
                    break;
            }

            if (SubscribeKind == wclGattSubscribeKind.skManual)
            {
                // In case if characteristic has both Indication and Notification properties
                // set to True we have to select one of them. Here we use Notifications but
                // you can use other one.
                if (Characteristic.IsNotifiable && Characteristic.IsIndicatable)
                {
                    // Change the code line below to
                    // Characteristic.IsNotifiable = false;
                    // if you want to receive Indications instead of notifications.
                    Characteristic.IsIndicatable = false;
                }
            }

            Int32 Res;
            if (cbFastSubscribe.IsChecked.Value)
            {
                if (SubscribeKind == wclGattSubscribeKind.skManual)
                    Res = Client.SubscribeForNotifications(Characteristic);
                else
                {
                    Res = Client.SubscribeForNotifications(Characteristic, OpFlag(),
                        Protection(), SubscribeKind);
                }
            }
            else
            {
                if (SubscribeKind == wclGattSubscribeKind.skManual)
                    Res = Client.Subscribe(Characteristic);
                else
                    Res = Client.Subscribe(Characteristic, SubscribeKind);
            }

            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButton.OK, MessageBoxImage.Error);
        }

        private void btUnsubscribe_Click(object sender, RoutedEventArgs e)
        {
            if (lvCharacteristics.SelectedItems.Count == 0)
            {
                MessageBox.Show("Select characteristic", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
                return;
            }

            wclGattCharacteristic Characteristic = FCharacteristics[lvCharacteristics.SelectedIndex];

            wclGattSubscribeKind SubscribeKind;
            switch (cbSubscribeKind.SelectedIndex)
            {
                case 0:
                    SubscribeKind = wclGattSubscribeKind.skNotification;
                    break;
                case 1:
                    SubscribeKind = wclGattSubscribeKind.skIndication;
                    break;
                default:
                    SubscribeKind = wclGattSubscribeKind.skManual;
                    break;
            }

            if (SubscribeKind == wclGattSubscribeKind.skManual)
            {
                // In case if characteristic has both Indication and Notification properties
                // set to True we have to select one of them. Here we use Notifications but
                // you can use other one.
                // YOU MUST USE THE SAME PROPERTY THAT IS USED IN SUBSCRIBE
                if (Characteristic.IsNotifiable && Characteristic.IsIndicatable)
                {
                    // Change the code line below to
                    // Characteristic.IsNotifiable = false;
                    // if you want to receive Indications instead of notifications.
                    Characteristic.IsIndicatable = false;
                }
            }

            Int32 Res;
            if (cbFastSubscribe.IsChecked.Value)
            {
                if (SubscribeKind == wclGattSubscribeKind.skManual)
                    Res = Client.UnsubscribeFromNotifications(Characteristic);
                else
                {
                    Res = Client.UnsubscribeFromNotifications(Characteristic, OpFlag(),
                        Protection(), SubscribeKind);
                }
            }
            else
            {
                if (SubscribeKind == wclGattSubscribeKind.skManual)
                    Res = Client.Unsubscribe(Characteristic);
                else
                    Res = Client.Unsubscribe(Characteristic, SubscribeKind);
            }

            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButton.OK, MessageBoxImage.Error);
        }

        private wclGattOperationFlag OpFlag()
        {
            switch (cbOpFlag.SelectedIndex)
            {
                case 1:
                    return wclGattOperationFlag.goReadFromDevice;
                case 2:
                    return wclGattOperationFlag.goReadFromCache;
                default:
                    return wclGattOperationFlag.goNone;
            }
        }

        private void btWriteCcdSubsc_Click(object sender, RoutedEventArgs e)
        {
            if (lvCharacteristics.SelectedItems.Count == 0)
            {
                MessageBox.Show("Select characteristic", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
                return;
            }

            wclGattCharacteristic Characteristic = FCharacteristics[lvCharacteristics.SelectedIndex];

            wclGattSubscribeKind SubscribeKind;
            switch (cbSubscribeKind.SelectedIndex)
            {
                case 0:
                    SubscribeKind = wclGattSubscribeKind.skNotification;
                    break;
                case 1:
                    SubscribeKind = wclGattSubscribeKind.skIndication;
                    break;
                default:
                    SubscribeKind = wclGattSubscribeKind.skManual;
                    break;
            }

            Int32 Res;
            if (SubscribeKind == wclGattSubscribeKind.skManual)
            {
                // In case if characteristic has both Indication and Notification properties
                // set to True we have to select one of them. Here we use Notifications but
                // you can use other one.
                if (Characteristic.IsNotifiable && Characteristic.IsIndicatable)
                {
                    // Change the code line below to
                    // Characteristic.IsNotifiable = false;
                    // if you want to receive Indications instead of notifications.
                    Characteristic.IsIndicatable = false;
                }
                Res = Client.WriteClientConfiguration(Characteristic, true, OpFlag(), Protection());
            }
            else
            {
                Res = Client.WriteClientConfiguration(Characteristic, true, OpFlag(),
                    Protection(), SubscribeKind);
            }

            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButton.OK, MessageBoxImage.Error);
        }

        private void btWriteCcdUnsubsc_Click(object sender, RoutedEventArgs e)
        {
            if (lvCharacteristics.SelectedItems.Count == 0)
            {
                MessageBox.Show("Select characteristic", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
                return;
            }

            wclGattCharacteristic Characteristic = FCharacteristics[lvCharacteristics.SelectedIndex];

            wclGattSubscribeKind SubscribeKind;
            switch (cbSubscribeKind.SelectedIndex)
            {
                case 0:
                    SubscribeKind = wclGattSubscribeKind.skNotification;
                    break;
                case 1:
                    SubscribeKind = wclGattSubscribeKind.skIndication;
                    break;
                default:
                    SubscribeKind = wclGattSubscribeKind.skManual;
                    break;
            }

            Int32 Res;
            if (SubscribeKind == wclGattSubscribeKind.skManual)
            {
                // In case if characteristic has both Indication and Notification properties
                // set to True we have to select one of them. Here we use Notifications but
                // you can use other one.
                // YOU MUST USE THE SAME PROPERTY THAT IS USED IN SUBSCRIBE
                if (Characteristic.IsNotifiable && Characteristic.IsIndicatable)
                {
                    // Change the code line below to
                    // Characteristic.IsNotifiable = false;
                    // if you want to receive Indications instead of notifications.
                    Characteristic.IsIndicatable = false;
                }
                Res = Client.WriteClientConfiguration(Characteristic, false, OpFlag(), Protection());
            }
            else
            {
                Res = Client.WriteClientConfiguration(Characteristic, false, OpFlag(),
                    Protection(), SubscribeKind);
            }

            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButton.OK, MessageBoxImage.Error);
        }

        void Manager_OnNumericComparison(object Sender, wclBluetoothRadio Radio, long Address, uint Number, out bool Confirm)
        {
            // Accept any pairing.
            Confirm = true;
            TraceEvent(Address, "Numeric comparison", "Number", Number.ToString());
        }

        void Manager_OnPasskeyNotification(object Sender, wclBluetoothRadio Radio, long Address, uint Passkey)
        {
            TraceEvent(Address, "Passkey notification", "Passkey", Passkey.ToString());
        }

        void Manager_OnPasskeyRequest(object Sender, wclBluetoothRadio Radio, long Address, out uint Passkey)
        {
            // Use 123456 as passkey.
            Passkey = 123456;
            TraceEvent(Address, "Passkey request", "Passkey", Passkey.ToString());
        }

        void Manager_OnPinRequest(object Sender, wclBluetoothRadio Radio, long Address, out string Pin)
        {
            // Use '0000' as PIN.
            Pin = "0000";
            TraceEvent(Address, "Pin request", "PIN", Pin);
        }

        void Manager_OnDeviceFound(object Sender, wclBluetoothRadio Radio, long Address)
        {
            DeviceItem Item = new DeviceItem();
            Item.Address = Address.ToString("X12");
            Item.Name = "";
            Item.Radio = Radio; // To use it later.

            wclBluetoothDeviceType DevType = wclBluetoothDeviceType.dtMixed;
            Int32 Res = Radio.GetRemoteDeviceType(Address, out DevType);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Item.Type = "Error: 0x" + Res.ToString("X8");
            else
            {
                switch (DevType)
                {
                    case wclBluetoothDeviceType.dtClassic:
                        Item.Type = "Classic";
                        break;
                    case wclBluetoothDeviceType.dtBle:
                        Item.Type = "BLE";
                        break;
                    case wclBluetoothDeviceType.dtMixed:
                        Item.Type = "Mixed";
                        break;
                    default:
                        Item.Type = "Unknown";
                        break;
                }
            }

            Boolean Paired;
            Res = Radio.GetRemotePaired(Address, out Paired);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Item.Paired = "Error: 0x" + Res.ToString("X8");
            else
            {
                if (Paired)
                    Item.Paired = "True";
                else
                    Item.Paired = "False";
            }

            lvDevices.Items.Add(Item);

            TraceEvent(Address, "Device found", "", "");
        }

        void Manager_OnDiscoveringCompleted(object Sender, wclBluetoothRadio Radio, int Error)
        {
            if (lvDevices.Items.Count == 0)
                MessageBox.Show("No BLE devices were found.", "Discovering for BLE devices", MessageBoxButton.OK, MessageBoxImage.Information);
            else
            {
                // Here we can update found devices names.
                for (Int32 i = 0; i < lvDevices.Items.Count; i++)
                {
                    DeviceItem Item = (DeviceItem)lvDevices.Items[i];

                    Int64 Address = Convert.ToInt64(Item.Address, 16);
                    String DevName;
                    Int32 Res = Radio.GetRemoteName(Address, out DevName);
                    if (Res != wclErrors.WCL_E_SUCCESS)
                        Item.Name = "Error: 0x" + Res.ToString("X8");
                    else
                        Item.Name = DevName;
                }

                lvDevices.Items.Refresh();
            }

            TraceEvent(0, "Discovering completed", "", "");
        }

        void Manager_OnDiscoveringStarted(object Sender, wclBluetoothRadio Radio)
        {
            lvDevices.Items.Clear();
            TraceEvent(0, "Discovering started", "", "");
        }

        void Client_OnCharacteristicChanged(object Sender, ushort Handle, byte[] Value)
        {
            TraceEvent(((wclGattClient)Sender).Address, "ValueChanged", "Handle", Handle.ToString("X4"));
            if (Value == null)
                TraceEvent(0, "", "Value", "");
            else
            {
                if (Value.Length == 0)
                    TraceEvent(0, "", "Value", "");
                else
                {
                    String Str = "";

                    for (Int32 i = 0; i < Value.Length; i++)
                        Str = Str + Value[i].ToString("X2");

                    TraceEvent(0, "", "Value", Str);
                }
            }
        }

        void Client_OnConnect(object Sender, int Error)
        {
            // Connection property is valid here.
            TraceEvent(((wclGattClient)Sender).Address, "Connected", "Error", "0x" + Error.ToString("X8"));

            if (Error == wclErrors.WCL_E_SUCCESS)
            {
                TraceEvent(Client.Address, "Max PDU", "", "");
                GetMaxPduSize();

                TraceEvent(Client.Address, "Connection params", "", "");
                GetConnectionParams();

                TraceEvent(Client.Address, "Connection PHY", "", "");
                GetConnectionPhy();
            }
        }

        void Client_OnDisconnect(object Sender, int Reason)
        {
            // Connection property is valid here.
            TraceEvent(((wclGattClient)Sender).Address, "Disconnected", "Reason", "0x" + Reason.ToString("X8"));

            Cleanup();
        }

        private void btTerminate_Click(object sender, RoutedEventArgs e)
        {
            wclBluetoothRadio Radio = GetRadio();
            if (Radio != null)
            {
                Int32 Res = Radio.Terminate();
                if (Res != wclErrors.WCL_E_SUCCESS)
                {
                    MessageBox.Show("Error terminating discovering: 0x" + Res.ToString("X8"),
                        "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                }
            }
        }

        private wclGattProtectionLevel Protection()
        {
            switch (cbProtection.SelectedIndex)
            {
                case 0:
                    return wclGattProtectionLevel.plNone;
                case 1:
                    return wclGattProtectionLevel.plAuthentication;
                case 2:
                    return wclGattProtectionLevel.plEncryption;
                case 3:
                    return wclGattProtectionLevel.plEncryptionAndAuthentication;
                default:
                    return wclGattProtectionLevel.plNone;
            }
        }

        void Manager_OnConfirm(object Sender, wclBluetoothRadio Radio, long Address, out bool Confirm)
        {
            // Accept any pairing.
            Confirm = true;
            TraceEvent(Address, "Just works pairing", "Accept", "True");
        }

        void Manager_OnAuthenticationCompleted(object Sender, wclBluetoothRadio Radio, long Address, int Error)
        {
            TraceEvent(Address, "Authentication completed", "Error", Error.ToString("X8"));
        }

        private void btPair_Click(object sender, RoutedEventArgs e)
        {
            if (lvDevices.SelectedItems.Count == 0)
                MessageBox.Show("Select device", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                DeviceItem Item = (DeviceItem)lvDevices.SelectedItems[0];
                Int32 Res = Item.Radio.RemotePair(Convert.ToInt64(Item.Address, 16),
                    wclBluetoothPairingMethod.pmLe);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButton.OK, MessageBoxImage.Error);
            }
        }

        private void cbFastSubscribe_Click(object sender, RoutedEventArgs e)
        {
            btWriteCcdSubsc.IsEnabled = (!cbFastSubscribe.IsChecked.Value);
            btWriteCcdUnsubsc.IsEnabled = (!cbFastSubscribe.IsChecked.Value);
        }

        private void btGetParams_Click(object sender, RoutedEventArgs e)
        {
            TraceEvent(Client.Address, "Connection params", "", "");
            GetConnectionParams();
        }

        void Client_OnConnectionParamsChanged(object sender, EventArgs e)
        {
            TraceEvent(Client.Address, "Connection params changed", "", "");
            GetConnectionParams();
        }

        private void btSetParams_Click(object sender, RoutedEventArgs e)
        {
            Int32 Res;
            if (cbParams.SelectedIndex == 3)
            {
                wclBluetoothLeConnectionParametersValue ParamsValue = new wclBluetoothLeConnectionParametersValue();
                ParamsValue.MinInterval = 84;
                ParamsValue.MaxInterval = 84;
                ParamsValue.Latency = 0;
                ParamsValue.LinkTimeout = 800;
                Res = Client.SetConnectionParams(ParamsValue);
            }
            else
            {
                wclBluetoothLeConnectionParametersType Params;
                switch (cbParams.SelectedIndex)
                {
                    case 0:
                        Params = wclBluetoothLeConnectionParametersType.ppBalanced;
                        break;

                    case 1:
                        Params = wclBluetoothLeConnectionParametersType.ppPowerOptimized;
                        break;
                    case 2:
                        Params = wclBluetoothLeConnectionParametersType.ppThroughputOptimized;
                        break;
                    default:
                        throw new wclEInvalidArgument("Invalid connection parameters.");
                }
                Res = Client.SetConnectionParams(Params);
            }
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Error: 0x" + Res.ToString("X8"));
        }

        void Client_OnMaxPduSizeChanged(object sender, EventArgs e)
        {
            TraceEvent(Client.Address, "Max PDU size changed", "", "");
            GetMaxPduSize();
        }

        private void btGatMaxPduSize_Click(object sender, RoutedEventArgs e)
        {
            TraceEvent(Client.Address, "Max PDU", "", "");
            GetMaxPduSize();
        }

        private void GetMaxPduSize()
        {
            UInt16 Size;
            Int32 Res = Client.GetMaxPduSize(out Size);
            if (Res != wclErrors.WCL_E_SUCCESS)
                TraceEvent(Client.Address, "", "Error", Res.ToString("X8"));
            else
                TraceEvent(Client.Address, "", "Size", Size.ToString());
        }

        private void GetConnectionParams()
        {
            wclBluetoothLeConnectionParameters Params;
            Int32 Res = Client.GetConnectionParams(out Params);
            if (Res != wclErrors.WCL_E_SUCCESS)
                TraceEvent(0, "", "Error", Res.ToString("X8"));
            else
            {
                TraceEvent(0, "", "Connection interval", Params.Interval.ToString());
                TraceEvent(0, "", "Connection latency", Params.Latency.ToString());
                TraceEvent(0, "", "Link timeout", Params.LinkTimeout.ToString());
            }
        }

        void Manager_OnProtectionLevelRequest(Object Sender, wclBluetoothRadio Radio,
            Int64 Address, out wclBluetoothLeProtectionLevel Protection)
        {
            switch (cbProtection.SelectedIndex)
            {
                case 0:
                    Protection = wclBluetoothLeProtectionLevel.pplNone;
                    break;
                case 1:
                    Protection = wclBluetoothLeProtectionLevel.pplDefault;
                    break;
                case 2:
                    Protection = wclBluetoothLeProtectionLevel.pplEncryption;
                    break;
                case 3:
                    Protection = wclBluetoothLeProtectionLevel.pplEncryptionAndAuthentication;
                    break;
                default:
                    Protection = wclBluetoothLeProtectionLevel.pplDefault;
                    break;
            }
        }

        private void btPhy_Click(object sender, RoutedEventArgs e)
        {
            TraceEvent(Client.Address, "Connection PHY", "", "");
            GetConnectionPhy();
        }

        void Client_OnConnectionPhyChanged(object sender, EventArgs e)
        {
            TraceEvent(Client.Address, "Connection PHY changed", "", "");
            GetConnectionPhy();
        }

        private void GetConnectionPhy()
        {
            wclBluetoothLeConnectionPhy Phy;
            Int32 Res = Client.GetConnectionPhyInfo(out Phy);
            if (Res != wclErrors.WCL_E_SUCCESS)
                TraceEvent(0, "", "Error", Res.ToString("X8"));
            else
            {
                TraceEvent(0, "Transmit PHY", "", "");
                TraceEvent(0, "", "Coded", Phy.Transmit.IsCoded.ToString());
                TraceEvent(0, "", "1M", Phy.Transmit.IsUncoded1MPhy.ToString());
                TraceEvent(0, "", "2M", Phy.Transmit.IsUncoded2MPhy.ToString());

                TraceEvent(0, "Receive PHY", "", "");
                TraceEvent(0, "", "Coded", Phy.Receive.IsCoded.ToString());
                TraceEvent(0, "", "1M", Phy.Receive.IsUncoded1MPhy.ToString());
                TraceEvent(0, "", "2M", Phy.Receive.IsUncoded2MPhy.ToString());
            }
        }

        private void btUnpair_Click(object sender, RoutedEventArgs e)
        {
            if (lvDevices.SelectedItems.Count == 0)
                MessageBox.Show("Select device", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                DeviceItem Item = (DeviceItem)lvDevices.SelectedItems[0];
                Int32 Res = Item.Radio.RemoteUnpair(Convert.ToInt64(Item.Address, 16),
                    wclBluetoothPairingMethod.pmLe);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButton.OK, MessageBoxImage.Error);
            }
        }

        private void btEnumPaired_Click(object sender, RoutedEventArgs e)
        {
            wclBluetoothRadio Radio = GetRadio();
            if (Radio != null)
            {
                lvDevices.Items.Clear();

                Int64[] Devices;
                Int32 Res = Radio.EnumPairedDevices(wclBluetoothDiscoverKind.dkBle, out Devices);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Enum paired failed: 0x" + Res.ToString("X8"));
                else
                {
                    if (Devices != null && Devices.Length > 0)
                    {
                        foreach (Int64 Device in Devices)
                        {
                            DeviceItem Item = new DeviceItem();
                            Item.Address = Device.ToString("X12");

                            String Name;
                            Res = Radio.GetRemoteName(Device, out Name);
                            if (Res == wclErrors.WCL_E_SUCCESS)
                                Item.Name = Name;
                            else
                                Item.Name = "Error: 0x" + Res.ToString("X8");
                            Item.Radio = Radio; // To use it later.

                            wclBluetoothDeviceType DevType = wclBluetoothDeviceType.dtMixed;
                            Res = Radio.GetRemoteDeviceType(Device, out DevType);
                            if (Res != wclErrors.WCL_E_SUCCESS)
                                Item.Type = "Error: 0x" + Res.ToString("X8");
                            else
                            {
                                switch (DevType)
                                {
                                    case wclBluetoothDeviceType.dtClassic:
                                        Item.Type = "Classic";
                                        break;
                                    case wclBluetoothDeviceType.dtBle:
                                        Item.Type = "BLE";
                                        break;
                                    case wclBluetoothDeviceType.dtMixed:
                                        Item.Type = "Mixed";
                                        break;
                                    default:
                                        Item.Type = "Unknown";
                                        break;
                                }
                            }

                            Boolean Paired;
                            Res = Radio.GetRemotePaired(Device, out Paired);
                            if (Res != wclErrors.WCL_E_SUCCESS)
                                Item.Paired = "Error: 0x" + Res.ToString("X8");
                            else
                            {
                                if (Paired)
                                    Item.Paired = "True";
                                else
                                    Item.Paired = "False";
                            }

                            lvDevices.Items.Add(Item);
                        }
                    }
                }
            }
        }
    }
}
