﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using System.IO;
using Microsoft.Win32;

using wclCommon;
using wclCommunication;
using wclBluetooth;

namespace OppClient
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        private wclBluetoothManager wclBluetoothManager;
        private wclRfCommClient wclRfCommClient;
        private wclObexOppClient FOpp;

        public MainWindow()
        {
            InitializeComponent();
        }

        private wclBluetoothRadio GetRadio()
        {
            wclBluetoothRadio Radio;
            Int32 Res = wclBluetoothManager.GetClassicRadio(out Radio);
            if (Res != wclErrors.WCL_E_SUCCESS)
            {
                MessageBox.Show("Get working radio failed: 0x" + Res.ToString("X8"));
                return null;
            }
            return Radio;
        }

        private void Window_Loaded(object sender, RoutedEventArgs e)
        {
            wclBluetoothManager = new wclBluetoothManager();
            wclBluetoothManager.AfterOpen += new EventHandler(wclBluetoothManager_AfterOpen);
            wclBluetoothManager.OnAuthenticationCompleted += new wclBluetoothDeviceResultEvent(wclBluetoothManager_OnAuthenticationCompleted);
            wclBluetoothManager.OnDiscoveringCompleted += new wclBluetoothResultEvent(wclBluetoothManager_OnDiscoveringCompleted);
            wclBluetoothManager.OnDiscoveringStarted += new wclBluetoothEvent(wclBluetoothManager_OnDiscoveringStarted);
            wclBluetoothManager.OnNumericComparison += new wclBluetoothNumericComparisonEvent(wclBluetoothManager_OnNumericComparison);
            wclBluetoothManager.OnPasskeyNotification += new wclBluetoothPasskeyNotificationEvent(wclBluetoothManager_OnPasskeyNotification);
            wclBluetoothManager.OnPasskeyRequest += new wclBluetoothPasskeyRequestEvent(wclBluetoothManager_OnPasskeyRequest);
            wclBluetoothManager.OnPinRequest += new wclBluetoothPinRequestEvent(wclBluetoothManager_OnPinRequest);
            wclBluetoothManager.OnDeviceFound += new wclBluetoothDeviceEvent(wclBluetoothManager_OnDeviceFound);

            wclRfCommClient = new wclRfCommClient();
            wclRfCommClient.OnDisconnect += new wclClientConnectionDisconnectEvent(wclRfCommClient_OnDisconnect);
            wclRfCommClient.OnConnect += new wclClientConnectionConnectEvent(wclRfCommClient_OnConnect);
            wclRfCommClient.OnDestroyProcessor += new wclClientConnectionProcessorEvent(wclRfCommClient_OnDestroyProcessor);
            wclRfCommClient.OnCreateProcessor += new wclClientConnectionProcessorEvent(wclRfCommClient_OnCreateProcessor);

            // We use ObjectPush Profile so must set the profile's UUID.
            wclRfCommClient.Service = wclUUIDs.OBEXObjectPushServiceClass_UUID;

            FOpp = null;

            // Try to open Bluetooth Manager.
            Int32 Res = wclBluetoothManager.Open();
            if (Res != wclErrors.WCL_E_SUCCESS)
                lbLog.Items.Add("Bluetooth Manager Open failed: 0x" + Res.ToString("X8"));
        }

        void wclBluetoothManager_AfterOpen(object sender, EventArgs e)
        {
            lbLog.Items.Add("Bluetooth Manager has been opened");
        }

        void wclBluetoothManager_OnAuthenticationCompleted(object Sender, wclBluetoothRadio Radio, long Address, int Error)
        {
            lbLog.Items.Add("Authentication completed with result: 0x" + Error.ToString("X8"));
        }

        void wclBluetoothManager_OnDiscoveringCompleted(object Sender, wclBluetoothRadio Radio, int Error)
        {
            lbLog.Items.Add("Discovering completed with result: 0x" + Error.ToString("X8"));

            for (Int32 i = 0; i < lvDevices.Items.Count; i++)
            {
                DeviceItem Item = (DeviceItem)lvDevices.Items[i];
                Int64 Addr = Convert.ToInt64(Item.Address, 16);
                String Name;
                Int32 Res = Radio.GetRemoteName(Addr, out Name);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    Name = "Error: 0x" + Res.ToString("X8");
                Item.Name = Name;
                lvDevices.Items.Refresh();
            }
        }

        void wclBluetoothManager_OnDiscoveringStarted(object Sender, wclBluetoothRadio Radio)
        {
            lbLog.Items.Add("Discovering started");

            lvDevices.Items.Clear();
        }

        void wclBluetoothManager_OnNumericComparison(object Sender, wclBluetoothRadio Radio, long Address, uint Number, out bool Confirm)
        {
            lbLog.Items.Add("Numeric comparison. Accept");
            Confirm = true;
        }

        void wclBluetoothManager_OnPasskeyNotification(object Sender, wclBluetoothRadio Radio, long Address, uint Passkey)
        {
            lbLog.Items.Add("Passkey notification. Passkey: " + Passkey.ToString());
        }

        void wclBluetoothManager_OnPasskeyRequest(object Sender, wclBluetoothRadio Radio, long Address, out uint Passkey)
        {
            lbLog.Items.Add("Passkey request. Passkey: 12345678");
            Passkey = 12345678;
        }

        void wclBluetoothManager_OnPinRequest(object Sender, wclBluetoothRadio Radio, long Address, out string Pin)
        {
            lbLog.Items.Add("PIN code request. PIN: 0000");
            Pin = "0000";
        }

        internal class DeviceItem
        {
            public String Address { get; set; }
            public String Name { get; set; }
        };

        void wclBluetoothManager_OnDeviceFound(object Sender, wclBluetoothRadio Radio, long address)
        {
            lvDevices.Items.Add(new DeviceItem {
                Address = address.ToString("X12"),
                Name = ""
            });
        }

        void wclRfCommClient_OnDisconnect(object Sender, int Reason)
        {
            lbLog.Items.Add("Disconnected with reason: 0x" + Reason.ToString("X8"));
        }

        void wclRfCommClient_OnConnect(object Sender, int Error)
        {
            lbLog.Items.Add("Connect. Operation result: 0x" + Error.ToString("X8"));
        }

        void wclRfCommClient_OnDestroyProcessor(object Sender, wclClientDataConnection Connection)
        {
            // Do we have any data processor created?
            if (FOpp != null)
            {
                // Make sure it is our connection is goind to destroy.
                if (Connection.Processor == FOpp)
                {
                    // Ok, destroy the data processor here.
                    FOpp.Dispose();
                    FOpp = null;
                }
            }
        }

        void wclRfCommClient_OnCreateProcessor(object Sender, wclClientDataConnection Connection)
        {
            // here we must create the data processor for the connection. In this demo
            // we use OPPClient.
            FOpp = new wclObexOppClient(Connection);
            FOpp.OnConnect += new wclObexOperationResultEvent(FOpp_OnConnect);
            FOpp.OnDisconnect += new wclObexOperationResultEvent(FOpp_OnDisconnect);
            FOpp.OnGetComplete += new wclObexObjectOperationResultEvent(FOpp_OnGetComplete);
            FOpp.OnPutComplete += new wclObexObjectOperationResultEvent(FOpp_OnPutComplete);
            FOpp.OnProgress += new wclObexOperationProgressEvent(FOpp_OnProgress);
        }

        void FOpp_OnGetComplete(object Sender, int Error, string Description, Stream Stream)
        {
            lbLog.Items.Add("File receiving completed with result: 0x" +
                Error.ToString("X8") + ". Description: " + Description);

            ProgressBar.Value = 0;
            ProgressBar.Maximum = 0;

            // We must destroy the stream object here.
            Stream.Close();
            Stream.Dispose();
        }

        void FOpp_OnProgress(object Sender, uint Size, uint Position)
        {
            ProgressBar.Maximum = (int)Size;
            ProgressBar.Value = (int)Position;
        }

        void FOpp_OnPutComplete(object Sender, int Error, string Description, System.IO.Stream Stream)
        {
            lbLog.Items.Add("File sending completed with result: 0x" + Error.ToString("X8") +
                ". Description: " + Description);

            ProgressBar.Value = 0;
            ProgressBar.Maximum = 0;

            // We must destroy the stream object here.
            Stream.Close();
            Stream.Dispose();
        }

        void FOpp_OnDisconnect(object Sender, int Error, string Description)
        {
            lbLog.Items.Add("OBEX session closed with result: 0x" + Error.ToString("X8") +
                ". Description: " + Description);
        }

        void FOpp_OnConnect(object Sender, int Error, string Description)
        {
            lbLog.Items.Add("OBEX session opened with result: 0x" + Error.ToString("X8") +
                ". Description: " + Description);
        }

        private void btClear_Click(object sender, RoutedEventArgs e)
        {
            lbLog.Items.Clear();
        }

        private void Window_Closed(object sender, EventArgs e)
        {
            wclRfCommClient.Disconnect();
            wclBluetoothManager.Close();
        }

        private void btDiscover_Click(object sender, RoutedEventArgs e)
        {
            wclBluetoothRadio Radio = GetRadio();
            if (Radio != null)
            {
                // Discover classic devices only.
                Int32 Res = Radio.Discover(10, wclBluetoothDiscoverKind.dkClassic);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Failed to start discovering: 0x" + Res.ToString("X8"));
            }
        }

        private void btDisconnect_Click(object sender, RoutedEventArgs e)
        {
            Int32 Res = wclRfCommClient.Disconnect();
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Disconnect failed: 0x" + Res.ToString("X8"));
        }

        private void btCloseSession_Click(object sender, RoutedEventArgs e)
        {
            if (FOpp == null)
                MessageBox.Show("Not connected");
            else
            {
                Int32 Res = FOpp.Disconnect("");
                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Close session failed: 0x" + Res.ToString("X8"));
            }
        }

        private void btOpenSession_Click(object sender, RoutedEventArgs e)
        {
            if (FOpp == null)
                MessageBox.Show("Not connected");
            else
            {
                Int32 Res = FOpp.Connect();
                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Open session failed: 0x" + Res.ToString("X8"));
            }
        }

        private void btConnect_Click(object sender, RoutedEventArgs e)
        {
            if (wclRfCommClient.State != wclClientState.csDisconnected)
                MessageBox.Show("Client is connected");
            else
            {
                if (lvDevices.SelectedItems.Count == 0)
                    MessageBox.Show("Select device");
                else
                {
                    wclBluetoothRadio Radio = GetRadio();
                    if (Radio != null)
                    {
                        Int64 Addr = Convert.ToInt64(((DeviceItem)lvDevices.SelectedItems[0]).Address, 16);
                        wclRfCommClient.Address = Addr;
                        Int32 Res = wclRfCommClient.Connect(Radio);
                        if (Res != wclErrors.WCL_E_SUCCESS)
                            MessageBox.Show("Failed to connect: 0x" + Res.ToString("X8"));
                    }
                }
            }
        }

        private void btAbort_Click(object sender, RoutedEventArgs e)
        {
            if (FOpp == null)
                MessageBox.Show("Not connected");
            else
            {
                Int32 Res = FOpp.Abort("");
                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Abort failed: 0x" + Res.ToString("X8"));
            }
        }

        private void btSelectFile_Click(object sender, RoutedEventArgs e)
        {
            OpenFileDialog openFileDialog = new OpenFileDialog();
            if (openFileDialog.ShowDialog() == true)
                edFileName.Text = openFileDialog.FileName;
        }

        private void btGetvCard_Click(object sender, RoutedEventArgs e)
        {
            if (FOpp == null)
                MessageBox.Show("Not connected");
            else
            {
                FileStream Stream = new FileStream("vcard.vcf", FileMode.CreateNew);
                // Request default vCard object.
                Int32 Res = FOpp.Get("text/x-vCard", Stream);
                if (Res != wclErrors.WCL_E_SUCCESS)
                {
                    MessageBox.Show("GetvCard failed: 0x" + Res.ToString("X8"));
                    Stream.Close();
                    Stream.Dispose();
                }
            }
        }

        private void btGetCaps_Click(object sender, RoutedEventArgs e)
        {
            if (FOpp == null)
                MessageBox.Show("Not connected");
            else
            {
                FileStream Stream = new FileStream("caps.xml", FileMode.CreateNew);
                // Request Device Capability object.
                Int32 Res = FOpp.Get("x-obex/capability", Stream);
                if (Res != wclErrors.WCL_E_SUCCESS)
                {
                    MessageBox.Show("GetCaps failed: 0x" + Res.ToString("X8"));
                    Stream.Close();
                    Stream.Dispose();
                }
            }
        }

        private void btSend_Click(object sender, RoutedEventArgs e)
        {
            if (FOpp == null)
                MessageBox.Show("Not connected");
            else
            {
                if (edFileName.Text == "")
                    MessageBox.Show("Select file");
                else
                {
                    FileStream Stream = new FileStream(edFileName.Text, FileMode.Open);
                    Int32 Res = FOpp.Put(System.IO.Path.GetFileName(edFileName.Text), "", Stream);
                    if (Res != wclErrors.WCL_E_SUCCESS)
                    {
                        MessageBox.Show("Send file failed: 0x" + Res.ToString("X8"));
                        Stream.Close();
                        Stream.Dispose();
                    }
                }
            }
        }
    }
}
