﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;

using wclCommon;
using wclCommunication;
using wclBluetooth;

namespace WiiRemote
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class fmMain : Window
    {
        private wclBluetoothManager FManager;
        private wclWiiRemoteClient FClient;

        private wclBluetoothRadio GetRadio()
        {
            wclBluetoothRadio Radio;
            Int32 Res = FManager.GetClassicRadio(out Radio);
            if (Res != wclErrors.WCL_E_SUCCESS)
            {
                MessageBox.Show("Get working radio failed: 0x" + Res.ToString("X8"));
                return null;
            }
            return Radio;
        }

        public fmMain()
        {
            InitializeComponent();
        }

        private void Window_Loaded(object sender, RoutedEventArgs e)
        {
            FManager = new wclBluetoothManager();
            FManager.OnDiscoveringStarted += new wclBluetoothEvent(FManager_OnDiscoveringStarted);
            FManager.OnDeviceFound += new wclBluetoothDeviceEvent(FManager_OnDeviceFound);
            FManager.OnDiscoveringCompleted += new wclBluetoothResultEvent(FManager_OnDiscoveringCompleted);

            FClient = new wclWiiRemoteClient();
            FClient.OnConnect += new wclClientConnectionConnectEvent(FClient_OnConnect);
            FClient.OnDisconnect += new wclClientConnectionDisconnectEvent(FClient_OnDisconnect);
            FClient.OnStatusChanged += new wclWiiRemoteStatusChangedEvent(FClient_OnStatusChanged);
            FClient.OnButtonsChanged += new wclWiiRemoteButtonsChangedEvent(FClient_OnButtonsChanged);
            FClient.OnAccelChanged += new wclWiiRemoteAccelChangedEvent(FClient_OnAccelChanged);
            FClient.OnIrChanged += new wclWiiRemoteIrChangedEvent(FClient_OnIrChanged);
            FClient.OnExtensionAttached += new EventHandler(FClient_OnExtensionAttached);
            FClient.OnExtensionDetached += new EventHandler(FClient_OnExtensionDetached);
            FClient.OnNunchukChanged += new wclWiiRemoteNunchukChangedEvent(FClient_OnNunchukChanged);
            FClient.OnBalanceBoardChanged += new wclWiiRemoteBalanceBoardChangedEvent(FClient_OnBalanceBoardChanged);

            FManager.Open();
            ClearAccel();
            ClearIr();
            ClearExtensions();
        }

        void  FClient_OnBalanceBoardChanged(object Sender, wclWiiRemoteBalanceBoard Board)
        {
            laBbTopRight.Content = Board.SensorsKg.TopRight.ToString("N2");
            laBbTopLeft.Content = Board.SensorsKg.TopLeft.ToString("N2");
            laBbBottomRight.Content = Board.SensorsKg.BottomRight.ToString("N2");
            laBbBottomLeft.Content = Board.SensorsKg.BottomLeft.ToString("N2");
        }
        
        void  FClient_OnNunchukChanged(object Sender, wclWiiRemoteNunchuk Nunchuk)
        {
            laNunchukX.Content = Nunchuk.Accel.Values.X.ToString();
            laNunchukY.Content = Nunchuk.Accel.Values.Y.ToString();
            laNunchukZ.Content = Nunchuk.Accel.Values.Z.ToString();
            
            laNunchukJoyX.Content = Nunchuk.Joystick.X.ToString();
            laNunchukJoyY.Content = Nunchuk.Joystick.Y.ToString();
            
            cbNunchukC.IsChecked = Nunchuk.C;
            cbNunchukZ.IsChecked = Nunchuk.Z;
        }

        private void ClearNunchuk()
        {
            laNunchukState.Content = "Detached";
            
            laNunchukX.Content = "0";
            laNunchukY.Content = "0";
            laNunchukZ.Content = "0";
            
            laNunchukJoyX.Content = "0";
            laNunchukJoyY.Content = "0";
            
            cbNunchukC.IsChecked = false;
            cbNunchukZ.IsChecked = false;
        }

        private void ClearExtensions()
        {
            ClearNunchuk();
            ClearBalanceBoard();
        }

        private void ClearBalanceBoard()
        {
            laBalanceBoardState.Content = "Detached";
            
            laBbTopRight.Content = "0";
            laBbTopLeft.Content = "0";
            laBbBottomRight.Content = "0";
            laBbBottomLeft.Content = "0";
        }

        void  FClient_OnExtensionDetached(object sender, EventArgs e)
        {
            Trace("Extension detached");
            ClearExtensions();
        }
        
        void  FClient_OnExtensionAttached(object sender, EventArgs e)
        {
            String Ext;
            switch (FClient.Extension)
            {
                case wclWiiRemoteExtension.wiiNoExtension:
                    Ext = "wiiNoExtension";
                    break;
                case wclWiiRemoteExtension.wiiNunchuk:
                    Ext = "wiiNunchuk";
                    laNunchukState.Content = "Attached";
                    break;
                case wclWiiRemoteExtension.wiiClassicController:
                    Ext = "wiiClassicController";
                    break;
                case wclWiiRemoteExtension.wiiGuitar:
                    Ext = "wiiGuitar";
                    break;
                case wclWiiRemoteExtension.wiiDrums:
                    Ext = "wiiDrums";
                    break;
                case wclWiiRemoteExtension.wiiBalanceBoard:
                    Ext = "wiiBalanceBoard";
                    laBalanceBoardState.Content = "Attached";
                    break;
                default:
                    Ext = "Unknonw";
                    break;
            }
            
            Trace("Extension attached: " + Ext);
        }

        void FClient_OnIrChanged(object Sender, wclWiiRemoteIrSensors Ir)
        {
            switch (Ir.Mode)
            {
                case wclWiiRemoteIrMode.wiiIrOff:
                    ClearIr();
                    break;
                case wclWiiRemoteIrMode.wiiIrBasicMode:
                    laIrMode.Content = "Basic";
                    break;
                case wclWiiRemoteIrMode.wiiIrExtendedMode:
                    laIrMode.Content = "Extended";
                    break;
                case wclWiiRemoteIrMode.wiiIrFullMode:
                    laIrMode.Content = "Full";
                    break;
                default:
                    laIrMode.Content = "Unknown";
                    break;
            }
            
            if (Ir.Mode != wclWiiRemoteIrMode.wiiIrOff)
            {
                switch (Ir.Sensitivity)
                {
                    case wclWiiRemoteIrSensitivity. wiiIrLevelOff:
                        laIrSensitivity.Content = "Off";
                        break;
                    case wclWiiRemoteIrSensitivity.wiiIrLevel1:
                        laIrSensitivity.Content = "Level 1";
                        break;
                    case wclWiiRemoteIrSensitivity.wiiIrLevel2:
                        laIrSensitivity.Content = "Level 2";
                        break;
                    case wclWiiRemoteIrSensitivity.wiiIrLevel3:
                        laIrSensitivity.Content = "Level 3";
                        break;
                    case wclWiiRemoteIrSensitivity.wiiIrLevel4:
                        laIrSensitivity.Content = "Level 4";
                        break;
                    case wclWiiRemoteIrSensitivity.wiiIrLevel5:
                        laIrSensitivity.Content = "Level 5";
                        break;
                    case wclWiiRemoteIrSensitivity.wiiIrLevelMax:
                        laIrSensitivity.Content = "Maximum";
                        break;
                    default:
                        laIrSensitivity.Content = "Unknown";
                        break;
                }
                
                if (Ir.Sensors[0].Found)
                {
                    laSensor0Position.Content = Ir.Sensors[0].Position.X.ToString() + "; " +
                        Ir.Sensors[0].Position.Y.ToString();
                    laSensor0Size.Content = Ir.Sensors[0].Size.ToString();
                    laSensor0Found.Content = "True";
                }
                else
                {
                    laSensor0Position.Content = "0; 0";
                    laSensor0Size.Content = "0";
                    laSensor0Found.Content = "False";
                }
                
                if (Ir.Sensors[1].Found)
                {
                    laSensor1Position.Content = Ir.Sensors[1].Position.X.ToString() + "; " +
                        Ir.Sensors[1].Position.Y.ToString();
                    laSensor1Size.Content = Ir.Sensors[1].Size.ToString();
                    laSensor1Found.Content = "True";
                }
                else
                {
                    laSensor1Position.Content = "0; 0";
                    laSensor1Size.Content = "0";
                    laSensor1Found.Content = "False";
                }
                
                if (Ir.Sensors[2].Found)
                {
                    laSensor2Position.Content = Ir.Sensors[2].Position.X.ToString() + "; " +
                        Ir.Sensors[2].Position.Y.ToString();
                    laSensor2Size.Content = Ir.Sensors[2].Size.ToString();
                    laSensor2Found.Content = "True";
                }
                else
                {
                    laSensor2Position.Content = "0; 0";
                    laSensor2Size.Content = "0";
                    laSensor2Found.Content = "False";
                }
                
                if (Ir.Sensors[3].Found)
                {
                    laSensor3Position.Content = Ir.Sensors[3].Position.X.ToString() + "; " +
                        Ir.Sensors[3].Position.Y.ToString();
                    laSensor3Size.Content = Ir.Sensors[3].Size.ToString();
                    laSensor3Found.Content = "True";
                }
                else
                {
                    laSensor3Position.Content = "0; 0";
                    laSensor3Size.Content = "0";
                    laSensor3Found.Content = "False";
                }
            }
        }

        private void ClearAccel()
        {
            laX.Content = "0";
            laY.Content = "0";
            laZ.Content = "0";
        }

        private void ClearIr()
        {
            laIrMode.Content = "Off";
            laIrSensitivity.Content = "Off";
            
            laSensor0Position.Content = "0; 0";
            laSensor0Size.Content = "0";
            laSensor0Found.Content = "False";
            
            laSensor1Position.Content = "0; 0";
            laSensor1Size.Content = "0";
            laSensor1Found.Content = "False";
            
            laSensor2Position.Content = "0; 0";
            laSensor2Size.Content = "0";
            laSensor2Found.Content = "False";
            
            laSensor3Position.Content = "0; 0";
            laSensor3Size.Content = "0";
            laSensor3Found.Content = "False";
        }

        void FClient_OnAccelChanged(object Sender, wclWiiRemoteAccel Accel)
        {
            laX.Content = Accel.Values.X.ToString();
            laY.Content = Accel.Values.Y.ToString();
            laZ.Content = Accel.Values.Z.ToString();
        }

        void FClient_OnButtonsChanged(object Sender, wclWiiRemoteButtons Buttons)
        {
            cbA.IsChecked = Buttons.A;
            cbB.IsChecked = Buttons.B;
            cbPlus.IsChecked = Buttons.Plus;
            cbHome.IsChecked = Buttons.Home;
            cbMinus.IsChecked = Buttons.Minus;
            cbOne.IsChecked = Buttons.One;
            cbTwo.IsChecked = Buttons.Two;
            cbUp.IsChecked = Buttons.Up;
            cbDown.IsChecked = Buttons.Down;
            cbLeft.IsChecked = Buttons.Left;
            cbRight.IsChecked = Buttons.Right;
        }

        void FClient_OnStatusChanged(object Sender, double Batt, wclWiiRemoteLeds Leds)
        {
            Trace("Status changed");

            laBattLevel.Content = Batt.ToString("N2");
            
            cbLed1.IsChecked = Leds.Led1;
            cbLed2.IsChecked = Leds.Led2;
            cbLed3.IsChecked = Leds.Led3;
            cbLed4.IsChecked = Leds.Led4;
        }

        void FClient_OnDisconnect(object Sender, int Reason)
        {
            Trace("Disconnected. Reason", Reason);
            
            laBattLevel.Content = "0.00";
            cbLed1.IsChecked = false;
            cbLed2.IsChecked = false;
            cbLed3.IsChecked = false;
            cbLed4.IsChecked = false;
            
            cbA.IsChecked = false;
            cbB.IsChecked = false;
            cbPlus.IsChecked = false;
            cbHome.IsChecked = false;
            cbMinus.IsChecked = false;
            cbOne.IsChecked = false;
            cbTwo.IsChecked = false;
            cbUp.IsChecked = false;
            cbDown.IsChecked = false;
            cbLeft.IsChecked = false;
            cbRight.IsChecked = false;
            
            ClearAccel();
            ClearIr();
            ClearExtensions();
        }

        void FClient_OnConnect(object Sender, int Error)
        {
            if (Error != wclErrors.WCL_E_SUCCESS)
                Trace("Connection failed", Error);
            else
            {
                Trace("Connected");
                SetLeds();
            }
        }

        internal class DeviceItem
        {
            public String Address { get; set; }
            public String Name { get; set; }
        };

        void FManager_OnDiscoveringCompleted(object Sender, wclBluetoothRadio Radio, int Error)
        {
            for (int i = 0; i < lvDevices.Items.Count; i++)
            {
                DeviceItem Item = (DeviceItem)lvDevices.Items[i];
                Int64 Addr = Convert.ToInt64(Item.Address, 16);

                String Name;
                Int32 Res = Radio.GetRemoteName(Addr, out Name);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    Name = "Error: 0x" + Res.ToString("X9");
                Item.Name = Name;
            }

            lvDevices.Items.Refresh();

            Trace("Discovering completed with result", Error);
        }

        void FManager_OnDeviceFound(object Sender, wclBluetoothRadio Radio, long Address)
        {
            lvDevices.Items.Add(new DeviceItem
            {
                Address = Address.ToString("X12"),
                Name = ""
            });
        }

        void FManager_OnDiscoveringStarted(object Sender, wclBluetoothRadio Radio)
        {
            lvDevices.Items.Clear();
            Trace("Discovering started");
        }

        private void btDiscover_Click(object sender, EventArgs e)
        {
            wclBluetoothRadio Radio = GetRadio();
            if (Radio != null)
            {
                Int32 Res = Radio.Discover(10, wclBluetoothDiscoverKind.dkClassic);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Discover failed; 0x" + Res.ToString("X8"));
            }
        }

        private void btConnect_Click(object sender, EventArgs e)
        {
            if (FClient.State != wclClientState.csDisconnected)
                MessageBox.Show("Wii Remote is connected");
            else
            {
                wclBluetoothRadio Radio = GetRadio();
                if (Radio != null)
                {
                    if (lvDevices.SelectedItems.Count == 0)
                        MessageBox.Show("Select device");
                    else
                    {
                        DeviceItem Item = (DeviceItem)lvDevices.SelectedItem;
                        FClient.Address = Convert.ToInt64(Item.Address, 16);
                        Int32 Res = FClient.Connect(Radio);
                        if (Res != wclErrors.WCL_E_SUCCESS)
                            MessageBox.Show("Unable connect: 0x" + Res.ToString("X8"));
                    }
                }
            }
        }

        private void btDisconnect_Click(object sender, EventArgs e)
        {
            Int32 Res = FClient.Disconnect();
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Unable disconnect: 0x" + Res.ToString("X8"));
        }

        private void btClear_Click(object sender, EventArgs e)
        {
            lbLog.Items.Clear();
        }

        private void Trace(String Msg)
        {
            lbLog.Items.Add(Msg);
        }

        private void Trace(String Msg, Int32 Res)
        {
            Trace(Msg + ": 0x" + Res.ToString("X8"));
        }

        private void btGetStatus_Click(object sender, EventArgs e)
        {
            Int32 Res = FClient.GetStatus();
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Get status failed; 0x" + Res.ToString("X8"));
        }

        private void btGetLeds_Click(object sender, EventArgs e)
        {
            if (FClient.State != wclClientState.csConnected)
                MessageBox.Show("Not connected");
            else
            {
                cbLed1.IsChecked = FClient.Leds.Led1;
                cbLed2.IsChecked = FClient.Leds.Led2;
                cbLed3.IsChecked = FClient.Leds.Led3;
                cbLed4.IsChecked = FClient.Leds.Led4;
            }
        }

        private void btSetLeds_Click(object sender, EventArgs e)
        {
            SetLeds();
        }

        private void SetLeds()
        {
            Int32 Res = FClient.SetLeds(cbLed1.IsChecked.Value, cbLed2.IsChecked.Value,
                cbLed3.IsChecked.Value, cbLed4.IsChecked.Value);
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Set LEDs failed: 0x" + Res.ToString("X8"));
        }

        private void btEnableAccel_Click(object sender, EventArgs e)
        {
            Int32 Res = FClient.EnableAccel();
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Enable accelerometer failed: 0x" + Res.ToString("X8"));
        }

        private void btDisableAccel_Click(object sender, EventArgs e)
        {
            Int32 Res = FClient.DisableAccel();
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Disable accelerometer failed: 0x" + Res.ToString("X8"));
            else
                ClearAccel();
        }

        private void btEnableIr_Click(object sender, EventArgs e)
        {
            Int32 Res = FClient.SetIrSensitivity(wclWiiRemoteIrSensitivity.wiiIrLevelMax);
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Enable IR failed: 0x" + Res.ToString("X8"));
        }

        private void btDisableIr_Click(object sender, EventArgs e)
        {
            Int32 Res = FClient.SetIrSensitivity(wclWiiRemoteIrSensitivity.wiiIrLevelOff);
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Disable IR failed: 0x" + Res.ToString("X8"));
            else
                ClearIr();
        }

        private void btRumbleOn_Click(object sender, EventArgs e)
        {
            Int32 Res = FClient.SetRumble(true);
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Set Rumble ON failed; 0x" + Res.ToString("X8"));
        }

        private void btRumbleOff_Click(object sender, EventArgs e)
        {
            Int32 Res = FClient.SetRumble(false);
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Set Rumble OFF failed; 0x" + Res.ToString("X8"));
        }

        private void Window_Closed(object sender, EventArgs e)
        {
            FClient.Disconnect();
            FManager.Close();
        }
    }
}
