﻿using System;
using System.Windows.Forms;

using wclCommon;
using wclAudio;

namespace AudioSwitcher
{
    public partial class fmMain : Form
    {
        private wclAudioSwitcher Switcher;

        public fmMain()
        {
            InitializeComponent();
        }

        private void fmMain_Load(Object sender, EventArgs e)
        {
            Switcher = new wclAudioSwitcher();
            Switcher.OnOpened += new EventHandler(Switcher_OnOpened);
            Switcher.OnClosed += new EventHandler(Switcher_OnClosed);
            Switcher.OnDefaultDeviceChanged += new wclAudioDefaultDeviceChangedEvent(Switcher_OnDefaultDeviceChanged);
            Switcher.OnDeviceAdded += new wclAudioDeviceEvent(Switcher_OnDeviceAdded);
            Switcher.OnDeviceRemoved += new wclAudioDeviceEvent(Switcher_OnDeviceRemoved);
            Switcher.OnStateChanged += new wclAudioDeviceStateChangedEvent(Switcher_OnStateChanged);

            cbActive.Checked = true;
            cbRole.SelectedIndex = 0;
            cbFlow.SelectedIndex = 0;

            Int32 Res = Switcher.Open();
            if (Res != wclErrors.WCL_E_SUCCESS)
                lbLog.Items.Add("Open failed: 0x" + Res.ToString("X8"));
        }

        private void Switcher_OnStateChanged(Object Sender, String Id, wclAudioDeviceState State)
        {
            String Str = "";
            switch (State)
            {
                case wclAudioDeviceState.asActive:
                    Str = "Active";
                    break;
                case wclAudioDeviceState.asDisabled:
                    Str = "Disabled";
                    break;
                case wclAudioDeviceState.asNotPresent:
                    Str = "NotPresent";
                    break;
                case wclAudioDeviceState.asUnplugged:
                    Str = "Unplugged";
                    break;
            }
            lbLog.Items.Add("Device " + Id + " state changed: " + Str);
        }

        private void Switcher_OnDeviceRemoved(Object Sender, String Id)
        {
            lbLog.Items.Add("Device removed: " + Id);
        }

        private void Switcher_OnDeviceAdded(Object Sender, String Id)
        {
            lbLog.Items.Add("Device added: " + Id);
        }

        private void Switcher_OnDefaultDeviceChanged(Object Sender, String Id, wclAudioDeviceDataFlow Flow,
            wclAudioDeviceRole Role)
        {
            String Str = "Default device changed: " + Id + "; Flow: ";
            if (Flow == wclAudioDeviceDataFlow.dfRender)
                Str += "Render";
            else
                Str += "Capture";
            Str += "; Role: ";
            switch (Role)
            {
                case wclAudioDeviceRole.drConsole:
                    Str += "Console";
                    break;
                case wclAudioDeviceRole.drMultimedia:
                    Str += "Multimedia";
                    break;
                case wclAudioDeviceRole.drCommunications:
                    Str += "Communications";
                    break;
            }
            lbLog.Items.Add(Str);
        }

        private void Switcher_OnClosed(Object sender, EventArgs e)
        {
            lvDevices.Clear();
            lbLog.Items.Add("Audio switcher closed");
        }

        private void Switcher_OnOpened(Object sender, EventArgs e)
        {
            lbLog.Items.Add("Audio switcher opened");
            EnumDevices();
        }

        private void fmMain_FormClosed(Object sender, FormClosedEventArgs e)
        {
            Switcher.Close();
        }

        private void btEnum_Click(Object sender, EventArgs e)
        {
            EnumDevices();
        }

        private void btSetDefault_Click(object sender, EventArgs e)
        {
            if (lvDevices.SelectedItems == null)
                MessageBox.Show("Select device");
            else
            {
                wclAudioDeviceRole Role;
                switch (cbRole.SelectedIndex)
                {
                    case 0:
                        Role = wclAudioDeviceRole.drConsole;
                        break;
                    case 1:
                        Role = wclAudioDeviceRole.drMultimedia;
                        break;
                    default:
                        Role = wclAudioDeviceRole.drCommunications;
                        break;
                }

                Int32 Res = Switcher.SetDefault(Role, lvDevices.SelectedItems[0].Text);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Unable to set default device: 0x" + Res.ToString("X8"));
            }
        }

        private void EnumDevices()
        {
            wclAudioDeviceState States = 0;
            if (cbActive.Checked)
                States |= wclAudioDeviceState.asActive;
            if (cbDisabled.Checked)
                States |= wclAudioDeviceState.asDisabled;
            if (cbNotPresent.Checked)
                States |= wclAudioDeviceState.asNotPresent;
            if (cbUnplugged.Checked)
                States |= wclAudioDeviceState.asUnplugged;
            if (States == 0)
                MessageBox.Show("Select device state.");
            else
            {
                lvDevices.Items.Clear();

                wclAudioDevice[] Devices;
                Int32 Res = Switcher.Enum(States, out Devices);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Enum failed: 0x" + Res.ToString("X8"));
                else
                {
                    if (Devices != null && Devices.Length > 0)
                    {
                        foreach (wclAudioDevice Device in Devices)
                        {
                            ListViewItem Item = lvDevices.Items.Add(Device.Id);
                            Item.SubItems.Add("");
                            Item.SubItems.Add("");
                            Item.SubItems.Add("");
                            Item.SubItems.Add("");
                            Item.SubItems.Add("");
                            Item.SubItems.Add("");
                            Item.SubItems.Add("");
                            UpdateDevice(Item, Device);
                        }
                    }
                }
            }
        }

        private void btClear_Click(object sender, EventArgs e)
        {
            lbLog.Items.Clear();
        }

        private void btRefresh_Click(object sender, EventArgs e)
        {
            if (lvDevices.SelectedItems == null)
                MessageBox.Show("Select device");
            else
            {
                wclAudioDevice Device;
                Int32 Res = Switcher.GetDeviceDetails(lvDevices.SelectedItems[0].Text,
                    out Device);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Refresh device error: 0x" + Res.ToString("X8"));
                else
                    UpdateDevice(lvDevices.SelectedItems[0], Device);
            }
        }

        private void UpdateDevice(ListViewItem Item, wclAudioDevice Device)
        {
            String FlowStr;
            if (Device.Flow == wclAudioDeviceDataFlow.dfRender)
                FlowStr = "Render";
            else
                FlowStr = "Capture";

            String RolesStr = "[ ";
            if ((Device.Roles & wclAudioDeviceRole.drConsole) != 0)
                RolesStr += "Console ";
            if ((Device.Roles & wclAudioDeviceRole.drMultimedia) != 0)
                RolesStr += "Multimedia ";
            if ((Device.Roles & wclAudioDeviceRole.drCommunications) != 0)
                RolesStr += "Communications ";
            RolesStr += "]";

            String StateStr = "";
            switch (Device.State)
            {
                case wclAudioDeviceState.asActive:
                    StateStr = "Active";
                    break;
                case wclAudioDeviceState.asDisabled:
                    StateStr = "Disabled";
                    break;
                case wclAudioDeviceState.asNotPresent:
                    StateStr = "NotPresent";
                    break;
                case wclAudioDeviceState.asUnplugged:
                    StateStr = "Unplugged";
                    break;
            }

            String MacStr;
            String ServiceStr;
            if (Device.IsBluetooth)
            {
                MacStr = Device.Mac.ToString("X12");
                ServiceStr = Device.Service.ToString();
            }
            else
            {
                MacStr = "";
                ServiceStr = "";
            }

            Item.SubItems[1].Text = Device.FriendlyName;
            Item.SubItems[2].Text = Device.Description;
            Item.SubItems[3].Text = FlowStr;
            Item.SubItems[4].Text = RolesStr;
            Item.SubItems[5].Text = StateStr;
            Item.SubItems[6].Text = MacStr;
            Item.SubItems[7].Text = ServiceStr;
        }

        private void btConnect_Click(object sender, EventArgs e)
        {
            if (lvDevices.SelectedItems == null)
                MessageBox.Show("Select device");
            else
            {
                Int32 Res;
                if (cbUseMac.Checked)
                {
                    if (lvDevices.SelectedItems[0].SubItems[6].Text == "")
                    {
                        MessageBox.Show("Selected device is not Bluetooth enabled");
                        return;
                    }
                    
                    Res = Switcher.Connect(Convert.ToInt64(lvDevices.SelectedItems[0].SubItems[6].Text, 16));
                }
                else
                    Res = Switcher.Connect(lvDevices.SelectedItems[0].Text);

                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Connect error: 0x" + Res.ToString("X8"));
            }
        }

        private void btDisconnect_Click(object sender, EventArgs e)
        {
            if (lvDevices.SelectedItems == null)
                MessageBox.Show("Select device");
            else
            {
                Int32 Res;
                if (cbUseMac.Checked)
                {
                    if (lvDevices.SelectedItems[0].SubItems[6].Text == "")
                    {
                        MessageBox.Show("Selected device is not Bluetooth enabled");
                        return;
                    }

                    Res = Switcher.Disconnect(Convert.ToInt64(lvDevices.SelectedItems[0].SubItems[6].Text, 16));
                }
                else
                    Res = Switcher.Disconnect(lvDevices.SelectedItems[0].Text);

                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Disconnect error: 0x" + Res.ToString("X8"));
            }
        }

        private void btGetDefault_Click(object sender, EventArgs e)
        {
            wclAudioDeviceRole Role;
            switch (cbRole.SelectedIndex)
            {
                case 0:
                    Role = wclAudioDeviceRole.drConsole;
                    break;
                case 1:
                    Role = wclAudioDeviceRole.drMultimedia;
                    break;
                default:
                    Role = wclAudioDeviceRole.drCommunications;
                    break;
            }
            
            wclAudioDeviceDataFlow Flow;
            if (cbFlow.SelectedIndex == 0)
                Flow = wclAudioDeviceDataFlow.dfCapture;
            else
                Flow = wclAudioDeviceDataFlow.dfRender;
            
            String Id;
            Int32 Res = Switcher.GetDefault(Role, Flow, out Id);
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Unable to get default device: 0x" + Res.ToString("X8"));
            else
                MessageBox.Show("Default device ID: " + Id);
        }
    }
}
