using System;
using System.Windows.Forms;

using wclCommon;
using wclBluetooth;

namespace BluetoothManager
{
    public partial class fmMain : Form
    {
        private wclBluetoothManager Manager;
        private wclPowerEventsMonitor FPowerMonitor;

        public fmMain()
        {
            InitializeComponent();
        }

        private void btClear_Click(object sender, EventArgs e)
        {
            lvEvents.Items.Clear();
        }

        private void TraceEvent(wclBluetoothRadio Radio, String Event,
            String Parameter, String Value)
        {
            String Str;
            if (Radio != null)
                Str = Radio.ApiName;
            else
                Str = "";

            ListViewItem Item = lvEvents.Items.Add(Str);

            Item.SubItems.Add(Event);
            Item.SubItems.Add(Parameter);
            Item.SubItems.Add(Value);
        }

        private void fmMain_Load(object sender, EventArgs e)
        {
            Manager = new wclBluetoothManager();

            Manager.AfterOpen += new EventHandler(Manager_AfterOpen);
            Manager.BeforeClose += new EventHandler(Manager_BeforeClose);
            Manager.OnAuthenticationCompleted += new wclBluetoothDeviceResultEvent(Manager_OnAuthenticationCompleted);
            Manager.OnDeviceFound += new wclBluetoothDeviceEvent(Manager_OnDeviceFound);
            Manager.OnDiscoveringCompleted += new wclBluetoothResultEvent(Manager_OnDiscoveringCompleted);
            Manager.OnDiscoveringStarted += new wclBluetoothEvent(Manager_OnDiscoveringStarted);
            Manager.OnNumericComparison += new wclBluetoothNumericComparisonEvent(Manager_OnNumericComparison);
            Manager.OnPasskeyNotification += new wclBluetoothPasskeyNotificationEvent(Manager_OnPasskeyNotification);
            Manager.OnPasskeyRequest += new wclBluetoothPasskeyRequestEvent(Manager_OnPasskeyRequest);
            Manager.OnPinRequest += new wclBluetoothPinRequestEvent(Manager_OnPinRequest);
            Manager.OnStatusChanged += new wclBluetoothEvent(Manager_OnStatusChanged);
            Manager.OnConfirm += new wclBluetoothConfirmEvent(Manager_OnConfirm);
            Manager.OnIoCapabilityRequest += new wclBluetoothIoCapabilityRequestEvent(Manager_OnIoCapabilityRequest);
            Manager.OnOobDataRequest += new wclBluetoothOobDataRequestEvent(Manager_OnOobDataRequest);
            Manager.OnProtectionLevelRequest += new wclBluetoothProtectionLevelRequestEvent(Manager_OnProtectionLevelRequest);
            Manager.OnClosed += new EventHandler(Manager_OnClosed);

            FPowerMonitor = new wclPowerEventsMonitor();
            FPowerMonitor.OnPowerStateChanged += new wclPowerStateChangedEvent(FPowerMonitor_OnPowerStateChanged);
            FPowerMonitor.Open();
            
            cbDiscoveringMode.SelectedIndex = 0;
            cbMitmProtection.SelectedIndex = 3;
            cbIoCap.SelectedIndex = 1;
            cbBleProtection.SelectedIndex = 0;
            cbPairingMethod.SelectedIndex = 0;
        }

        void Manager_OnClosed(object sender, EventArgs e)
        {
            ClearRadios();

            TraceEvent(null, "Closed", "", "");
        }

        private Boolean WasOpened = false;

        void FPowerMonitor_OnPowerStateChanged(object Sender, wclPowerState State)
        {
            switch (State)
            {
                case wclPowerState.psResumeAutomatic:
                    TraceEvent(null, "Power", "State", "psResumeAutomatic");
                    if (WasOpened)
                    {
                        WasOpened = false;
                        Manager.Open();
                    }
                    break;

                case wclPowerState.psResume:
                    TraceEvent(null, "Power", "State", "psResume");
                    break;

                case wclPowerState.psSuspend:
                    TraceEvent(null, "Power", "State", "psSuspend");
                    if (Manager.Active)
                    {
                        WasOpened = true;
                        Manager.Close();
                    }
                    break;

                case wclPowerState.psUnknown:
                    TraceEvent(null, "Power", "State", "psUnknown");
                    break;
            }
        }

        void Manager_OnProtectionLevelRequest(object Sender, wclBluetoothRadio Radio, long Address,
            out wclBluetoothLeProtectionLevel Protection)
        {
            switch (cbBleProtection.SelectedIndex)
            {
                case 0:
                    Protection = wclBluetoothLeProtectionLevel.pplDefault;
                    break;
                case 1:
                    Protection = wclBluetoothLeProtectionLevel.pplNone;
                    break;
                case 2:
                    Protection = wclBluetoothLeProtectionLevel.pplEncryption;
                    break;
                case 3:
                    Protection = wclBluetoothLeProtectionLevel.pplEncryptionAndAuthentication;
                    break;
                default:
                    Protection = wclBluetoothLeProtectionLevel.pplDefault;
                    break;
            }
        }

        void Manager_OnOobDataRequest(object Sender, wclBluetoothRadio Radio, long Address, out wclBluetoothOobData OobData)
        {
            // This event fires when a remote device requests OOB data.
            OobData = null;
        }

        void Manager_OnIoCapabilityRequest(object Sender, wclBluetoothRadio Radio, long Address,
            out wclBluetoothMitmProtection Mitm, out wclBluetoothIoCapability IoCapability, out bool OobPresent)
        {
            switch (cbMitmProtection.SelectedIndex)
            {
                case 0:
                    Mitm = wclBluetoothMitmProtection.mitmProtectionNotRequired;
                    break;
                case 1:
                    Mitm = wclBluetoothMitmProtection.mitmProtectionRequired;
                    break;
                case 2:
                    Mitm = wclBluetoothMitmProtection.mitmProtectionNotRequiredBonding;
                    break;
                case 3:
                    Mitm = wclBluetoothMitmProtection.mitmProtectionRequiredBonding;
                    break;
                case 4:
                    Mitm = wclBluetoothMitmProtection.mitmProtectionNotRequiredGeneralBonding;
                    break;
                case 5:
                    Mitm = wclBluetoothMitmProtection.mitmProtectionRequiredGeneralBonding;
                    break;
                default:
                    Mitm = wclBluetoothMitmProtection.mitmProtectionNotDefined;
                    break;
            }
            
            switch (cbIoCap.SelectedIndex)
            {
                case 0:
                    IoCapability = wclBluetoothIoCapability.iocapDisplayOnly;
                    break;
                case 1:
                    IoCapability = wclBluetoothIoCapability.iocapDisplayYesNo;
                    break;
                case 2:
                    IoCapability = wclBluetoothIoCapability.iocapKeyboardOnly;
                    break;
                case 3:
                    IoCapability = wclBluetoothIoCapability.iocapNoInputNoOutput;
                    break;
                case 4:
                    IoCapability = wclBluetoothIoCapability.iocapDisplayKeyboard;
                    break;
                default:
                    IoCapability = wclBluetoothIoCapability.iocapNotDefined;
                    break;
            }

            // To accept OOB pairing set this parameter to True.
            OobPresent = false;
        }

        void Manager_OnConfirm(object Sender, wclBluetoothRadio Radio, long Address, out bool Confirm)
        {
            Confirm = true;
            
            TraceEvent(Radio, "Just Works pairing", "Address", Address.ToString("X12"));
        }

        void Manager_OnStatusChanged(object Sender, wclBluetoothRadio Radio)
        {
            String Status;
            if (Radio.Available)
                Status = "Available";
            else
            {
                if (Radio.Plugged)
                    Status = "Not available, Plugged";
                else
                    Status = "Unplugged";
            }
            TraceEvent(Radio, "StatusChanged", "New status", Status);
            
            Boolean Found = false;
            foreach (ListViewItem Item in lvRadios.Items)
            {
                if (((wclBluetoothRadio)Item.Tag) == Radio)
                {
                    Found = true;
                    RefreshRadio(Item);
                    break;
                }
            }

            if (!Found)
            {
                ListViewItem Item = lvRadios.Items.Add(""); // Api
                Item.Tag = Radio;

                Item.SubItems.Add(""); // Available
                Item.SubItems.Add(""); // Address
                Item.SubItems.Add(""); // Connectable
                Item.SubItems.Add(""); // Discoverable
                Item.SubItems.Add(""); // HCI version
                Item.SubItems.Add(""); // LMP version
                Item.SubItems.Add(""); // Manufacturer
                Item.SubItems.Add(""); // Name
                Item.SubItems.Add(""); // COD

                RefreshRadio(Item);
            }
        }

        void Manager_OnPinRequest(object Sender, wclBluetoothRadio Radio, long Address, out string Pin)
        {
            Pin = edPIN.Text;
            
            TraceEvent(Radio, "PinRequest", "Address", Address.ToString("X12"));
            TraceEvent(null, "", "PIN", Pin);
        }

        void Manager_OnPasskeyRequest(object Sender, wclBluetoothRadio Radio, long Address, out uint Passkey)
        {
            Passkey = 123456;
            
            TraceEvent(Radio, "PasskeyRequest", "Address", Address.ToString());
            TraceEvent(null, "", "Passkey", Passkey.ToString());
        }

        void Manager_OnPasskeyNotification(object Sender, wclBluetoothRadio Radio, long Address, uint Passkey)
        {
            TraceEvent(Radio, "PasskeyNotification", "Address", Address.ToString("X12"));
            TraceEvent(null, "", "Passkey", Passkey.ToString());
        }

        void Manager_OnNumericComparison(object Sender, wclBluetoothRadio Radio, long Address, uint Number, out bool Confirm)
        {
            Confirm = true;

            TraceEvent(Radio, "NumericComparison", "Address", Address.ToString("X12"));
            TraceEvent(null, "", "Number", Number.ToString());
            TraceEvent(null, "", "Confirm", Confirm.ToString());
        }

        void Manager_OnDiscoveringStarted(object Sender, wclBluetoothRadio Radio)
        {
            ClearDevices();
            
            TraceEvent(Radio, "DiscoveringStarted", "", "");
        }

        void Manager_OnDiscoveringCompleted(object Sender, wclBluetoothRadio Radio, int Error)
        {
            TraceEvent(Radio, "DiscoveringCompleted", "Error", "0x" + Error.ToString("X8"));
            
            RefreshDevices();
        }

        void Manager_OnDeviceFound(object Sender, wclBluetoothRadio Radio, long Address)
        {
            TraceEvent(Radio, "DeviceFound", "Address", Address.ToString("X12"));
            
            AddDevice(Radio, Address);
        }

        void Manager_OnAuthenticationCompleted(object Sender, wclBluetoothRadio Radio, long Address, int Error)
        {
            TraceEvent(Radio, "AuthenticationCompleted", "Address", Address.ToString("X12"));
            TraceEvent(null, "", "Error", "0x" + Error.ToString("X8"));
        }

        void Manager_BeforeClose(object sender, EventArgs e)
        {
            TraceEvent(null, "BeforeClose", "", "");
        }

        void Manager_AfterOpen(object sender, EventArgs e)
        {
            TraceEvent(null, "AfterOpen", "Handle pairing", Manager.HandlePairing.ToString());
        }

        private void fmMain_FormClosed(object sender, FormClosedEventArgs e)
        {
            Manager.Close();
            Manager = null;

            FPowerMonitor.Close();
            FPowerMonitor = null;
        }

        private void ClearRadios()
        {
            lvRadios.Items.Clear();
            ClearDevices();
        }

        private void RefreshRadio(ListViewItem Item)
        {
            wclBluetoothRadio Radio = (wclBluetoothRadio)Item.Tag;

            Item.Text = Radio.ApiName;
            Item.SubItems[1].Text = Radio.Available.ToString();

            Int64 Address = 0;
            Int32 Res = Radio.GetAddress(out Address);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Item.SubItems[2].Text = "Error: 0x" + Res.ToString("X8");
            else
                Item.SubItems[2].Text = Address.ToString("X12");

            Boolean BoolVal;
            Res = Radio.GetConnectable(out BoolVal);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Item.SubItems[3].Text = "Error: 0x" + Res.ToString("X8");
            else
                Item.SubItems[3].Text = BoolVal.ToString();

            Res = Radio.GetDiscoverable(out BoolVal);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Item.SubItems[4].Text = "Error: 0x" + Res.ToString("X8");
            else
                Item.SubItems[4].Text = BoolVal.ToString();

            Byte ByteVal;
            UInt16 WordVal;
            Res = Radio.GetHciVersion(out ByteVal, out WordVal);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Item.SubItems[5].Text = "Error: 0x" + Res.ToString("X8");
            else
                Item.SubItems[5].Text = ByteVal.ToString() + "." + WordVal.ToString();

            Res = Radio.GetLmpVersion(out ByteVal, out WordVal);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Item.SubItems[6].Text = "Error: 0x" + Res.ToString("X8");
            else
                Item.SubItems[6].Text = ByteVal.ToString() + "." + WordVal.ToString();

            Res = Radio.GetManufacturer(out WordVal);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Item.SubItems[7].Text = "Error: 0x" + Res.ToString("X8");
            else
                Item.SubItems[7].Text = WordVal.ToString();

            String StrVal;
            Res = Radio.GetName(out StrVal);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Item.SubItems[8].Text = "Error: 0x" + Res.ToString("X8");
            else
                Item.SubItems[8].Text = StrVal;

            UInt32 Cod;
            Res = Radio.GetCod(out Cod);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Item.SubItems[9].Text = "Error: 0x" + Res.ToString("X8");
            else
                Item.SubItems[9].Text = "0x" + Cod.ToString("X8");
        }

        private void btRefreshRadio_Click(object sender, EventArgs e)
        {
            if (lvRadios.SelectedItems.Count == 0)
                MessageBox.Show("Select radio", "Warning", MessageBoxButtons.OK,
                    MessageBoxIcon.Warning);
            else
                RefreshRadio(lvRadios.SelectedItems[0]);
        }

        private void btOpen_Click(object sender, EventArgs e)
        {
            Int32 Res = Manager.Open(cbHandlePairing.Checked);
            if (Res != wclErrors.WCL_E_SUCCESS)
                ShowError(Res);
        }

        private void btClose_Click(object sender, EventArgs e)
        {
            Int32 Res = Manager.Close();
            if (Res != wclErrors.WCL_E_SUCCESS)
                ShowError(Res);
        }

        private void ShowError(Int32 Error)
        {
            MessageBox.Show("Error: 0x" + Error.ToString("X8"), "Error", MessageBoxButtons.OK,
                MessageBoxIcon.Error);
        }

        private void btSetConnectable_Click(object sender, EventArgs e)
        {
            if (lvRadios.SelectedItems.Count == 0)
                MessageBox.Show("Select radio", "Warning", MessageBoxButtons.OK,
                    MessageBoxIcon.Warning);
            else
            {
                DialogResult Mr = MessageBox.Show("[YES] - turn connectable ON, [NO] - turn connectable OFF",
                    "Set Connectable", MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question);
                if (Mr != DialogResult.Cancel)
                {
                    ListViewItem Item = lvRadios.SelectedItems[0];
                    wclBluetoothRadio Radio = (wclBluetoothRadio)Item.Tag;

                    Int32 Res = Radio.SetConnectable(Mr == DialogResult.Yes);
                    if (Res != wclErrors.WCL_E_SUCCESS)
                        ShowError(Res);

                    RefreshRadio(Item);
                }
            }
        }

        private void btSetDiscoverable_Click(object sender, EventArgs e)
        {
            if (lvRadios.SelectedItems.Count == 0)
                MessageBox.Show("Select radio", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning);
            else
            {
                DialogResult Mr = MessageBox.Show("[YES] - turn discoverable ON, [NO] - turn discoverable OFF",
                  "Set discoverable", MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question);
                if (Mr != DialogResult.Cancel)
                {
                    ListViewItem Item = lvRadios.SelectedItems[0];
                    wclBluetoothRadio Radio = (wclBluetoothRadio)Item.Tag;

                    Int32 Res = Radio.SetDiscoverable(Mr == DialogResult.Yes);
                    if (Res != wclErrors.WCL_E_SUCCESS)
                        ShowError(Res);

                    RefreshRadio(Item);
                }
            }
        }

        private void ntSetName_Click(object sender, EventArgs e)
        {
            if (lvRadios.SelectedItems.Count == 0)
                MessageBox.Show("Select radio", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning);
            else
            {
                ListViewItem Item = lvRadios.SelectedItems[0];
                wclBluetoothRadio Radio = (wclBluetoothRadio)Item.Tag;

                Int32 Res = Radio.SetName(edName.Text);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    ShowError(Res);

                RefreshRadio(Item);
            }
        }

        private void btEnumPaired_Click(object sender, EventArgs e)
        {
            if (lvRadios.SelectedItems.Count == 0)
                MessageBox.Show("Select radio", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning);
            else
            {
                ClearDevices();

                ListViewItem Item = lvRadios.SelectedItems[0];
                wclBluetoothRadio Radio = (wclBluetoothRadio)Item.Tag;

                Int64[] Devices;
                Int32 Res = Radio.EnumPairedDevices(out Devices);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    ShowError(Res);
                else
                {
                    if (Devices != null)
                    {
                        for (Int32 i = 0; i < Devices.Length; i++)
                            AddDevice(Radio, Devices[i]);

                        RefreshDevices();

                    }
                }
            }
        }

        private void ClearDevices()
        {
            lvDevices.Items.Clear();
            ClearServices();
        }

        private void RefreshDevice(ListViewItem Item)
        {
            wclBluetoothRadio Radio = (wclBluetoothRadio)Item.Tag;
            Int64 Address = Convert.ToInt64(Item.SubItems[1].Text, 16);

            UInt32 Cod;
            Int32 Res = Radio.GetRemoteCod(Address, out Cod);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Item.SubItems[2].Text = "Error: 0x" + Res.ToString("X8");
            else
                Item.SubItems[2].Text = Cod.ToString("X8");

            wclBluetoothDeviceType DevType;
            Res = Radio.GetRemoteDeviceType(Address, out DevType);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Item.SubItems[3].Text = "Error: 0x" + Res.ToString("X8");
            else
            {
                switch (DevType)
                {
                    case wclBluetoothDeviceType.dtClassic:
                        Item.SubItems[3].Text = "Classic";
                        break;
                    case wclBluetoothDeviceType.dtBle:
                        Item.SubItems[3].Text = "BLE";
                        break;
                    case wclBluetoothDeviceType.dtMixed:
                        Item.SubItems[3].Text = "Mixed";
                        break;
                    default:
                        Item.SubItems[3].Text = "Unknown";
                        break;
                }
            }

            String DevName;
            Res = Radio.GetRemoteName(Address, out DevName);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Item.SubItems[4].Text = "Error: 0x" + Res.ToString("X8");
            else
                Item.SubItems[4].Text = DevName;

            Boolean Paired;
            Res = Radio.GetRemotePaired(Address, out Paired);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Item.SubItems[5].Text = "Error: 0x" + Res.ToString("X8");
            else
                Item.SubItems[5].Text = Paired.ToString();

            Boolean Connected;
            Res = Radio.GetRemoteConnectedStatus(Address, out Connected);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Item.SubItems[6].Text = "Error: 0x" + Res.ToString("X8");
            else
                Item.SubItems[6].Text = Connected.ToString();

            wclBluetoothAddressType AddrType;
            Res = Radio.GetRemoteAddressType(Address, out AddrType);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Item.SubItems[7].Text = "Error: 0x" + Res.ToString("X8");
            else
            {
                switch (AddrType)
                {
                    case wclBluetoothAddressType.atClassic:
                        Item.SubItems[7].Text = "Classic";
                        break;
                    case wclBluetoothAddressType.atPublic:
                        Item.SubItems[7].Text = "LE Public";
                        break;
                    case wclBluetoothAddressType.atRandom:
                        Item.SubItems[7].Text = "LE Random";
                        break;
                    case wclBluetoothAddressType.atUnspecified:
                        Item.SubItems[7].Text = "Unspecified";
                        break;
                    default:
                        Item.SubItems[7].Text = "Unknown";
                        break;
                }
            }
        }

        private void btRefreshDevice_Click(object sender, EventArgs e)
        {
            if (lvDevices.SelectedItems.Count == 0)
                MessageBox.Show("Select device", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning);
            else
                RefreshDevice(lvDevices.SelectedItems[0]);
        }

        private void ClearServices()
        {
            lvServices.Items.Clear();
            ClearComPorts();
        }

        private void btEnumServices_Click(object sender, EventArgs e)
        {
            if (lvDevices.SelectedItems.Count == 0)
                MessageBox.Show("Select device", "Warning", MessageBoxButtons.OK,
                    MessageBoxIcon.Warning);
            else 
            {
                ClearServices();
                
                ListViewItem Item = lvDevices.SelectedItems[0];
                wclBluetoothRadio Radio = (wclBluetoothRadio)Item.Tag;
                Int64 Address = Convert.ToInt64(Item.SubItems[1].Text, 16);
                
                wclBluetoothService[] Services;
                Guid g = Guid.Empty;
                Int32 Res = Radio.EnumRemoteServices(Address, g, out Services);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    ShowError(Res);
                else
                {
                    if (Services != null)
                    {
                        foreach (wclBluetoothService Service in Services)
                        {
                            Item = lvServices.Items.Add(Radio.ApiName);
                            Item.SubItems.Add(Address.ToString("X12"));
                            Item.SubItems.Add(Service.Handle.ToString("X8"));
                            Item.SubItems.Add(Service.Uuid.ToString());
                            Item.SubItems.Add(Service.Channel.ToString());
                            Item.SubItems.Add(Service.Name);
                            Item.SubItems.Add(Service.Comment);
                        }
                    }
                }
            }
        }

        private void btDiscover_Click(object sender, EventArgs e)
        {
            if (lvRadios.SelectedItems.Count == 0)
                MessageBox.Show("Select radio", "Warning", MessageBoxButtons.OK,
                    MessageBoxIcon.Warning);
            else
            {
                ListViewItem Item = lvRadios.SelectedItems[0];
                wclBluetoothRadio Radio = (wclBluetoothRadio)Item.Tag;

                wclBluetoothDiscoverKind Kind;
                if (cbDiscoveringMode.SelectedIndex == 0)
                    Kind = wclBluetoothDiscoverKind.dkClassic;
                else
                    Kind = wclBluetoothDiscoverKind.dkBle;

                Int32 Res = Radio.Discover(Convert.ToByte(edTimeout.Text), Kind);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    ShowError(Res);
            }
        }

        private void btTerminate_Click(object sender, EventArgs e)
        {
            if (lvRadios.SelectedItems.Count == 0)
                MessageBox.Show("Select radio", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning);
            else
            {
                ListViewItem Item = lvRadios.SelectedItems[0];
                wclBluetoothRadio Radio = (wclBluetoothRadio)Item.Tag;

                Int32 Res = Radio.Terminate();
                if (Res != wclErrors.WCL_E_SUCCESS)
                    ShowError(Res);
            }
        }

        private void AddDevice(wclBluetoothRadio Radio, Int64 Address)
        {
            ListViewItem Item = lvDevices.Items.Add(Radio.ApiName);
            Item.Tag = Radio;

            Item.SubItems.Add(Address.ToString("X12"));
            Item.SubItems.Add(""); // Class Of Device
            Item.SubItems.Add(""); // Device type
            Item.SubItems.Add(""); // Name
            Item.SubItems.Add(""); // Paired
            Item.SubItems.Add(""); // Connected
            Item.SubItems.Add(""); // Address type
        }

        private void RefreshDevices()
        {
            foreach (ListViewItem Item in lvDevices.Items)
                RefreshDevice(Item);
        }

        private void btUnpair_Click(object sender, EventArgs e)
        {
            if (lvDevices.SelectedItems.Count == 0)
                MessageBox.Show("Select device", "Warning", MessageBoxButtons.OK,
                    MessageBoxIcon.Warning);
            else
            {
                ListViewItem Item = lvDevices.SelectedItems[0];
                wclBluetoothRadio Radio = (wclBluetoothRadio)Item.Tag;
                Int64 Address = Convert.ToInt64(Item.SubItems[1].Text, 16);

                wclBluetoothPairingMethod Method;
                switch (cbPairingMethod.SelectedIndex)
                {
                    case 1:
                        Method = wclBluetoothPairingMethod.pmClassic;
                        break;
                    case 2:
                        Method = wclBluetoothPairingMethod.pmLe;
                        break;
                    default:
                        Method = wclBluetoothPairingMethod.pmAuto;
                        break;
                }

                Int32 Res = Radio.RemoteUnpair(Address, Method);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    ShowError(Res);

                RefreshDevice(Item);
            }
        }

        private void btPair_Click(object sender, EventArgs e)
        {
            if (lvDevices.SelectedItems.Count == 0)
                MessageBox.Show("Select device", "Warning", MessageBoxButtons.OK,
                    MessageBoxIcon.Warning);
            else
            {
                ListViewItem Item = lvDevices.SelectedItems[0];
                wclBluetoothRadio Radio = (wclBluetoothRadio)Item.Tag;
                Int64 Address = Convert.ToInt64(Item.SubItems[1].Text, 16);

                wclBluetoothPairingMethod Method;
                switch (cbPairingMethod.SelectedIndex)
                {
                    case 1:
                        Method = wclBluetoothPairingMethod.pmClassic;
                        break;
                    case 2:
                        Method = wclBluetoothPairingMethod.pmLe;
                        break;
                    default:
                        Method = wclBluetoothPairingMethod.pmAuto;
                        break;
                }

                Int32 Res = Radio.RemotePair(Address, Method);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    ShowError(Res);
            }
        }

        private void btRssi_Click(object sender, EventArgs e)
        {
            if (lvDevices.SelectedItems.Count == 0)
                MessageBox.Show("Select device", "Warning", MessageBoxButtons.OK,
                    MessageBoxIcon.Warning);
            else
            {
                ListViewItem Item = lvDevices.SelectedItems[0];
                wclBluetoothRadio Radio = (wclBluetoothRadio)Item.Tag;
                Int64 Address = Convert.ToInt64(Item.SubItems[1].Text, 16);

                SByte Rssi;
                Int32 Res = Radio.GetRemoteRssi(Address, out Rssi);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    ShowError(Res);
                else
                {
                    MessageBox.Show("RSSI: " + Rssi.ToString() + "dB", "RSSI", MessageBoxButtons.OK,
                        MessageBoxIcon.Information);
                }
            }
        }

        private void btTurnOn_Click(object sender, EventArgs e)
        {
            if (lvRadios.SelectedItems.Count == 0)
                MessageBox.Show("Select radio", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning);
            else
            {
                ListViewItem Item = lvRadios.SelectedItems[0];
                wclBluetoothRadio Radio = (wclBluetoothRadio)Item.Tag;

                Int32 Res = Radio.TurnOn();
                if (Res != wclErrors.WCL_E_SUCCESS)
                    ShowError(Res);

                RefreshRadio(Item);
            }
        }

        private void ntTurnOff_Click(object sender, EventArgs e)
        {
            if (lvRadios.SelectedItems.Count == 0)
                MessageBox.Show("Select radio", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning);
            else
            {
                ListViewItem Item = lvRadios.SelectedItems[0];
                wclBluetoothRadio Radio = (wclBluetoothRadio)Item.Tag;

                Int32 Res = Radio.TurnOff();
                if (Res != wclErrors.WCL_E_SUCCESS)
                    ShowError(Res);

                RefreshRadio(Item);
            }
        }

        private void ClearComPorts()
        {
            lvComPorts.Items.Clear();
        }

        private void btEnumVCOMs_Click(object sender, EventArgs e)
        {
            if (lvRadios.SelectedItems.Count == 0)
                MessageBox.Show("Select radio", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning);
            else
                RefreshVComs();
        }

        private void RefreshVComs()
        {
            ClearComPorts();

            ListViewItem Item = lvRadios.SelectedItems[0];
            wclBluetoothRadio Radio = (wclBluetoothRadio)Item.Tag;

            Int32 Res;
            wclVirtualComPort[] Ports;
            if (cbEnumPreinstalled.Checked)
                Res = Radio.EnumComPorts(out Ports);
            else
            {
                Ports = Radio.ComPorts;
                Res = wclErrors.WCL_E_SUCCESS;
            }

            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Error enumerating vCOMs: 0x" + Res.ToString("X8"));
            else
            {
                if (Ports != null && Ports.Length > 0)
                {
                    for (int i = 0; i < Ports.Length; i++)
                    {
                        wclVirtualComPort Port = Ports[i];

                        Item = lvComPorts.Items.Add(Radio.ApiName);
                        Item.SubItems.Add(Port.Address.ToString("X12"));
                        Item.SubItems.Add(Port.Service.ToString());
                        Item.SubItems.Add(Port.Number.ToString());

                        Item.Tag = Radio;
                    }
                }
            }
        }

        private void btCreateVCOM_Click(object sender, EventArgs e)
        {
            if (lvDevices.SelectedItems.Count == 0)
                MessageBox.Show("Select device", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning);
            else
            {
                if (lvServices.SelectedItems.Count == 0)
                    MessageBox.Show("Select service", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                else
                {
                    ListViewItem Item = lvDevices.SelectedItems[0];
                    wclBluetoothRadio Radio = (wclBluetoothRadio)Item.Tag;
                    Int64 Address = Convert.ToInt64(Item.SubItems[1].Text, 16);
                    Guid Service = new Guid(lvServices.SelectedItems[0].SubItems[3].Text);

                    UInt16 Number;
                    Int32 Res = Radio.CreateComPort(Address, Service, out Number);
                    if (Res != wclErrors.WCL_E_SUCCESS)
                        ShowError(Res);
                    else
                    {
                        TraceEvent(Radio, "Create vCOM", "Number", Number.ToString());

                        RefreshVComs();
                    }
                }
            }
        }

        private void btDestroyVCOM_Click(object sender, EventArgs e)
        {
            if (lvComPorts.SelectedItems.Count == 0)
                MessageBox.Show("Select vCOM", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning);
            else
            {
                ListViewItem Item = lvComPorts.SelectedItems[0];
                wclBluetoothRadio Radio = (wclBluetoothRadio)Item.Tag;
                UInt16 Number = Convert.ToUInt16(Item.SubItems[3].Text);

                Int32 Res = Radio.DestroyComPort(Number);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    ShowError(Res);
                else
                {
                    TraceEvent(Radio, "Destroy vCOM", "Number", Number.ToString());

                    RefreshVComs();
                }
            }
        }

        private void btIntsallDevice_Click(object sender, EventArgs e)
        {
            if (lvDevices.SelectedItems.Count == 0)
                MessageBox.Show("Select device", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning);
            else
            {
                if (lvServices.SelectedItems.Count == 0)
                    MessageBox.Show("Select service", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                else
                {
                    ListViewItem Item = lvDevices.SelectedItems[0];
                    wclBluetoothRadio Radio = (wclBluetoothRadio)Item.Tag;
                    Int64 Address = Convert.ToInt64(Item.SubItems[1].Text, 16);
                    Guid Service = new Guid(lvServices.SelectedItems[0].SubItems[3].Text);

                    Int32 Res = Radio.InstallDevice(Address, Service);
                    if (Res == wclBluetoothErrors.WCL_E_BLUETOOTH_DEVICE_ALREADY_INSTALLED)
                    {
                        // Reinstall to reconnect!
                        Radio.UninstallDevice(Address, Service);
                        Res = Radio.InstallDevice(Address, Service);
                    }

                    if (Res != wclErrors.WCL_E_SUCCESS)
                        ShowError(Res);
                    else
                        MessageBox.Show("Device installed");
                }
            }
        }

        private void btUninstallDevice_Click(object sender, EventArgs e)
        {
            if (lvDevices.SelectedItems.Count == 0)
                MessageBox.Show("Select device", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning);
            else
            {
                if (lvServices.SelectedItems.Count == 0)
                    MessageBox.Show("Select service", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                else
                {
                    ListViewItem Item = lvDevices.SelectedItems[0];
                    wclBluetoothRadio Radio = (wclBluetoothRadio)Item.Tag;
                    Int64 Address = Convert.ToInt64(Item.SubItems[1].Text, 16);
                    Guid Service = new Guid(lvServices.SelectedItems[0].SubItems[3].Text);

                    Int32 Res = Radio.UninstallDevice(Address, Service);
                    if (Res != wclErrors.WCL_E_SUCCESS)
                        ShowError(Res);
                    else
                        MessageBox.Show("Device uninstalled");
                }
            }
        }

        private void btIsInRange_Click(object sender, EventArgs e)
        {
            if (lvDevices.SelectedItems.Count == 0)
                MessageBox.Show("Select device", "Warning", MessageBoxButtons.OK,
                    MessageBoxIcon.Warning);
            else
            {
                ListViewItem Item = lvDevices.SelectedItems[0];
                wclBluetoothRadio Radio = (wclBluetoothRadio)Item.Tag;
                Int64 Address = Convert.ToInt64(Item.SubItems[1].Text, 16);

                Boolean InRange;
                Int32 Res = Radio.IsRemoteDeviceInRange(Address, out InRange);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    ShowError(Res);
                else
                {
                    if (InRange)
                        MessageBox.Show("Device is in range");
                    else
                        MessageBox.Show("Device is NOT in range");
                }
            }
        }

        private void btDisconnect_Click(object sender, EventArgs e)
        {
            if (lvDevices.SelectedItems.Count == 0)
                MessageBox.Show("Select device", "Warning", MessageBoxButtons.OK,
                    MessageBoxIcon.Warning);
            else
            {
                ListViewItem Item = lvDevices.SelectedItems[0];
                wclBluetoothRadio Radio = (wclBluetoothRadio)Item.Tag;
                Int64 Address = Convert.ToInt64(Item.SubItems[1].Text, 16);

                Int32 Res = Radio.RemoteDisconnect(Address);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    ShowError(Res);
            }
        }

        private void btEnumConnected_Click(object sender, EventArgs e)
        {
            if (lvRadios.SelectedItems.Count == 0)
                MessageBox.Show("Select radio", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning);
            else
            {
                ClearDevices();

                ListViewItem Item = lvRadios.SelectedItems[0];
                wclBluetoothRadio Radio = (wclBluetoothRadio)Item.Tag;

                Int64[] Devices;
                Int32 Res = Radio.EnumConnectedDevices(out Devices);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    ShowError(Res);
                else
                {
                    if (Devices != null)
                    {
                        for (Int32 i = 0; i < Devices.Length; i++)
                            AddDevice(Radio, Devices[i]);

                        RefreshDevices();
                    }
                }
            }
        }

        private void btEnumInstalledServices_Click(object sender, EventArgs e)
        {
            if (lvDevices.SelectedItems.Count == 0)
                MessageBox.Show("Select device", "Warning", MessageBoxButtons.OK,
                    MessageBoxIcon.Warning);
            else
            {
                ClearServices();

                ListViewItem Item = lvDevices.SelectedItems[0];
                wclBluetoothRadio Radio = (wclBluetoothRadio)Item.Tag;
                Int64 Address = Convert.ToInt64(Item.SubItems[1].Text, 16);

                Guid[] Services;
                Int32 Res = Radio.EnumInstalledServices(Address, out Services);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    ShowError(Res);
                else
                {
                    if (Services != null && Services.Length > 0)
                    {
                        foreach (Guid Service in Services)
                        {
                            Item = lvServices.Items.Add(Radio.ApiName);
                            Item.SubItems.Add(Address.ToString("X12"));
                            Item.SubItems.Add("");
                            Item.SubItems.Add(Service.ToString());
                            Item.SubItems.Add("");
                            Item.SubItems.Add("");
                            Item.SubItems.Add("");
                        }
                    }
                }
            }
        }
    }
}