using System;
using System.Threading;
using wclCommon;
using wclBluetooth;

namespace Console
{
    class Program
    {
        private AutoResetEvent FEvent;
        
        private void WriteLn(String s)
        {
            System.Console.WriteLine(s);
        }

        private void ReadLn()
        {
            System.Console.ReadKey(true);
        }

        private void Manager_AfterOpen(object sender, EventArgs e)
        {
            FEvent.Set();
        }

        private void Manager_OnClosed(object sender, EventArgs e)
        {
            FEvent.Set();
        }

        private void Manager_OnDeviceFound(object Sender, wclBluetoothRadio Radio, long Address)
        {
            WriteLn("Device " + Address.ToString("X12") + " found");
        }

        private void Manager_OnDiscoveringCompleted(object Sender, wclBluetoothRadio Radio, int Error)
        {
            WriteLn("Discoverign has been compledted");
            FEvent.Set();
        }

        private void Manager_OnDiscoveringStarted(object Sender, wclBluetoothRadio Radio)
        {
            WriteLn("Discovering has been started");
        }

        public void Run()
        {
            WriteLn("Bluetooth Framework Console Demo application");
            WriteLn("  The demo is very simple: it just discovers nearby classic and LE");
            WriteLn("  devices.");
            WriteLn("  The demo shows how to use Bluetooth Framework in console");
            WriteLn("  applications with APC synchronization.");
            WriteLn("Press ENTER to continue");

            ReadLn();

            // Change default synchronization method.
            wclMessageBroadcaster.SetSyncMethod(wclMessageSynchronizationKind.skApc);

            // Create classes we have to use.
            wclBluetoothManager Manager = new wclBluetoothManager();
            Manager.AfterOpen += new EventHandler(Manager_AfterOpen);
            Manager.OnClosed += new EventHandler(Manager_OnClosed);
            Manager.OnDeviceFound += new wclBluetoothDeviceEvent(Manager_OnDeviceFound);
            Manager.OnDiscoveringCompleted += new wclBluetoothResultEvent(Manager_OnDiscoveringCompleted);
            Manager.OnDiscoveringStarted += new wclBluetoothEvent(Manager_OnDiscoveringStarted);

            // The event used to notify about event completion.
            try { FEvent = new AutoResetEvent(false); } catch { FEvent = null; }
            if (FEvent == null)
                WriteLn("Unable to create synchronization event!");
            else
            {
                // Open Bluetooth Manager
                Int32 Res = Manager.Open();
                if (Res != wclErrors.WCL_E_SUCCESS)
                    WriteLn("Open Bluetooth Manager failed: 0x" + Res.ToString("X8"));
                else
                {
                    // Wait for Open completion.
                    wclMessageBroadcaster.Wait(FEvent);

                    // Try to find available Bluetooth Radio. Use first one found.
                    if (Manager.Count == 0)
                        WriteLn("No Bluetooth Hardware connected.");
                    else
                    {
                        wclBluetoothRadio Radio = null;
                        for (int i = 0; i < Manager.Count; i++)
                        {
                            if (Manager[i].Available)
                            {
                                Radio = Manager[i];
                                break;
                            }
                        }

                        if (Radio == null)
                            WriteLn("Available BluetoothRadio not found.");
                        else
                        {
                            // Discover classic devices
                            WriteLn("Try to start classic discovering");
                            Res = Radio.Discover(10, wclBluetoothDiscoverKind.dkClassic);
                            if (Res != wclErrors.WCL_E_SUCCESS)
                                WriteLn("Start classic discovering failed: 0x" + Res.ToString("X8"));
                            else
                                // Wait for discovering completion.
                                wclMessageBroadcaster.Wait(FEvent);

                            // Discover LE devices
                            WriteLn("Try to start LE discovering");
                            Res = Radio.Discover(10, wclBluetoothDiscoverKind.dkBle);
                            if (Res != wclErrors.WCL_E_SUCCESS)
                                WriteLn("Start LE discovering failed: 0x" + Res.ToString("X8"));
                            else
                                // Wait for discovering completion.
                                wclMessageBroadcaster.Wait(FEvent);
                        }
                    }

                    // Close Bluetooth Manager.
                    Res = Manager.Close();
                    if (Res != wclErrors.WCL_E_SUCCESS)
                        WriteLn("Open Bluetooth Manager failed: 0x" + Res.ToString("X8"));
                    else
                        // Wait for Close completion.
                        wclMessageBroadcaster.Wait(FEvent);
                }

                // Do not forget about event.
                FEvent.Close();
            }

            // Do not forget to destroy Bluetooth Manager.
            Manager = null;

            WriteLn("Demo finished. Press ENTER");
            ReadLn();
        }

        static void Main(string[] args)
        {
            Program p = new Program();
            p.Run();
        }
    }
}