﻿using System;
using System.Windows.Forms;
using System.IO;

using wclCommon;
using wclCommunication;
using wclBluetooth;

namespace FtpClient
{
    public partial class fmMain : Form
    {
        private wclBluetoothManager wclBluetoothManager;
        private wclRfCommClient wclRfCommClient;
        private wclObexFtpClient FFtp;

        private wclBluetoothRadio GetRadio()
        {
            wclBluetoothRadio Radio;
            Int32 Res = wclBluetoothManager.GetClassicRadio(out Radio);
            if (Res != wclErrors.WCL_E_SUCCESS)
            {
                MessageBox.Show("Get working radio failed: 0x" + Res.ToString("X8"));
                return null;
            }
            return Radio;
        }

        public fmMain()
        {
            InitializeComponent();
        }

        private void fmMain_Load(object sender, EventArgs e)
        {
            wclBluetoothManager = new wclBluetoothManager();
            wclBluetoothManager.AfterOpen +=new EventHandler(wclBluetoothManager_AfterOpen);
            wclBluetoothManager.OnAuthenticationCompleted +=new wclBluetoothDeviceResultEvent(wclBluetoothManager_OnAuthenticationCompleted);
            wclBluetoothManager.OnDiscoveringCompleted +=new wclBluetoothResultEvent(wclBluetoothManager_OnDiscoveringCompleted);
            wclBluetoothManager.OnDiscoveringStarted +=new wclBluetoothEvent(wclBluetoothManager_OnDiscoveringStarted);
            wclBluetoothManager.OnNumericComparison +=new wclBluetoothNumericComparisonEvent(wclBluetoothManager_OnNumericComparison);
            wclBluetoothManager.OnPasskeyNotification +=new wclBluetoothPasskeyNotificationEvent(wclBluetoothManager_OnPasskeyNotification);
            wclBluetoothManager.OnPasskeyRequest +=new wclBluetoothPasskeyRequestEvent(wclBluetoothManager_OnPasskeyRequest);
            wclBluetoothManager.OnPinRequest +=new wclBluetoothPinRequestEvent(wclBluetoothManager_OnPinRequest);
            wclBluetoothManager.OnDeviceFound +=new wclBluetoothDeviceEvent(wclBluetoothManager_OnDeviceFound);

            wclRfCommClient = new wclRfCommClient();
            wclRfCommClient.OnDisconnect +=new wclClientConnectionDisconnectEvent(wclRfCommClient_OnDisconnect);
            wclRfCommClient.OnConnect +=new wclClientConnectionConnectEvent(wclRfCommClient_OnConnect);
            wclRfCommClient.OnDestroyProcessor +=new wclClientConnectionProcessorEvent(wclRfCommClient_OnDestroyProcessor);
            wclRfCommClient.OnCreateProcessor +=new wclClientConnectionProcessorEvent(wclRfCommClient_OnCreateProcessor);
            
            // We use ObjectPush Profile so must set the profile's UUID.
            wclRfCommClient.Service = wclUUIDs.OBEXFileTransferServiceClass_UUID;
            
            FFtp = null;
            
            // Try to open Bluetooth Manager.
            Int32 Res = wclBluetoothManager.Open();
            if (Res != wclErrors.WCL_E_SUCCESS)
                lbLog.Items.Add("Bluetooth Manager Open failed: 0x" + Res.ToString("X8"));
        }

        void wclRfCommClient_OnCreateProcessor(object Sender, wclClientDataConnection Connection)
        {
            // here we must create the data processor for the connection. In this demo
            // we use OPPClient.
            FFtp = new wclObexFtpClient(Connection);
            FFtp.OnConnect += new wclObexOperationResultEvent(FFtp_OnConnect);
            FFtp.OnDisconnect += new wclObexOperationResultEvent(FFtp_OnDisconnect);
            FFtp.OnGetComplete += new wclObexObjectOperationResultEvent(FFtp_OnGetComplete);
            FFtp.OnPutComplete += new wclObexObjectOperationResultEvent(FFtp_OnPutComplete);
            FFtp.OnProgress += new wclObexOperationProgressEvent(FFtp_OnProgress);
            FFtp.OnDirComplete += new wclObexFtpDirCompleteEvent(FFtp_OnDirComplete);
            FFtp.OnChangeDirComplete += new wclObexOperationResultEvent(FFtp_OnChangeDirComplete);
            FFtp.OnDeleteComplete += new wclObexOperationResultEvent(FFtp_OnDeleteComplete);
            FFtp.OnMakeDirComplete += new wclObexOperationResultEvent(FFtp_OnMakeDirComplete);
        }

        void FFtp_OnMakeDirComplete(object Sender, int Error, string Description)
        {
            lbLog.Items.Add("Make dir completed with result: 0x" + Error.ToString("X8") + ". Description: " + Description);

            ProgressBar.Value = 0;
            ProgressBar.Maximum = 0;

            lvFiles.Items.Clear();
            FFtp.Dir();
        }

        void FFtp_OnDeleteComplete(object Sender, int Error, string Description)
        {
            lbLog.Items.Add("Delete completed with result: 0x" + Error.ToString("X8") + ". Description: " + Description);

            ProgressBar.Value = 0;
            ProgressBar.Maximum = 0;

            lvFiles.Items.Clear();
            FFtp.Dir();
        }

        void FFtp_OnChangeDirComplete(object Sender, int Error, string Description)
        {
            lbLog.Items.Add("Change dir completed with result: 0x" + Error.ToString("X8") + ". Description: " + Description);
            
            ProgressBar.Value = 0;
            ProgressBar.Maximum = 0;
            
            lvFiles.Items.Clear();
            FFtp.Dir();
        }

        void FFtp_OnDirComplete(object Sender, int Error, string Description, wclObexFileObject[] Dirs)
        {
            lbLog.Items.Add("Dir completed with result: 0x" + Error.ToString("X8") + ". Description: " + Description);
            
            ProgressBar.Value = 0;
            ProgressBar.Maximum = 0;

            if (Dirs != null)
            {
                foreach (wclObexFileObject Obj in Dirs)
                {
                    ListViewItem Item = lvFiles.Items.Add(Obj.Name);
                    if (Obj.IsDirectory)
                        Item.SubItems.Add("Folder");
                    else
                        Item.SubItems.Add("File");
                    Item.SubItems.Add(Obj.Size.ToString());

                    String Perm = "";
                    if ((Obj.Permissions & wclObexFilePermission.opRead) != 0)
                        Perm = Perm + "R";
                    if ((Obj.Permissions & wclObexFilePermission.opWrite) != 0)
                        Perm = Perm + "W";
                    if ((Obj.Permissions & wclObexFilePermission.opDelete) != 0)
                        Perm = Perm + "D";
                    Item.SubItems.Add(Perm);
                    if (!Obj.Modified.Equals(DateTime.MinValue))
                        Item.SubItems.Add(Obj.Modified.ToString());
                    else
                        Item.SubItems.Add("");
                    if (!Obj.Created.Equals(DateTime.MinValue))
                        Item.SubItems.Add(Obj.Created.ToString());
                    else
                        Item.SubItems.Add("");
                    if (!Obj.Accessed.Equals(DateTime.MinValue))
                        Item.SubItems.Add(Obj.Accessed.ToString());
                    else
                        Item.SubItems.Add("");

                    Item.SubItems.Add(Obj.Description);
                }
            }
        }

        void FFtp_OnGetComplete(object Sender, int Error, string Description, Stream Stream)
        {
            lbLog.Items.Add("File receiving completed with result: 0x" + Error.ToString("X8") + ". Description: " + Description);
            
            ProgressBar.Value = 0;
            ProgressBar.Maximum = 0;
            
            // We must destroy the stream object here.
            Stream.Close();
            Stream.Dispose();
        }

        void FFtp_OnProgress(object Sender, uint Size, uint Position)
        {
            ProgressBar.Maximum = (int)Size;
            ProgressBar.Value = (int)Position;
        }

        void FFtp_OnPutComplete(object Sender, int Error, string Description, Stream Stream)
        {
            lbLog.Items.Add("File sending completed with result: 0x" + Error.ToString("X8") + ". Description: " + Description);

            ProgressBar.Value = 0;
            ProgressBar.Maximum = 0;

            // We must destroy the stream object here.
            Stream.Close();
            Stream.Dispose();
        }

        void FFtp_OnDisconnect(object Sender, int Error, string Description)
        {
            lbLog.Items.Add("OBEX session closed with result: 0x" + Error.ToString("X8") + ". Description: " + Description);
        }

        void FFtp_OnConnect(object Sender, int Error, string Description)
        {
            lbLog.Items.Add("OBEX session opened with result: 0x" + Error.ToString("X8") + ". Description: " + Description);
        }

        void wclRfCommClient_OnDestroyProcessor(object Sender, wclClientDataConnection Connection)
        {
            // Do we have any data processor created?
            if (FFtp != null)
            {
                // Make sure it is our connection is goind to destroy.
                if (Connection.Processor == FFtp)
                {
                    // Ok, destroy the data processor here.
                    FFtp.Dispose();
                    FFtp = null;
                }
            }
        }

        void wclRfCommClient_OnConnect(object Sender, int Error)
        {
            lbLog.Items.Add("Connect. Operation result: 0x" + Error.ToString("X8"));
        }

        void wclRfCommClient_OnDisconnect(object Sender, int Reason)
        {
            lbLog.Items.Add("Disconnected with reason: 0x" + Reason.ToString("X8"));
        }

        void wclBluetoothManager_OnDeviceFound(object Sender, wclBluetoothRadio Radio, long Address)
        {
            ListViewItem Item = lvDevices.Items.Add(Address.ToString("X12"));
            Item.SubItems.Add("");
        }

        void wclBluetoothManager_OnPinRequest(object Sender, wclBluetoothRadio Radio, long Address, out string Pin)
        {
            lbLog.Items.Add("PIN code request. PIN: 0000");
            Pin = "0000";
        }

        void wclBluetoothManager_OnPasskeyRequest(object Sender, wclBluetoothRadio Radio, long Address, out uint Passkey)
        {
            lbLog.Items.Add("Passkey request. Passkey: 12345678");
            Passkey = 12345678;
        }

        void wclBluetoothManager_OnPasskeyNotification(object Sender, wclBluetoothRadio Radio, long Address, uint Passkey)
        {
            lbLog.Items.Add("Passkey notification. Passkey: " + Passkey.ToString());
        }

        void wclBluetoothManager_OnNumericComparison(object Sender, wclBluetoothRadio Radio, long Address, uint Number, out bool Confirm)
        {
            lbLog.Items.Add("Numeric comparison. Accept");
            Confirm = true;
        }

        void wclBluetoothManager_OnDiscoveringStarted(object Sender, wclBluetoothRadio Radio)
        {
            lbLog.Items.Add("Discovering started");

            lvDevices.Items.Clear();
        }

        void wclBluetoothManager_OnDiscoveringCompleted(object Sender, wclBluetoothRadio Radio, int Error)
        {
            lbLog.Items.Add("Discovering completed with result: 0x" + Error.ToString("X8"));

            for (Int32 i = 0; i < lvDevices.Items.Count; i++)
            {
                ListViewItem Item = lvDevices.Items[i];
                Int64 Addr = Convert.ToInt64(Item.Text, 16);
                String Name;
                Int32 Res = Radio.GetRemoteName(Addr, out Name);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    Name = "Error: 0x" + Res.ToString("X8");
                Item.SubItems[1].Text = Name;
            }
        }

        void wclBluetoothManager_OnAuthenticationCompleted(object Sender, wclBluetoothRadio Radio, long Address, int Error)
        {
            lbLog.Items.Add("Authentication completed with result: 0x" + Error.ToString("X8"));
        }

        void wclBluetoothManager_AfterOpen(object sender, EventArgs e)
        {
            lbLog.Items.Add("Bluetooth Manager has been opened");
        }

        private void btClear_Click(object sender, EventArgs e)
        {
            lbLog.Items.Clear();
        }

        private void fmMain_FormClosing(object sender, FormClosingEventArgs e)
        {
            wclRfCommClient.Disconnect();
            wclBluetoothManager.Close();
        }

        private void btDiscover_Click(object sender, EventArgs e)
        {
            wclBluetoothRadio Radio = GetRadio();
            if (Radio != null)
            {
                // Discover classic devices only.
                Int32 Res = Radio.Discover(10, wclBluetoothDiscoverKind.dkClassic);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Failed to start discovering: 0x" + Res.ToString("X8"));
            }
        }

        private void btDisconnect_Click(object sender, EventArgs e)
        {
            Int32 Res = wclRfCommClient.Disconnect();
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Disconnect failed: 0x" + Res.ToString("X8"));
        }

        private void btCloseSession_Click(object sender, EventArgs e)
        {
            if (FFtp == null)
                MessageBox.Show("Not connected");
            else
            {
                Int32 Res = FFtp.Disconnect("");
                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Close session failed: 0x" + Res.ToString("X8"));
            }
        }

        private void btOpenSession_Click(object sender, EventArgs e)
        {
            if (FFtp == null)
                MessageBox.Show("Not connected");
            else
            {
                Int32 Res = FFtp.Connect();
                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Open session failed: 0x" + Res.ToString("X8"));
            }
        }

        private void btConnect_Click(object sender, EventArgs e)
        {
            if (wclRfCommClient.State != wclClientState.csDisconnected)
                MessageBox.Show("Client is connected");
            else
            {
                if (lvDevices.SelectedItems.Count == 0)
                    MessageBox.Show("Select device");
                else
                {
                    wclBluetoothRadio Radio = GetRadio();
                    if (Radio != null)
                    {
                        Int64 Addr = Convert.ToInt64(lvDevices.SelectedItems[0].Text, 16);
                        wclRfCommClient.Address = Addr;
                        Int32 Res = wclRfCommClient.Connect(Radio);
                        if (Res != wclErrors.WCL_E_SUCCESS)
                            MessageBox.Show("Failed to connect: 0x" + Res.ToString("X8"));
                    }
                }
            }
        }

        private void btSelectFile_Click(object sender, EventArgs e)
        {
            if (openFileDialog.ShowDialog() == DialogResult.OK)
                edFileName.Text = openFileDialog.FileName;
        }

        private void btAbort_Click(object sender, EventArgs e)
        {
            if (FFtp == null)
                MessageBox.Show("Not connected");
            else
            {
                Int32 Res = FFtp.Abort("");
                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Abort failed: 0x" + Res.ToString("X8"));
            }
        }

        private void btSendFile_Click(object sender, EventArgs e)
        {
            if (FFtp == null)
                MessageBox.Show("Not connected");
            else
            {
                if (edFileName.Text == "")
                    MessageBox.Show("Select file");
                else
                {
                    FileStream Stream = new FileStream(edFileName.Text, FileMode.Open);
                    Int32 Res = FFtp.Put(Path.GetFileName(edFileName.Text), "", Stream);
                    if (Res != wclErrors.WCL_E_SUCCESS)
                    {
                        MessageBox.Show("Send file failed: 0x" + Res.ToString("X8"));
                        Stream.Close();
                        Stream.Dispose();
                    }
                }
            }
        }

        private void btDir_Click(object sender, EventArgs e)
        {
            if (FFtp == null)
                MessageBox.Show("Not connected");
            else
            {
                Int32 Res = FFtp.Dir();
                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Dir failed: 0x" + Res.ToString("X8"));
                else
                    lvFiles.Items.Clear();
            }
        }

        private void lvFiles_DoubleClick(object sender, EventArgs e)
        {
            if (lvFiles.SelectedItems.Count > 0)
            {
                if (FFtp == null)
                    MessageBox.Show("Not connected");
                else
                {
                    if (lvFiles.SelectedItems[0].SubItems[1].Text == "Folder")
                    {
                        Int32 Res = FFtp.ChangeDir(lvFiles.SelectedItems[0].Text);
                        if (Res != wclErrors.WCL_E_SUCCESS)
                            MessageBox.Show("Change path failed 0x" + Res.ToString("X8"));
                    }
                    else
                    {
                        FileStream Stream = new FileStream(".\\" + lvFiles.SelectedItems[0].Text, FileMode.OpenOrCreate, FileAccess.ReadWrite);
                        Int32 Res = FFtp.Get(lvFiles.SelectedItems[0].Text, Stream);
                        if (Res != wclErrors.WCL_E_SUCCESS)
                        {
                            MessageBox.Show("Get failed 0x" + Res.ToString("X8"));
                            Stream.Close();
                            Stream.Dispose();
                        }
                    }
                }
            }
        }

        private void btRoot_Click(object sender, EventArgs e)
        {
            if (FFtp == null)
                MessageBox.Show("Not connected");
            else
            {
                Int32 Res = FFtp.ChangeDir("");
                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Root failed 0x" + Res.ToString("X8"));
            }
        }

        private void btDelete_Click(object sender, EventArgs e)
        {
            if (lvFiles.SelectedItems.Count > 0)
            {
                if (FFtp == null)
                    MessageBox.Show("Not connected");
                else
                {
                    Int32 Res = FFtp.Delete(lvFiles.SelectedItems[0].Text);
                    if (Res != wclErrors.WCL_E_SUCCESS)
                        MessageBox.Show("Delete failed 0x" + Res.ToString("X8"));
                }
            }
        }

        private void btMkDir_Click(object sender, EventArgs e)
        {
            if (FFtp == null)
                MessageBox.Show("Not connected");
            else
            {
                Int32 Res = FFtp.MkDir(edNewDirName.Text);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Make sire failed 0x" + Res.ToString("X8"));
            }
        }
    }
}
