using System;
using System.Windows.Forms;
using System.IO;

using wclCommon;
using wclCommunication;
using wclBluetooth;

namespace OppServer
{
    public partial class fmMain : Form
    {
        private wclBluetoothManager wclBluetoothManager;
        private wclRfCommServer wclRfCommServer;
        private String FFileName;

        private void Trace(String Msg)
        {
            lbLog.Items.Add(Msg);
        }

        public fmMain()
        {
            InitializeComponent();
        }

        private void fmMain_Load(object sender, EventArgs e)
        {
            wclBluetoothManager = new wclBluetoothManager();

            wclRfCommServer = new wclRfCommServer();
            wclRfCommServer.Authentication = false;
            wclRfCommServer.Service = wclUUIDs.OBEXObjectPushServiceClass_UUID;
            wclRfCommServer.ServiceName = "WCL OPP Server";
            wclRfCommServer.OnListen += new EventHandler(wclRfCommServer_OnListen);
            wclRfCommServer.OnClosed += new wclClientConnectionDisconnectEvent(wclRfCommServer_OnClosed);
            wclRfCommServer.OnDisconnect += new wclRfCommServerDisconnectEvent(wclRfCommServer_OnDisconnect);
            wclRfCommServer.OnDestroyProcessor += new wclServerConnectionProcessorEvent(wclRfCommServer_OnDestroyProcessor);
            wclRfCommServer.OnConnect += new wclRfCommServerConnectEvent(wclRfCommServer_OnConnect);
            wclRfCommServer.OnCreateProcessor += new wclServerConnectionProcessorEvent(wclRfCommServer_OnCreateProcessor);
            wclRfCommServer.GetSdpAttributes += new wclBluetoothGetSdpAttributesEvent(wclRfCommServer_GetSdpAttributes);
        }

        void wclRfCommServer_GetSdpAttributes(object Sender, out ushort[] Protocols, out wclBluetoothSdpProfile[] Profiles, out byte[] Formats, out uint Cod)
        {
            // We have to provide additional information for SDP record.
            Protocols = new ushort[1];
            Protocols[0] = wclUUIDs.OBEX_PROTOCOL_UUID16;
            Profiles = new wclBluetoothSdpProfile[1];
            Profiles[0].Uuid = wclUUIDs.OBEXObjectPushServiceClass_UUID16;
            Profiles[0].Version = 0x0100;
            Formats = new byte[1];
            Formats[0] = 0xFF; // All formats.
            Cod = wclCODs.COD_SERVICE_OBJECT_XFER;
        }

        void wclRfCommServer_OnCreateProcessor(object Sender, wclServerClientDataConnection Connection)
        {
            // Here we must create OBEX OPP server data processor.
            wclObexOppServer Proc = new wclObexOppServer(Connection);
            Proc.OnConnect += new wclObexOppServerConnectEvent(Proc_OnConnect);
            Proc.OnDisconnected += new wclObexServerClientDisconnectedEvent(Proc_OnDisconnected);
            Proc.OnPutBegin += new wclObexServerClientPutBeginEvent(Proc_OnPutBegin);
            Proc.OnPutCompleted += new wclObexServerClientPutCompletedEvent(Proc_OnPutCompleted);
            Proc.OnPutProgress += new wclObexServerClientOperationProgressEvent(Proc_OnPutProgress);
            Proc.OnGetRequest += new wclObexOppServerGetRequestEvent(Proc_OnGetRequest);
            Proc.OnProgress += new wclObexOperationProgressEvent(Proc_OnProgress);
            Proc.OnGetCompleted += new wclObexServerClientGetCompletedEvent(Proc_OnGetCompleted);
        }

        void Proc_OnGetCompleted(object Sender, int Error, Stream Stream)
        {
            Trace("Get operation completed with result: 0x" + Error.ToString("X8"));
            if (Stream != null)
                Stream.Close();
        }

        void Proc_OnProgress(object Sender, uint Length, uint Position)
        {
            Trace("Sending file: " + Position.ToString() + " from " + Length.ToString());
        }

        void Proc_OnGetRequest(object Sender, string aType, out wclObexServerOperationResult Result, out Stream Stream)
        {
            if (aType == "text/x-vCard" || aType == "x-obex/capability")
            {
                Trace("Requested file type: " + aType);

                if (openFileDialog.ShowDialog() != DialogResult.OK)
                {
                    Result = wclObexServerOperationResult.orForbidden;
                    Stream = null;
                }
                else
                {
                    Stream = new FileStream(openFileDialog.FileName, FileMode.Open);
                    Result = wclObexServerOperationResult.orSuccess;
                }

            }
            else
            {
                Result = wclObexServerOperationResult.orUnsupportedMedia;
                Stream = null;
            }
        }

        void Proc_OnPutProgress(object Sender, uint Position, uint Length, out bool Continue)
        {
            Trace("Sending file: " + Position.ToString() + " from " + Length.ToString());
            // Do not foget to set Continue flag!
            Continue = true;
        }

        void Proc_OnPutCompleted(object Sender, int Error, Stream Stream, out bool Accept)
        {
            wclObexOppServer Proc = (wclObexOppServer)Sender;
            Int64 Address = ((wclRfCommServerClientConnection)Proc.Connection).Address;
            if (Error == wclErrors.WCL_E_SUCCESS)
            {
                String Str = "Client " + Address.ToString("X12") + " completed sent file. Accept?";
                Accept = MessageBox.Show(Str, "Accept?", MessageBoxButtons.YesNo, MessageBoxIcon.Question) == DialogResult.Yes;

                if (Accept)
                {
                    saveFileDialog.FileName = FFileName;
                    if (saveFileDialog.ShowDialog() == DialogResult.OK)
                    {
                        Stream.Seek(0, SeekOrigin.Begin);
                        FileStream fs = new FileStream(saveFileDialog.FileName, FileMode.CreateNew, FileAccess.Write);
                        Byte[] Bytes = new Byte[Stream.Length];
                        Stream.Read(Bytes, 0, (int)Stream.Length);
                        fs.Write(Bytes, 0, Bytes.Length);
                        fs.Close();
                     }
                }
            }
            else
            {
                Trace("Send file error: 0x" + Error.ToString("X8"));
                Accept = false; // does not matter.
            }
        }

        void Proc_OnPutBegin(object Sender, string Name, string Description, string Mime, uint Length, out bool Accept)
        {
            wclObexOppServer Proc = (wclObexOppServer)Sender;
            Int64 Address = ((wclRfCommServerClientConnection)Proc.Connection).Address;
            String Str = "Client " + Address.ToString("X12") + " send file " + Name +
                ". Accept?";
            Accept = MessageBox.Show(Str, "Accept?", MessageBoxButtons.YesNo, MessageBoxIcon.Question) == DialogResult.Yes;
            if (Accept)
                FFileName = Name;
        }

        void Proc_OnDisconnected(object Sender, int Reason, string Description)
        {
            wclObexOppServer Proc = (wclObexOppServer)Sender;
            Int64 Address = ((wclRfCommServerClientConnection)Proc.Connection).Address;
            
            Trace("Client " + Address.ToString("X12") + " closed OBEX session with reason: 0x" + Reason.ToString("X8") +
                ". Description: " + Description);
            
            for (Int32 i = 0; i < lvClients.Items.Count; i++)
            {
                if (lvClients.Items[i].Tag == Sender)
                {
                    lvClients.Items.RemoveAt(i);
                    break;
                }
            }
        }

        void Proc_OnConnect(object Sender, string Description)
        {
            wclObexOppServer Proc = (wclObexOppServer)Sender;
            Int64 Address = ((wclRfCommServerClientConnection)Proc.Connection).Address;
            Boolean Accept = MessageBox.Show("Client " + Address.ToString("X12") +
                " requsts for connection: " + Description + ". Accept?", "Accept?",
                MessageBoxButtons.YesNo, MessageBoxIcon.Question) == DialogResult.Yes;
            
            Int32 Res;
            if (Accept)
            {
                Res = Proc.Accept("Accepted by user");
                if (Res == wclErrors.WCL_E_SUCCESS)
                {
                    ListViewItem Item = lvClients.Items.Add(Address.ToString("X12"));
                    Item.Tag = Sender;
                    Trace("OBEX session opened for client: " + Address.ToString("X12"));
                }
            }
            else
                Res = Proc.Reject("Rejected by user");

            if (Res != wclErrors.WCL_E_SUCCESS)
            {
                MessageBox.Show("Operation failed: 0x" + Res.ToString("X8"));
                // Disconnect client!
                Proc.Connection.Disconnect();
            }
        }

        void wclRfCommServer_OnConnect(object Sender, wclRfCommServerClientConnection Client, int Error)
        {
            if (Error == wclErrors.WCL_E_SUCCESS)
                Trace("Client " + Client.Address.ToString("X12") + " connected");
            else
                Trace("Client " + Client.Address.ToString("X12") + " connect failed: 0x" + Error.ToString("X8"));
        }

        void wclRfCommServer_OnDestroyProcessor(object Sender, wclServerClientDataConnection Connection)
        {
            Trace("Data processor for client " + ((wclRfCommServerClientConnection)Connection).Address.ToString("X12") + " is destroing");
            Connection.Processor.Dispose();
        }

        void wclRfCommServer_OnDisconnect(object Sender, wclRfCommServerClientConnection Client, int Reason)
        {
            Trace("Client " + Client.Address.ToString("X12") + " disconnected. Reason: 0x" + Reason.ToString("X8"));
        }

        void wclRfCommServer_OnClosed(object Sender, int Reason)
        {
            Trace("Server closed. Reason:: 0x" + Reason.ToString("X8"));
        }

        void wclRfCommServer_OnListen(object sender, EventArgs e)
        {
            Trace("Server listening");
        }

        private void fmMain_FormClosed(object sender, FormClosedEventArgs e)
        {
            wclRfCommServer.Close();
            wclBluetoothManager.Close();

            wclRfCommServer = null;
            wclBluetoothManager = null;
        }

        private void btClear_Click(object sender, EventArgs e)
        {
            lbLog.Items.Clear();
        }

        private void btListen_Click(object sender, EventArgs e)
        {
            Int32 Res = wclBluetoothManager.Open();
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Bluetooth manager open failed: 0x" + Res.ToString("X8"));
            else
            {
                wclBluetoothRadio Radio;
                Res = wclBluetoothManager.GetClassicRadio(out Radio);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Get working radio failed: 0x" + Res.ToString("X8"));
                else
                {
                    // Switch radio in discoverable and connectable mode.
                    // Ignore any errors. In real application you need to check the result.
                    Radio.SetDiscoverable(true);
                    Radio.SetConnectable(true);

                    // As wrote above it is better to start listening in AfterOpen event
                    // handler. But again, for demo app its OK to do it here.
                    Res = wclRfCommServer.Listen(Radio);
                    if (Res != wclErrors.WCL_E_SUCCESS)
                        MessageBox.Show("Failed to start listening: 0x" + Res.ToString("X8"));
                }

                if (Res != wclErrors.WCL_E_SUCCESS)
                    // We must close Bluetooth Manager here!
                    wclBluetoothManager.Close();
            }
        }

        private void btClose_Click(object sender, EventArgs e)
        {
            Int32 Res = wclRfCommServer.Close();
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Close failed; 0x" + Res.ToString("X8"));
            else
                wclBluetoothManager.Close();
        }

        private void btDisconnect_Click(object sender, EventArgs e)
        {
            if (lvClients.SelectedItems == null)
                MessageBox.Show("Select client");
            else
            {
                wclObexOppServer Proc = (wclObexOppServer)lvClients.SelectedItems[0].Tag;
                Int32 Res = Proc.Connection.Disconnect();
                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Disconnect failed: 0x" + Res.ToString("X8"));
            }
        }
    }
}