﻿using System;
using System.ComponentModel;
using System.Drawing;
using System.Windows.Forms;

using wclCommon;
using wclCommunication;
using wclBluetooth;

namespace WiiRemote
{
    public partial class fmMain : Form
    {
        private wclBluetoothManager FManager;
        private wclWiiRemoteClient FClient;

        public fmMain()
        {
            InitializeComponent();
        }

        private wclBluetoothRadio GetRadio()
        {
            wclBluetoothRadio Radio;
            Int32 Res = FManager.GetClassicRadio(out Radio);
            if (Res != wclErrors.WCL_E_SUCCESS)
            {
                MessageBox.Show("Get working radio failed: 0x" + Res.ToString("X8"));
                return null;
            }
            return Radio;
        }

        private void fmMain_Load(object sender, EventArgs e)
        {
            FManager = new wclBluetoothManager();
            FManager.OnDiscoveringStarted += new wclBluetoothEvent(FManager_OnDiscoveringStarted);
            FManager.OnDeviceFound += new wclBluetoothDeviceEvent(FManager_OnDeviceFound);
            FManager.OnDiscoveringCompleted += new wclBluetoothResultEvent(FManager_OnDiscoveringCompleted);

            FClient = new wclWiiRemoteClient();
            FClient.OnConnect += new wclClientConnectionConnectEvent(FClient_OnConnect);
            FClient.OnDisconnect += new wclClientConnectionDisconnectEvent(FClient_OnDisconnect);
            FClient.OnStatusChanged += new wclWiiRemoteStatusChangedEvent(FClient_OnStatusChanged);
            FClient.OnButtonsChanged += new wclWiiRemoteButtonsChangedEvent(FClient_OnButtonsChanged);
            FClient.OnAccelChanged += new wclWiiRemoteAccelChangedEvent(FClient_OnAccelChanged);
            FClient.OnIrChanged += new wclWiiRemoteIrChangedEvent(FClient_OnIrChanged);
            FClient.OnExtensionAttached += new EventHandler(FClient_OnExtensionAttached);
            FClient.OnExtensionDetached += new EventHandler(FClient_OnExtensionDetached);
            FClient.OnNunchukChanged += new wclWiiRemoteNunchukChangedEvent(FClient_OnNunchukChanged);
            FClient.OnBalanceBoardChanged += new wclWiiRemoteBalanceBoardChangedEvent(FClient_OnBalanceBoardChanged);

            FManager.Open();
            ClearAccel();
            ClearIr();
            ClearExtensions();
        }
        
        void  FClient_OnBalanceBoardChanged(object Sender, wclWiiRemoteBalanceBoard Board)
        {
            laBbTopRight.Text = Board.SensorsKg.TopRight.ToString("N2");
            laBbTopLeft.Text = Board.SensorsKg.TopLeft.ToString("N2");
            laBbBottomRight.Text = Board.SensorsKg.BottomRight.ToString("N2");
            laBbBottomLeft.Text = Board.SensorsKg.BottomLeft.ToString("N2");
        }
        
        void  FClient_OnNunchukChanged(object Sender, wclWiiRemoteNunchuk Nunchuk)
        {
            laNunchukX.Text = Nunchuk.Accel.Values.X.ToString();
            laNunchukY.Text = Nunchuk.Accel.Values.Y.ToString();
            laNunchukZ.Text = Nunchuk.Accel.Values.Z.ToString();
            
            laNunchukJoyX.Text = Nunchuk.Joystick.X.ToString();
            laNunchukJoyY.Text = Nunchuk.Joystick.Y.ToString();
            
            cbNunchukC.Checked = Nunchuk.C;
            cbNunchukZ.Checked = Nunchuk.Z;
        }

        private void ClearNunchuk()
        {
            laNunchukState.Text = "Detached";
            
            laNunchukX.Text = "0";
            laNunchukY.Text = "0";
            laNunchukZ.Text = "0";
            
            laNunchukJoyX.Text = "0";
            laNunchukJoyY.Text = "0";
            
            cbNunchukC.Checked = false;
            cbNunchukZ.Checked = false;
        }

        private void ClearExtensions()
        {
            ClearNunchuk();
            ClearBalanceBoard();
        }

        private void ClearBalanceBoard()
        {
            laBalanceBoardState.Text = "Detached";
            
            laBbTopRight.Text = "0";
            laBbTopLeft.Text = "0";
            laBbBottomRight.Text = "0";
            laBbBottomLeft.Text = "0";
        }

        void  FClient_OnExtensionDetached(object sender, EventArgs e)
        {
            Trace("Extension detached");
            ClearExtensions();
        }
        
        void  FClient_OnExtensionAttached(object sender, EventArgs e)
        {
            String Ext;
            switch (FClient.Extension)
            {
                case wclWiiRemoteExtension.wiiNoExtension:
                    Ext = "wiiNoExtension";
                    break;
                case wclWiiRemoteExtension.wiiNunchuk:
                    Ext = "wiiNunchuk";
                    laNunchukState.Text = "Attached";
                    break;
                case wclWiiRemoteExtension.wiiClassicController:
                    Ext = "wiiClassicController";
                    break;
                case wclWiiRemoteExtension.wiiGuitar:
                    Ext = "wiiGuitar";
                    break;
                case wclWiiRemoteExtension.wiiDrums:
                    Ext = "wiiDrums";
                    break;
                case wclWiiRemoteExtension.wiiBalanceBoard:
                    Ext = "wiiBalanceBoard";
                    laBalanceBoardState.Text = "Attached";
                    break;
                default:
                    Ext = "Unknonw";
                    break;
            }
            
            Trace("Extension attached: " + Ext);
        }

        void FClient_OnIrChanged(object Sender, wclWiiRemoteIrSensors Ir)
        {
            switch (Ir.Mode)
            {
                case wclWiiRemoteIrMode.wiiIrOff:
                    ClearIr();
                    break;
                case wclWiiRemoteIrMode.wiiIrBasicMode:
                    laIrMode.Text = "Basic";
                    break;
                case wclWiiRemoteIrMode.wiiIrExtendedMode:
                    laIrMode.Text = "Extended";
                    break;
                case wclWiiRemoteIrMode.wiiIrFullMode:
                    laIrMode.Text = "Full";
                    break;
                default:
                    laIrMode.Text = "Unknown";
                    break;
            }
            
            if (Ir.Mode != wclWiiRemoteIrMode.wiiIrOff)
            {
                switch (Ir.Sensitivity)
                {
                    case wclWiiRemoteIrSensitivity. wiiIrLevelOff:
                        laIrSensitivity.Text = "Off";
                        break;
                    case wclWiiRemoteIrSensitivity.wiiIrLevel1:
                        laIrSensitivity.Text = "Level 1";
                        break;
                    case wclWiiRemoteIrSensitivity.wiiIrLevel2:
                        laIrSensitivity.Text = "Level 2";
                        break;
                    case wclWiiRemoteIrSensitivity.wiiIrLevel3:
                        laIrSensitivity.Text = "Level 3";
                        break;
                    case wclWiiRemoteIrSensitivity.wiiIrLevel4:
                        laIrSensitivity.Text = "Level 4";
                        break;
                    case wclWiiRemoteIrSensitivity.wiiIrLevel5:
                        laIrSensitivity.Text = "Level 5";
                        break;
                    case wclWiiRemoteIrSensitivity.wiiIrLevelMax:
                        laIrSensitivity.Text = "Maximum";
                        break;
                    default:
                        laIrSensitivity.Text = "Unknown";
                        break;
                }
                
                if (Ir.Sensors[0].Found)
                {
                    laSensor0Position.Text = Ir.Sensors[0].Position.X.ToString() + "; " +
                        Ir.Sensors[0].Position.Y.ToString();
                    laSensor0Size.Text = Ir.Sensors[0].Size.ToString();
                    laSensor0Found.Text = "True";
                }
                else
                {
                    laSensor0Position.Text = "0; 0";
                    laSensor0Size.Text = "0";
                    laSensor0Found.Text = "False";
                }
                
                if (Ir.Sensors[1].Found)
                {
                    laSensor1Position.Text = Ir.Sensors[1].Position.X.ToString() + "; " +
                        Ir.Sensors[1].Position.Y.ToString();
                    laSensor1Size.Text = Ir.Sensors[1].Size.ToString();
                    laSensor1Found.Text = "True";
                }
                else
                {
                    laSensor1Position.Text = "0; 0";
                    laSensor1Size.Text = "0";
                    laSensor1Found.Text = "False";
                }
                
                if (Ir.Sensors[2].Found)
                {
                    laSensor2Position.Text = Ir.Sensors[2].Position.X.ToString() + "; " +
                        Ir.Sensors[2].Position.Y.ToString();
                    laSensor2Size.Text = Ir.Sensors[2].Size.ToString();
                    laSensor2Found.Text = "True";
                }
                else
                {
                    laSensor2Position.Text = "0; 0";
                    laSensor2Size.Text = "0";
                    laSensor2Found.Text = "False";
                }
                
                if (Ir.Sensors[3].Found)
                {
                    laSensor3Position.Text = Ir.Sensors[3].Position.X.ToString() + "; " +
                        Ir.Sensors[3].Position.Y.ToString();
                    laSensor3Size.Text = Ir.Sensors[3].Size.ToString();
                    laSensor3Found.Text = "True";
                }
                else
                {
                    laSensor3Position.Text = "0; 0";
                    laSensor3Size.Text = "0";
                    laSensor3Found.Text = "False";
                }
            }
        }

        private void ClearAccel()
        {
            laX.Text = "0";
            laY.Text = "0";
            laZ.Text = "0";
        }

        private void ClearIr()
        {
            laIrMode.Text = "Off";
            laIrSensitivity.Text = "Off";
            
            laSensor0Position.Text = "0; 0";
            laSensor0Size.Text = "0";
            laSensor0Found.Text = "False";
            
            laSensor1Position.Text = "0; 0";
            laSensor1Size.Text = "0";
            laSensor1Found.Text = "False";
            
            laSensor2Position.Text = "0; 0";
            laSensor2Size.Text = "0";
            laSensor2Found.Text = "False";
            
            laSensor3Position.Text = "0; 0";
            laSensor3Size.Text = "0";
            laSensor3Found.Text = "False";
        }

        void FClient_OnAccelChanged(object Sender, wclWiiRemoteAccel Accel)
        {
            laX.Text = Accel.Values.X.ToString();
            laY.Text = Accel.Values.Y.ToString();
            laZ.Text = Accel.Values.Z.ToString();
        }

        void FClient_OnButtonsChanged(object Sender, wclWiiRemoteButtons Buttons)
        {
            cbA.Checked = Buttons.A;
            cbB.Checked = Buttons.B;
            cbPlus.Checked = Buttons.Plus;
            cbHome.Checked = Buttons.Home;
            cbMinus.Checked = Buttons.Minus;
            cbOne.Checked = Buttons.One;
            cbTwo.Checked = Buttons.Two;
            cbUp.Checked = Buttons.Up;
            cbDown.Checked = Buttons.Down;
            cbLeft.Checked = Buttons.Left;
            cbRight.Checked = Buttons.Right;
        }

        void FClient_OnStatusChanged(object Sender, double Batt, wclWiiRemoteLeds Leds)
        {
            Trace("Status changed");

            laBattLevel.Text = Batt.ToString("N2");
            
            cbLed1.Checked = Leds.Led1;
            cbLed2.Checked = Leds.Led2;
            cbLed3.Checked = Leds.Led3;
            cbLed4.Checked = Leds.Led4;
        }

        void FClient_OnDisconnect(object Sender, int Reason)
        {
            Trace("Disconnected. Reason", Reason);
            
            laBattLevel.Text = "0.00";
            cbLed1.Checked = false;
            cbLed2.Checked = false;
            cbLed3.Checked = false;
            cbLed4.Checked = false;
            
            cbA.Checked = false;
            cbB.Checked = false;
            cbPlus.Checked = false;
            cbHome.Checked = false;
            cbMinus.Checked = false;
            cbOne.Checked = false;
            cbTwo.Checked = false;
            cbUp.Checked = false;
            cbDown.Checked = false;
            cbLeft.Checked = false;
            cbRight.Checked = false;
            
            ClearAccel();
            ClearIr();
            ClearExtensions();
        }

        void FClient_OnConnect(object Sender, int Error)
        {
            if (Error != wclErrors.WCL_E_SUCCESS)
                Trace("Connection failed", Error);
            else
            {
                Trace("Connected");
                SetLeds();
            }
        }

        void FManager_OnDiscoveringCompleted(object Sender, wclBluetoothRadio Radio, int Error)
        {
            for (int i = 0; i < lvDevices.Items.Count; i++)
            {
                ListViewItem Item = lvDevices.Items[i];
                Int64 Addr = Convert.ToInt64(Item.Text, 16);

                String Name;
                Int32 Res = Radio.GetRemoteName(Addr, out Name);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    Name = "Error: 0x" + Res.ToString("X9");
                Item.SubItems[1].Text = Name;
            }

            Trace("Discovering completed with result", Error);
        }

        void FManager_OnDeviceFound(object Sender, wclBluetoothRadio Radio, long Address)
        {
            ListViewItem Item = lvDevices.Items.Add(Address.ToString("X12"));
            Item.SubItems.Add("");
        }

        void FManager_OnDiscoveringStarted(object Sender, wclBluetoothRadio Radio)
        {
            lvDevices.Items.Clear();
            Trace("Discovering started");
        }

        private void fmMain_FormClosed(object sender, FormClosedEventArgs e)
        {
            FClient.Disconnect();
            FManager.Close();
        }

        private void btDiscover_Click(object sender, EventArgs e)
        {
            wclBluetoothRadio Radio = GetRadio();
            if (Radio != null)
            {
                Int32 Res = Radio.Discover(10, wclBluetoothDiscoverKind.dkClassic);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Discover failed; 0x" + Res.ToString("X8"));
            }
        }

        private void btConnect_Click(object sender, EventArgs e)
        {
            if (FClient.State != wclClientState.csDisconnected)
                MessageBox.Show("Wii Remote is connected");
            else
            {
                wclBluetoothRadio Radio = GetRadio();
                if (Radio != null)
                {
                    if (lvDevices.SelectedItems.Count == 0)
                        MessageBox.Show("Select device");
                    else
                    {
                        FClient.Address = Convert.ToInt64(lvDevices.SelectedItems[0].Text, 16);
                        Int32 Res = FClient.Connect(Radio);
                        if (Res != wclErrors.WCL_E_SUCCESS)
                            MessageBox.Show("Unable connect: 0x" + Res.ToString("X8"));
                    }
                }
            }
        }

        private void btDisconnect_Click(object sender, EventArgs e)
        {
            Int32 Res = FClient.Disconnect();
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Unable disconnect: 0x" + Res.ToString("X8"));
        }

        private void btClear_Click(object sender, EventArgs e)
        {
            lbLog.Items.Clear();
        }

        private void Trace(String Msg)
        {
            lbLog.Items.Add(Msg);
        }

        private void Trace(String Msg, Int32 Res)
        {
            Trace(Msg + ": 0x" + Res.ToString("X8"));
        }

        private void btGetStatus_Click(object sender, EventArgs e)
        {
            Int32 Res = FClient.GetStatus();
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Get status failed; 0x" + Res.ToString("X8"));
        }

        private void btGetLeds_Click(object sender, EventArgs e)
        {
            if (FClient.State != wclClientState.csConnected)
                MessageBox.Show("Not connected");
            else
            {
                cbLed1.Checked = FClient.Leds.Led1;
                cbLed2.Checked = FClient.Leds.Led2;
                cbLed3.Checked = FClient.Leds.Led3;
                cbLed4.Checked = FClient.Leds.Led4;
            }
        }

        private void btSetLeds_Click(object sender, EventArgs e)
        {
            SetLeds();
        }

        private void SetLeds()
        {
            Int32 Res = FClient.SetLeds(cbLed1.Checked, cbLed2.Checked,
                cbLed3.Checked, cbLed4.Checked);
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Set LEDs failed: 0x" + Res.ToString("X8"));
        }

        private void btEnableAccel_Click(object sender, EventArgs e)
        {
            Int32 Res = FClient.EnableAccel();
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Enable accelerometer failed: 0x" + Res.ToString("X8"));
        }

        private void btDisableAccel_Click(object sender, EventArgs e)
        {
            Int32 Res = FClient.DisableAccel();
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Disable accelerometer failed: 0x" + Res.ToString("X8"));
            else
                ClearAccel();
        }

        private void btEnableIr_Click(object sender, EventArgs e)
        {
            Int32 Res = FClient.SetIrSensitivity(wclWiiRemoteIrSensitivity.wiiIrLevelMax);
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Enable IR failed: 0x" + Res.ToString("X8"));
        }

        private void btDisableIr_Click(object sender, EventArgs e)
        {
            Int32 Res = FClient.SetIrSensitivity(wclWiiRemoteIrSensitivity.wiiIrLevelOff);
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Disable IR failed: 0x" + Res.ToString("X8"));
            else
                ClearIr();
        }

        private void btRumbleOn_Click(object sender, EventArgs e)
        {
            Int32 Res = FClient.SetRumble(true);
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Set Rumble ON failed; 0x" + Res.ToString("X8"));
        }

        private void btRumbleOff_Click(object sender, EventArgs e)
        {
            Int32 Res = FClient.SetRumble(false);
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Set Rumble OFF failed; 0x" + Res.ToString("X8"));
        }
    }
}
