﻿Public Class fmMain
    Private WithEvents Switcher As wclAudioSwitcher

    Private Sub fmMain_Load(sender As System.Object, e As System.EventArgs) Handles MyBase.Load
        Switcher = New wclAudioSwitcher()

        cbActive.Checked = True
        cbRole.SelectedIndex = 0
        cbFlow.SelectedIndex = 0

        Dim Res As Int32 = Switcher.Open()
        If Res <> wclErrors.WCL_E_SUCCESS Then lbLog.Items.Add("Open failed: 0x" + Res.ToString("X8"))
    End Sub

    Private Sub Switcher_OnClosed(sender As Object, e As System.EventArgs) Handles Switcher.OnClosed
        lvDevices.Clear()
        lbLog.Items.Add("Audio switcher closed")
    End Sub

    Private Sub Switcher_OnDefaultDeviceChanged(Sender As Object, Id As String, Flow As wclAudio.wclAudioDeviceDataFlow, Role As wclAudio.wclAudioDeviceRole) Handles Switcher.OnDefaultDeviceChanged
        Dim Str As String = "Default device changed: " + Id + "; Flow: "
        If Flow = wclAudioDeviceDataFlow.dfRender Then
            Str += "Render"
        Else
            Str += "Capture"
        End If
        Str += "; Role: "
        Select Case Role
            Case wclAudioDeviceRole.drConsole
                Str += "Console"
            Case wclAudioDeviceRole.drMultimedia
                Str += "Multimedia"
            Case wclAudioDeviceRole.drCommunications
                Str += "Communications"
        End Select
        lbLog.Items.Add(Str)
    End Sub

    Private Sub Switcher_OnDeviceAdded(Sender As Object, Id As String) Handles Switcher.OnDeviceAdded
        lbLog.Items.Add("Device added: " + Id)
    End Sub

    Private Sub Switcher_OnDeviceRemoved(Sender As Object, Id As String) Handles Switcher.OnDeviceRemoved
        lbLog.Items.Add("Device removed: " + Id)
    End Sub

    Private Sub Switcher_OnOpened(sender As Object, e As System.EventArgs) Handles Switcher.OnOpened
        lbLog.Items.Add("Audio switcher opened")
        EnumDevices()
    End Sub

    Private Sub Switcher_OnStateChanged(Sender As Object, Id As String, State As wclAudio.wclAudioDeviceState) Handles Switcher.OnStateChanged
        Dim Str As String = ""
        Select Case State
            Case wclAudioDeviceState.asActive
                Str = "Active"
            Case wclAudioDeviceState.asDisabled
                Str = "Disabled"
            Case wclAudioDeviceState.asNotPresent
                Str = "NotPresent"
            Case wclAudioDeviceState.asUnplugged
                Str = "Unplugged"
        End Select
        lbLog.Items.Add("Device " + Id + " state changed: " + Str)
    End Sub

    Private Sub fmMain_FormClosed(sender As System.Object, e As System.Windows.Forms.FormClosedEventArgs) Handles MyBase.FormClosed
        Switcher.Close()
    End Sub

    Private Sub btEnum_Click(sender As System.Object, e As System.EventArgs) Handles btEnum.Click
        EnumDevices()
    End Sub

    Private Sub btSetDefault_Click(sender As System.Object, e As System.EventArgs) Handles btSetDefault.Click
        If lvDevices.SelectedItems Is Nothing Then
            MessageBox.Show("Select device")
        Else
            Dim Role As wclAudioDeviceRole
            Select cbRole.SelectedIndex
                Case 0
                    Role = wclAudioDeviceRole.drConsole
                Case 1
                    Role = wclAudioDeviceRole.drMultimedia
                Case Else
                    Role = wclAudioDeviceRole.drCommunications
            End Select

            Dim Res As Int32 = Switcher.SetDefault(Role, lvDevices.SelectedItems(0).Text)
            If Res <> wclErrors.WCL_E_SUCCESS Then MessageBox.Show("Unable to set default device: 0x" + Res.ToString("X8"))
        End If
    End Sub

    Private Sub EnumDevices()
        Dim States As wclAudioDeviceState = 0
        If cbActive.Checked Then States = States Or wclAudioDeviceState.asActive
        If cbDisabled.Checked Then States = States Or wclAudioDeviceState.asDisabled
        If cbNotPresent.Checked Then States = States Or wclAudioDeviceState.asNotPresent
        If cbUnplugged.Checked Then States = States Or wclAudioDeviceState.asUnplugged
        If States = 0 Then
            MessageBox.Show("Select device state.")
        Else
            lvDevices.Items.Clear()

            Dim Devices As wclAudioDevice() = Nothing
            Dim Res As Int32 = Switcher.Enum(States, Devices)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                MessageBox.Show("Enum failed: 0x" + Res.ToString("X8"))
            Else
                If Devices IsNot Nothing AndAlso Devices.Length > 0 Then
                    For Each Device As wclAudioDevice In Devices
                        Dim Item As ListViewItem = lvDevices.Items.Add(Device.Id)
                        Item.SubItems.Add("")
                        Item.SubItems.Add("")
                        Item.SubItems.Add("")
                        Item.SubItems.Add("")
                        Item.SubItems.Add("")
                        Item.SubItems.Add("")
                        Item.SubItems.Add("")
                        UpdateDevice(Item, Device)
                    Next Device
                End If
            End If
        End If
    End Sub

    Private Sub btClear_Click(sender As System.Object, e As System.EventArgs) Handles btClear.Click
        lbLog.Items.Clear()
    End Sub

    Private Sub UpdateDevice(Item As ListViewItem, Device As wclAudioDevice)
        Dim FlowStr As String
        If Device.Flow = wclAudioDeviceDataFlow.dfRender Then
            FlowStr = "Render"
        Else
            FlowStr = "Capture"
        End If

        Dim RolesStr As String = "[ "
        If (Device.Roles And wclAudioDeviceRole.drConsole) <> 0 Then RolesStr += "Console "
        If (Device.Roles And wclAudioDeviceRole.drMultimedia) <> 0 Then RolesStr += "Multimedia "
        If (Device.Roles And wclAudioDeviceRole.drCommunications) <> 0 Then RolesStr += "Communications "
        RolesStr += "]"

        Dim StateStr As String = ""
        Select Case Device.State
            Case wclAudioDeviceState.asActive
                StateStr = "Active"
            Case wclAudioDeviceState.asDisabled
                StateStr = "Disabled"
            Case wclAudioDeviceState.asNotPresent
                StateStr = "NotPresent"
            Case wclAudioDeviceState.asUnplugged
                StateStr = "Unplugged"
        End Select


        Dim MacStr As String
        Dim ServiceStr As String
        If Device.IsBluetooth Then
            MacStr = Device.Mac.ToString("X12")
            ServiceStr = Device.Service.ToString()
        Else
            MacStr = ""
            ServiceStr = ""
        End If

        Item.SubItems(1).Text = Device.FriendlyName
        Item.SubItems(2).Text = Device.Description
        Item.SubItems(3).Text = FlowStr
        Item.SubItems(4).Text = RolesStr
        Item.SubItems(5).Text = StateStr
        Item.SubItems(6).Text = MacStr
        Item.SubItems(7).Text = ServiceStr
    End Sub

    Private Sub btRefresh_Click(sender As System.Object, e As System.EventArgs) Handles btRefresh.Click
        If lvDevices.SelectedItems Is Nothing Then
            MessageBox.Show("Select device")
        Else
            Dim Device As wclAudioDevice = New wclAudioDevice()
            Dim Res As Int32 = Switcher.GetDeviceDetails(lvDevices.SelectedItems(0).Text, Device)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                MessageBox.Show("Refresh device error: 0x" + Res.ToString("X8"))
            Else
                UpdateDevice(lvDevices.SelectedItems(0), Device)
            End If
        End If
    End Sub

    Private Sub btCionnect_Click(sender As System.Object, e As System.EventArgs) Handles btCionnect.Click
        If lvDevices.SelectedItems Is Nothing Then
            MessageBox.Show("Select device")
        Else
            Dim Res As Int32
            If cbUseMac.Checked Then
                If lvDevices.SelectedItems(0).SubItems(6).Text = "" Then
                    MessageBox.Show("Selected device is not Bluetooth enabled")
                    Return
                End If

                Res = Switcher.Connect(Convert.ToInt64(lvDevices.SelectedItems(0).SubItems(6).Text, 16))
            Else
                Res = Switcher.Connect(lvDevices.SelectedItems(0).Text)
            End If

            If Res <> wclErrors.WCL_E_SUCCESS Then MessageBox.Show("Connect error: 0x" + Res.ToString("X8"))
        End If
    End Sub

    Private Sub btDisconnect_Click(sender As System.Object, e As System.EventArgs) Handles btDisconnect.Click
        If lvDevices.SelectedItems Is Nothing Then
            MessageBox.Show("Select device")
        Else
            Dim Res As Int32
            If cbUseMac.Checked Then
                If lvDevices.SelectedItems(0).SubItems(6).Text = "" Then
                    MessageBox.Show("Selected device is not Bluetooth enabled")
                    Return
                End If

                Res = Switcher.Disconnect(Convert.ToInt64(lvDevices.SelectedItems(0).SubItems(6).Text, 16))
            Else
                Res = Switcher.Disconnect(lvDevices.SelectedItems(0).Text)
            End If

            If Res <> wclErrors.WCL_E_SUCCESS Then MessageBox.Show("Disconnect error: 0x" + Res.ToString("X8"))
        End If
    End Sub

    Private Sub btGetDefault_Click(sender As System.Object, e As System.EventArgs) Handles btGetDefault.Click
        Dim Role As wclAudioDeviceRole
        Select cbRole.SelectedIndex
            Case 0
                    Role = wclAudioDeviceRole.drConsole
            Case 1
                    Role = wclAudioDeviceRole.drMultimedia
            Case Else
                Role = wclAudioDeviceRole.drCommunications
        End Select

        Dim Flow As wclAudioDeviceDataFlow
        If cbFlow.SelectedIndex = 0 Then
            Flow = wclAudioDeviceDataFlow.dfCapture
        Else
            Flow = wclAudioDeviceDataFlow.dfRender
        End If

        Dim Id As String
        Dim Res As Integer = Switcher.GetDefault(Role, Flow, Id)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Unable to get default device: 0x" + Res.ToString("X8"))
        Else
            MessageBox.Show("Default device ID: " + Id)
        End If
    End Sub
End Class
