﻿Public Class fmMain
    Public Class VerticalProgressBar
        Inherits System.Windows.Forms.ProgressBar
        Protected Overrides ReadOnly Property CreateParams As System.Windows.Forms.CreateParams
            Get
                Dim cp As System.Windows.Forms.CreateParams = MyBase.CreateParams
                cp.Style = cp.Style Or &H4
                Return cp
            End Get
        End Property
    End Class

    Private Class AudioDeviceStorage
        Public Device As wclAudioDevice

        Public Overrides Function ToString() As String
            Return Device.FriendlyName
        End Function
    End Class

    Private WithEvents AudioSwitcher As wclAudioSwitcher
    Private WithEvents AudioMeter As wclAudioMeter
    Private WithEvents AudioVolume As wclAudioVolume
    Private FBars() As VerticalProgressBar

    Private Sub fmMain_Load(sender As System.Object, e As System.EventArgs) Handles MyBase.Load
        FBars = Nothing

        AudioSwitcher = New wclAudioSwitcher()
        AudioMeter = New wclAudioMeter()
        AudioVolume = New wclAudioVolume()

        Dim Res As Int32 = AudioSwitcher.Open()
        If Res <> wclErrors.WCL_E_SUCCESS Then Trace("Audio switcher open failed", Res)
    End Sub

    Private Sub AudioSwitcher_OnClosed(sender As Object, e As System.EventArgs) Handles AudioSwitcher.OnClosed
        Trace("Audio switcher closed")

        cbEndpoints.Items.Clear()
    End Sub

    Private Sub AudioSwitcher_OnDefaultDeviceChanged(Sender As Object, Id As String, Flow As wclAudio.wclAudioDeviceDataFlow, Role As wclAudio.wclAudioDeviceRole) Handles AudioSwitcher.OnDefaultDeviceChanged
        Trace("Active device changed")

        If Not AudioMeter.Active Then SelectActiveDevice()
    End Sub

    Private Sub AudioSwitcher_OnDeviceAdded(Sender As Object, Id As String) Handles AudioSwitcher.OnDeviceAdded
        Dim Device As wclAudioDevice = New wclAudioDevice()
        If AudioSwitcher.GetDeviceDetails(Id, Device) = wclErrors.WCL_E_SUCCESS Then AddDevice(Device)
    End Sub

    Private Sub AudioSwitcher_OnDeviceRemoved(Sender As Object, Id As String) Handles AudioSwitcher.OnDeviceRemoved
        DeleteDevice(Id)
    End Sub

    Private Sub AudioSwitcher_OnOpened(sender As Object, e As System.EventArgs) Handles AudioSwitcher.OnOpened
        Trace("Audio switcher opened")

        Dim Devices() As wclAudioDevice = Nothing
        Dim Res As Int32 = AudioSwitcher.Enum(wclAudioDeviceState.asActive, Devices)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            Trace("Enum audio endpoints failed", Res)
        Else
            If Devices Is Nothing OrElse Devices.Length = 0 Then
                Trace("No active audio endpoints found")
            Else
                For Each Device As wclAudioDevice In Devices
                    AddDevice(Device)
                Next

                If cbEndpoints.Items.Count = 0 Then
                    Trace("No output audio endpoints found")
                Else
                    SelectActiveDevice()
                End If
            End If
        End If
    End Sub

    Private Sub AudioSwitcher_OnStateChanged(Sender As Object, Id As String, State As wclAudio.wclAudioDeviceState) Handles AudioSwitcher.OnStateChanged
        If State = wclAudioDeviceState.asActive Then
            AudioSwitcher_OnDeviceAdded(AudioSwitcher, Id)
        Else
            DeleteDevice(Id)
        End If
    End Sub

    Private Sub AudioMeter_OnClosed(sender As Object, e As System.EventArgs) Handles AudioMeter.OnClosed
        Trace("Audio meter closed")
        Timer.Enabled = False
        cbEndpoints.Enabled = True

        If FBars IsNot Nothing AndAlso FBars.Length > 0 Then
            For i As Int32 = 0 To FBars.Length - 1
                Controls.Remove(FBars(i))
                FBars(i).Dispose()
            Next i
            FBars = Nothing
        End If

        pbMaster.Value = 0
        pbMaster.Visible = False
    End Sub

    Private Sub AudioMeter_OnDisconnected(sender As Object, e As System.EventArgs) Handles AudioMeter.OnDisconnected
        Trace("Audio meter device " + AudioMeter.Id + " disconnected")
    End Sub

    Private Sub AudioMeter_OnOpened(sender As Object, e As System.EventArgs) Handles AudioMeter.OnOpened
        Trace("Audio meter opened")
        cbEndpoints.Enabled = False
        pbMaster.Visible = True

        Dim Features As wclAudioHardwareFeature
        Dim Res As Int32 = AudioMeter.GetHardwareFeatures(Features)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            Trace("Get hardware features failed", Res)
        Else
            Trace("Hardware features")
            If (Features And wclAudioHardwareFeature.hfVolume) <> 0 Then Trace("  Volume")
            If (Features And wclAudioHardwareFeature.hfMute) <> 0 Then Trace("  Mute")
            If (Features And wclAudioHardwareFeature.hfMeter) <> 0 Then Trace("  Meter")
        End If

        Dim Channels As UInt32
        Res = AudioMeter.GetChannels(Channels)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            Trace("Get channels count failed", Res)
        Else
            If Channels = 0 Then
                Trace("No mettering channels available")
                AudioMeter.Close()
            Else
                Trace("Channels: " + Channels.ToString())

                Dim Pos As Int32 = pbMaster.Left + pbMaster.Width + 5
                ReDim FBars(Channels - 1)
                For i As UInt32 = 0 To Channels - 1
                    Dim Bar As VerticalProgressBar = New VerticalProgressBar()
                    Bar.Width = pbMaster.Width
                    Bar.Height = pbMaster.Height
                    Bar.Top = pbMaster.Top
                    Bar.Left = Pos
                    Bar.Style = System.Windows.Forms.ProgressBarStyle.Continuous
                    Pos = Bar.Left + Bar.Width + 5
                    Controls.Add(Bar)

                    FBars(i) = Bar
                Next i

                UpdateChannelsValue()

                Timer.Enabled = True
            End If
        End If
    End Sub

    Private Sub btClear_Click(sender As System.Object, e As System.EventArgs) Handles btClear.Click
        lbLog.Items.Clear()
    End Sub

    Private Sub Trace(Msg As String)
        lbLog.Items.Add(Msg)
    End Sub

    Private Sub Trace(Msg As String, [Error] As Int32)
        Trace(Msg + ": 0x" + [Error].ToString("X8"))
    End Sub

    Private Sub DeleteDevice(Id As String)
        If cbEndpoints.Items.Count > 0 Then
            For i As Int32 = 0 To cbEndpoints.Items.Count - 1
                Dim Storage As AudioDeviceStorage = CType(cbEndpoints.Items(i), AudioDeviceStorage)
                If Storage.Device.Id = Id Then
                    cbEndpoints.Items.RemoveAt(i)
                    If cbEndpoints.Items.Count > 0 Then SelectActiveDevice()
                    Exit For
                End If
            Next i
        End If
    End Sub

    Private Sub SelectActiveDevice()
        If cbEndpoints.Items.Count > 0 Then
            For i As Int32 = 0 To cbEndpoints.Items.Count - 1
                Dim Storage As AudioDeviceStorage = CType(cbEndpoints.Items(i), AudioDeviceStorage)
                Dim Device As wclAudioDevice = New wclAudioDevice()
                Dim Res As Int32 = AudioSwitcher.GetDeviceDetails(Storage.Device.Id, Device)
                If Res = wclErrors.WCL_E_SUCCESS Then Storage.Device = Device
                If (Storage.Device.Roles And wclAudioDeviceRole.drMultimedia) <> 0 Then
                    cbEndpoints.SelectedIndex = i
                    Return
                End If
            Next i
        End If

        cbEndpoints.SelectedIndex = -1
    End Sub

    Private Sub AddDevice(Device As wclAudioDevice)
        If Device.State = wclAudioDeviceState.asActive AndAlso Device.Flow = wclAudioDeviceDataFlow.dfRender Then
            Dim Storage As AudioDeviceStorage = New AudioDeviceStorage()
            Storage.Device = Device
            cbEndpoints.Items.Add(Storage)

            SelectActiveDevice()
        End If
    End Sub

    Private Sub fmMain_FormClosed(sender As System.Object, e As System.Windows.Forms.FormClosedEventArgs) Handles MyBase.FormClosed
        AudioVolume.Close()
        AudioMeter.Close()
        AudioSwitcher.Close()
    End Sub

    Private Sub btClose_Click(sender As System.Object, e As System.EventArgs) Handles btClose.Click
        Dim Res As Integer = AudioVolume.Close()
        If Res <> wclErrors.WCL_E_SUCCESS Then Trace("Audio volume close failed", Res)

        Res = AudioMeter.Close()
        If Res <> wclErrors.WCL_E_SUCCESS Then Trace("Audio meter close failed", Res)
    End Sub

    Private Sub btOpen_Click(sender As System.Object, e As System.EventArgs) Handles btOpen.Click
        If cbEndpoints.SelectedIndex = -1 Then
            Trace("No audio endpoints found")
        Else
            Dim Id As String = CType(cbEndpoints.Items(cbEndpoints.SelectedIndex), AudioDeviceStorage).Device.Id

            Dim Res As Int32 = AudioMeter.Open(Id)
            If Res <> wclErrors.WCL_E_SUCCESS Then Trace("Audio meter open failed", Res)

            Res = AudioVolume.Open(Id)
            If Res <> wclErrors.WCL_E_SUCCESS Then Trace("Audio volume open failed", Res)
        End If
    End Sub

    Private Sub UpdateChannelsValue()
        Dim Value As Single
        Dim Res As Int32 = AudioMeter.GetPeak(Value)
        If Res = wclErrors.WCL_E_SUCCESS Then
            If tbMaster.Visible Then
                Value = Value * 100.0 * (tbMaster.Value / 100.0)
                pbMaster.Value = Value
            Else
                pbMaster.Value = Value * 100
            End If

            If FBars IsNot Nothing AndAlso FBars.Length > 0 Then
                Dim Values() As Single = Nothing
                Res = AudioMeter.GetChannelsPeak(Values)
                If Res = wclErrors.WCL_E_SUCCESS AndAlso Values IsNot Nothing AndAlso Values.Length > 0 Then
                    For i As Int32 = 0 To Values.Length - 1
                        FBars(i).Value = Values(i) * 100
                    Next i
                End If
            End If
        End If
    End Sub

    Private Sub Timer_Tick(sender As System.Object, e As System.EventArgs) Handles Timer.Tick
        UpdateChannelsValue()
    End Sub

    Private Sub AudioVolume_OnChanged(Sender As Object, Muted As Boolean, Volume As Single, Volumes() As Single) Handles AudioVolume.OnChanged
        If AudioVolume.Active Then
            Trace("Volume changed")

            cbMute.Checked = Muted
            tbMaster.Value = Volume * 100
        End If
    End Sub

    Private Sub AudioVolume_OnClosed(sender As Object, e As System.EventArgs) Handles AudioVolume.OnClosed
        Trace("Audio volume closed")
        cbEndpoints.Enabled = True
        tbMaster.Visible = False
        tbMaster.Value = 0
        cbMute.Visible = False
    End Sub

    Private Sub AudioVolume_OnDisconnected(sender As Object, e As System.EventArgs) Handles AudioVolume.OnDisconnected
        Trace("Audio volume device " + AudioVolume.Id + " disconnected")
    End Sub

    Private Sub AudioVolume_OnOpened(sender As Object, e As System.EventArgs) Handles AudioVolume.OnOpened
        Trace("Audio volume opened")
        cbEndpoints.Enabled = False

        Dim Features As wclAudioHardwareFeature
        Dim Res As Integer = AudioVolume.GetHardwareFeatures(Features)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            Trace("Get hardware features failed", Res)
        Else
            Trace("Hardware features")
            If (Features And wclAudioHardwareFeature.hfVolume) <> 0 Then Trace("  Volume")
            If (Features And wclAudioHardwareFeature.hfMute) <> 0 Then Trace("  Mute")
            If (Features And wclAudioHardwareFeature.hfMeter) <> 0 Then Trace("  Meter")
        End If

        Dim Volume As Single
        Res = AudioVolume.GetVolume(Volume)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            Trace("Get volume failed", Res)
        Else
            tbMaster.Visible = True
            tbMaster.Value = Volume * 100
        End If

        Dim Mute As Boolean
        Res = AudioVolume.GetMute(Mute)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            Trace("Get mute failed", Res)
        Else
            cbMute.Visible = True
            cbMute.Checked = Mute
        End If
    End Sub

    Private Sub tbMaster_ValueChanged(sender As System.Object, e As System.EventArgs) Handles tbMaster.ValueChanged
        If AudioVolume.Active Then
            Dim Res As Integer = AudioVolume.SetVolume(tbMaster.Value / 100.0)
            If Res <> wclErrors.WCL_E_SUCCESS Then Trace("Set volume failed", Res)
        End If
    End Sub

    Private Sub cbMute_CheckedChanged(sender As System.Object, e As System.EventArgs) Handles cbMute.CheckedChanged
        If AudioVolume.Active Then
            Dim Res As Integer = AudioVolume.SetMute(cbMute.Checked)
            If Res <> wclErrors.WCL_E_SUCCESS Then Trace("Set mute failed", Res)
        End If
    End Sub
End Class
