Public Class fmMain
    Dim WithEvents wclBluetoothManager As wclBluetoothManager
    Dim WithEvents wclBluetoothLeBeaconWatcher As wclBluetoothLeBeaconWatcher
    Dim WithEvents wclBluetoothLeAdvertiser As wclBluetoothLeAdvertiser

    Private Function GetRadio() As wclBluetoothRadio
        Dim Radio As wclBluetoothRadio = Nothing
        Dim Res As Int32 = wclBluetoothManager.GetLeRadio(Radio)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Get working radio failed: &H" + Res.ToString("X8"))
            Return Nothing
        End If

        Return Radio
    End Function

    Private Sub DumpData(ByVal Data() As Byte)
        If Data IsNot Nothing AndAlso Data.Length > 0 Then
            Dim Str As String = ""
            Dim c As Byte = 0
            For Each b As Byte In Data
                Str = Str + b.ToString("X2")
                If c = 15 Then
                    lbFrames.Items.Add("    " + Str)
                    Str = ""
                    c = 0
                Else
                    c = c + 1
                End If
            Next
            If Str <> "" Then
                lbFrames.Items.Add("    " + Str)
            End If
        End If
    End Sub

    Private Sub fmMain_FormClosing(ByVal sender As Object, ByVal e As System.Windows.Forms.FormClosingEventArgs) Handles Me.FormClosing
        wclBluetoothManager.Close()
    End Sub

    Private Sub fmMain_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load
        wclBluetoothManager = New wclBluetoothManager()
        wclBluetoothLeBeaconWatcher = New wclBluetoothLeBeaconWatcher()
        wclBluetoothLeAdvertiser = New wclBluetoothLeAdvertiser()

        edAdvInterval.Text = wclBluetoothConstants.WCL_BLE_DEFAULT_ADVERTISING_INTERVAL.ToString()

        DefaultScanParams()
    End Sub

    Private Sub btOpen_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btOpen.Click
        Dim Res As Int32 = wclBluetoothManager.Open()
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Error: &H" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
        End If
    End Sub

    Private Sub AddAdvertisement(ByVal Adv As wclBluetoothLeAdvertisement)
        Dim Res As Integer = wclBluetoothLeAdvertiser.Add(Adv)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            ListBox.Items.Add("Add advertisement failed: &H" + Res.ToString("8"))
        End If
    End Sub

    Private Sub btClose_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btClose.Click
        Dim Res As Int32 = wclBluetoothManager.Close()
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Error: &H" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
        End If
    End Sub

    Private Sub btStart_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btStart.Click
        Dim Radio As wclBluetoothRadio = GetRadio()
        If Not (Radio Is Nothing) Then
            Dim Mode As wclBluetoothLeScanningMode
            If cbScanMode.Checked Then
                Mode = wclBluetoothLeScanningMode.smActive
            Else
                Mode = wclBluetoothLeScanningMode.smPassive
            End If

            Dim Res As Int32 = wclBluetoothLeBeaconWatcher.Start(Radio, Mode, Convert.ToUInt16(edScanInterval.Text), Convert.ToUInt16(edScanWindow.Text))
            If Res <> wclErrors.WCL_E_SUCCESS Then
                MessageBox.Show("Error: &H" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
            End If
        End If
    End Sub

    Private Sub btStop_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btStop.Click
        If Not wclBluetoothManager.Active Then
            MessageBox.Show("Bluetooth Manager was not opened", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            Dim Res As Int32 = wclBluetoothLeBeaconWatcher.Stop()
            If Res <> wclErrors.WCL_E_SUCCESS Then
                MessageBox.Show("Error: &H" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
            End If
        End If
    End Sub

    Private Sub wclBluetoothManager_AfterOpen(ByVal sender As Object, ByVal e As System.EventArgs) Handles wclBluetoothManager.AfterOpen
        ListBox.Items.Add("Bluetooth Manager Opened")
    End Sub

    Private Sub wclBluetoothManager_BeforeClose(ByVal sender As Object, ByVal e As System.EventArgs) Handles wclBluetoothManager.BeforeClose
        ListBox.Items.Add("Bluetooth Manager Closed")
    End Sub

    Private Sub wclBluetoothManager_OnStatusChanged(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio) Handles wclBluetoothManager.OnStatusChanged
        Dim Str As String
        If Radio.Available Then
            Str = "AVAILABLE"
        Else
            Str = "UNAVAILABLE"
        End If
        ListBox.Items.Add("Radio " + Radio.ApiName + " status changed: " + Str)
    End Sub

    Private Sub btClear_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btClear.Click
        lbFrames.Items.Clear()
    End Sub

    Private Sub btClearLog_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btClearLog.Click
        ListBox.Items.Clear()
    End Sub

    Private Sub wclBluetoothLeBeaconWatcher_OnAdvertisementAppearanceFrame(Sender As Object, Address As Long, Timestamp As Long, Rssi As SByte, Appearance As UShort) Handles wclBluetoothLeBeaconWatcher.OnAdvertisementAppearanceFrame
        lbFrames.Items.Add("APPEARANCE FRAME")

        lbFrames.Items.Add("  Address: " + Address.ToString("X12"))
        lbFrames.Items.Add("  Timestamp: " + DateTime.FromFileTime(Timestamp).ToString())
        lbFrames.Items.Add("  RSSI: " + Rssi.ToString())

        lbFrames.Items.Add("  Appearance: " + Appearance.ToString("X4"))

        lbFrames.Items.Add("-------------------------------------------------------")
    End Sub

    Private Sub wclBluetoothLeBeaconWatcher_OnAdvertisementExtFrameInformation(Sender As Object, Address As Long, Timestamp As Long, Rssi As SByte, AddressType As wclBluetooth.wclBluetoothAddressType, TxPower As SByte, Flags As wclBluetooth.wclBluetoothLeExtendedFrameFlag) Handles wclBluetoothLeBeaconWatcher.OnAdvertisementExtFrameInformation
        lbFrames.Items.Add("EXTENDED FRAME INFORMATION")

        lbFrames.Items.Add("  Address: " + Address.ToString("X12"))
        lbFrames.Items.Add("  Timestamp: " + DateTime.FromFileTime(Timestamp).ToString())
        lbFrames.Items.Add("  RSSI: " + Rssi.ToString())

        Dim Str As String
        Select Case AddressType
            Case wclBluetoothAddressType.atPublic
                Str = "Public"
            Case wclBluetoothAddressType.atRandom
                Str = "Random"
            Case wclBluetoothAddressType.atUnspecified
                Str = "Unspecified"
            Case Else
                Str = "Unknown"
        End Select

        lbFrames.Items.Add("  Address type: " + Str)

        lbFrames.Items.Add("  TX Power: " + TxPower.ToString())

        Str = "("
        If (Flags And wclBluetoothLeExtendedFrameFlag.efAnonymous) <> 0 Then
            Str = Str + " efAnonymous"
        End If
        If (Flags And wclBluetoothLeExtendedFrameFlag.efConnectable) <> 0 Then
            Str = Str + " efConnectable"
        End If
        If (Flags And wclBluetoothLeExtendedFrameFlag.efDirected) <> 0 Then
            Str = Str + " efDirected"
        End If
        If (Flags And wclBluetoothLeExtendedFrameFlag.efScannable) <> 0 Then
            Str = Str + " efScannable"
        End If
        If (Flags And wclBluetoothLeExtendedFrameFlag.efScanResponse) <> 0 Then
            Str = Str + " efScanResponse"
        End If
        Str = Str + " )"
        lbFrames.Items.Add("  Flags: " + Str)

        lbFrames.Items.Add("-------------------------------------------------------")
    End Sub

    Private Sub wclBluetoothLeBeaconWatcher_OnAdvertisementFrameInformation(ByVal Sender As Object, ByVal Address As Long, ByVal Timestamp As Long, ByVal Rssi As SByte, ByVal Name As String, ByVal PacketType As wclBluetooth.wclBluetoothLeAdvertisementType, ByVal Flags As wclBluetooth.wclBluetoothLeAdvertisementFlag) Handles wclBluetoothLeBeaconWatcher.OnAdvertisementFrameInformation
        lbFrames.Items.Add("FRAME INFORMATION")

        lbFrames.Items.Add("  Address: " + Address.ToString("X12"))
        lbFrames.Items.Add("  Timestamp: " + DateTime.FromFileTime(Timestamp).ToString())
        lbFrames.Items.Add("  RSSI: " + Rssi.ToString())
        lbFrames.Items.Add("  Name: " + Name)

        Dim Str As String
        Select Case PacketType
            Case wclBluetoothLeAdvertisementType.atConnectableUndirected
                Str = "Connectable Undirected"
            Case wclBluetoothLeAdvertisementType.atConnectableDirected
                Str = "Connectable Directed"
            Case wclBluetoothLeAdvertisementType.atScannableUndirected
                Str = "Scannable Undirecte"
            Case wclBluetoothLeAdvertisementType.atNonConnectableUndirected
                Str = "Non Connectable Undirected"
            Case wclBluetoothLeAdvertisementType.atScanResponse
                Str = "Scan Response"
            Case Else
                Str = "Unknown"
        End Select
        lbFrames.Items.Add("  Frame type: " + Str)

        Str = "("
        If (Flags And wclBluetoothLeAdvertisementFlag.afLimitedDiscoverableMode) <> 0 Then
            Str = Str + " afLimitedDiscoverableMode"
        End If
        If (Flags And wclBluetoothLeAdvertisementFlag.afGeneralDiscoverableMode) <> 0 Then
            Str = Str + " afGeneralDiscoverableMode"
        End If
        If (Flags And wclBluetoothLeAdvertisementFlag.afClassicNotSupported) <> 0 Then
            Str = Str + " afClassicNotSupported"
        End If
        If (Flags And wclBluetoothLeAdvertisementFlag.afDualModeControllerCapable) <> 0 Then
            Str = Str + " afDualModeControllerCapable"
        End If
        If (Flags And wclBluetoothLeAdvertisementFlag.afDualModeHostCapable) <> 0 Then
            Str = Str + " afDualModeHostCapable"
        End If
        Str = Str + " )"
        lbFrames.Items.Add("  Flags: " + Str)

        lbFrames.Items.Add("-------------------------------------------------------")
    End Sub

    Private Sub wclBluetoothLeBeaconWatcher_OnAdvertisementRawFrame(ByVal Sender As Object, ByVal Address As Long, ByVal Timestamp As Long, ByVal Rssi As SByte, ByVal DataType As Byte, ByVal Data() As Byte) Handles wclBluetoothLeBeaconWatcher.OnAdvertisementRawFrame
        lbFrames.Items.Add("UNKNOWN RAW FRAME")

        lbFrames.Items.Add("  Address: " + Address.ToString("X12"))
        lbFrames.Items.Add("  Timestamp: " + DateTime.FromFileTime(Timestamp).ToString())
        lbFrames.Items.Add("  RSSI: " + Rssi.ToString())

        lbFrames.Items.Add("  Data type: " + DataType.ToString("X2"))

        DumpData(Data)

        lbFrames.Items.Add("-------------------------------------------------------")
    End Sub

    Private Sub wclBluetoothLeBeaconWatcher_OnEddystoneTlmFrame(ByVal Sender As Object, ByVal Address As Long, ByVal Timestamp As Long, ByVal Rssi As SByte, ByVal AdvCnt As UInteger, ByVal Batt As UShort, ByVal SecCnt As UInteger, ByVal Temp As Double, Data() As Byte) Handles wclBluetoothLeBeaconWatcher.OnEddystoneTlmFrame
        lbFrames.Items.Add("EDDYSTONE TLM FRAME")

        lbFrames.Items.Add("  Address: " + Address.ToString("X12"))
        lbFrames.Items.Add("  Timestamp: " + DateTime.FromFileTime(Timestamp).ToString())
        lbFrames.Items.Add("  RSSI: " + Rssi.ToString())

        lbFrames.Items.Add("  Advertisements counter: " + AdvCnt.ToString())
        lbFrames.Items.Add("  Battery (mv): " + Batt.ToString())
        lbFrames.Items.Add("  Seconds running: " + (SecCnt / 10).ToString())
        lbFrames.Items.Add("  Temperature: " + Temp.ToString())

        DumpData(Data)

        lbFrames.Items.Add("-------------------------------------------------------")
    End Sub

    Private Sub wclBluetoothLeBeaconWatcher_OnEddystoneUidFrame(ByVal Sender As Object, ByVal Address As Long, ByVal Timestamp As Long, ByVal Rssi As SByte, ByVal TxRssi As SByte, ByVal Uuid As System.Guid, Data() As Byte) Handles wclBluetoothLeBeaconWatcher.OnEddystoneUidFrame
        ' Calculate distance.
        Dim Dist As Double = Math.Pow(10.0, (TxRssi - 41.0 - Rssi) / 20.0)

        lbFrames.Items.Add("EDDYSTONE UID FRAME")

        lbFrames.Items.Add("  Address: " + Address.ToString("X12"))
        lbFrames.Items.Add("  Timestamp: " + DateTime.FromFileTime(Timestamp).ToString())
        lbFrames.Items.Add("  RSSI: " + Rssi.ToString())

        lbFrames.Items.Add("  UUID: " + Uuid.ToString())

        lbFrames.Items.Add("  TX RSSI: " + TxRssi.ToString())
        lbFrames.Items.Add("  Distance: " + Dist.ToString())

        DumpData(Data)

        lbFrames.Items.Add("-------------------------------------------------------")
    End Sub

    Private Sub wclBluetoothLeBeaconWatcher_OnAdvertisementUuidFrame(Sender As Object, Address As Long, Timestamp As Long, Rssi As SByte, Uuid As System.Guid) Handles wclBluetoothLeBeaconWatcher.OnAdvertisementUuidFrame
        lbFrames.Items.Add("UUID FRAME")

        lbFrames.Items.Add("  Address: " + Address.ToString("X12"))
        lbFrames.Items.Add("  Timestamp: " + DateTime.FromFileTime(Timestamp).ToString())
        lbFrames.Items.Add("  RSSI: " + Rssi.ToString())
        lbFrames.Items.Add("  UUID: " + Uuid.ToString())

        lbFrames.Items.Add("-------------------------------------------------------")
    End Sub

    Private Sub wclBluetoothLeBeaconWatcher_OnAltBeaconFrame(Sender As Object, Address As Long, Timestamp As Long, Rssi As SByte, CompanyId As UShort, Major As UShort, Minor As UShort, Uuid As System.Guid, TxRssi As SByte, Reserved As Byte, Data() As Byte) Handles wclBluetoothLeBeaconWatcher.OnAltBeaconFrame
        ' Clculate distance.
        Dim Dist As Double = Math.Pow(10.0, (TxRssi - Rssi) / 20.0)

        lbFrames.Items.Add("ALT BEACON FRAME")

        lbFrames.Items.Add("  Address: " + Address.ToString("X12"))
        lbFrames.Items.Add("  Company ID: " + CompanyId.ToString("X4"))
        lbFrames.Items.Add("  Timestamp: " + DateTime.FromFileTime(Timestamp).ToString())
        lbFrames.Items.Add("  RSSI: " + Rssi.ToString())

        lbFrames.Items.Add("  UUID: " + Uuid.ToString())
        lbFrames.Items.Add("  Major: " + Major.ToString("X4"))
        lbFrames.Items.Add("  Minor: " + Minor.ToString("X4"))
        lbFrames.Items.Add("  Reserved: " + Reserved.ToString("X2"))
        lbFrames.Items.Add("  TX RSSI: " + TxRssi.ToString())
        lbFrames.Items.Add("  Distance: " + Dist.ToString())

        DumpData(Data)

        lbFrames.Items.Add("-------------------------------------------------------")
    End Sub

    Private Sub wclBluetoothLeBeaconWatcher_OnEddystoneUrlFrame(ByVal Sender As Object, ByVal Address As Long, ByVal Timestamp As Long, ByVal Rssi As SByte, ByVal TxRssi As SByte, ByVal Url As String) Handles wclBluetoothLeBeaconWatcher.OnEddystoneUrlFrame
        ' Calculate distance.
        Dim Dist As Double = Math.Pow(10.0, (TxRssi - 41.0 - Rssi) / 20.0)

        lbFrames.Items.Add("EDDYSTONE URL FRAME")

        lbFrames.Items.Add("  Address: " + Address.ToString("X12"))
        lbFrames.Items.Add("  Timestamp: " + DateTime.FromFileTime(Timestamp).ToString())
        lbFrames.Items.Add("  RSSI: " + Rssi.ToString())

        lbFrames.Items.Add("  URL: " + Url)

        lbFrames.Items.Add("  TX RSSI: " + TxRssi.ToString())
        lbFrames.Items.Add("  Distance: " + Dist.ToString())

        lbFrames.Items.Add("-------------------------------------------------------")
    End Sub

    Private Sub wclBluetoothLeBeaconWatcher_OnManufacturerRawFrame(ByVal Sender As Object, ByVal Address As Long, ByVal Timestamp As Long, ByVal Rssi As SByte, ByVal CompanyId As UShort, ByVal Data() As Byte) Handles wclBluetoothLeBeaconWatcher.OnManufacturerRawFrame
        lbFrames.Items.Add("UNKNOWN MANUFACTURER RAW FRAME")

        lbFrames.Items.Add("  Address: " + Address.ToString("X12"))
        lbFrames.Items.Add("  Company ID: " + CompanyId.ToString("X4"))
        lbFrames.Items.Add("  Timestamp: " + DateTime.FromFileTime(Timestamp).ToString())
        lbFrames.Items.Add("  RSSI: " + Rssi.ToString())

        DumpData(Data)

        lbFrames.Items.Add("-------------------------------------------------------")
    End Sub

    Private Sub wclBluetoothLeBeaconWatcher_OnProximityBeaconFrame(ByVal Sender As Object, ByVal Address As Long, ByVal Timestamp As Long, ByVal Rssi As SByte, ByVal CompanyId As UShort, ByVal Major As UShort, ByVal Minor As UShort, ByVal Uuid As System.Guid, ByVal TxRssi As SByte, Data() As Byte) Handles wclBluetoothLeBeaconWatcher.OnProximityBeaconFrame
        ' Clculate distance.
        Dim Dist As Double = Math.Pow(10.0, (TxRssi - Rssi) / 20.0)

        lbFrames.Items.Add("PROXIMITY BEACON FRAME")

        lbFrames.Items.Add("  Address: " + Address.ToString("X12"))
        lbFrames.Items.Add("  Company ID: " + CompanyId.ToString("X4"))
        lbFrames.Items.Add("  Timestamp: " + DateTime.FromFileTime(Timestamp).ToString())
        lbFrames.Items.Add("  RSSI: " + Rssi.ToString())

        lbFrames.Items.Add("  UUID: " + Uuid.ToString())
        lbFrames.Items.Add("  Major: " + Major.ToString("X4"))
        lbFrames.Items.Add("  Minor: " + Minor.ToString("X4"))
        lbFrames.Items.Add("  TX RSSI: " + TxRssi.ToString())
        lbFrames.Items.Add("  Distance: " + Dist.ToString())

        DumpData(Data)

        lbFrames.Items.Add("-------------------------------------------------------")
    End Sub

    Private Sub wclBluetoothLeBeaconWatcher_OnStarted(ByVal sender As Object, ByVal e As System.EventArgs) Handles wclBluetoothLeBeaconWatcher.OnStarted
        ListBox.Items.Add("Beacons Monitoring Started")
    End Sub

    Private Sub wclBluetoothLeBeaconWatcher_OnStopped(ByVal sender As Object, ByVal e As System.EventArgs) Handles wclBluetoothLeBeaconWatcher.OnStopped
        ListBox.Items.Add("Beacons Monitoring Stopped")
    End Sub

    Private Sub btStartAdvertising_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btStartAdvertising.Click
        Dim BEACON_UUID As Guid = New Guid("{09039835-4A80-443B-87AA-DC565D09EA61}")

        wclBluetoothLeAdvertiser.Stop()

        Dim Radio As wclBluetoothRadio = GetRadio()
        If Not Radio Is Nothing Then
            Dim Adv As wclBluetoothLeAdvertisement

            ' Create and add iBeacon advdertisement.
            If cbIBeacon.Checked Then
                Adv = New wclBluetoothLeiBeaconAdvertisement(-5, &H101, &H202, BEACON_UUID)
                AddAdvertisement(Adv)
            End If

            ' Create and add Proximity Beacon advertisement.
            If cbProximityBeacon.Checked Then
                Adv = New wclBluetoothLeProximityBeaconAdvertisement(-5, &H101, &H202, BEACON_UUID, &HFFFE)
                AddAdvertisement(Adv)
            End If

            ' Create and add Alt Beacon advertisement.
            If cbAltBeacon.Checked Then
            Adv = New wclBluetoothLeAltBeaconAdvertisement(-5, &H101, &H202, BEACON_UUID, &HFFFE, &H11)
                AddAdvertisement(Adv)
            End If

            ' Create and add Eddystone UID advertisement.
            If cbEddystoneUid.Checked Then
                Adv = New wclBluetoothLeEddystoneUidBeaconAdvertisement(-5, BEACON_UUID)
                AddAdvertisement(Adv)
            End If

            ' Create and add Eddystone URL.
            If cbEddystoneUrl.Checked Then
                Adv = New wclBluetoothLeEddystoneUrlBeaconAdvertisement(-5, "https:'www.btframework.com")
                AddAdvertisement(Adv)
            End If

            ' Create and add custom advertisement.
            Dim Data As Byte()

            If cb128SolUuid.Checked Then
                ReDim Data(15)
                Data(0) = &HD0
                Data(1) = &H0
                Data(2) = &H2D
                Data(3) = &H12
                Data(4) = &H1E
                Data(5) = &H4B
                Data(6) = &HF
                Data(7) = &HA4
                Data(8) = &H99
                Data(9) = &H4E
                Data(10) = &HCE
                Data(11) = &HB5
                Data(12) = &H31
                Data(13) = &HF4
                Data(14) = &H5
                Data(15) = &H79
                Adv = New wclBluetoothLeCustomAdvertisement(wclUUIDs.LE_GAP_AD_TYPE_SERVICE_SOL_UUIDS_128, Data)
                AddAdvertisement(Adv)
            End If

            ' Create and add manufacturer specific advertisement.
            If cbManufacturer.Checked Then
                ReDim Data(1)
                Data(0) = &H12
                Data(1) = &H34
                Adv = New wclBluetoothLeManufacturerAdvertisement(&H10E, Data)
                AddAdvertisement(Adv)
            End If

            If cb16UuidService.Checked Then
                Adv = New wclBluetoothLe16ServiceAdvertisement(&H9835)
                AddAdvertisement(Adv)
            End If

            If cb32UuidService.Checked Then
                Adv = New wclBluetoothLe32ServiceAdvertisement(&H9039835)
                AddAdvertisement(Adv)
            End If

            If cb128UuidService.Checked Then
                Adv = New wclBluetoothLe128ServiceAdvertisement(BEACON_UUID)
                AddAdvertisement(Adv)
            End If

            If cb16UuidData.Checked Then
                ReDim Data(1)
                Data(0) = &H12
                Data(1) = &H34
                Adv = New wclBluetoothLe16ServiceDataAdvertisement(&H9835, Data)
                AddAdvertisement(Adv)
            End If

            If cb32UuidData.Checked Then
                ReDim Data(1)
                Data(0) = &H12
                Data(1) = &H34
                Adv = New wclBluetoothLe32ServiceDataAdvertisement(&H9039835, Data)
                AddAdvertisement(Adv)
            End If

            If cb128UuidData.Checked Then
                ReDim Data(1)
                Data(0) = &H12
                Data(1) = &H34
                Adv = New wclBluetoothLe128ServiceDataAdvertisement(BEACON_UUID, Data)
                AddAdvertisement(Adv)
            End If

            If cbCustomRaw.Checked Then
                ReDim Data(19)

                ' 16 bit Service Data
                Data(0) = &H9 ' Length.
                Data(1) = wclUUIDs.LE_GAP_AD_TYPE_SERVICE_DATA_16 ' Data type.
                Data(2) = &H12 ' 16 UUID LO BYTE
                Data(3) = &H34 ' 16 UUID HI BYTE
                Data(4) = &HFF ' Data 1st byte
                Data(5) = &HFE ' Data 2st byte
                Data(6) = &HFD ' Data 3st byte
                Data(7) = &HFC ' Data 4st byte
                Data(8) = &HFB ' Data 5st byte
                Data(9) = &HFA ' Data 6st byte
                ' Manufacturer specific data.
                Data(10) = &H9 ' Length.
                Data(11) = wclUUIDs.LE_GAP_AD_TYPE_MANUFACTURER ' Data type.
                Data(12) = &H10 ' Company ID.
                Data(13) = &H12 ' Company ID.
                Data(14) = &H1 ' Data 1st byte
                Data(15) = &H2 ' Data 2st byte
                Data(16) = &H3 ' Data 3st byte
                Data(17) = &H4 ' Data 4st byte
                Data(18) = &H5 ' Data 5st byte
                Data(19) = &H6 ' Data 6st byte

                Adv = New wclBluetoothLeRawAdvertisement(Data)
                AddAdvertisement(Adv)
            End If

            If wclBluetoothLeAdvertiser.Count = 0 Then
                MessageBox.Show("Select at least one advertisement type.")
            Else
                wclBluetoothLeAdvertiser.UseExtendedAdvertisement = cbUseExtended.Checked
                wclBluetoothLeAdvertiser.Anonymous = cbAnonymous.Checked
                wclBluetoothLeAdvertiser.IncludeTxRssi = cbIncludeTxRssi.Checked
                wclBluetoothLeAdvertiser.PreferredTxRssi = 10

                Dim Res As Int32 = wclBluetoothLeAdvertiser.Start(Radio, Convert.ToUInt16(edAdvInterval.Text))
                If Res <> wclErrors.WCL_E_SUCCESS Then
                    ListBox.Items.Add("Start advertiser failed: &H" + Res.ToString("8"))
                    wclBluetoothLeAdvertiser.Clear()
                End If
            End If
        End If
    End Sub

    Private Sub btStopAdvertising_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btStopAdvertising.Click
        Dim Res As Integer = wclBluetoothLeAdvertiser.Stop()
        If Res <> wclErrors.WCL_E_SUCCESS Then
            ListBox.Items.Add("Stop advertiser failed: &H" + Res.ToString("8"))
        End If
    End Sub

    Private Sub wclBluetoothLeBeaconWatcher_OnAdvertisementService128DataFrame(Sender As Object, Address As Long, Timestamp As Long, Rssi As SByte, Uuid As System.Guid, Data() As Byte) Handles wclBluetoothLeBeaconWatcher.OnAdvertisementService128DataFrame
        lbFrames.Items.Add("128 UUID SERVICE DATA")

        lbFrames.Items.Add("  Address: " + Address.ToString("X12"))
        lbFrames.Items.Add("  Timestamp: " + DateTime.FromFileTime(Timestamp).ToString())
        lbFrames.Items.Add("  RSSI: " + Rssi.ToString())
        lbFrames.Items.Add("  UUID: " + Uuid.ToString())

        DumpData(Data)

        lbFrames.Items.Add("-------------------------------------------------------")
    End Sub

    Private Sub wclBluetoothLeBeaconWatcher_OnAdvertisementService16DataFrame(Sender As Object, Address As Long, Timestamp As Long, Rssi As SByte, Uuid As UShort, Data() As Byte) Handles wclBluetoothLeBeaconWatcher.OnAdvertisementService16DataFrame
        lbFrames.Items.Add("16 UUID SERVICE DATA")

        lbFrames.Items.Add("  Address: " + Address.ToString("X12"))
        lbFrames.Items.Add("  Timestamp: " + DateTime.FromFileTime(Timestamp).ToString())
        lbFrames.Items.Add("  RSSI: " + Rssi.ToString())
        lbFrames.Items.Add("  UUID: " + Uuid.ToString("X4"))

        DumpData(Data)

        lbFrames.Items.Add("-------------------------------------------------------")
    End Sub

    Private Sub wclBluetoothLeBeaconWatcher_OnAdvertisementService32DataFrame(Sender As Object, Address As Long, Timestamp As Long, Rssi As SByte, Uuid As UInteger, Data() As Byte) Handles wclBluetoothLeBeaconWatcher.OnAdvertisementService32DataFrame
        lbFrames.Items.Add("32 UUID SERVICE DATA")

        lbFrames.Items.Add("  Address: " + Address.ToString("X12"))
        lbFrames.Items.Add("  Timestamp: " + DateTime.FromFileTime(Timestamp).ToString())
        lbFrames.Items.Add("  RSSI: " + Rssi.ToString())
        lbFrames.Items.Add("  UUID: " + Uuid.ToString("X8"))

        DumpData(Data)

        lbFrames.Items.Add("-------------------------------------------------------")
    End Sub

    Private Sub wclBluetoothLeBeaconWatcher_OnAdvertisementServiceSol128Frame(Sender As Object, Address As Long, Timestamp As Long, Rssi As SByte, Uuid As System.Guid) Handles wclBluetoothLeBeaconWatcher.OnAdvertisementServiceSol128Frame
        lbFrames.Items.Add("128 UUID SOL SERVICE")

        lbFrames.Items.Add("  Address: " + Address.ToString("X12"))
        lbFrames.Items.Add("  Timestamp: " + DateTime.FromFileTime(Timestamp).ToString())
        lbFrames.Items.Add("  RSSI: " + Rssi.ToString())
        lbFrames.Items.Add("  UUID: " + Uuid.ToString())

        lbFrames.Items.Add("-------------------------------------------------------")
    End Sub

    Private Sub wclBluetoothLeBeaconWatcher_OnAdvertisementServiceSol16Frame(Sender As Object, Address As Long, Timestamp As Long, Rssi As SByte, Uuid As UShort) Handles wclBluetoothLeBeaconWatcher.OnAdvertisementServiceSol16Frame
        lbFrames.Items.Add("16 UUID SOL SERVICE")

        lbFrames.Items.Add("  Address: " + Address.ToString("X12"))
        lbFrames.Items.Add("  Timestamp: " + DateTime.FromFileTime(Timestamp).ToString())
        lbFrames.Items.Add("  RSSI: " + Rssi.ToString())
        lbFrames.Items.Add("  UUID: " + Uuid.ToString("X4"))

        lbFrames.Items.Add("-------------------------------------------------------")
    End Sub

    Private Sub wclBluetoothLeBeaconWatcher_OnAdvertisementServiceSol32Frame(Sender As Object, Address As Long, Timestamp As Long, Rssi As SByte, Uuid As UInteger) Handles wclBluetoothLeBeaconWatcher.OnAdvertisementServiceSol32Frame
        lbFrames.Items.Add("32 UUID SOL SERVICE")

        lbFrames.Items.Add("  Address: " + Address.ToString("X12"))
        lbFrames.Items.Add("  Timestamp: " + DateTime.FromFileTime(Timestamp).ToString())
        lbFrames.Items.Add("  RSSI: " + Rssi.ToString())
        lbFrames.Items.Add("  UUID: " + Uuid.ToString("X8"))

        lbFrames.Items.Add("-------------------------------------------------------")
    End Sub

    Private Sub wclBluetoothLeAdvertiser_OnStarted(sender As Object, e As System.EventArgs) Handles wclBluetoothLeAdvertiser.OnStarted
        ListBox.Items.Add("Bluetooth LE Advertising Started")

        EnableAdvertisements(False)
    End Sub

    Private Sub wclBluetoothLeAdvertiser_OnStopped(sender As Object, e As System.EventArgs) Handles wclBluetoothLeAdvertiser.OnStopped
        ListBox.Items.Add("Bluetooth LE Advertising Stopped")
        wclBluetoothLeAdvertiser.Clear()

        EnableAdvertisements(True)
    End Sub

    Private Sub EnableAdvertisements(Enable As Boolean)
        cbIBeacon.Enabled = Enable
        cbProximityBeacon.Enabled = Enable
        cbAltBeacon.Enabled = Enable
        cbEddystoneUid.Enabled = Enable
        cbEddystoneUrl.Enabled = Enable
        cb128SolUuid.Enabled = Enable
        cbManufacturer.Enabled = Enable
        cb16UuidService.Enabled = Enable
        cb32UuidService.Enabled = Enable
        cb128UuidService.Enabled = Enable
        cb16UuidData.Enabled = Enable
        cb32UuidData.Enabled = Enable
        cb128UuidData.Enabled = Enable
    End Sub

    Private Sub wclBluetoothLeBeaconWatcher_OnAdvertisementReceived(Sender As Object, Address As Long, Timestamp As Long, Rssi As SByte, Data() As Byte) Handles wclBluetoothLeBeaconWatcher.OnAdvertisementReceived
        lbFrames.Items.Add("RAW ADVERTISEMENT FRAME")

        lbFrames.Items.Add("  Address: " + Address.ToString("X12"))
        lbFrames.Items.Add("  Timestamp: " + DateTime.FromFileTime(Timestamp).ToString())
        lbFrames.Items.Add("  RSSI: " + Rssi.ToString())

        DumpData(Data)

        lbFrames.Items.Add("-------------------------------------------------------")
    End Sub

    Private Sub DefaultScanParams()
        cbScanMode.Checked = True
        edScanInterval.Text = wclBluetoothConstants.WCL_BLE_DEFAULT_SCAN_INTERVAL.ToString()
        edScanWindow.Text = wclBluetoothConstants.WCL_BLE_DEFAULT_SCAN_WINDOW.ToString()
    End Sub

    Private Sub btDefault_Click(sender As System.Object, e As System.EventArgs) Handles btDefault.Click
        DefaultScanParams()
    End Sub

    Private Sub Button1_Click(sender As System.Object, e As System.EventArgs) Handles Button1.Click
        edAdvInterval.Text = wclBluetoothConstants.WCL_BLE_DEFAULT_ADVERTISING_INTERVAL.ToString()
    End Sub

    Private Sub wclBluetoothLeBeaconWatcher_OnAdvertisementTxPowerLevelFrame(Sender As Object, Address As Long, Timestamp As Long, Rssi As SByte, TxPower As SByte) Handles wclBluetoothLeBeaconWatcher.OnAdvertisementTxPowerLevelFrame
        lbFrames.Items.Add("TX POWER FRAME")

        lbFrames.Items.Add("  Address: " + Address.ToString("X12"))
        lbFrames.Items.Add("  Timestamp: " + DateTime.FromFileTime(Timestamp).ToString())
        lbFrames.Items.Add("  RSSI: " + Rssi.ToString())

        lbFrames.Items.Add("  TX Power: " + TxPower.ToString())

        lbFrames.Items.Add("-------------------------------------------------------")
    End Sub
End Class
