﻿Public Class fmMain
    Private WithEvents FSniffer As wclBleSniffer

    Private Sub fmMain_Load(sender As System.Object, e As System.EventArgs) Handles MyBase.Load
        cbChannel.SelectedIndex = 0
        FSniffer = New wclBleSniffer()
    End Sub

    Private Sub DumpHeader(Header As wclBluetoothLePacketHeader)
        lbPackets.Items.Add("   HEADER")
        lbPackets.Items.Add("  ========")
        lbPackets.Items.Add("    Timestamp: " + Header.Timestamp.ToString("X8"))
        lbPackets.Items.Add("    Access Address: " + Header.AccessAddress.ToString("X8"))
        lbPackets.Items.Add("    Channel: " + Header.Channel.ToString("X2"))
        lbPackets.Items.Add("    Crc: " + Header.Crc.ToString("X8"))
        lbPackets.Items.Add("    Rssi: " + Header.Rssi.ToString())
        If (Header.Valid) Then
            lbPackets.Items.Add("    PACKET: OK")
        Else
            lbPackets.Items.Add("    PACKET: ERROR")
        End If
    End Sub

    Private Sub DumpPduHeader(PduHeader As wclBluetoothLeAdvertisingPduHeader)
        DumpHeader(PduHeader.Header)

        lbPackets.Items.Add("   PDU HEADER")
        lbPackets.Items.Add("  ============")
        lbPackets.Items.Add("    ChSel: " + PduHeader.ChSel.ToString())
        lbPackets.Items.Add("    TxAdd: " + PduHeader.TxAdd.ToString())
        lbPackets.Items.Add("    RxAdd: " + PduHeader.RxAdd.ToString())
    End Sub

    Private Sub DumpPayload(Payload() As Byte)
        If Payload IsNot Nothing AndAlso Payload.Length > 0 Then
            Dim Str As String = ""
            Dim Ndx As Byte = 1
            For i = 0 To Payload.Length - 1
                Str = Str + Payload(i).ToString("X2") + " "
                Ndx = Ndx + 1
                If Ndx = 16 Then
                    lbPackets.Items.Add("    " + Str)
                    Str = ""
                    Ndx = 1
                End If
            Next i
            If Str <> "" Then
                lbPackets.Items.Add("    " + Str)
            End If
        End If
        lbPackets.Items.Add("")
    End Sub

    Private Sub FSniffer_OnAdvDirectIndReceived(Sender As Object, PduHeader As wclBluetooth.wclBluetoothLeAdvertisingPduHeader, AdvA As Long, TargetA As Long) Handles FSniffer.OnAdvDirectIndReceived
        lbPackets.Items.Add("ADV_DIRECT_IND RECEIVED")
        lbPackets.Items.Add("  AdvA: " + AdvA.ToString("X12"))
        lbPackets.Items.Add("  TargetA: " + TargetA.ToString("X12"))
        DumpPduHeader(PduHeader)
    End Sub

    Private Sub FSniffer_OnAdvIndReceived(Sender As Object, PduHeader As wclBluetooth.wclBluetoothLeAdvertisingPduHeader, AdvA As Long, AdvData() As Byte) Handles FSniffer.OnAdvIndReceived
        lbPackets.Items.Add("ADV_IND RECEIVED")
        lbPackets.Items.Add("  AdvA: " + AdvA.ToString("X12"))
        DumpPduHeader(PduHeader)
        DumpPayload(AdvData)
    End Sub

    Private Sub FSniffer_OnAdvNonConnIndReceived(Sender As Object, PduHeader As wclBluetooth.wclBluetoothLeAdvertisingPduHeader, AdvA As Long, AdvData() As Byte) Handles FSniffer.OnAdvNonConnIndReceived
        lbPackets.Items.Add("ADV_NONCONN_IND RECEIVED")
        lbPackets.Items.Add("  AdvA: " + AdvA.ToString("X12"))
        DumpPduHeader(PduHeader)
        DumpPayload(AdvData)
    End Sub

    Private Sub FSniffer_OnAdvScanIndReceived(Sender As Object, PduHeader As wclBluetooth.wclBluetoothLeAdvertisingPduHeader, AdvA As Long, AdvData() As Byte) Handles FSniffer.OnAdvScanIndReceived
        lbPackets.Items.Add("ADV_SCAN_IND RECEIVED")
        lbPackets.Items.Add("  AdvA: " + AdvA.ToString("X12"))
        DumpPduHeader(PduHeader)
        DumpPayload(AdvData)
    End Sub

    Private Sub FSniffer_OnConnectIndReceived(Sender As Object, PduHeader As wclBluetooth.wclBluetoothLeAdvertisingPduHeader, InitA As Long, AdvA As Long, LlData As wclBluetooth.wclBleSnifferLlData) Handles FSniffer.OnConnectIndReceived
        lbPackets.Items.Add("CONNECT_IND RECEIVED")
        lbPackets.Items.Add("  InitA: " + InitA.ToString("X12"))
        lbPackets.Items.Add("  AdvA: " + AdvA.ToString("X12"))
        lbPackets.Items.Add("  LLDATA")
        lbPackets.Items.Add("    Aa: " + LlData.Aa.ToString("X8"))
        lbPackets.Items.Add("    CrcInit: " + LlData.CrcInit.ToString("X8"))
        lbPackets.Items.Add("    WinSize: " + LlData.WinSize.ToString())
        lbPackets.Items.Add("    WinOffset: " + LlData.WinOffset.ToString())
        lbPackets.Items.Add("    Interval: " + LlData.Interval.ToString())
        lbPackets.Items.Add("    Latency: " + LlData.Latency.ToString())
        lbPackets.Items.Add("    Timeout: " + LlData.Timeout.ToString())
        lbPackets.Items.Add("    ChM: " + LlData.ChM.ToString("X16"))
        lbPackets.Items.Add("    Hop: " + LlData.Hop.ToString())
        lbPackets.Items.Add("    Sca: " + LlData.Sca.ToString())
        DumpPduHeader(PduHeader)
    End Sub

    Private Sub FSniffer_OnRawPacketReceived(Sender As Object, Header As wclBluetooth.wclBluetoothLePacketHeader, Payload() As Byte) Handles FSniffer.OnRawPacketReceived
        lbPackets.Items.Add("RAW PACKET RECEIVED")
        DumpHeader(Header)
        DumpPayload(Payload)
    End Sub

    Private Sub FSniffer_OnScanReqReceived(Sender As Object, PduHeader As wclBluetooth.wclBluetoothLeAdvertisingPduHeader, ScanA As Long, AdvA As Long) Handles FSniffer.OnScanReqReceived
        lbPackets.Items.Add("SCAN_REQ RECEIVED")
        lbPackets.Items.Add("  ScanA: " + ScanA.ToString("X12"))
        lbPackets.Items.Add("  AdvA: " + AdvA.ToString("X12"))
        DumpPduHeader(PduHeader)
    End Sub

    Private Sub FSniffer_OnScanRspReceived(Sender As Object, PduHeader As wclBluetooth.wclBluetoothLeAdvertisingPduHeader, AdvA As Long, ScanRspData() As Byte) Handles FSniffer.OnScanRspReceived
        lbPackets.Items.Add("SCAN_RSP RECEIVED")
        lbPackets.Items.Add("  AdvA: " + AdvA.ToString("X12"))
        DumpPduHeader(PduHeader)
        DumpPayload(ScanRspData)
    End Sub

    Private Sub FSniffer_OnStarted(sender As Object, e As System.EventArgs) Handles FSniffer.OnStarted
        lbPackets.Items.Add("Started")
    End Sub

    Private Sub FSniffer_OnStopped(sender As Object, e As System.EventArgs) Handles FSniffer.OnStopped
        lbPackets.Items.Add("Stopped")
    End Sub

    Private Sub fmMain_FormClosed(sender As System.Object, e As System.Windows.Forms.FormClosedEventArgs) Handles MyBase.FormClosed
        FSniffer.Stop()
        FSniffer = Nothing
    End Sub

    Private Sub btClear_Click(sender As System.Object, e As System.EventArgs) Handles btClear.Click
        lbPackets.Items.Clear()
    End Sub

    Private Sub btStart_Click(sender As System.Object, e As System.EventArgs) Handles btStart.Click
        Dim Res As Int32 = FSniffer.Start(cbChannel.SelectedIndex + 37)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Start failed: 0x" + Res.ToString("X8"))
        End If
    End Sub

    Private Sub btStop_Click(sender As System.Object, e As System.EventArgs) Handles btStop.Click
        Dim Res As Int32 = FSniffer.Stop()
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Stop failed: 0x" + Res.ToString("X8"))
        End If
    End Sub
End Class
