Public Class fmMain
    Private WithEvents Manager As wclBluetoothManager
    Private WithEvents FPowerMonitor As wclPowerEventsMonitor

    Private Sub btClear_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btClear.Click
        lvEvents.Items.Clear()
    End Sub

    Private Sub TraceEvent(ByVal Radio As wclBluetoothRadio, ByVal Ev As String, ByVal Parameter As String, ByVal Value As String)
        Dim Str As String
        If Radio IsNot Nothing Then
            Str = Radio.ApiName
        Else
            Str = ""
        End If

        Dim Item As ListViewItem = lvEvents.Items.Add(Str)

        Item.SubItems.Add(Ev)
        Item.SubItems.Add(Parameter)
        Item.SubItems.Add(Value)
    End Sub

    Private Sub fmMain_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load
        Manager = New wclBluetoothManager()

        FPowerMonitor = New wclPowerEventsMonitor()
        FPowerMonitor.Open()

        cbDiscoveringMode.SelectedIndex = 0
        cbMitmProtection.SelectedIndex = 3
        cbIoCap.SelectedIndex = 1
        cbBleProtection.SelectedIndex = 0
        cbPairingMethod.SelectedIndex = 0
    End Sub

    Private Sub Manager_AfterOpen(ByVal sender As Object, ByVal e As System.EventArgs) Handles Manager.AfterOpen
        TraceEvent(Nothing, "AfterOpen", "Handle pairing", Manager.HandlePairing.ToString())
    End Sub

    Private Sub Manager_BeforeClose(ByVal sender As Object, ByVal e As System.EventArgs) Handles Manager.BeforeClose
        TraceEvent(Nothing, "BeforeClose", "", "")
    End Sub

    Private Sub Manager_OnAuthenticationCompleted(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal Address As Long, ByVal [Error] As Integer) Handles Manager.OnAuthenticationCompleted
        TraceEvent(Radio, "AuthenticationCompleted", "Address", Address.ToString("X12"))
        TraceEvent(Nothing, "", "Error", "0x" + [Error].ToString("X8"))
    End Sub

    Private Sub Manager_OnClosed(sender As Object, e As System.EventArgs) Handles Manager.OnClosed
        ClearRadios()

        TraceEvent(Nothing, "Closed", "", "")
    End Sub

    Private Sub Manager_OnConfirm(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal Address As Long, ByRef Confirm As Boolean) Handles Manager.OnConfirm
        Confirm = True

        TraceEvent(Radio, "Just Works pairing", "Address", Address.ToString("X12"))
    End Sub

    Private Sub Manager_OnDeviceFound(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal Address As Long) Handles Manager.OnDeviceFound
        TraceEvent(Radio, "DeviceFound", "Address", Address.ToString("X12"))

        AddDevice(Radio, Address)
    End Sub

    Private Sub Manager_OnDiscoveringCompleted(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal [Error] As Integer) Handles Manager.OnDiscoveringCompleted
        TraceEvent(Radio, "DiscoveringCompleted", "Error", "0x" + [Error].ToString("X8"))

        RefreshDevices()
    End Sub

    Private Sub Manager_OnDiscoveringStarted(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio) Handles Manager.OnDiscoveringStarted
        ClearDevices()

        TraceEvent(Radio, "DiscoveringStarted", "", "")
    End Sub

    Private Sub Manager_OnIoCapabilityRequest(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal Address As Long, ByRef Mitm As wclBluetooth.wclBluetoothMitmProtection, ByRef IoCapability As wclBluetooth.wclBluetoothIoCapability, ByRef OobPresent As Boolean) Handles Manager.OnIoCapabilityRequest
        Select Case cbMitmProtection.SelectedIndex
            Case 0
                    Mitm = wclBluetoothMitmProtection.mitmProtectionNotRequired
            Case 1
                    Mitm = wclBluetoothMitmProtection.mitmProtectionRequired
            Case 2
                    Mitm = wclBluetoothMitmProtection.mitmProtectionNotRequiredBonding
            Case 3
                    Mitm = wclBluetoothMitmProtection.mitmProtectionRequiredBonding
            Case 4
                    Mitm = wclBluetoothMitmProtection.mitmProtectionNotRequiredGeneralBonding
            Case 5
                    Mitm = wclBluetoothMitmProtection.mitmProtectionRequiredGeneralBonding
            Case Else
                Mitm = wclBluetoothMitmProtection.mitmProtectionNotDefined
        End Select

        Select Case cbIoCap.SelectedIndex
            Case 0
                    IoCapability = wclBluetoothIoCapability.iocapDisplayOnly
            Case 1
                    IoCapability = wclBluetoothIoCapability.iocapDisplayYesNo
            Case 2
                    IoCapability = wclBluetoothIoCapability.iocapKeyboardOnly
            Case 3
                    IoCapability = wclBluetoothIoCapability.iocapNoInputNoOutput
            Case 4
                    IoCapability = wclBluetoothIoCapability.iocapDisplayKeyboard
            Case Else
                    IoCapability = wclBluetoothIoCapability.iocapNotDefined
        End Select

        ' To accept OOB pairing set this parameter to True.
        OobPresent = False
    End Sub

    Private Sub Manager_OnNumericComparison(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal Address As Long, ByVal Number As UInteger, ByRef Confirm As Boolean) Handles Manager.OnNumericComparison
        Confirm = True

        TraceEvent(Radio, "NumericComparison", "Address", Address.ToString("X12"))
        TraceEvent(Nothing, "", "Number", Number.ToString())
        TraceEvent(Nothing, "", "Confirm", Confirm.ToString())
    End Sub

    Private Sub Manager_OnOobDataRequest(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal Address As Long, ByRef OobData As wclBluetooth.wclBluetoothOobData) Handles Manager.OnOobDataRequest
        ' This event fires when a remote device requests OOB data.
        OobData = Nothing
    End Sub

    Private Sub Manager_OnPasskeyNotification(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal Address As Long, ByVal Passkey As UInteger) Handles Manager.OnPasskeyNotification
        TraceEvent(Radio, "PasskeyNotification", "Address", Address.ToString("X12"))
        TraceEvent(Nothing, "", "Passkey", Passkey.ToString())
    End Sub

    Private Sub Manager_OnPasskeyRequest(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal Address As Long, ByRef Passkey As UInteger) Handles Manager.OnPasskeyRequest
        Passkey = 123456

        TraceEvent(Radio, "PasskeyRequest", "Address", Address.ToString())
        TraceEvent(Nothing, "", "Passkey", Passkey.ToString())
    End Sub

    Private Sub Manager_OnPinRequest(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal Address As Long, ByRef Pin As String) Handles Manager.OnPinRequest
        Pin = edPIN.Text

        TraceEvent(Radio, "PinRequest", "Address", Address.ToString("X12"))
        TraceEvent(Nothing, "", "PIN", Pin)
    End Sub

    Private Sub Manager_OnProtectionLevelRequest(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal Address As Long, ByRef Protection As wclBluetooth.wclBluetoothLeProtectionLevel) Handles Manager.OnProtectionLevelRequest
        Select Case cbBleProtection.SelectedIndex
            Case 0
                Protection = wclBluetoothLeProtectionLevel.pplDefault
            Case 1
                Protection = wclBluetoothLeProtectionLevel.pplNone
            Case 2
                Protection = wclBluetoothLeProtectionLevel.pplEncryption
            Case 3
                Protection = wclBluetoothLeProtectionLevel.pplEncryptionAndAuthentication
            Case Else
                Protection = wclBluetoothLeProtectionLevel.pplDefault
        End Select
    End Sub

    Private Sub Manager_OnStatusChanged(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio) Handles Manager.OnStatusChanged
        Dim Status As String
        If Radio.Available Then
            Status = "Available"
        Else
            If Radio.Plugged Then
                Status = "Not available, Plugged"
            Else
                Status = "Unplugged"
            End If
        End If
        TraceEvent(Radio, "StatusChanged", "New status", Status)

        Dim Found As Boolean = False
        For Each Item As ListViewItem In lvRadios.Items
            If Item.Tag.Equals(Radio) Then
                Found = True
                RefreshRadio(Item)
                Exit For
            End If
        Next Item

        If Not Found Then
            Dim Item As ListViewItem = lvRadios.Items.Add("") ' Api
            Item.Tag = Radio

            Item.SubItems.Add("") ' Available
            Item.SubItems.Add("") ' Address
            Item.SubItems.Add("") ' Connectable
            Item.SubItems.Add("") ' Discoverable
            Item.SubItems.Add("") ' HCI version
            Item.SubItems.Add("") ' LMP version
            Item.SubItems.Add("") ' Manufacturer
            Item.SubItems.Add("") ' Name
            Item.SubItems.Add("") ' COD

            RefreshRadio(Item)
        End If
    End Sub

    Private Sub fmMain_FormClosed(ByVal sender As System.Object, ByVal e As System.Windows.Forms.FormClosedEventArgs) Handles MyBase.FormClosed
        Manager.Close()
        Manager = Nothing

        FPowerMonitor.Close()
        FPowerMonitor = Nothing
    End Sub

    Private Sub ClearRadios()
        lvRadios.Items.Clear()
        ClearDevices()
    End Sub

    Private Sub RefreshRadio(ByVal Item As ListViewItem)
        Dim Radio As wclBluetoothRadio = CType(Item.Tag, wclBluetoothRadio)

        Item.Text = Radio.ApiName
        Item.SubItems(1).Text = Radio.Available.ToString()

        Dim Address As Int64 = 0
        Dim Res As Int32 = Radio.GetAddress(Address)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            Item.SubItems(2).Text = "Error: 0x" + Res.ToString("X8")
        Else
            Item.SubItems(2).Text = Address.ToString("X12")
        End If

        Dim BoolVal As Boolean
        Res = Radio.GetConnectable(BoolVal)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            Item.SubItems(3).Text = "Error: 0x" + Res.ToString("X8")
        Else
            Item.SubItems(3).Text = BoolVal.ToString()
        End If

        Res = Radio.GetDiscoverable(BoolVal)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            Item.SubItems(4).Text = "Error: 0x" + Res.ToString("X8")
        Else
            Item.SubItems(4).Text = BoolVal.ToString()
        End If

        Dim ByteVal As Byte
        Dim WordVal As UInt16
        Res = Radio.GetHciVersion(ByteVal, WordVal)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            Item.SubItems(5).Text = "Error: 0x" + Res.ToString("X8")
        Else
            Item.SubItems(5).Text = ByteVal.ToString() + "." + WordVal.ToString()
        End If

        Res = Radio.GetLmpVersion(ByteVal, WordVal)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            Item.SubItems(6).Text = "Error: 0x" + Res.ToString("X8")
        Else
            Item.SubItems(6).Text = ByteVal.ToString() + "." + WordVal.ToString()
        End If

        Res = Radio.GetManufacturer(WordVal)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            Item.SubItems(7).Text = "Error: 0x" + Res.ToString("X8")
        Else
            Item.SubItems(7).Text = WordVal.ToString()
        End If

        Dim StrVal As String = ""
        Res = Radio.GetName(StrVal)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            Item.SubItems(8).Text = "Error: 0x" + Res.ToString("X8")
        Else
            Item.SubItems(8).Text = StrVal
        End If

        Dim Cod As UInt32 = 0
        Res = Radio.GetCod(Cod)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            Item.SubItems(9).Text = "Error: 0x" + Res.ToString("X8")
        Else
            Item.SubItems(9).Text = "0x" + Cod.ToString("X8")
        End If
    End Sub

    Private Sub btRefreshRadio_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btRefreshRadio.Click
        If lvRadios.SelectedItems.Count = 0 Then
            MessageBox.Show("Select radio", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            RefreshRadio(lvRadios.SelectedItems(0))
        End If
    End Sub

    Private Sub btOpen_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btOpen.Click
        Dim Res As Int32 = Manager.Open(cbHandlePairing.Checked)
        If Res <> wclErrors.WCL_E_SUCCESS Then ShowError(Res)
    End Sub

    Private Sub btClose_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btClose.Click
        Dim Res As Int32 = Manager.Close()
        If Res <> wclErrors.WCL_E_SUCCESS Then ShowError(Res)
    End Sub

    Private Sub ShowError(ByVal [Error] As Int32)
        MessageBox.Show("Error: 0x" + [Error].ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
    End Sub

    Private Sub btSetConnectable_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btSetConnectable.Click
        If lvRadios.SelectedItems.Count = 0 Then
            MessageBox.Show("Select radio", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            Dim Mr As DialogResult = MessageBox.Show("(YES) - turn connectable ON, (NO) - turn connectable OFF", "Set Connectable", MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question)
            If Mr <> DialogResult.Cancel Then
                Dim Item As ListViewItem = lvRadios.SelectedItems(0)
                Dim Radio As wclBluetoothRadio = CType(Item.Tag, wclBluetoothRadio)

                Dim Res As Int32 = Radio.SetConnectable(Mr = DialogResult.Yes)
                If Res <> wclErrors.WCL_E_SUCCESS Then
                    ShowError(Res)
                End If

                RefreshRadio(Item)
            End If
        End If
    End Sub

    Private Sub btSetDiscoverable_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btSetDiscoverable.Click
        If lvRadios.SelectedItems.Count = 0 Then
            MessageBox.Show("Select radio", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            Dim Mr As DialogResult = MessageBox.Show("(YES) - turn discoverable ON, (NO) - turn discoverable OFF", "Set discoverable", MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question)
            If Mr <> DialogResult.Cancel Then
                Dim Item As ListViewItem = lvRadios.SelectedItems(0)
                Dim Radio As wclBluetoothRadio = CType(Item.Tag, wclBluetoothRadio)

                Dim Res As Int32 = Radio.SetDiscoverable(Mr = DialogResult.Yes)
                If Res <> wclErrors.WCL_E_SUCCESS Then
                    ShowError(Res)
                End If

                RefreshRadio(Item)
            End If
        End If
    End Sub

    Private Sub ntSetName_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles ntSetName.Click
        If lvRadios.SelectedItems.Count = 0 Then
            MessageBox.Show("Select radio", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            Dim Item As ListViewItem = lvRadios.SelectedItems(0)
            Dim Radio As wclBluetoothRadio = CType(Item.Tag, wclBluetoothRadio)

            Dim Res As Int32 = Radio.SetName(edName.Text)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                ShowError(Res)
            End If

            RefreshRadio(Item)
        End If
    End Sub

    Private Sub btEnumPaired_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btEnumPaired.Click
        If lvRadios.SelectedItems.Count = 0 Then
            MessageBox.Show("Select radio", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            ClearDevices()

            Dim Item As ListViewItem = lvRadios.SelectedItems(0)
            Dim Radio As wclBluetoothRadio = CType(Item.Tag, wclBluetoothRadio)

            Dim Devices As Int64() = Nothing
            Dim Res As Int32 = Radio.EnumPairedDevices(Devices)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                ShowError(Res)
            Else
                If Devices IsNot Nothing Then
                    For i As Int32 = 0 To Devices.Length - 1
                        AddDevice(Radio, Devices(i))
                    Next i

                    RefreshDevices()
                End If
            End If
        End If
    End Sub

    Private Sub ClearDevices()
        lvDevices.Items.Clear()
        ClearServices()
    End Sub

    Private Sub RefreshDevice(ByVal Item As ListViewItem)
        Dim Radio As wclBluetoothRadio = CType(Item.Tag, wclBluetoothRadio)
        Dim Address As Int64 = Convert.ToInt64(Item.SubItems(1).Text, 16)

        Dim Cod As UInt32
        Dim Res As Int32 = Radio.GetRemoteCod(Address, Cod)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            Item.SubItems(2).Text = "Error: 0x" + Res.ToString("X8")
        Else
            Item.SubItems(2).Text = Cod.ToString("X8")
        End If

        Dim DevType As wclBluetoothDeviceType
        Res = Radio.GetRemoteDeviceType(Address, DevType)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            Item.SubItems(3).Text = "Error: 0x" + Res.ToString("X8")
        Else
            Select Case DevType
                Case wclBluetoothDeviceType.dtClassic
                    Item.SubItems(3).Text = "Classic"
                Case wclBluetoothDeviceType.dtBle
                    Item.SubItems(3).Text = "BLE"
                Case wclBluetoothDeviceType.dtMixed
                    Item.SubItems(3).Text = "Mixed"
                Case Else
                    Item.SubItems(3).Text = "Unknown"
            End Select
        End If

        Dim DevName As String = ""
        Res = Radio.GetRemoteName(Address, DevName)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            Item.SubItems(4).Text = "Error: 0x" + Res.ToString("X8")
        Else
            Item.SubItems(4).Text = DevName
        End If

        Dim Paired As Boolean
        Res = Radio.GetRemotePaired(Address, Paired)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            Item.SubItems(5).Text = "Error: 0x" + Res.ToString("X8")
        Else
            Item.SubItems(5).Text = Paired.ToString()
        End If

        Dim Connected As Boolean
        Res = Radio.GetRemoteConnectedStatus(Address, Connected)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            Item.SubItems(6).Text = "Error: 0x" + Res.ToString("X8")
        Else
            Item.SubItems(6).Text = Connected.ToString()
        End If

        Dim AddrType As wclBluetoothAddressType = wclBluetoothAddressType.atUnspecified
        Res = Radio.GetRemoteAddressType(Address, AddrType)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            Item.SubItems(7).Text = "Error: 0x" + Res.ToString("X8")
        Else
            Select Case AddrType
                Case wclBluetoothAddressType.atClassic
                    Item.SubItems(7).Text = "Classic"
                Case wclBluetoothAddressType.atPublic
                    Item.SubItems(7).Text = "LE Public:"
                Case wclBluetoothAddressType.atRandom
                    Item.SubItems(7).Text = "LE Random"
                Case wclBluetoothAddressType.atUnspecified
                    Item.SubItems(7).Text = "Unspecified"
                Case Else
                    Item.SubItems(7).Text = "Unknown"
            End Select
        End If
    End Sub

    Private Sub btRefreshDevice_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btRefreshDevice.Click
        If lvDevices.SelectedItems.Count = 0 Then
            MessageBox.Show("Select device", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            RefreshDevice(lvDevices.SelectedItems(0))
        End If
    End Sub

    Private Sub ClearServices()
        lvServices.Items.Clear()
        ClearComPorts()
    End Sub

    Private Sub btEnumServices_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btEnumServices.Click
        If lvDevices.SelectedItems.Count = 0 Then
            MessageBox.Show("Select device", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            ClearServices()

            Dim Item As ListViewItem = lvDevices.SelectedItems(0)
            Dim Radio As wclBluetoothRadio = CType(Item.Tag, wclBluetoothRadio)
            Dim Address As Int64 = Convert.ToInt64(Item.SubItems(1).Text, 16)

            Dim Services() As wclBluetoothService = Nothing
            Dim g As Guid = Guid.Empty
            Dim Res As Int32 = Radio.EnumRemoteServices(Address, g, Services)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                ShowError(Res)
            Else
                If Services IsNot Nothing AndAlso Services.Length > 0 Then
                    For Each Service As wclBluetoothService In Services
                        Item = lvServices.Items.Add(Radio.ApiName)
                        Item.SubItems.Add(Address.ToString("X12"))
                        Item.SubItems.Add(Service.Handle.ToString("X8"))
                        Item.SubItems.Add(Service.Uuid.ToString())
                        Item.SubItems.Add(Service.Channel.ToString())
                        Item.SubItems.Add(Service.Name)
                        Item.SubItems.Add(Service.Comment)
                    Next Service
                End If
            End If
        End If
    End Sub

    Private Sub btDiscover_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btDiscover.Click
        If lvRadios.SelectedItems.Count = 0 Then
            MessageBox.Show("Select radio", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            Dim Item As ListViewItem = lvRadios.SelectedItems(0)
            Dim Radio As wclBluetoothRadio = CType(Item.Tag, wclBluetoothRadio)

            Dim Kind As wclBluetoothDiscoverKind
            If cbDiscoveringMode.SelectedIndex = 0 Then
                Kind = wclBluetoothDiscoverKind.dkClassic
            Else
                Kind = wclBluetoothDiscoverKind.dkBle
            End If
            Dim Res As Int32 = Radio.Discover(Convert.ToByte(edTimeout.Text), Kind)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                ShowError(Res)
            End If
        End If
    End Sub

    Private Sub btTerminate_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btTerminate.Click
        If lvRadios.SelectedItems.Count = 0 Then
            MessageBox.Show("Select radio", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            Dim Item As ListViewItem = lvRadios.SelectedItems(0)
            Dim Radio As wclBluetoothRadio = CType(Item.Tag, wclBluetoothRadio)

            Dim Res As Int32 = Radio.Terminate()
            If Res <> wclErrors.WCL_E_SUCCESS Then
                ShowError(Res)
            End If
        End If
    End Sub

    Private Sub AddDevice(ByVal Radio As wclBluetoothRadio, ByVal Address As Int64)
        Dim Item As ListViewItem = lvDevices.Items.Add(Radio.ApiName)
        Item.Tag = Radio

        Item.SubItems.Add(Address.ToString("X12"))
        Item.SubItems.Add("") ' Class Of Device
        Item.SubItems.Add("") ' Device type
        Item.SubItems.Add("") ' Name
        Item.SubItems.Add("") ' Paired
        Item.SubItems.Add("") ' Connected
        Item.SubItems.Add("") ' Address type
    End Sub

    Private Sub RefreshDevices()
        For Each Item As ListViewItem In lvDevices.Items
            RefreshDevice(Item)
        Next Item
    End Sub

    Private Sub btUnpair_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btUnpair.Click
        If lvDevices.SelectedItems.Count = 0 Then
            MessageBox.Show("Select device", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            Dim Item As ListViewItem = lvDevices.SelectedItems(0)
            Dim Radio As wclBluetoothRadio = CType(Item.Tag, wclBluetoothRadio)
            Dim Address As Int64 = Convert.ToInt64(Item.SubItems(1).Text, 16)

            Dim Method As wclBluetoothPairingMethod
            Select Case cbPairingMethod.SelectedIndex
                Case 1
                    Method = wclBluetoothPairingMethod.pmClassic
                Case 2
                    Method = wclBluetoothPairingMethod.pmLe
                Case Else
                    Method = wclBluetoothPairingMethod.pmAuto
            End Select

            Dim Res As Int32 = Radio.RemoteUnpair(Address, Method)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                ShowError(Res)
            End If

            RefreshDevice(Item)
        End If
    End Sub

    Private Sub btPair_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btPair.Click
        If lvDevices.SelectedItems.Count = 0 Then
            MessageBox.Show("Select device", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            Dim Item As ListViewItem = lvDevices.SelectedItems(0)
            Dim Radio As wclBluetoothRadio = CType(Item.Tag, wclBluetoothRadio)
            Dim Address As Int64 = Convert.ToInt64(Item.SubItems(1).Text, 16)

            Dim Method As wclBluetoothPairingMethod
            Select Case cbPairingMethod.SelectedIndex
                Case 1
                    Method = wclBluetoothPairingMethod.pmClassic
                Case 2
                    Method = wclBluetoothPairingMethod.pmLe
                Case Else
                    Method = wclBluetoothPairingMethod.pmAuto
            End Select

            Dim Res As Int32 = Radio.RemotePair(Address, Method)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                ShowError(Res)
            End If
        End If
    End Sub

    Private Sub btRssi_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btRssi.Click
        If lvDevices.SelectedItems.Count = 0 Then
            MessageBox.Show("Select device", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            Dim Item As ListViewItem = lvDevices.SelectedItems(0)
            Dim Radio As wclBluetoothRadio = CType(Item.Tag, wclBluetoothRadio)
            Dim Address As Int64 = Convert.ToInt64(Item.SubItems(1).Text, 16)

            Dim Rssi As SByte
            Dim Res As Int32 = Radio.GetRemoteRssi(Address, Rssi)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                ShowError(Res)
            Else
                MessageBox.Show("RSSI: " + Rssi.ToString() + "dB", "RSSI", MessageBoxButtons.OK, MessageBoxIcon.Information)
            End If
        End If
    End Sub

    Private Sub ntTurnOff_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles ntTurnOff.Click
        If lvRadios.SelectedItems.Count = 0 Then
            MessageBox.Show("Select radio", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            Dim Item As ListViewItem = lvRadios.SelectedItems(0)
            Dim Radio As wclBluetoothRadio = CType(Item.Tag, wclBluetoothRadio)

            Dim Res As Int32 = Radio.TurnOff()
            If Res <> wclErrors.WCL_E_SUCCESS Then
                ShowError(Res)
            End If

            RefreshRadio(Item)
        End If
    End Sub

    Private Sub btTurnOn_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btTurnOn.Click
        If lvRadios.SelectedItems.Count = 0 Then
            MessageBox.Show("Select radio", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            Dim Item As ListViewItem = lvRadios.SelectedItems(0)
            Dim Radio As wclBluetoothRadio = CType(Item.Tag, wclBluetoothRadio)

            Dim Res As Int32 = Radio.TurnOn()
            If Res <> wclErrors.WCL_E_SUCCESS Then
                ShowError(Res)
            End If

            RefreshRadio(Item)
        End If
    End Sub

    Private Sub ClearComPorts()
        lvComPorts.Items.Clear()
    End Sub

    Private Sub btEnumVCOMs_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btEnumVCOMs.Click
        If lvRadios.SelectedItems.Count = 0 Then
            MessageBox.Show("Select radio", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            RefreshVComs()
        End If
    End Sub

    Private Sub RefreshVComs()
        ClearComPorts()

        Dim Item As ListViewItem = lvRadios.SelectedItems(0)
        Dim Radio As wclBluetoothRadio = CType(Item.Tag, wclBluetoothRadio)

        Dim Res As Int32
        Dim Ports As wclVirtualComPort() = Nothing
        If cbEnumPreinstalled.Checked Then
            Res = Radio.EnumComPorts(Ports)
        Else
            Ports = Radio.ComPorts
            Res = wclErrors.WCL_E_SUCCESS
        End If

        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Error enumerating vCOMs: 0x" + Res.ToString("X8"))
        Else
            If Ports IsNot Nothing AndAlso Ports.Length > 0 Then
                For i As Integer = 0 To Ports.Length - 1
                    Dim Port As wclVirtualComPort = Ports(i)

                    Item = lvComPorts.Items.Add(Radio.ApiName)
                    Item.SubItems.Add(Port.Address.ToString("X12"))
                    Item.SubItems.Add(Port.Service.ToString())
                    Item.SubItems.Add(Port.Number.ToString())

                    Item.Tag = Radio
                Next i
            End If
        End If
    End Sub

    Private Sub btCreateVCOM_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btCreateVCOM.Click
        If lvDevices.SelectedItems.Count = 0 Then
            MessageBox.Show("Select device", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            If lvServices.SelectedItems.Count = 0 Then
                MessageBox.Show("Select service", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
            Else
                Dim Item As ListViewItem = lvDevices.SelectedItems(0)
                Dim Radio As wclBluetoothRadio = CType(Item.Tag, wclBluetoothRadio)
                Dim Address As Int64 = Convert.ToInt64(Item.SubItems(1).Text, 16)
                Dim Service As Guid = New Guid(lvServices.SelectedItems(0).SubItems(3).Text)

                Dim Number As UInt16
                Dim Res As Int32 = Radio.CreateComPort(Address, Service, Number)
                If Res <> wclErrors.WCL_E_SUCCESS Then
                    ShowError(Res)
                Else
                    TraceEvent(Radio, "Create vCOM", "Number", Number.ToString())

                    RefreshVComs()
                End If
            End If
        End If
    End Sub

    Private Sub btDestroyVCOM_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btDestroyVCOM.Click
        If lvComPorts.SelectedItems.Count = 0 Then
            MessageBox.Show("Select vCOM", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            Dim Item As ListViewItem = lvComPorts.SelectedItems(0)
            Dim Radio As wclBluetoothRadio = CType(Item.Tag, wclBluetoothRadio)
            Dim Number As UInt16 = Convert.ToUInt16(Item.SubItems(3).Text)

            Dim Res As Int32 = Radio.DestroyComPort(Number)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                ShowError(Res)
            Else
                TraceEvent(Radio, "Destroy vCOM", "Number", Number.ToString())

                RefreshVComs()
            End If
        End If
    End Sub

    Private Sub btIntsallDevice_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btIntsallDevice.Click
        If lvDevices.SelectedItems.Count = 0 Then
            MessageBox.Show("Select device", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            If lvServices.SelectedItems.Count = 0 Then
                MessageBox.Show("Select service", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
            Else
                Dim Item As ListViewItem = lvDevices.SelectedItems(0)
                Dim Radio As wclBluetoothRadio = CType(Item.Tag, wclBluetoothRadio)
                Dim Address As Int64 = Convert.ToInt64(Item.SubItems(1).Text, 16)
                Dim Service As Guid = New Guid(lvServices.SelectedItems(0).SubItems(3).Text)

                Dim Res As Int32 = Radio.InstallDevice(Address, Service)
                If Res = wclBluetoothErrors.WCL_E_BLUETOOTH_DEVICE_ALREADY_INSTALLED Then
                    ' Reinstall to reconnect!
                    Radio.UninstallDevice(Address, Service)
                    Res = Radio.InstallDevice(Address, Service)
                End If

                If Res <> wclErrors.WCL_E_SUCCESS Then
                    ShowError(Res)
                Else
                    MessageBox.Show("Device installed")
                End If
            End If
        End If
    End Sub

    Private Sub btUninstallDevice_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btUninstallDevice.Click
        If lvDevices.SelectedItems.Count = 0 Then
            MessageBox.Show("Select device", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            If lvServices.SelectedItems.Count = 0 Then
                MessageBox.Show("Select service", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
            Else
                Dim Item As ListViewItem = lvDevices.SelectedItems(0)
                Dim Radio As wclBluetoothRadio = CType(Item.Tag, wclBluetoothRadio)
                Dim Address As Int64 = Convert.ToInt64(Item.SubItems(1).Text, 16)
                Dim Service As Guid = New Guid(lvServices.SelectedItems(0).SubItems(3).Text)

                Dim Res As Int32 = Radio.UninstallDevice(Address, Service)
                If Res <> wclErrors.WCL_E_SUCCESS Then
                    ShowError(Res)
                Else
                    MessageBox.Show("Device uninstalled")
                End If
            End If
        End If
    End Sub

    Private Sub btIsInRange_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btIsInRange.Click
        If lvDevices.SelectedItems.Count = 0 Then
            MessageBox.Show("Select device", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            Dim Item As ListViewItem = lvDevices.SelectedItems(0)
            Dim Radio As wclBluetoothRadio = CType(Item.Tag, wclBluetoothRadio)
            Dim Address As Int64 = Convert.ToInt64(Item.SubItems(1).Text, 16)

            Dim InRange As Boolean
            Dim Res As Int32 = Radio.IsRemoteDeviceInRange(Address, InRange)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                ShowError(Res)
            Else
                If InRange Then
                    MessageBox.Show("Device is in range")
                Else
                    MessageBox.Show("Device is NOT in range")
                End If
            End If
        End If
    End Sub

    Private WasOpened As Boolean = False

    Private Sub FPowerMonitor_OnPowerStateChanged(Sender As Object, State As wclCommon.wclPowerState) Handles FPowerMonitor.OnPowerStateChanged
        Select Case State
            Case wclPowerState.psResumeAutomatic
                TraceEvent(Nothing, "Power", "State", "psResumeAutomatic")
                If WasOpened Then
                    WasOpened = False
                    Manager.Open()
                End If

            Case wclPowerState.psResume
                TraceEvent(Nothing, "Power", "State", "psResume")

            Case wclPowerState.psSuspend
                TraceEvent(Nothing, "Power", "State", "psSuspend")
                If Manager.Active Then
                    WasOpened = True
                    Manager.Close()
                End If

            Case wclPowerState.psUnknown
                TraceEvent(Nothing, "Power", "State", "psUnknown")
        End Select
    End Sub

    Private Sub btDisconnect_Click(sender As System.Object, e As System.EventArgs) Handles btDisconnect.Click
        If lvDevices.SelectedItems.Count = 0 Then
            MessageBox.Show("Select device", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            Dim Item As ListViewItem = lvDevices.SelectedItems(0)
            Dim Radio As wclBluetoothRadio = CType(Item.Tag, wclBluetoothRadio)
            Dim Address As Int64 = Convert.ToInt64(Item.SubItems(1).Text, 16)

            Dim Res As Int32 = Radio.RemoteDisconnect(Address)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                ShowError(Res)
            End If
        End If
    End Sub

    Private Sub btEnumConnected_Click(sender As System.Object, e As System.EventArgs) Handles btEnumConnected.Click
        If lvRadios.SelectedItems.Count = 0 Then
            MessageBox.Show("Select radio", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            ClearDevices()

            Dim Item As ListViewItem = lvRadios.SelectedItems(0)
            Dim Radio As wclBluetoothRadio = CType(Item.Tag, wclBluetoothRadio)

            Dim Devices As Int64() = Nothing
            Dim Res As Int32 = Radio.EnumConnectedDevices(Devices)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                ShowError(Res)
            Else
                If Devices IsNot Nothing Then
                    For i As Int32 = 0 To Devices.Length - 1
                        AddDevice(Radio, Devices(i))
                    Next i

                    RefreshDevices()
                End If
            End If
        End If
    End Sub

    Private Sub btEnumInstalledServices_Click(sender As System.Object, e As System.EventArgs) Handles btEnumInstalledServices.Click
        If lvDevices.SelectedItems.Count = 0 Then
            MessageBox.Show("Select device", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            ClearServices()

            Dim Item As ListViewItem = lvDevices.SelectedItems(0)
            Dim Radio As wclBluetoothRadio = CType(Item.Tag, wclBluetoothRadio)
            Dim Address As Int64 = Convert.ToInt64(Item.SubItems(1).Text, 16)

            Dim Services As Guid() = Nothing
            Dim Res As Int32 = Radio.EnumInstalledServices(Address, Services)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                ShowError(Res)
            Else
                If Services IsNot Nothing AndAlso Services.Length > 0 Then
                    For Each Service As Guid In Services
                        Item = lvServices.Items.Add(Radio.ApiName)
                        Item.SubItems.Add(Address.ToString("X12"))
                        Item.SubItems.Add("")
                        Item.SubItems.Add(Service.ToString())
                        Item.SubItems.Add("")
                        Item.SubItems.Add("")
                        Item.SubItems.Add("")
                    Next
                End If
            End If
        End If
    End Sub
End Class
