Module main

    Private Class Console
        Private WithEvents Manager As wclBluetoothManager
        Private FEvent As AutoResetEvent

        Private Sub WriteLn(ByVal s As String)
            System.Console.WriteLine(s)
        End Sub

        Private Sub ReadLn()
            System.Console.ReadKey(True)
        End Sub

        Public Sub Run()
            WriteLn("Bluetooth Framework Console Demo application")
            WriteLn("  The demo is very simple: it just discovers nearby classic and LE")
            WriteLn("  devices.")
            WriteLn("  The demo shows how to use Bluetooth Framework in console")
            WriteLn("  applications with APC synchronization.")
            WriteLn("Press ENTER to continue")

            ReadLn()

            ' Change default synchronization method.
            wclMessageBroadcaster.SetSyncMethod(wclMessageSynchronizationKind.skApc)

            ' Create classes we have to use.
            Manager = New wclBluetoothManager()

            ' The event used to notify about event completion.
            Try
                FEvent = New AutoResetEvent(False)
            Catch
                FEvent = Nothing
            End Try

            If FEvent Is Nothing Then
                WriteLn("Unable to create synchronization event!")
            Else
                ' Open Bluetooth Manager
                Dim Res As Int32 = Manager.Open()
                If Res <> wclErrors.WCL_E_SUCCESS Then
                    WriteLn("Open Bluetooth Manager failed: 0x" + Res.ToString("X8"))
                Else
                    ' Wait for Open completion.
                    wclMessageBroadcaster.Wait(FEvent)

                    ' Try to find available Bluetooth Radio. Use first one found.
                    If Manager.Count = 0 Then
                        WriteLn("No Bluetooth Hardware connected.")
                    Else
                        Dim Radio As wclBluetoothRadio = Nothing
                        Dim i As Integer
                        For i = 0 To Manager.Count - 1
                            If Manager(i).Available Then
                                Radio = Manager(i)
                                Exit For
                            End If
                        Next i

                        If Radio Is Nothing Then
                            WriteLn("Available BluetoothRadio not found.")
                        Else
                            ' Discover classic devices
                            WriteLn("Try to start classic discovering")
                            Res = Radio.Discover(10, wclBluetoothDiscoverKind.dkClassic)
                            If Res <> wclErrors.WCL_E_SUCCESS Then
                                WriteLn("Start classic discovering failed: 0x" + Res.ToString("X8"))
                            Else
                                ' Wait for discovering completion.
                                wclMessageBroadcaster.Wait(FEvent)
                            End If

                            ' Discover LE devices
                            WriteLn("Try to start LE discovering")
                            Res = Radio.Discover(10, wclBluetoothDiscoverKind.dkBle)
                            If Res <> wclErrors.WCL_E_SUCCESS Then
                                WriteLn("Start LE discovering failed: 0x" + Res.ToString("X8"))
                            Else
                                ' Wait for discovering completion.
                                wclMessageBroadcaster.Wait(FEvent)
                            End If
                        End If
                    End If

                    ' Close Bluetooth Manager.
                    Res = Manager.Close()
                    If Res <> wclErrors.WCL_E_SUCCESS Then
                        WriteLn("Open Bluetooth Manager failed: 0x" + Res.ToString("X8"))
                    Else
                        ' Wait for Close completion.
                        wclMessageBroadcaster.Wait(FEvent)
                    End If
                End If

                ' Do not forget about event.
                FEvent.Close()
            End If

            ' Do not forget to destroy Bluetooth Manager.
            Manager = Nothing


            WriteLn("Demo finished. Press ENTER")
            ReadLn()
        End Sub

        Private Sub FManager_OnDeviceFound(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal Address As Long) Handles Manager.OnDeviceFound
            WriteLn("Device " + Address.ToString("X12") + " found")
        End Sub

        Private Sub FManager_OnDiscoveringCompleted(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal [Error] As Integer) Handles Manager.OnDiscoveringCompleted
            WriteLn("Discoverign has been compledted")
            FEvent.Set()
        End Sub

        Private Sub FManager_OnDiscoveringStarted(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio) Handles Manager.OnDiscoveringStarted
            WriteLn("Discovering has been started")
        End Sub

        Private Sub Manager_AfterOpen(sender As Object, e As System.EventArgs) Handles Manager.AfterOpen
            FEvent.Set()
        End Sub

        Private Sub Manager_OnClosed(sender As Object, e As System.EventArgs) Handles Manager.OnClosed
            FEvent.Set()
        End Sub
    End Class

    Sub Main()
        Dim c As Console = New Console()
        c.Run()
    End Sub

End Module
