﻿Public Class fmMain
    Dim WithEvents wclBluetoothManager As wclBluetoothManager
    Dim WithEvents wclRfCommClient As wclRfCommClient
    Dim WithEvents FFtp As wclObexFtpClient

    Private Function GetRadio() As wclBluetoothRadio
        Dim Radio As wclBluetoothRadio = Nothing
        Dim Res As Int32 = wclBluetoothManager.GetClassicRadio(Radio)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Get working radio failed: 0x" + Res.ToString("X8"))
            Return Nothing
        End If
        Return Radio
    End Function

    Private Sub fmMain_Load(sender As System.Object, e As System.EventArgs) Handles MyBase.Load
        wclBluetoothManager = New wclBluetoothManager()
        wclRfCommClient = New wclRfCommClient()
            ' We use ObjectPush Profile so must set the profile's UUID.
        wclRfCommClient.Service = wclUUIDs.OBEXFileTransferServiceClass_UUID

        FFtp = Nothing

        ' Try to open Bluetooth Manager.
        Dim Res As Integer = wclBluetoothManager.Open()
        If Res <> wclErrors.WCL_E_SUCCESS Then
            lbLog.Items.Add("Bluetooth Manager Open failed: 0x" + Res.ToString("X8"))
        End If
    End Sub

    Private Sub wclRfCommClient_OnConnect(Sender As Object, [Error] As Integer) Handles wclRfCommClient.OnConnect
        lbLog.Items.Add("Connect. Operation result: 0x" + [Error].ToString("X8"))
    End Sub

    Private Sub wclRfCommClient_OnCreateProcessor(Sender As Object, Connection As wclCommunication.wclClientDataConnection) Handles wclRfCommClient.OnCreateProcessor
        ' here we must create the data processor for the connection. In this demo
        ' we use OPPClient.
        FFtp = New wclObexFtpClient(Connection)
    End Sub

    Private Sub FFtp_OnChangeDirComplete(Sender As Object, [Error] As Integer, Description As String) Handles FFtp.OnChangeDirComplete
        lbLog.Items.Add("Change dir completed with result: 0x" + [Error].ToString("X8") + ". Description: " + Description)

        ProgressBar.Value = 0
        ProgressBar.Maximum = 0

        lvFiles.Items.Clear()
        FFtp.Dir()
    End Sub

    Private Sub FFtp_OnConnect(Sender As Object, [Error] As Integer, Description As String) Handles FFtp.OnConnect
        lbLog.Items.Add("OBEX session opened with result: 0x" + [Error].ToString("X8") + ". Description: " + Description)
    End Sub

    Private Sub FFtp_OnDeleteComplete(Sender As Object, [Error] As Integer, Description As String) Handles FFtp.OnDeleteComplete
        lbLog.Items.Add("Delete completed with result: 0x" + [Error].ToString("X8") + ". Description: " + Description)

        ProgressBar.Value = 0
        ProgressBar.Maximum = 0

        lvFiles.Items.Clear()
        FFtp.Dir()
    End Sub

    Private Sub FFtp_OnDirComplete(Sender As Object, [Error] As Integer, Description As String, Dirs() As wclCommunication.wclObexFileObject) Handles FFtp.OnDirComplete
        lbLog.Items.Add("Dir completed with result: 0x" + [Error].ToString("X8") + ". Description: " + Description)

        ProgressBar.Value = 0
        ProgressBar.Maximum = 0

        If Not (Dirs Is Nothing) Then
            For Each Obj As wclObexFileObject In Dirs
                Dim Item As ListViewItem = lvFiles.Items.Add(Obj.Name)
                If Obj.IsDirectory Then
                    Item.SubItems.Add("Folder")
                Else
                    Item.SubItems.Add("File")
                End If
                Item.SubItems.Add(Obj.Size.ToString())

                Dim Perm As String = ""
                If (Obj.Permissions And wclObexFilePermission.opRead) <> 0 Then
                    Perm = Perm + "R"
                End If
                If (Obj.Permissions And wclObexFilePermission.opWrite) <> 0 Then
                    Perm = Perm + "W"
                End If
                If (Obj.Permissions And wclObexFilePermission.opDelete) <> 0 Then
                    Perm = Perm + "D"
                End If
                Item.SubItems.Add(Perm)
                If Not Obj.Modified.Equals(DateTime.MinValue) Then
                    Item.SubItems.Add(Obj.Modified.ToString())
                Else
                    Item.SubItems.Add("")
                End If
                If Not Obj.Created.Equals(DateTime.MinValue) Then
                    Item.SubItems.Add(Obj.Created.ToString())
                Else
                    Item.SubItems.Add("")
                End If
                If Not Obj.Accessed.Equals(DateTime.MinValue) Then
                    Item.SubItems.Add(Obj.Accessed.ToString())
                Else
                    Item.SubItems.Add("")
                End If

                Item.SubItems.Add(Obj.Description)
            Next Obj
        End If
    End Sub

    Private Sub FFtp_OnDisconnect(Sender As Object, [Error] As Integer, Description As String) Handles FFtp.OnDisconnect
        lbLog.Items.Add("OBEX session closed with result: 0x" + [Error].ToString("X8") + ". Description: " + Description)
    End Sub

    Private Sub FFtp_OnGetComplete(Sender As Object, [Error] As Integer, Description As String, Stream As System.IO.Stream) Handles FFtp.OnGetComplete
        lbLog.Items.Add("File receiving completed with result: 0x" + [Error].ToString("X8") + ". Description: " + Description)

        ProgressBar.Value = 0
        ProgressBar.Maximum = 0

        ' We must destroy the stream object here.
        Stream.Close()
        Stream.Dispose()
    End Sub

    Private Sub FFtp_OnMakeDirComplete(Sender As Object, [Error] As Integer, Description As String) Handles FFtp.OnMakeDirComplete
        lbLog.Items.Add("Make dir completed with result: 0x" + [Error].ToString("X8") + ". Description: " + Description)

        ProgressBar.Value = 0
        ProgressBar.Maximum = 0

        lvFiles.Items.Clear()
        FFtp.Dir()
    End Sub

    Private Sub FFtp_OnProgress(Sender As Object, Size As UInteger, Position As UInteger) Handles FFtp.OnProgress
        ProgressBar.Maximum = Size
        ProgressBar.Value = Position
    End Sub

    Private Sub FFtp_OnPutComplete(Sender As Object, [Error] As Integer, Description As String, Stream As System.IO.Stream) Handles FFtp.OnPutComplete
        lbLog.Items.Add("File sending completed with result: 0x" + [Error].ToString("X8") + ". Description: " + Description)

        ProgressBar.Value = 0
        ProgressBar.Maximum = 0

        ' We must destroy the stream object here.
        Stream.Close()
        Stream.Dispose()
    End Sub

    Private Sub wclRfCommClient_OnDestroyProcessor(Sender As Object, Connection As wclCommunication.wclClientDataConnection) Handles wclRfCommClient.OnDestroyProcessor
        ' Do we have any data processor created?
        If Not (FFtp Is Nothing) Then
            ' Make sure it is our connection is goind to destroy.
            If Connection.Processor Is FFtp Then
                ' Ok, destroy the data processor here.
                FFtp.Dispose()
                FFtp = Nothing
            End If
        End If
    End Sub

    Private Sub wclRfCommClient_OnDisconnect(Sender As Object, Reason As Integer) Handles wclRfCommClient.OnDisconnect
        lbLog.Items.Add("Disconnected with reason: 0x" + Reason.ToString("X8"))
    End Sub

    Private Sub wclBluetoothManager_AfterOpen(sender As Object, e As System.EventArgs) Handles wclBluetoothManager.AfterOpen
        lbLog.Items.Add("Bluetooth Manager has been opened")
    End Sub

    Private Sub wclBluetoothManager_OnAuthenticationCompleted(Sender As Object, Radio As wclBluetooth.wclBluetoothRadio, Address As Long, [Error] As Integer) Handles wclBluetoothManager.OnAuthenticationCompleted
        lbLog.Items.Add("Authentication completed with result: 0x" + [Error].ToString("X8"))
    End Sub

    Private Sub wclBluetoothManager_OnDeviceFound(Sender As Object, Radio As wclBluetooth.wclBluetoothRadio, Address As Long) Handles wclBluetoothManager.OnDeviceFound
        Dim Item As ListViewItem = lvDevices.Items.Add(Address.ToString("X12"))
        Item.SubItems.Add("")
    End Sub

    Private Sub wclBluetoothManager_OnDiscoveringCompleted(Sender As Object, Radio As wclBluetooth.wclBluetoothRadio, [Error] As Integer) Handles wclBluetoothManager.OnDiscoveringCompleted
        lbLog.Items.Add("Discovering completed with result: 0x" + [Error].ToString("X8"))

        For i = 0 To lvDevices.Items.Count - 1
            Dim Item As ListViewItem = lvDevices.Items(i)
            Dim Addr As Int64 = Convert.ToInt64(Item.Text, 16)
            Dim Name As String = ""
            Dim Res As Int32 = Radio.GetRemoteName(Addr, Name)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                Name = "Error: 0x" + Res.ToString("X8")
            End If
            Item.SubItems(1).Text = Name
        Next i
    End Sub

    Private Sub wclBluetoothManager_OnDiscoveringStarted(Sender As Object, Radio As wclBluetooth.wclBluetoothRadio) Handles wclBluetoothManager.OnDiscoveringStarted
        lbLog.Items.Add("Discovering started")

        lvDevices.Items.Clear()
    End Sub

    Private Sub wclBluetoothManager_OnNumericComparison(Sender As Object, Radio As wclBluetooth.wclBluetoothRadio, Address As Long, Number As UInteger, ByRef Confirm As Boolean) Handles wclBluetoothManager.OnNumericComparison
        lbLog.Items.Add("Numeric comparison. Accept")
        Confirm = True
    End Sub

    Private Sub wclBluetoothManager_OnPasskeyNotification(Sender As Object, Radio As wclBluetooth.wclBluetoothRadio, Address As Long, Passkey As UInteger) Handles wclBluetoothManager.OnPasskeyNotification
        lbLog.Items.Add("Passkey notification. Passkey: " + Passkey.ToString())
    End Sub

    Private Sub wclBluetoothManager_OnPasskeyRequest(Sender As Object, Radio As wclBluetooth.wclBluetoothRadio, Address As Long, ByRef Passkey As UInteger) Handles wclBluetoothManager.OnPasskeyRequest
        lbLog.Items.Add("Passkey request. Passkey: 12345678")
        Passkey = 12345678
    End Sub

    Private Sub wclBluetoothManager_OnPinRequest(Sender As Object, Radio As wclBluetooth.wclBluetoothRadio, Address As Long, ByRef Pin As String) Handles wclBluetoothManager.OnPinRequest
        lbLog.Items.Add("PIN code request. PIN: 0000")
        Pin = "0000"
    End Sub

    Private Sub btClear_Click(sender As System.Object, e As System.EventArgs) Handles btClear.Click
        lbLog.Items.Clear()
    End Sub

    Private Sub fmMain_FormClosing(sender As System.Object, e As System.Windows.Forms.FormClosingEventArgs) Handles MyBase.FormClosing
        wclRfCommClient.Disconnect()
        wclBluetoothManager.Close()
    End Sub

    Private Sub btDiscover_Click(sender As System.Object, e As System.EventArgs) Handles btDiscover.Click
        Dim Radio As wclBluetoothRadio = GetRadio()
        If Not (Radio Is Nothing) Then
            ' Discover classic devices only.
            Dim Res As Int32 = Radio.Discover(10, wclBluetoothDiscoverKind.dkClassic)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                MessageBox.Show("Failed to start discovering: 0x" + Res.ToString("X8"))
            End If
        End If
    End Sub

    Private Sub btDisconnect_Click(sender As System.Object, e As System.EventArgs) Handles btDisconnect.Click
        Dim Res As Int32 = wclRfCommClient.Disconnect()
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Disconnect failed: 0x" + Res.ToString("X8"))
        End If
    End Sub

    Private Sub btCloseSession_Click(sender As System.Object, e As System.EventArgs) Handles btCloseSession.Click
        If FFtp Is Nothing Then
            MessageBox.Show("Not connected")
        Else
            Dim Res As Int32 = FFtp.Disconnect("")
            If Res <> wclErrors.WCL_E_SUCCESS Then
                MessageBox.Show("Close session failed: 0x" + Res.ToString("X8"))
            End If
        End If
    End Sub

    Private Sub btOpenSession_Click(sender As System.Object, e As System.EventArgs) Handles btOpenSession.Click
        If FFtp Is Nothing Then
            MessageBox.Show("Not connected")
        Else
            Dim Res As Int32 = FFtp.Connect()
            If Res <> wclErrors.WCL_E_SUCCESS Then
                MessageBox.Show("Open session failed: 0x" + Res.ToString("X8"))
            End If
        End If
    End Sub

    Private Sub btConnect_Click(sender As System.Object, e As System.EventArgs) Handles btConnect.Click
        If wclRfCommClient.State <> wclClientState.csDisconnected Then
            MessageBox.Show("Client is connected")
        Else
            If lvDevices.SelectedItems.Count = 0 Then
                MessageBox.Show("Select device")
            Else
                Dim Radio As wclBluetoothRadio = GetRadio()
                If Not (Radio Is Nothing) Then
                    Dim Addr As Int64 = Convert.ToInt64(lvDevices.SelectedItems(0).Text, 16)
                    wclRfCommClient.Address = Addr
                    Dim Res As Int32 = wclRfCommClient.Connect(Radio)
                    If Res <> wclErrors.WCL_E_SUCCESS Then
                        MessageBox.Show("Failed to connect: 0x" + Res.ToString("X8"))
                    End If
                End If
            End If
        End If
    End Sub

    Private Sub btSelectFile_Click(sender As System.Object, e As System.EventArgs) Handles btSelectFile.Click
        If openFileDialog.ShowDialog() = DialogResult.OK Then
            edFileName.Text = openFileDialog.FileName
        End If
    End Sub

    Private Sub btAbort_Click(sender As System.Object, e As System.EventArgs) Handles btAbort.Click
        If FFtp Is Nothing Then
            MessageBox.Show("Not connected")
        Else
            Dim Res As Int32 = FFtp.Abort("")
            If Res <> wclErrors.WCL_E_SUCCESS Then
                MessageBox.Show("Abort failed: 0x" + Res.ToString("X8"))
            End If
        End If
    End Sub

    Private Sub btSendFile_Click(sender As System.Object, e As System.EventArgs) Handles btSendFile.Click
        If FFtp Is Nothing Then
            MessageBox.Show("Not connected")
        Else
            If edFileName.Text = "" Then
                MessageBox.Show("Select file")
            Else
                Dim Stream As System.IO.FileStream = New System.IO.FileStream(edFileName.Text, System.IO.FileMode.Open)
                Dim Res As Int32 = FFtp.Put(System.IO.Path.GetFileName(edFileName.Text), "", Stream)
                If Res <> wclErrors.WCL_E_SUCCESS Then
                    MessageBox.Show("Send file failed: 0x" + Res.ToString("X8"))
                    Stream.Close()
                    Stream.Dispose()
                End If
            End If
        End If
    End Sub

    Private Sub btDir_Click(sender As System.Object, e As System.EventArgs) Handles btDir.Click
        If FFtp Is Nothing Then
            MessageBox.Show("Not connected")
        Else
            Dim Res As Int32 = FFtp.Dir()
            If Res <> wclErrors.WCL_E_SUCCESS Then
                MessageBox.Show("Dir failed: 0x" + Res.ToString("X8"))
            Else
                lvFiles.Items.Clear()
            End If
        End If
    End Sub

    Private Sub lvFiles_DoubleClick(sender As System.Object, e As System.EventArgs) Handles lvFiles.DoubleClick
        If lvFiles.SelectedItems.Count > 0 Then
            If FFtp Is Nothing Then
                MessageBox.Show("Not connected")
            Else
                If lvFiles.SelectedItems(0).SubItems(1).Text = "Folder" Then
                    Dim Res As Int32 = FFtp.ChangeDir(lvFiles.SelectedItems(0).Text)
                    If Res <> wclErrors.WCL_E_SUCCESS Then
                        MessageBox.Show("Change path failed 0x" + Res.ToString("X8"))
                    End If
                Else
                    Dim Stream As System.IO.FileStream = New System.IO.FileStream(".\" + lvFiles.SelectedItems(0).Text, System.IO.FileMode.OpenOrCreate, System.IO.FileAccess.ReadWrite)
                    Dim Res As Int32 = FFtp.Get(lvFiles.SelectedItems(0).Text, Stream)
                    If Res <> wclErrors.WCL_E_SUCCESS Then
                        MessageBox.Show("Get failed 0x" + Res.ToString("X8"))
                        Stream.Close()
                        Stream.Dispose()
                    End If
                End If
            End If
        End If
    End Sub

    Private Sub btRoot_Click(sender As System.Object, e As System.EventArgs) Handles btRoot.Click
        If FFtp Is Nothing Then
            MessageBox.Show("Not connected")
        Else
            Dim Res As Int32 = FFtp.ChangeDir("")
            If Res <> wclErrors.WCL_E_SUCCESS Then
                MessageBox.Show("Root failed 0x" + Res.ToString("X8"))
            End If
        End If
    End Sub

    Private Sub btDelete_Click(sender As System.Object, e As System.EventArgs) Handles btDelete.Click
        If lvFiles.SelectedItems.Count > 0 Then
            If FFtp Is Nothing Then
                MessageBox.Show("Not connected")
        Else
                Dim Res As Int32 = FFtp.Delete(lvFiles.SelectedItems(0).Text)
                If Res <> wclErrors.WCL_E_SUCCESS Then
                    MessageBox.Show("Delete failed 0x" + Res.ToString("X8"))
                End If
            End If
        End If
    End Sub

    Private Sub btMkDir_Click(sender As System.Object, e As System.EventArgs) Handles btMkDir.Click
        If FFtp Is Nothing Then
            MessageBox.Show("Not connected")
        Else
            Dim Res As Int32 = FFtp.MkDir(edNewDirName.Text)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                MessageBox.Show("Make sire failed 0x" + Res.ToString("X8"))
            End If
        End If
    End Sub
End Class