Public Class main
    Private WithEvents Manager As wclBluetoothManager
    Private WithEvents Client As wclGattClient

    Private FCharacteristics As wclGattCharacteristic()
    Private FDescriptors As wclGattDescriptor()
    Private FServices As wclGattService()

    Private Sub main_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load
        Manager = New wclBluetoothManager()
        Client = New wclGattClient()

        cbOpFlag.SelectedIndex = 0
        cbProtection.SelectedIndex = 0
        cbParams.SelectedIndex = 0
        cbWriteKind.SelectedIndex = 2
        cbSubscribeKind.SelectedIndex = 2
        cbMitmProtection.SelectedIndex = 0
        cbIoCap.SelectedIndex = 0

        Manager.Open()
        Cleanup()
    End Sub

    Private Sub main_FormClosed(ByVal sender As System.Object, ByVal e As System.Windows.Forms.FormClosedEventArgs) Handles MyBase.FormClosed
        Client.Disconnect()
        Client = Nothing

        Manager.Close()
        Manager = Nothing
        Cleanup()
    End Sub

    Private Function GetRadio() As wclBluetoothRadio
        Dim Radio As wclBluetoothRadio = Nothing
        Dim Res As Int32 = Manager.GetLeRadio(Radio)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Get working radio failed: 0x" + Res.ToString("X8"))
            Return Nothing
        End If
        Return Radio
    End Function

    Private Sub btDiscover_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btDiscover.Click
        Dim Radio As wclBluetoothRadio = GetRadio()
        If Radio IsNot Nothing Then
            Dim Res As Int32 = Radio.Discover(10, wclBluetoothDiscoverKind.dkBle)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                MessageBox.Show("Error starting discovering: 0x" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
            End If
        End If
    End Sub

    Private Sub btClear_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btClear.Click
        lvEvents.Items.Clear()
    End Sub

    Private Sub TraceEvent(ByVal Address As Int64, ByVal [Event] As String, ByVal Param As String, ByVal Value As String)
        Dim s As String = ""
        If Address <> 0 Then
            s = Address.ToString("X12")
        Else
            s = ""
        End If
        Dim Item As ListViewItem = lvEvents.Items.Add(s)
        Item.SubItems.Add([Event])
        Item.SubItems.Add(Param)
        Item.SubItems.Add(Value)
    End Sub

    Private Sub btConnect_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btConnect.Click
        If lvDevices.SelectedItems.Count = 0 Then
            MessageBox.Show("Select device", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            Dim Item As ListViewItem = lvDevices.SelectedItems(0)
            Try
                Client.Address = Convert.ToInt64(Item.Text, 16)
                Client.ConnectOnRead = cbConnectOnRead.Checked
                Client.ForceNotifications = cbForceIndications.Checked
                Dim Res As Int32 = Client.Connect(CType(Item.Tag, wclBluetoothRadio))
                If Res <> wclErrors.WCL_E_SUCCESS Then
                    MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
                End If
            Catch ex As Exception
                MessageBox.Show(ex.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
            End Try
        End If
    End Sub

    Private Sub btDisconnect_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btDisconnect.Click
        Dim Res As Int32 = Client.Disconnect()
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
        End If
    End Sub

    Private Sub btGetServices_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btGetServices.Click
        lvServices.Items.Clear()
        FServices = Nothing

        Dim Res As Int32 = Client.ReadServices(OpFlag(), FServices)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
            Return
        End If

        If FServices Is Nothing Then
            Return
        End If

        For Each Service As wclGattService In FServices
            Dim s As String
            If Service.Uuid.IsShortUuid Then
                s = Service.Uuid.ShortUuid.ToString("X4")
            Else
                s = Service.Uuid.LongUuid.ToString()
            End If
            Dim Item As ListViewItem = lvServices.Items.Add(s)
            Item.SubItems.Add(Service.Uuid.IsShortUuid.ToString())
            Item.SubItems.Add(Service.Handle.ToString("X4"))
        Next Service
    End Sub

    Private Sub btGetIncludedServices_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btGetIncludedServices.Click
        If lvServices.SelectedItems.Count = 0 Then
            MessageBox.Show("Select service", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
            Return
        End If

        Dim Service As wclGattService = FServices(lvServices.SelectedItems(0).Index)

        lvServices.Items.Clear()
        FServices = Nothing

        Dim Res As Int32 = Client.ReadIncludedServices(Service, OpFlag(), FServices)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
            Return
        End If

        If FServices Is Nothing Then
            Return
        End If

        For Each IncService As wclGattService In FServices
            Dim s As String
            If IncService.Uuid.IsShortUuid Then
                s = IncService.Uuid.ShortUuid.ToString("X4")
            Else
                s = IncService.Uuid.LongUuid.ToString()
            End If

            Dim Item As ListViewItem = lvServices.Items.Add(s)
            Item.SubItems.Add(IncService.Uuid.IsShortUuid.ToString())
            Item.SubItems.Add(IncService.Handle.ToString("X4"))
        Next IncService
    End Sub

    Private Sub btGetCharacteristics_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btGetCharacteristics.Click
        FCharacteristics = Nothing
        lvCharacteristics.Items.Clear()

        If lvServices.SelectedItems.Count = 0 Then
            MessageBox.Show("Select service", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
            Return
        End If

        Dim Service As wclGattService = FServices(lvServices.SelectedItems(0).Index)
        Dim Res As Int32 = Client.ReadCharacteristics(Service, OpFlag(), FCharacteristics)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
            Return
        End If

        If FCharacteristics Is Nothing Then
            Return
        End If

        For Each Character As wclGattCharacteristic In FCharacteristics
            Dim s As String
            If Character.Uuid.IsShortUuid Then
                s = Character.Uuid.ShortUuid.ToString("X4")
            Else
                s = Character.Uuid.LongUuid.ToString()
            End If
            Dim Item As ListViewItem = lvCharacteristics.Items.Add(s)

            Item.SubItems.Add(Character.Uuid.IsShortUuid.ToString())
            Item.SubItems.Add(Character.ServiceHandle.ToString("X4"))
            Item.SubItems.Add(Character.Handle.ToString("X4"))
            Item.SubItems.Add(Character.ValueHandle.ToString("X4"))
            Item.SubItems.Add(Character.IsBroadcastable.ToString())
            Item.SubItems.Add(Character.IsReadable.ToString())
            Item.SubItems.Add(Character.IsWritable.ToString())
            Item.SubItems.Add(Character.IsWritableWithoutResponse.ToString())
            Item.SubItems.Add(Character.IsSignedWritable.ToString())
            Item.SubItems.Add(Character.IsNotifiable.ToString())
            Item.SubItems.Add(Character.IsIndicatable.ToString())
            Item.SubItems.Add(Character.HasExtendedProperties.ToString())
        Next Character
    End Sub

    Private Sub btGetCharValue_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btGetCharValue.Click
        If lvCharacteristics.SelectedItems.Count = 0 Then
            MessageBox.Show("Select characteristic", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
            Return
        End If

        Dim Characteristic As wclGattCharacteristic = FCharacteristics(lvCharacteristics.SelectedItems(0).Index)
        Dim Value As Byte() = Nothing
        Dim Res As Int32 = Client.ReadCharacteristicValue(Characteristic, OpFlag(), Value, Protection())
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
            Return
        End If

        If Value Is Nothing Then
            Return
        End If

        If Value.Length = 0 Then
            MessageBox.Show("Value is empty", "Information", MessageBoxButtons.OK, MessageBoxIcon.Information)
        Else
            Dim Str As String = ""
            For i As Int32 = 0 To Value.Length - 1
                Str = Str + Value(i).ToString("X2")
            Next i
            MessageBox.Show("Value: " + Str, "Characterist value", MessageBoxButtons.OK, MessageBoxIcon.Information)
        End If
    End Sub

    Private Sub btGetDesc_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btGetDesc.Click
        FDescriptors = Nothing
        lvDescriptors.Items.Clear()

        If lvCharacteristics.SelectedItems.Count = 0 Then
            MessageBox.Show("Select characteristic", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
            Return
        End If

        Dim Characteristic As wclGattCharacteristic = FCharacteristics(lvCharacteristics.SelectedItems(0).Index)
        Dim Res As Int32 = Client.ReadDescriptors(Characteristic, OpFlag(), FDescriptors)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
            Return
        End If

        If FDescriptors Is Nothing Then
            Return
        End If

        For Each Descriptor As wclGattDescriptor In FDescriptors
            Dim s As String
            If Descriptor.Uuid.IsShortUuid Then
                s = Descriptor.Uuid.ShortUuid.ToString("X4")
            Else
                s = Descriptor.Uuid.LongUuid.ToString()
            End If
            Dim Item As ListViewItem = lvDescriptors.Items.Add(s)

            Item.SubItems.Add(Descriptor.Uuid.IsShortUuid.ToString())
            Item.SubItems.Add(Descriptor.ServiceHandle.ToString("X4"))
            Item.SubItems.Add(Descriptor.CharacteristicHandle.ToString("X4"))
            Item.SubItems.Add(Descriptor.Handle.ToString("X4"))
            Select Case Descriptor.DescriptorType
                Case wclGattDescriptorType.dtCharacteristicExtendedProperties
                    Item.SubItems.Add("dtCharacteristicExtendedProperties")
                Case wclGattDescriptorType.dtCharacteristicUserDescription
                    Item.SubItems.Add("dtCharacteristicUserDescription")
                Case wclGattDescriptorType.dtClientCharacteristicConfiguration
                    Item.SubItems.Add("dtClientCharacteristicConfiguration")
                Case wclGattDescriptorType.dtServerCharacteristicConfiguration
                    Item.SubItems.Add("dtServerCharacteristicConfiguration")
                Case wclGattDescriptorType.dtCharacteristicFormat
                    Item.SubItems.Add("dtCharacteristicFormat")
                Case wclGattDescriptorType.dtCharacteristicAggregateFormat
                    Item.SubItems.Add("dtCharacteristicAggregateFormat")
                Case Else
                    Item.SubItems.Add("dtCustomDescriptor")
            End Select
        Next Descriptor
    End Sub

    Private Sub Cleanup()
        FCharacteristics = Nothing
        FDescriptors = Nothing
        FServices = Nothing

        lvServices.Items.Clear()
        lvCharacteristics.Items.Clear()
        lvDescriptors.Items.Clear()
        lvDescriptorValue.Items.Clear()
    End Sub

    Private Sub btGetDescVal_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btGetDescVal.Click
        lvDescriptorValue.Items.Clear()

        If lvDescriptors.SelectedItems.Count = 0 Then
            MessageBox.Show("Select descriptor", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
            Return
        End If

        Dim Descriptor As wclGattDescriptor = FDescriptors(lvDescriptors.SelectedItems(0).Index)
        Dim Value As wclGattDescriptorValue = Nothing
        Dim Res As Int32 = Client.ReadDescriptorValue(Descriptor, OpFlag(), Value, Protection())
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
        Else
            Select Case Value.AType
                Case wclGattDescriptorType.dtCharacteristicExtendedProperties
                    ListProp("Type", "dtCharacteristicExtendedProperties")
                    ListProp("IsReliableWriteEnabled", Value.CharacteristicExtendedProperties.IsReliableWriteEnabled.ToString())
                    ListProp("IsAuxiliariesWritable", Value.CharacteristicExtendedProperties.IsAuxiliariesWritable.ToString())

                Case wclGattDescriptorType.dtCharacteristicUserDescription
                    ListProp("Type", "dtCharacteristicUserDescription")
                    ListProp("Descriptipon", Value.UserDescription.Description)

                Case wclGattDescriptorType.dtClientCharacteristicConfiguration
                    ListProp("Type", "dtClientCharacteristicConfiguration")
                    ListProp("IsSubscribeToNotification", Value.ClientCharacteristicConfiguration.IsSubscribeToNotification.ToString())
                    ListProp("IsSubscribeToIndication", Value.ClientCharacteristicConfiguration.IsSubscribeToIndication.ToString())

                Case wclGattDescriptorType.dtServerCharacteristicConfiguration
                    ListProp("Type", "dtServerCharacteristicConfiguration")
                    ListProp("IsBroadcast", Value.ServerCharacteristicConfiguration.IsBroadcast.ToString())

                Case wclGattDescriptorType.dtCharacteristicFormat
                    ListProp("Type", "dtCharacteristicFormat")
                    ListProp("Format", Value.CharacteristicFormat.Format.ToString())
                    ListProp("Exponent", Value.CharacteristicFormat.Exponent.ToString("X2"))
                    ListProp("Unit", Value.CharacteristicFormat.AUnit.ToString("X4"))
                    ListProp("NameSpace", Value.CharacteristicFormat.NameSpace.ToString("X2"))
                    ListProp("Description", Value.CharacteristicFormat.Description.ToString("X4"))

                Case wclGattDescriptorType.dtCharacteristicAggregateFormat
                    ListProp("Type", "dtCharacteristicAggregateFormat")
                    If Value.CharacteristicAggregateFormat.Handles IsNot Nothing AndAlso Value.CharacteristicAggregateFormat.Handles.Length > 0 Then
                        For i As Integer = 0 To Value.CharacteristicAggregateFormat.Handles.Length - 1
                            ListProp("Handle [" + i.ToString() + "]", Value.CharacteristicAggregateFormat.Handles(i).ToString("X4"))
                        Next i
                    End If

                Case Else
                    ListProp("Type", "UNKNOWN")
            End Select

            If Value.Data IsNot Nothing Then
                Dim Str As String = ""
                For i As Int32 = 0 To Value.Data.Length - 1
                    Str = Str + Value.Data(i).ToString("X2")
                Next i
                ListProp("Data", Str)
            End If
        End If
    End Sub

    Private Sub ListProp(ByVal Name As String, ByVal Value As String)
        Dim Item As ListViewItem = lvDescriptorValue.Items.Add(Name)
        Item.SubItems.Add(Value)
    End Sub

    Private Sub btSetValue_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btSetValue.Click
        If lvCharacteristics.SelectedItems.Count = 0 Then
            MessageBox.Show("Select characteristic", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
            Return
        End If

        Dim Characteristic As wclGattCharacteristic = FCharacteristics(lvCharacteristics.SelectedItems(0).Index)

        Dim Str As String = edCharVal.Text
        If Str.Length Mod 2 <> 0 Then
            Str = "0" + Str
        End If

        Dim Val(Str.Length \ 2 - 1) As Byte
        For i As Int32 = 0 To Val.Length - 1
            Dim b As String = Str.Substring(i * 2, 2)
            Val(i) = Convert.ToByte(b, 16)
        Next i

        Dim Res As Int32
        Dim WriteKind As wclGattWriteKind
        Select Case cbWriteKind.SelectedIndex
            Case 0
                WriteKind = wclGattWriteKind.wkWithResponse
            Case 1
                WriteKind = wclGattWriteKind.wkWithoutResponse
            Case Else
                WriteKind = wclGattWriteKind.wkAuto
        End Select

        Res = Client.WriteCharacteristicValue(Characteristic, Val, Protection, WriteKind)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
        End If
    End Sub

    Private Sub btSubscribe_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btSubscribe.Click
        If lvCharacteristics.SelectedItems.Count = 0 Then
            MessageBox.Show("Select characteristic", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
            Return
        End If

        Dim Characteristic As wclGattCharacteristic = FCharacteristics(lvCharacteristics.SelectedItems(0).Index)

        Dim SubscribeKind As wclGattSubscribeKind
        Select Case cbSubscribeKind.SelectedIndex
            Case 0
                SubscribeKind = wclGattSubscribeKind.skNotification
            Case 1
                SubscribeKind = wclGattSubscribeKind.skIndication
            Case Else
                SubscribeKind = wclGattSubscribeKind.skManual
        End Select

        If SubscribeKind = wclGattSubscribeKind.skManual Then
            ' In case if characteristic has both Indication and Notification properties
            ' set to True we have to select one of them. Here we use Notifications but
            ' you can use other one.
            If Characteristic.IsNotifiable AndAlso Characteristic.IsIndicatable Then
                ' Change the code line below to
                ' Characteristic.IsNotifiable = false;
                ' if you want to receive Indications instead of notifications.
                Characteristic.IsIndicatable = False
            End If
        End If

        Dim Res As Int32
        If cbFastSubscribe.Checked Then
            If SubscribeKind = wclGattSubscribeKind.skManual Then
                Res = Client.SubscribeForNotifications(Characteristic)
            Else
                Res = Client.SubscribeForNotifications(Characteristic, OpFlag(),
                                                       Protection(), SubscribeKind)
            End If
        Else
            If SubscribeKind = wclGattSubscribeKind.skManual Then
                Res = Client.Subscribe(Characteristic)
            Else
                Res = Client.Subscribe(Characteristic, SubscribeKind)
            End If
        End If

        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
        End If
    End Sub

    Private Sub btUnsubscribe_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btUnsubscribe.Click
        If lvCharacteristics.SelectedItems.Count = 0 Then
            MessageBox.Show("Select characteristic", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
            Return
        End If

        Dim Characteristic As wclGattCharacteristic = FCharacteristics(lvCharacteristics.SelectedItems(0).Index)

        Dim SubscribeKind As wclGattSubscribeKind
        Select Case cbSubscribeKind.SelectedIndex
            Case 0
                SubscribeKind = wclGattSubscribeKind.skNotification
            Case 1
                SubscribeKind = wclGattSubscribeKind.skIndication
            Case Else
                SubscribeKind = wclGattSubscribeKind.skManual
        End Select

        If SubscribeKind = wclGattSubscribeKind.skManual Then
            ' In case if characteristic has both Indication and Notification properties
            ' set to True we have to select one of them. Here we use Notifications but
            ' you can use other one.
            ' YOU MUST USE THE SAME PROPERTY THAT IS USED IN SUBSCRIBE
            If Characteristic.IsNotifiable AndAlso Characteristic.IsIndicatable Then
                ' Change the code line below to
                ' Characteristic.IsNotifiable = false;
                ' if you want to receive Indications instead of notifications.
                Characteristic.IsIndicatable = False
            End If
        End If

        Dim Res As Int32
        If cbFastSubscribe.Checked Then
            If SubscribeKind = wclGattSubscribeKind.skManual Then
                Res = Client.UnsubscribeFromNotifications(Characteristic)
            Else
                Res = Client.UnsubscribeFromNotifications(Characteristic, OpFlag(),
                                                          Protection(), SubscribeKind)
            End If
        Else
            If SubscribeKind = wclGattSubscribeKind.skManual Then
                Res = Client.Unsubscribe(Characteristic)
            Else
                Res = Client.Unsubscribe(Characteristic, SubscribeKind)
            End If
        End If

        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
        End If
    End Sub

    Private Function OpFlag() As wclGattOperationFlag
        Select Case cbOpFlag.SelectedIndex
            Case 1
                Return wclGattOperationFlag.goReadFromDevice
            Case 2
                Return wclGattOperationFlag.goReadFromCache
            Case Else
                Return wclGattOperationFlag.goNone
        End Select
    End Function

    Private Sub btWriteSubsc_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btWriteSubsc.Click
        If lvCharacteristics.SelectedItems.Count = 0 Then
            MessageBox.Show("Select characteristic", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
            Return
        End If

        Dim Characteristic As wclGattCharacteristic = FCharacteristics(lvCharacteristics.SelectedItems(0).Index)

        Dim SubscribeKind As wclGattSubscribeKind
        Select Case cbSubscribeKind.SelectedIndex
            Case 0
                SubscribeKind = wclGattSubscribeKind.skNotification
            Case 1
                SubscribeKind = wclGattSubscribeKind.skIndication
            Case Else
                SubscribeKind = wclGattSubscribeKind.skManual
        End Select

        Dim Res As Int32
        If SubscribeKind = wclGattSubscribeKind.skManual Then
            ' In case if characteristic has both Indication and Notification properties
            ' set to True we have to select one of them. Here we use Notifications but
            ' you can use other one.
            If Characteristic.IsNotifiable AndAlso Characteristic.IsIndicatable Then
                ' Change the code line below to
                ' Characteristic.IsNotifiable = false;
                ' if you want to receive Indications instead of notifications.
                Characteristic.IsIndicatable = False
            End If
            Res = Client.WriteClientConfiguration(Characteristic, True, OpFlag(), Protection())
        Else
            Res = Client.WriteClientConfiguration(Characteristic, True, OpFlag(),
                                                  Protection(), SubscribeKind)
        End If

        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
        End If
    End Sub

    Private Sub btWriteUnsubsc_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btWriteUnsubsc.Click
        If lvCharacteristics.SelectedItems.Count = 0 Then
            MessageBox.Show("Select characteristic", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
            Return
        End If

        Dim Characteristic As wclGattCharacteristic = FCharacteristics(lvCharacteristics.SelectedItems(0).Index)

        Dim SubscribeKind As wclGattSubscribeKind
        Select Case cbSubscribeKind.SelectedIndex
            Case 0
                SubscribeKind = wclGattSubscribeKind.skNotification
            Case 1
                SubscribeKind = wclGattSubscribeKind.skIndication
            Case Else
                SubscribeKind = wclGattSubscribeKind.skManual
        End Select

        Dim Res As Int32
        If SubscribeKind = wclGattSubscribeKind.skManual Then
            ' In case if characteristic has both Indication and Notification properties
            ' set to True we have to select one of them. Here we use Notifications but
            ' you can use other one.
            ' YOU MUST USE THE SAME PROPERTY THAT IS USED IN SUBSCRIBE
            If Characteristic.IsNotifiable AndAlso Characteristic.IsIndicatable Then
                ' Change the code line below to
                ' Characteristic.IsNotifiable = false;
                ' if you want to receive Indications instead of notifications.
                Characteristic.IsIndicatable = False
            End If
            Res = Client.WriteClientConfiguration(Characteristic, False, OpFlag(), Protection())
        Else
            Res = Client.WriteClientConfiguration(Characteristic, False, OpFlag(),
                                                  Protection(), SubscribeKind)
        End If

        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
        End If
    End Sub

    Private Sub Manager_OnNumericComparison(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal Address As Long, ByVal Number As UInteger, ByRef Confirm As Boolean) Handles Manager.OnNumericComparison
        ' Accept any pairing.
        Confirm = True
        TraceEvent(Address, "Numeric comparison", "Number", Number.ToString())
    End Sub

    Private Sub Manager_OnPasskeyNotification(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal Address As Long, ByVal Passkey As UInteger) Handles Manager.OnPasskeyNotification
        TraceEvent(Address, "Passkey notification", "Passkey", Passkey.ToString())
    End Sub

    Private Sub Manager_OnPasskeyRequest(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal Address As Long, ByRef Passkey As UInteger) Handles Manager.OnPasskeyRequest
        ' Use 123456 as passkey.
        Passkey = 123456
        TraceEvent(Address, "Passkey request", "Passkey", Passkey.ToString())
    End Sub

    Private Sub Manager_OnPinRequest(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal Address As Long, ByRef Pin As String) Handles Manager.OnPinRequest
        ' Use '0000' as PIN.
        Pin = "0000"
        TraceEvent(Address, "Pin request", "PIN", Pin)
    End Sub

    Private Sub Manager_OnDeviceFound(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal Address As Long) Handles Manager.OnDeviceFound
        Dim Item As ListViewItem = lvDevices.Items.Add(Address.ToString("X12"))
        Item.SubItems.Add("") ' We can not read a device's name here.
        Item.Tag = Radio 'To use it later.

        Dim DevType As wclBluetoothDeviceType = wclBluetoothDeviceType.dtMixed
        Dim Res As Int32 = Radio.GetRemoteDeviceType(Address, DevType)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            Item.SubItems.Add("Error: 0x" + Res.ToString("X8"))
        Else
            Select Case DevType
                Case wclBluetoothDeviceType.dtClassic
                    Item.SubItems.Add("Classic")
                Case wclBluetoothDeviceType.dtBle
                    Item.SubItems.Add("BLE")
                Case wclBluetoothDeviceType.dtMixed
                    Item.SubItems.Add("Mixed")
                Case Else
                    Item.SubItems.Add("Unknown")
            End Select
        End If

        Dim Paired As Boolean = False
        Res = Radio.GetRemotePaired(Address, Paired)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            Item.SubItems.Add("Error: 0x" + Res.ToString("X8"))
        Else
            If Paired Then
                Item.SubItems.Add("True")
            Else
                Item.SubItems.Add("False")
            End If
        End If

        TraceEvent(Address, "Device found", "", "")
    End Sub

    Private Sub Manager_OnDiscoveringCompleted(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal [Error] As Integer) Handles Manager.OnDiscoveringCompleted
        If lvDevices.Items.Count = 0 Then
            MessageBox.Show("No BLE devices were found.", "Discovering for BLE devices", MessageBoxButtons.OK, MessageBoxIcon.Information)
        Else
            ' Here we can update found devices names.
            Dim i As Int32
            For i = 0 To lvDevices.Items.Count - 1
                Dim Item As ListViewItem = lvDevices.Items(i)

                Dim Address As Int64 = Convert.ToInt64(Item.Text, 16)
                Dim DevName As String = ""
                Dim Res As Int32 = Radio.GetRemoteName(Address, DevName)
                If Res <> wclErrors.WCL_E_SUCCESS Then
                    Item.SubItems(1).Text = "Error: 0x" + Res.ToString("X8")
                Else
                    Item.SubItems(1).Text = DevName
                End If
            Next i
        End If

        TraceEvent(0, "Discovering completed", "", "")
    End Sub

    Private Sub Manager_OnDiscoveringStarted(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio) Handles Manager.OnDiscoveringStarted
        lvDevices.Items.Clear()
        TraceEvent(0, "Discovering started", "", "")
    End Sub

    Private Sub Client_OnCharacteristicChanged(ByVal Sender As Object, ByVal Handle As UShort, ByVal Value() As Byte) Handles Client.OnCharacteristicChanged
        TraceEvent(CType(Sender, wclGattClient).Address, "ValueChanged", "Handle", Handle.ToString("X4"))
        If Value Is Nothing Then
            TraceEvent(0, "", "Value", "")
        Else
            If Value.Length = 0 Then
                TraceEvent(0, "", "Value", "")
            Else
                Dim Str As String = ""

                Dim i As Int32
                For i = 0 To Value.Length - 1
                    Str = Str + Value(i).ToString("X2")
                Next i

                TraceEvent(0, "", "Value", Str)
            End If
        End If
    End Sub

    Private Sub Client_OnConnect(ByVal Sender As Object, ByVal [Error] As Integer) Handles Client.OnConnect
        ' Connection property is valid here.
        TraceEvent(CType(Sender, wclGattClient).Address, "Connected", "Error", "0x" + [Error].ToString("X8"))

        If [Error] = wclErrors.WCL_E_SUCCESS Then
            TraceEvent(Client.Address, "Max PDU", "", "")
            GetMaxPduSize()
            TraceEvent(Client.Address, "Connection params", "", "")
            GetConnectionParams()
        End If
    End Sub

    Private Sub Client_OnDisconnect(ByVal Sender As Object, ByVal Reason As Integer) Handles Client.OnDisconnect
        ' Connection property is valid here.
        TraceEvent(CType(Sender, wclGattClient).Address, "Disconnected", "Reason", "0x" + Reason.ToString("X8"))

        Cleanup()
    End Sub

    Private Sub btTerminate_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btTerminate.Click
        Dim Radio As wclBluetoothRadio = GetRadio()
        If Radio IsNot Nothing Then
            Dim Res As Int32 = Radio.Terminate()
            If Res <> wclErrors.WCL_E_SUCCESS Then
                MessageBox.Show("Error terminating discovering: 0x" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
            End If
        End If
    End Sub

    Private Function Protection() As wclGattProtectionLevel
        Select Case cbProtection.SelectedIndex
            Case 0
                Return wclGattProtectionLevel.plNone
            Case 1
                Return wclGattProtectionLevel.plAuthentication
            Case 2
                Return wclGattProtectionLevel.plEncryption
            Case 3
                Return wclGattProtectionLevel.plEncryptionAndAuthentication
            Case Else
                Return wclGattProtectionLevel.plNone
        End Select
    End Function

    Private Sub Manager_OnConfirm(Sender As Object, Radio As wclBluetooth.wclBluetoothRadio, Address As Long, ByRef Confirm As Boolean) Handles Manager.OnConfirm
        ' Accept any pairing.
        Confirm = True
        TraceEvent(Address, "Just works pairing", "Accept", "True")
    End Sub

    Private Sub Manager_OnAuthenticationCompleted(Sender As Object, Radio As wclBluetooth.wclBluetoothRadio, Address As Long, [Error] As Integer) Handles Manager.OnAuthenticationCompleted
        TraceEvent(Address, "Authentication completed", "Error", [Error].ToString("X8"))
    End Sub

    Private Sub btPair_Click(sender As System.Object, e As System.EventArgs) Handles btPair.Click
        If lvDevices.SelectedItems.Count = 0 Then
            MessageBox.Show("Select device", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            Dim Item As ListViewItem = lvDevices.SelectedItems(0)
            Dim Res As Integer = (CType(Item.Tag, wclBluetoothRadio)).RemotePair(Convert.ToInt64(Item.Text, 16),
                                                                                 wclBluetoothPairingMethod.pmLe)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
            End If
        End If
    End Sub

    Private Sub cbFastSubscribe_CheckedChanged(sender As System.Object, e As System.EventArgs) Handles cbFastSubscribe.CheckedChanged
        btWriteSubsc.Enabled = Not cbFastSubscribe.Checked
        btWriteUnsubsc.Enabled = Not cbFastSubscribe.Checked
    End Sub

    Private Sub btGetParams_Click(sender As System.Object, e As System.EventArgs) Handles btGetParams.Click
        TraceEvent(Client.Address, "Connection params", "", "")
        GetConnectionParams()
    End Sub

    Private Sub Client_OnConnectionParamsChanged(sender As Object, e As System.EventArgs) Handles Client.OnConnectionParamsChanged
        TraceEvent(Client.Address, "Connection params changed", "", "")
        GetConnectionParams()
    End Sub

    Private Sub btSetParams_Click(sender As System.Object, e As System.EventArgs) Handles btSetParams.Click
        Dim Res As Int32
        If cbParams.SelectedIndex = 3 Then
            Dim ParamsValue As wclBluetoothLeConnectionParametersValue = New wclBluetoothLeConnectionParametersValue()
            ParamsValue.MinInterval = 84
            ParamsValue.MaxInterval = 84
            ParamsValue.Latency = 0
            ParamsValue.LinkTimeout = 800
            Res = Client.SetConnectionParams(ParamsValue)
        Else
            Dim Params As wclBluetoothLeConnectionParametersType
            Select Case cbParams.SelectedIndex
                Case 0
                    Params = wclBluetoothLeConnectionParametersType.ppBalanced
                Case 1
                    Params = wclBluetoothLeConnectionParametersType.ppPowerOptimized
                Case 2
                    Params = wclBluetoothLeConnectionParametersType.ppThroughputOptimized
                Case Else
                    Throw New wclEInvalidArgument("Invalid connection parameters.")
            End Select
            Res = Client.SetConnectionParams(Params)
        End If
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Error: 0x" + Res.ToString("X8"))
        End If
    End Sub

    Private Sub Client_OnMaxPduSizeChanged(sender As Object, e As System.EventArgs) Handles Client.OnMaxPduSizeChanged
        TraceEvent(Client.Address, "Max PDU size changed", "", "")
        GetMaxPduSize()
    End Sub

    Private Sub btGetMaxPduSize_Click(sender As System.Object, e As System.EventArgs) Handles btGetMaxPduSize.Click
        TraceEvent(Client.Address, "Max PDU", "", "")
        GetMaxPduSize()
    End Sub

    Private Sub GetMaxPduSize()
        Dim Size As UInt16
        Dim Res As Int32 = Client.GetMaxPduSize(Size)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            TraceEvent(Client.Address, "", "Error", Res.ToString("X8"))
        Else
            TraceEvent(Client.Address, "", "Size", Size.ToString())
        End If
    End Sub

    Private Sub GetConnectionParams()
        Dim Params As wclBluetoothLeConnectionParameters
        Dim Res As Int32 = Client.GetConnectionParams(Params)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            TraceEvent(0, "", "Error", Res.ToString("X8"))
        Else
            TraceEvent(0, "", "Connection interval", Params.Interval.ToString())
            TraceEvent(0, "", "Connection latency", Params.Latency.ToString())
            TraceEvent(0, "", "Link timeout", Params.LinkTimeout.ToString())
        End If
    End Sub

    Private Sub Manager_OnProtectionLevelRequest(Sender As Object, Radio As wclBluetooth.wclBluetoothRadio, Address As Int64, ByRef Protection As wclBluetooth.wclBluetoothLeProtectionLevel) Handles Manager.OnProtectionLevelRequest
        Select Case cbProtection.SelectedIndex
            Case 0
                Protection = wclBluetoothLeProtectionLevel.pplNone
            Case 1
                Protection = wclBluetoothLeProtectionLevel.pplDefault
            Case 2
                Protection = wclBluetoothLeProtectionLevel.pplEncryption
            Case 3
                Protection = wclBluetoothLeProtectionLevel.pplEncryptionAndAuthentication
            Case Else
                Protection = wclBluetoothLeProtectionLevel.pplDefault
        End Select
    End Sub

    Private Sub btPhy_Click(sender As System.Object, e As System.EventArgs) Handles btPhy.Click
        TraceEvent(Client.Address, "Connection PHY", "", "")
        GetConnectionPhy()
    End Sub

    Private Sub Client_OnConnectionPhyChanged(sender As Object, e As System.EventArgs) Handles Client.OnConnectionPhyChanged
        TraceEvent(Client.Address, "Connection PHY changed", "", "")
        GetConnectionPhy()
    End Sub

    Private Sub GetConnectionPhy()
        Dim Phy As wclBluetoothLeConnectionPhy
        Dim Res As Int32 = Client.GetConnectionPhyInfo(Phy)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            TraceEvent(0, "", "Error", Res.ToString("X8"))
        Else
            TraceEvent(0, "Transmit PHY", "", "")
            TraceEvent(0, "", "Coded", Phy.Transmit.IsCoded.ToString())
            TraceEvent(0, "", "1M", Phy.Transmit.IsUncoded1MPhy.ToString())
            TraceEvent(0, "", "2M", Phy.Transmit.IsUncoded2MPhy.ToString())

            TraceEvent(0, "Receive PHY", "", "")
            TraceEvent(0, "", "Coded", Phy.Receive.IsCoded.ToString())
            TraceEvent(0, "", "1M", Phy.Receive.IsUncoded1MPhy.ToString())
            TraceEvent(0, "", "2M", Phy.Receive.IsUncoded2MPhy.ToString())
        End If
    End Sub

    Private Sub Manager_OnIoCapabilityRequest(Sender As Object, Radio As wclBluetooth.wclBluetoothRadio, Address As Long, ByRef Mitm As wclBluetooth.wclBluetoothMitmProtection, ByRef IoCapability As wclBluetooth.wclBluetoothIoCapability, ByRef OobPresent As Boolean) Handles Manager.OnIoCapabilityRequest
        Select cbMitmProtection.SelectedIndex
            Case 0
                Mitm = wclBluetoothMitmProtection.mitmProtectionNotRequired
            Case 1
                Mitm = wclBluetoothMitmProtection.mitmProtectionRequired
            Case 2
                Mitm = wclBluetoothMitmProtection.mitmProtectionNotRequiredBonding
            Case 3
                Mitm = wclBluetoothMitmProtection.mitmProtectionRequiredBonding
            Case 4
                Mitm = wclBluetoothMitmProtection.mitmProtectionNotRequiredGeneralBonding
            Case 5
                Mitm = wclBluetoothMitmProtection.mitmProtectionRequiredGeneralBonding
            Case Else
                Mitm = wclBluetoothMitmProtection.mitmProtectionNotDefined
        End Select

        Select Case cbIoCap.SelectedIndex
            Case 0
                IoCapability = wclBluetoothIoCapability.iocapDisplayOnly
            Case 1
                IoCapability = wclBluetoothIoCapability.iocapDisplayYesNo
            Case 2
                IoCapability = wclBluetoothIoCapability.iocapKeyboardOnly
            Case 3
                IoCapability = wclBluetoothIoCapability.iocapNoInputNoOutput
            Case 4
                IoCapability = wclBluetoothIoCapability.iocapDisplayKeyboard
            Case Else
                IoCapability = wclBluetoothIoCapability.iocapNotDefined
        End Select

        OobPresent = False
    End Sub

    Private Sub btUnpair_Click(sender As System.Object, e As System.EventArgs) Handles btUnpair.Click
        If lvDevices.SelectedItems.Count = 0 Then
            MessageBox.Show("Select device", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            Dim Item As ListViewItem = lvDevices.SelectedItems(0)
            Dim Res As Integer = (CType(Item.Tag, wclBluetoothRadio)).RemoteUnpair(Convert.ToInt64(Item.Text, 16),
                                                                                  wclBluetoothPairingMethod.pmLe)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
            End If
        End If
    End Sub

    Private Sub btEnumPaired_Click(sender As System.Object, e As System.EventArgs) Handles btEnumPaired.Click
        Dim Radio As wclBluetoothRadio = GetRadio()
        If Radio IsNot Nothing Then
            lvDevices.Items.Clear()

            Dim Devices() As Int64 = Nothing
            Dim Res As Integer = Radio.EnumPairedDevices(wclBluetoothDiscoverKind.dkBle, Devices)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                MessageBox.Show("Enum paired failed: 0x" + Res.ToString("X8"))
            Else
                If Devices IsNot Nothing AndAlso Devices.Length > 0 Then
                    For Each Device As Int64 In Devices
                        Dim Item As ListViewItem = lvDevices.Items.Add(Device.ToString("X12"))
                        Item.Tag = Radio

                        Dim Name As String = ""
                        Res = Radio.GetRemoteName(Device, Name)
                        If Res = wclErrors.WCL_E_SUCCESS Then
                            Item.SubItems.Add(Name)
                        Else
                            Item.SubItems.Add("Error: 0x" + Res.ToString("X8"))
                        End If

                        Dim DevType As wclBluetoothDeviceType = wclBluetoothDeviceType.dtMixed
                        If Res <> wclErrors.WCL_E_SUCCESS Then
                            Item.SubItems.Add("Error: 0x" + Res.ToString("X8"))
                        Else
                            Select Case DevType
                                Case wclBluetoothDeviceType.dtClassic
                                    Item.SubItems.Add("Classic")
                                Case wclBluetoothDeviceType.dtBle
                                    Item.SubItems.Add("BLE")
                                Case wclBluetoothDeviceType.dtMixed
                                    Item.SubItems.Add("Mixed")
                                Case Else
                                    Item.SubItems.Add("Unknown")
                            End Select
                        End If

                        Dim Paired As Boolean = False
                        Res = Radio.GetRemotePaired(Device, Paired)
                        If Res <> wclErrors.WCL_E_SUCCESS Then
                            Item.SubItems.Add("Error: 0x" + Res.ToString("X8"))
                        Else
                            If Paired Then
                                Item.SubItems.Add("True")
                            Else
                                Item.SubItems.Add("False")
                            End If
                        End If
                    Next
                End If
            End If
        End If
    End Sub
End Class
