Public Class fmMain
    Private Const SHORT_DATA_LEN As UInt16 = 5
    Private Const LONG_DATA_LEN As UInt16 = 1024

    Private WithEvents FManager As wclBluetoothManager
    Private WithEvents FServer As wclGattServer
    Private FStarted As Boolean
    Private FCounter As UInt32

    Private FLongData(LONG_DATA_LEN) As Byte
    Private FShortData(SHORT_DATA_LEN) As Byte

    Private Sub btStart_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btStart.Click
        Start()
    End Sub

    Private Sub FManager_AfterOpen(ByVal sender As Object, ByVal e As System.EventArgs) Handles FManager.AfterOpen
        lbLog.Items.Add("Bluetooth Manager has been opened")
    End Sub

    Private Sub FManager_BeforeClose(ByVal sender As Object, ByVal e As System.EventArgs) Handles FManager.BeforeClose
        lbLog.Items.Add("Bluetooth Manager is closing")
    End Sub

    Private Sub FServer_OnClientConnected(Sender As Object, Client As wclBluetooth.wclGattServerClient) Handles FServer.OnClientConnected
        lbLog.Items.Add("Client connected " + Client.Address.ToString("X12"))

        Dim Size As UInt16
        Dim Res As Integer = Client.GetMaxPduSize(Size)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            lbLog.Items.Add("Get max PDU size failed: 0x" + Res.ToString("X8"))
        Else
            lbLog.Items.Add("Max PDU size: " + Size.ToString())
        End If

        GetConParams(Client)
        GetConPhy(Client)
    End Sub

    Private Sub FServer_OnClientDisconnected(Sender As Object, Client As wclBluetooth.wclGattServerClient) Handles FServer.OnClientDisconnected
        lbLog.Items.Add("Client disconnected " + Client.Address.ToString("X12"))
    End Sub

    Private Sub FServer_OnConnectionParamsChanged(Sender As Object, Client As wclBluetooth.wclGattServerClient) Handles FServer.OnConnectionParamsChanged
        GetConParams(Client)
    End Sub

    Private Sub FServer_OnConnectionPhyChanged(Sender As Object, Client As wclBluetooth.wclGattServerClient) Handles FServer.OnConnectionPhyChanged
        GetConPhy(Client)
    End Sub

    Private Sub FServer_OnMaxPduSizeChanged(Sender As Object, Client As wclBluetooth.wclGattServerClient) Handles FServer.OnMaxPduSizeChanged
        Dim Size As UInt16
        Dim Res As Int32 = Client.GetMaxPduSize(Size)
        If Res = wclErrors.WCL_E_SUCCESS Then
            lbLog.Items.Add(Client.Address.ToString("X12") + " PDU size changed: " + Size.ToString())
        Else
            lbLog.Items.Add(Client.Address.ToString("X12") + " get max PDU size failed: 0x" + Res.ToString("X8"))
        End If
    End Sub

    Private Sub FServer_OnNotificationSizeChanged(Sender As Object, Client As wclBluetooth.wclGattServerClient) Handles FServer.OnNotificationSizeChanged
        Dim Size As UInt16
        Dim Res As UInt32 = Client.GetMaxNotificationSize(Size)
        If Res = wclErrors.WCL_E_SUCCESS Then
            lbLog.Items.Add(Client.Address.ToString("X12") + " notification size changed: " +
                Size.ToString())
        Else
            lbLog.Items.Add(Client.Address.ToString("X12") + " get max notification size failed: 0x" +
                Res.ToString("X8"))
        End If
    End Sub

    Private Sub FServer_OnRead(ByVal Sender As Object, ByVal Client As wclGattServerClient, ByVal Characteristic As wclBluetooth.wclGattLocalCharacteristic, ByVal Request As wclBluetooth.wclGattLocalCharacteristicReadRequest) Handles FServer.OnRead
        lbLog.Items.Add("Read request from " + Client.Address.ToString("X12"))
        lbLog.Items.Add("  Offset: " + Request.Offset.ToString())
        lbLog.Items.Add("  Read Buffer Size: " + Request.Size.ToString())

        Dim Res As Integer
        Dim Tmp() As Byte
        Dim IsShort As Boolean = (Characteristic.Uuid.IsShortUuid And
            Characteristic.Uuid.ShortUuid = &HFFF2)
        If IsShort Then
            Res = Request.Respond(FShortData)
        Else
            ReDim Tmp(LONG_DATA_LEN - Request.Offset)
            Array.Copy(FLongData, Request.Offset, Tmp, 0, LONG_DATA_LEN - Request.Offset)
            Res = Request.Respond(Tmp)
        End If
        If Res <> wclErrors.WCL_E_SUCCESS Then
            lbLog.Items.Add("  Set data failed: 0x" + Res.ToString("X8"))
            lbLog.Items.Add("  respond with error")
            Res = Request.RespondWithError(wclBluetoothErrors.WCL_E_BLUETOOTH_LE_UNLIKELY)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                lbLog.Items.Add("  Respond failed: 0x" + Res.ToString("X8"))
            End If
        End If
    End Sub

    Private Sub FServer_OnStarted(ByVal sender As Object, ByVal e As System.EventArgs) Handles FServer.OnStarted
        lbLog.Items.Add("Server has been started")
        FStarted = True
        FCounter = 0
    End Sub

    Private Sub FServer_OnStopped(ByVal sender As Object, ByVal e As System.EventArgs) Handles FServer.OnStopped
        lbLog.Items.Add("Server has been stopped")
        FStarted = False
    End Sub

    Private Sub FServer_OnSubscribed(ByVal Sender As Object, ByVal Client As wclGattServerClient, ByVal Characteristic As wclBluetooth.wclGattLocalCharacteristic) Handles FServer.OnSubscribed
        Dim Str As String = "Client subscribed: " + Client.Address.ToString("X12")

        Dim Size As UInt16
        Dim Res As Int32 = Client.GetMaxNotificationSize(Size)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            Str = Str + " Get max notification size failed: 0x" + Res.ToString("X8")
        Else
            Str = Str + " Max notification size: " + Size.ToString()
        End If
        lbLog.Items.Add(Str)

        Dim a As Byte() = BitConverter.GetBytes(FCounter)
        Res = Characteristic.Notify(Client.Address, a)
        FCounter = FCounter + 1
        If Res <> wclErrors.WCL_E_SUCCESS Then
            lbLog.Items.Add("Notification failed: " + Res.ToString("X8"))
        End If
    End Sub

    Private Sub FServer_OnUnsubscribed(ByVal Sender As Object, ByVal Client As wclGattServerClient, ByVal Characteristic As wclBluetooth.wclGattLocalCharacteristic) Handles FServer.OnUnsubscribed
        lbLog.Items.Add("Client unsubscribed :" + Client.Address.ToString("X12"))
    End Sub

    Private Sub FServer_OnWrite(ByVal Sender As Object, ByVal Client As wclGattServerClient, ByVal Characteristic As wclBluetooth.wclGattLocalCharacteristic, ByVal Request As wclGattLocalCharacteristicWriteRequest) Handles FServer.OnWrite
        lbLog.Items.Add("Data received from " + Client.Address.ToString("X12"))
        lbLog.Items.Add("  Size: " + Request.Data.Length.ToString())
        lbLog.Items.Add("  Offset: " + Request.Offset.ToString())

        Dim s As String = ""
        For Each b As Byte In Request.Data
            s = s + b.ToString("X2")
        Next b
        lbLog.Items.Add(s)

        If Request.WithResponse Then
            lbLog.Items.Add("  Write With Response. Sending response")
            Dim Res As Integer = Request.Respond()
            If Res <> wclErrors.WCL_E_SUCCESS Then
                lbLog.Items.Add("  Respond failed: 0x" + Res.ToString("X8"))
            End If
        End If
    End Sub

    Private Function InitBluetooth(ByRef Radio As wclBluetoothRadio) As Boolean
        Radio = Nothing

        FManager = New wclBluetoothManager()

        Dim Result As Boolean = False

        Dim Res As Int32 = FManager.Open()
        If Res <> wclErrors.WCL_E_SUCCESS Then
            lbLog.Items.Add("Bluetooth Manager open failed: 0x" + Res.ToString("X8"))
        Else
            lbLog.Items.Add("Checking for working Radio")
            Res = FManager.GetLeRadio(Radio)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                lbLog.Items.Add("Get workign radio failed: 0x" + Res.ToString("X8"))
            Else
                lbLog.Items.Add("Use " + Radio.ApiName + " radio")
                Result = True
            End If

            If Not Result Then
                FManager.Close()
            End If
        End If

        If Not Result Then
            FManager = Nothing
        End If

        Return Result
    End Function

    Private Sub UninitBluetooth()
        lbLog.Items.Add("Close Bluetooth manager")
        If FManager.Active Then
            FManager.Close()
        End If
        FManager = Nothing
    End Sub

    Private Function InitServer(ByVal Radio As wclBluetoothRadio) As Boolean
        FServer = New wclGattServer()

        lbLog.Items.Add("Initialize server")
        Dim Res As Int32 = FServer.Initialize(Radio)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            lbLog.Items.Add("Initialize server failed: 0x" + Res.ToString("X8"))
            FServer = Nothing
            Return False
        End If

        lbLog.Items.Add("Server initialized")
        Return True
    End Function

    Private Sub UninitServer()
        lbLog.Items.Add("Uninitializing server")
        Dim Res As Int32 = FServer.Uninitialize()
        If Res <> wclErrors.WCL_E_SUCCESS Then
            lbLog.Items.Add("Uninitialize server failed: 0x" + Res.ToString("X8"))
        End If

        FServer = Nothing
    End Sub

    Private Function AddCharacteristics(ByVal Service As wclGattLocalService) As Boolean
        lbLog.Items.Add("Adding characteristics services")
        Dim Uuid As wclGattUuid = New wclGattUuid()
        Uuid.IsShortUuid = True

        Dim Params As wclGattLocalCharacteristicParameters = New wclGattLocalCharacteristicParameters()
        Dim [Char] As wclGattLocalCharacteristic = Nothing

        Params.Descriptors = Nothing
        Params.PresentationFormats = Nothing

        lbLog.Items.Add("Add empty characteristic")
        Uuid.ShortUuid = &HFFF1
        Params.Props = wclGattLocalCharacteristicProperty.cpNone
        Params.UserDescription = "Empty characteristic"
        Dim Res As Int32 = Service.AddCharacteristic(Uuid, Params, [Char])
        If Res <> wclErrors.WCL_E_SUCCESS Then
            lbLog.Items.Add("Failed to add empty characteristic: 0x" + Res.ToString("X8"))
            Return False
        End If

        lbLog.Items.Add("Add short readable characteristic")
        Uuid.ShortUuid = &HFFF2
        Params.Props = wclGattLocalCharacteristicProperty.cpReadable
        Params.UserDescription = "Short readable characteristic"
        Res = Service.AddCharacteristic(Uuid, Params, [Char])
        If Res <> wclErrors.WCL_E_SUCCESS Then
            lbLog.Items.Add("Failed to add short readable characteristic: 0x" + Res.ToString("X8"))
            Return False
        End If

        lbLog.Items.Add("Add long readable characteristic")
        Uuid.ShortUuid = &HFFF3
        Params.Props = wclGattLocalCharacteristicProperty.cpReadable
        Params.UserDescription = "Long readable characteristic"
        Res = Service.AddCharacteristic(Uuid, Params, [Char])
        If Res <> wclErrors.WCL_E_SUCCESS Then
            lbLog.Items.Add("Failed to add long readable characteristic: 0x" + Res.ToString("X8"))
            Return False
        End If

        lbLog.Items.Add("Add writable characteristic")
        Uuid.ShortUuid = &HFFF4
        Params.Props = wclGattLocalCharacteristicProperty.cpWritable Or wclGattLocalCharacteristicProperty.cpWritableWithoutResponse
        Params.UserDescription = "Writable characteristic"
        Res = Service.AddCharacteristic(Uuid, Params, [Char])
        If Res <> wclErrors.WCL_E_SUCCESS Then
            lbLog.Items.Add("Failed to add writable characteristic: 0x" + Res.ToString("X8"))
            Return False
        End If

        lbLog.Items.Add("Add notifiable characteristic")
        Uuid.ShortUuid = &HFFF5
        Params.Props = wclGattLocalCharacteristicProperty.cpNotifiable
        Res = Service.AddCharacteristic(Uuid, Params, [Char])
        Params.UserDescription = "Notifiable characteristic"
        If Res <> wclErrors.WCL_E_SUCCESS Then
            lbLog.Items.Add("Failed to add notifiable characteristic: 0x" + Res.ToString("X8"))
            Return False
        End If

        lbLog.Items.Add("Add indicatable characteristic")
        Uuid.ShortUuid = &HFFF6
        Params.Props = wclGattLocalCharacteristicProperty.cpIndicatable
        Params.UserDescription = "Indicatable characteristic"
        Res = Service.AddCharacteristic(Uuid, Params, [Char])
        If Res <> wclErrors.WCL_E_SUCCESS Then
            lbLog.Items.Add("Failed to add indicatable characteristic: 0x" + Res.ToString("X8"))
            Return False
        End If

        lbLog.Items.Add("All charactertistics were added")
        Return True
    End Function

    Private Function AddServices() As Boolean
        lbLog.Items.Add("Preparing services")
        Dim Uuid As wclGattUuid = New wclGattUuid()
        Uuid.IsShortUuid = True
        Uuid.ShortUuid = &HFFF0

        lbLog.Items.Add("Add service")
        Dim Service As wclGattLocalService = Nothing
        Dim Res As Int32 = FServer.AddService(Uuid, Service)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            lbLog.Items.Add("Failed to add service: 0x" + Res.ToString("X8"))
            Return False
        End If

        lbLog.Items.Add("Service has been added")
        Return AddCharacteristics(Service)
    End Function

    Private Sub Start()
        If FStarted Then
            lbLog.Items.Add("Already started")
        Else
            Dim Radio As wclBluetoothRadio = Nothing
            If InitBluetooth(Radio) Then
                Dim Result As Boolean = InitServer(Radio)
                If Result Then
                    Result = AddServices()
                    If Result Then
                        lbLog.Items.Add("Starting server")
                        Dim Res As Int32 = FServer.Start()
                        If Res <> wclErrors.WCL_E_SUCCESS Then
                            lbLog.Items.Add("Start server failed: 0x" + Res.ToString("X8"))
                            Result = False
                        End If
                    End If
                    If Not Result Then
                        UninitServer()
                    End If
                End If
                If Not Result Then
                    UninitBluetooth()
                End If
            End If
        End If
    End Sub

    Private Sub [Stop]()
        If Not FStarted Then
            lbLog.Items.Add("Not started")
        Else
            lbLog.Items.Add("Stopping server")
            Dim Res As Int32 = FServer.Stop()
            If Res <> wclErrors.WCL_E_SUCCESS Then
                lbLog.Items.Add("Stop server failed: 0x" + Res.ToString("X8"))
            End If

            UninitServer()
            UninitBluetooth()
        End If
    End Sub

    Private Sub btStop_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btStop.Click
        [Stop]()
    End Sub

    Private Sub fmMain_FormClosing(ByVal sender As System.Object, ByVal e As System.Windows.Forms.FormClosingEventArgs) Handles MyBase.FormClosing
        [Stop]()
    End Sub

    Private Sub fmMain_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load
        FStarted = False
        
        For i As UInt16 = 0 To SHORT_DATA_LEN - 1
            FShortData(i) = i + 1
        Next i
        For i As UInt16 = 0 To LONG_DATA_LEN - 1
            FLongData(i) = i And 255
        Next i
    End Sub

    Private Sub GetConParams(Client As wclGattServerClient)
        Dim Params As wclBluetoothLeConnectionParameters
        Dim Res As Int32 = Client.GetConnectionParams(Params)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            lbLog.Items.Add("Get connection params failed: 0x" + Res.ToString("X8"))
        Else
            lbLog.Items.Add("Connection params [" + Client.Address.ToString("X12") + "]")
            lbLog.Items.Add("  Interval     : " + Params.Interval.ToString())
            lbLog.Items.Add("  Latency      : " + Params.Latency.ToString())
            lbLog.Items.Add("  Link Timeout : " + Params.LinkTimeout.ToString())
        End If
    End Sub

    Private Sub GetConPhy(Client As wclGattServerClient)
        Dim Phy As wclBluetoothLeConnectionPhy
        Dim Res As Int32 = Client.GetConnectionPhyInfo(Phy)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            lbLog.Items.Add("Get connection PHY failed: 0x" + Res.ToString("X8"))
        Else
            lbLog.Items.Add("Connection PHY [" + Client.Address.ToString("X12") + "]")
            lbLog.Items.Add("  Receive")
            lbLog.Items.Add("    IsCoded        : " + Phy.Receive.IsCoded.ToString())
            lbLog.Items.Add("    IsUncoded1MPhy : " + Phy.Receive.IsUncoded1MPhy.ToString())
            lbLog.Items.Add("    IsUncoded2MPhy : " + Phy.Receive.IsUncoded2MPhy.ToString())
            lbLog.Items.Add("  Transmit")
            lbLog.Items.Add("    IsCoded        : " + Phy.Transmit.IsCoded.ToString())
            lbLog.Items.Add("    IsUncoded1MPhy : " + Phy.Transmit.IsUncoded1MPhy.ToString())
            lbLog.Items.Add("    IsUncoded2MPhy : " + Phy.Transmit.IsUncoded2MPhy.ToString())
        End If
    End Sub

    Private Sub btClear_Click(sender As System.Object, e As System.EventArgs) Handles btClear.Click
        lbLog.Items.Clear()
    End Sub
End Class
