Public Class fmMain
    Private WithEvents wclBluetoothManager As wclBluetoothManager
    Private WithEvents wclRfCommClient As wclRfCommClient
    Private WithEvents FOpp As wclObexOppClient

    Private Function GetRadio() As wclBluetoothRadio
        Dim Radio As wclBluetoothRadio = Nothing
        Dim Res As Int32 = wclBluetoothManager.GetClassicRadio(Radio)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Get working radio failed: 0x" + Res.ToString("X8"))
            Return Nothing
        End If
        Return Radio
    End Function

    Private Sub fmMain_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load
        wclBluetoothManager = New wclBluetoothManager()
        wclRfCommClient = New wclRfCommClient()
            
        ' We use ObjectPush Profile so must set the profile's UUID.
        wclRfCommClient.Service = wclUUIDs.OBEXObjectPushServiceClass_UUID

        FOpp = Nothing

        ' Try to open Bluetooth Manager.
        Dim Res As Int32 = wclBluetoothManager.Open()
        If Res <> wclErrors.WCL_E_SUCCESS Then
            lbLog.Items.Add("Bluetooth Manager Open failed: 0x" + Res.ToString("X8"))
        End If
    End Sub

    Private Sub wclRfCommClient_OnConnect(ByVal Sender As Object, ByVal [Error] As Integer) Handles wclRfCommClient.OnConnect
        lbLog.Items.Add("Connect. Operation result: 0x" + [Error].ToString("X8"))
    End Sub

    Private Sub wclRfCommClient_OnCreateProcessor(ByVal Sender As Object, ByVal Connection As wclCommunication.wclClientDataConnection) Handles wclRfCommClient.OnCreateProcessor
        ' here we must create the data processor for the connection. In this demo
        ' we use OPPClient.
        FOpp = New wclObexOppClient(Connection)
    End Sub

    Private Sub FOpp_OnConnect(ByVal Sender As Object, ByVal [Error] As Integer, ByVal Description As String) Handles FOpp.OnConnect
        lbLog.Items.Add("OBEX session opened with result: 0x" + [Error].ToString("X8") + ". Description: " + Description)
    End Sub

    Private Sub FOpp_OnDisconnect(ByVal Sender As Object, ByVal [Error] As Integer, ByVal Description As String) Handles FOpp.OnDisconnect
        lbLog.Items.Add("OBEX session closed with result: 0x" + [Error].ToString("X8") + ". Description: " + Description)
    End Sub

    Private Sub FOpp_OnGetComplete(ByVal Sender As Object, ByVal [Error] As Integer, ByVal Description As String, ByVal Stream As System.IO.Stream) Handles FOpp.OnGetComplete
        lbLog.Items.Add("File receiving completed with result: 0x" + [Error].ToString("X8") + ". Description: " + Description)

        ProgressBar.Value = 0
        ProgressBar.Maximum = 0

        ' We must destroy the stream object here.
        Stream.Close()
        Stream.Dispose()
    End Sub

    Private Sub FOpp_OnProgress(ByVal Sender As Object, ByVal Size As UInteger, ByVal Position As UInteger) Handles FOpp.OnProgress
        ProgressBar.Maximum = Size
        ProgressBar.Value = Position
    End Sub

    Private Sub FOpp_OnPutComplete(ByVal Sender As Object, ByVal [Error] As Integer, ByVal Description As String, ByVal Stream As System.IO.Stream) Handles FOpp.OnPutComplete
        lbLog.Items.Add("File sending completed with result: 0x" + [Error].ToString("X8") + ". Description: " + Description)

        ProgressBar.Value = 0
        ProgressBar.Maximum = 0

        ' We must destroy the stream object here.
        Stream.Close()
        Stream.Dispose()
    End Sub

    Private Sub wclRfCommClient_OnDestroyProcessor(ByVal Sender As Object, ByVal Connection As wclCommunication.wclClientDataConnection) Handles wclRfCommClient.OnDestroyProcessor
        ' Do we have any data processor created?
        If Not FOpp Is Nothing Then
            ' Make sure it is our connection is goind to destroy.
            If Connection.Processor Is FOpp Then
                ' Ok, destroy the data processor here.
                FOpp.Dispose()
                FOpp = Nothing
            End If
        End If
    End Sub

    Private Sub wclRfCommClient_OnDisconnect(ByVal Sender As Object, ByVal Reason As Integer) Handles wclRfCommClient.OnDisconnect
        lbLog.Items.Add("Disconnected with reason: 0x" + Reason.ToString("X8"))
    End Sub

    Private Sub wclBluetoothManager_AfterOpen(ByVal sender As Object, ByVal e As System.EventArgs) Handles wclBluetoothManager.AfterOpen
        lbLog.Items.Add("Bluetooth Manager has been opened")
    End Sub

    Private Sub wclBluetoothManager_OnAuthenticationCompleted(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal Address As Long, ByVal [Error] As Integer) Handles wclBluetoothManager.OnAuthenticationCompleted
        lbLog.Items.Add("Authentication completed with result: 0x" + [Error].ToString("X8"))
    End Sub

    Private Sub wclBluetoothManager_OnDeviceFound(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal Address As Long) Handles wclBluetoothManager.OnDeviceFound
        Dim Item As ListViewItem = lvDevices.Items.Add(Address.ToString("X12"))
        Item.SubItems.Add("")
    End Sub

    Private Sub wclBluetoothManager_OnDiscoveringCompleted(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal [Error] As Integer) Handles wclBluetoothManager.OnDiscoveringCompleted
        lbLog.Items.Add("Discovering completed with result: 0x" + [Error].ToString("X8"))

        For i As Int32 = 0 To lvDevices.Items.Count - 1
            Dim Item As ListViewItem = lvDevices.Items(i)
            Dim Addr As Int64 = Convert.ToInt64(Item.Text, 16)
            Dim Name As String = ""
            Dim Res As Int32 = Radio.GetRemoteName(Addr, Name)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                Name = "Error: 0x" + Res.ToString("X8")
            End If
            Item.SubItems(1).Text = Name
        Next i
    End Sub

    Private Sub wclBluetoothManager_OnDiscoveringStarted(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio) Handles wclBluetoothManager.OnDiscoveringStarted
        lbLog.Items.Add("Discovering started")

        lvDevices.Items.Clear()
    End Sub

    Private Sub wclBluetoothManager_OnNumericComparison(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal Address As Long, ByVal Number As UInteger, ByRef Confirm As Boolean) Handles wclBluetoothManager.OnNumericComparison
        lbLog.Items.Add("Numeric comparison. Accept")
        Confirm = True
    End Sub

    Private Sub wclBluetoothManager_OnPasskeyNotification(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal Address As Long, ByVal Passkey As UInteger) Handles wclBluetoothManager.OnPasskeyNotification
        lbLog.Items.Add("Passkey notification. Passkey: " + Passkey.ToString())
    End Sub

    Private Sub wclBluetoothManager_OnPasskeyRequest(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal Address As Long, ByRef Passkey As UInteger) Handles wclBluetoothManager.OnPasskeyRequest
        lbLog.Items.Add("Passkey request. Passkey: 12345678")
        Passkey = 12345678
    End Sub

    Private Sub wclBluetoothManager_OnPinRequest(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal Address As Long, ByRef Pin As String) Handles wclBluetoothManager.OnPinRequest
        lbLog.Items.Add("PIN code request. PIN: 0000")
        Pin = "0000"
    End Sub

    Private Sub btClear_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btClear.Click
        lbLog.Items.Clear()
    End Sub

    Private Sub fmMain_FormClosed(ByVal sender As System.Object, ByVal e As System.Windows.Forms.FormClosedEventArgs) Handles MyBase.FormClosed
        wclRfCommClient.Disconnect()
        wclBluetoothManager.Close()
    End Sub

    Private Sub btDiscover_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btDiscover.Click
        Dim Radio As wclBluetoothRadio = GetRadio()
        If Not Radio Is Nothing Then
            ' Discover classic devices only.
            Dim Res As Int32 = Radio.Discover(10, wclBluetoothDiscoverKind.dkClassic)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                MessageBox.Show("Failed to start discovering: 0x" + Res.ToString("X8"))
            End If
        End If
    End Sub

    Private Sub btDisconnect_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btDisconnect.Click
        Dim Res As Int32 = wclRfCommClient.Disconnect()
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Disconnect failed: 0x" + Res.ToString("X8"))
        End If
    End Sub

    Private Sub btCloseSession_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btCloseSession.Click
        If FOpp Is Nothing Then
            MessageBox.Show("Not connected")
        Else
            Dim Res As Int32 = FOpp.Disconnect("")
            If Res <> wclErrors.WCL_E_SUCCESS Then
                MessageBox.Show("Close session failed: 0x" + Res.ToString("X8"))
            End If
        End If
    End Sub

    Private Sub btOpenSession_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btOpenSession.Click
        If FOpp Is Nothing Then
            MessageBox.Show("Not connected")
        Else
            Dim Res As Int32 = FOpp.Connect()
            If Res <> wclErrors.WCL_E_SUCCESS Then
                MessageBox.Show("Open session failed: 0x" + Res.ToString("X8"))
            End If
        End If
    End Sub

    Private Sub btConnect_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btConnect.Click
        If wclRfCommClient.State <> wclClientState.csDisconnected Then
            MessageBox.Show("Client is connected")
        Else
            If lvDevices.SelectedItems.Count = 0 Then
                MessageBox.Show("Select device")
            Else
                Dim Radio As wclBluetoothRadio = GetRadio()
                If Not Radio Is Nothing Then
                    Dim Addr As Int64 = Convert.ToInt64(lvDevices.SelectedItems(0).Text, 16)
                    wclRfCommClient.Address = Addr
                    Dim Res As Int32 = wclRfCommClient.Connect(Radio)
                    If Res <> wclErrors.WCL_E_SUCCESS Then
                        MessageBox.Show("Failed to connect: 0x" + Res.ToString("X8"))
                    End If
                End If
            End If
        End If
    End Sub

    Private Sub btAbort_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btAbort.Click
        If FOpp Is Nothing Then
            MessageBox.Show("Not connected")
        Else
            Dim Res As Int32 = FOpp.Abort("")
            If Res <> wclErrors.WCL_E_SUCCESS Then
                MessageBox.Show("Abort failed: 0x" + Res.ToString("X8"))
            End If
        End If
    End Sub

    Private Sub btSendFile_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btSendFile.Click
        If FOpp Is Nothing Then
            MessageBox.Show("Not connected")
        Else
            If edFileName.Text = "" Then
                MessageBox.Show("Select file")
            Else
                Dim Stream As FileStream = New FileStream(edFileName.Text, FileMode.Open)
                Dim Res As Int32 = FOpp.Put(Path.GetFileName(edFileName.Text), "", Stream)
                If Res <> wclErrors.WCL_E_SUCCESS Then
                    MessageBox.Show("Send file failed: 0x" + Res.ToString("X8"))
                    Stream.Close()
                    Stream.Dispose()
                End If
            End If
        End If
    End Sub

    Private Sub btGetVCard_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btGetVCard.Click
        If FOpp Is Nothing Then
            MessageBox.Show("Not connected")
        Else
            Dim Stream As FileStream = New FileStream("vcard.vcf", FileMode.CreateNew)
            ' Request default vCard object.
            Dim Res As Int32 = FOpp.Get("text/x-vCard", Stream)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                MessageBox.Show("GetvCard failed: 0x" + Res.ToString("X8"))
                Stream.Close()
                Stream.Dispose()
            End If
        End If
    End Sub

    Private Sub btGetCaps_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btGetCaps.Click
        If FOpp Is Nothing Then
            MessageBox.Show("Not connected")
        Else
            Dim Stream As FileStream = New FileStream("caps.xml", FileMode.CreateNew)
            ' Request Device Capability object.
            Dim Res As Int32 = FOpp.Get("x-obex/capability", Stream)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                MessageBox.Show("GetCaps failed: 0x" + Res.ToString("X8"))
                Stream.Close()
                Stream.Dispose()
            End If
        End If
    End Sub

    Private Sub btSelectFile_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btSelectFile.Click
        If openFileDialog.ShowDialog() = DialogResult.OK Then
            edFileName.Text = openFileDialog.FileName
        End If
    End Sub
End Class
