Imports System.IO

Public Class fmMain
    Private WithEvents wclBluetoothManager As wclBluetoothManager
    Private WithEvents wclRfCommServer As wclRfCommServer
    Private FFileName As String

    Private Sub Trace(ByVal Msg As String)
        lbLog.Items.Add(Msg)
    End Sub

    Private Sub fmMain_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load
        wclBluetoothManager = New wclBluetoothManager()

        wclRfCommServer = New wclRfCommServer()
        wclRfCommServer.Authentication = False
        wclRfCommServer.Service = wclUUIDs.OBEXObjectPushServiceClass_UUID
        wclRfCommServer.ServiceName = "WCL OPP Server"
    End Sub

    Private Sub wclRfCommServer_GetSdpAttributes(ByVal Sender As Object, ByRef Protocols() As UShort, ByRef Profiles() As wclBluetooth.wclBluetoothSdpProfile, ByRef Formats() As Byte, ByRef Cod As UInteger) Handles wclRfCommServer.GetSdpAttributes
        ' We have to provide additional information for SDP record.
        Protocols = New UShort(1) {}
        Protocols(0) = wclUUIDs.OBEX_PROTOCOL_UUID16
        Profiles = New wclBluetoothSdpProfile(1) {}
        Profiles(0).Uuid = wclUUIDs.OBEXObjectPushServiceClass_UUID16
        Profiles(0).Version = &H100
        Formats = New Byte(1) {}
        Formats(0) = &HFF ' All formats.
        Cod = wclCODs.COD_SERVICE_OBJECT_XFER
    End Sub

    Private Sub wclRfCommServer_OnClosed(ByVal Sender As Object, ByVal Reason As Integer) Handles wclRfCommServer.OnClosed
        Trace("Server closed. Reason:: 0x" + Reason.ToString("X8"))
    End Sub

    Private Sub wclRfCommServer_OnConnect(ByVal Sender As Object, ByVal Client As wclBluetooth.wclRfCommServerClientConnection, ByVal [Error] As Integer) Handles wclRfCommServer.OnConnect
        If [Error] = wclErrors.WCL_E_SUCCESS Then
            Trace("Client " + Client.Address.ToString("X12") + " connected")
        Else
            Trace("Client " + Client.Address.ToString("X12") + " connect failed: 0x" + [Error].ToString("X8"))
        End If
    End Sub

    Private Sub wclRfCommServer_OnCreateProcessor(ByVal Sender As Object, ByVal Connection As wclCommunication.wclServerClientDataConnection) Handles wclRfCommServer.OnCreateProcessor
        ' Here we must create OBEX OPP server data processor.
        Dim Proc As wclObexOppServer = New wclObexOppServer(Connection)
        AddHandler Proc.OnPutProgress, AddressOf Proc_OnPutProgress
        AddHandler Proc.OnPutCompleted, AddressOf Proc_OnPutCompleted
        AddHandler Proc.OnPutBegin, AddressOf Proc_OnPutBegin
        AddHandler Proc.OnDisconnected, AddressOf Proc_OnDisconnected
        AddHandler Proc.OnConnect, AddressOf Proc_OnConnect
        AddHandler Proc.OnGetRequest, AddressOf Proc_OnGetRequest
        AddHandler Proc.OnGetCompleted, AddressOf Proc_OnGetCompleted
        AddHandler Proc.OnProgress, AddressOf Proc_OnProgress
    End Sub

    Private Sub Proc_OnGetCompleted(Sender As Object, [Error] As Integer, Stream As Stream)
        Trace("Get operation completed with result: 0x" + [Error].ToString("X8"))
        If Not Stream Is Nothing Then
            Stream.Close()
        End If
    End Sub

    Private Sub Proc_OnProgress(Sender As Object, Length As UInteger, Position As UInteger)
        Trace("Sending file: " + Position.ToString() + " from " + Length.ToString())
    End Sub

    Private Sub Proc_OnGetRequest(Sender As Object, aType As String, ByRef Result As wclObexServerOperationResult, ByRef Stream As Stream)
        If aType = "text/x-vCard" OrElse aType = "x-obex/capability" Then
            Trace("Requested file type: " + aType)

            If openFileDialog.ShowDialog() <> DialogResult.OK Then
                Result = wclObexServerOperationResult.orForbidden
                Stream = Nothing
            Else
                Stream = New FileStream(openFileDialog.FileName, FileMode.Open)
                Result = wclObexServerOperationResult.orSuccess
            End If
        Else
            Result = wclObexServerOperationResult.orUnsupportedMedia
            Stream = Nothing
        End If
    End Sub

    Private Sub Proc_OnPutCompleted(ByVal Sender As Object, ByVal [Error] As Integer, ByVal Stream As System.IO.Stream, ByRef Accept As Boolean)
        Dim Proc As wclObexOppServer = CType(Sender, wclObexOppServer)
        Dim Address As Int64 = CType(Proc.Connection, wclRfCommServerClientConnection).Address
        If [Error] = wclErrors.WCL_E_SUCCESS Then
            Dim Str As String = "Client " + Address.ToString("X12") + " completed sent file. Accept?"
            Accept = MessageBox.Show(Str, "Accept?", MessageBoxButtons.YesNo, MessageBoxIcon.Question) = DialogResult.Yes

            If Accept Then
                SaveFileDialog.FileName = FFileName
                If SaveFileDialog.ShowDialog() = DialogResult.OK Then
                    Stream.Seek(0, IO.SeekOrigin.Begin)
                    Dim fs As IO.FileStream = New System.IO.FileStream(SaveFileDialog.FileName, IO.FileMode.CreateNew, IO.FileAccess.Write)
                    Dim Bytes(Stream.Length) As Byte
                    Stream.Read(Bytes, 0, Stream.Length)
                    fs.Write(Bytes, 0, Bytes.Length)
                    fs.Close()
                End If
            End If
        Else
            Trace("Send file error: 0x" + [Error].ToString("X8"))
            Accept = False ' does not matter.
        End If
    End Sub

    Private Sub Proc_OnPutProgress(ByVal Sender As Object, ByVal Position As UInteger, ByVal Length As UInteger, ByRef [Continue] As Boolean)
        Trace("Sending file: " + Position.ToString() + " from " + Length.ToString())
        ' Do not foget to set Continue flag!
        [Continue] = True
    End Sub

    Private Sub Proc_OnConnect(ByVal Sender As Object, ByVal Description As String)
        Dim Proc As wclObexOppServer = CType(Sender, wclObexOppServer)
        Dim Address As Int64 = CType(Proc.Connection, wclRfCommServerClientConnection).Address
        Dim Accept As Boolean = MessageBox.Show("Client " + Address.ToString("X12") + " requsts for connection: " + Description + ". Accept?", "Accept?", MessageBoxButtons.YesNo, MessageBoxIcon.Question) = DialogResult.Yes

        Dim Res As Integer
        If Accept Then
            Res = Proc.Accept("Accepted by user")
            If Res = wclErrors.WCL_E_SUCCESS Then
                Dim Item As ListViewItem = lvClients.Items.Add(Address.ToString("X12"))
                Item.Tag = Sender
                Trace("OBEX session opened for client: " + Address.ToString("X12"))
            End If
        Else
            Res = Proc.Reject("Rejected by user")
        End If

        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Operation failed: 0x" + Res.ToString("X8"))
            ' Disconnect client!
            Proc.Connection.Disconnect()
        End If
    End Sub

    Private Sub Proc_OnDisconnected(ByVal Sender As Object, ByVal Reason As Integer, ByVal Description As String)
        Dim Proc As wclObexOppServer = CType(Sender, wclObexOppServer)
        Dim Address As Int64 = CType(Proc.Connection, wclRfCommServerClientConnection).Address

        Trace("Client " + Address.ToString("X12") + " closed OBEX session with reason: 0x" + Reason.ToString("X8") + ". Description: " + Description)

        For i As Integer = 0 To lvClients.Items.Count - 1
            If lvClients.Items(i).Tag Is Sender Then
                lvClients.Items.RemoveAt(i)
                Exit For

            End If
        Next i
    End Sub

    Private Sub Proc_OnPutBegin(ByVal Sender As Object, ByVal Name As String, ByVal Description As String, ByVal Mime As String, ByVal Length As UInteger, ByRef Accept As Boolean)
        Dim Proc As wclObexOppServer = CType(Sender, wclObexOppServer)
        Dim Address As Int64 = CType(Proc.Connection, wclRfCommServerClientConnection).Address
        Dim Str As String = "Client " + Address.ToString("X12") + " send file " + Name + ". Accept?"
        Accept = MessageBox.Show(Str, "Accept?", MessageBoxButtons.YesNo, MessageBoxIcon.Question) = DialogResult.Yes
        If Accept Then
            FFileName = Name
        End If
    End Sub

    Private Sub wclRfCommServer_OnDestroyProcessor(ByVal Sender As Object, ByVal Connection As wclCommunication.wclServerClientDataConnection) Handles wclRfCommServer.OnDestroyProcessor
        Trace("Data processor for client " + CType(Connection, wclRfCommServerClientConnection).Address.ToString("X12") + " is destroing")
        Connection.Processor.Dispose()
    End Sub

    Private Sub wclRfCommServer_OnDisconnect(ByVal Sender As Object, ByVal Client As wclBluetooth.wclRfCommServerClientConnection, ByVal Reason As Integer) Handles wclRfCommServer.OnDisconnect
        Trace("Client " + Client.Address.ToString("X12") + " disconnected. Reason: 0x" + Reason.ToString("X8"))
    End Sub

    Private Sub wclRfCommServer_OnListen(ByVal sender As Object, ByVal e As System.EventArgs) Handles wclRfCommServer.OnListen
        Trace("Server listening")
    End Sub

    Private Sub fmMain_FormClosed(ByVal sender As System.Object, ByVal e As System.Windows.Forms.FormClosedEventArgs) Handles MyBase.FormClosed
        wclRfCommServer.Close()
        wclBluetoothManager.Close()

        wclRfCommServer = Nothing
        wclBluetoothManager = Nothing
    End Sub

    Private Sub btClear_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btClear.Click
        lbLog.Items.Clear()
    End Sub

    Private Sub btListen_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btListen.Click
        Dim Res As Int32 = wclBluetoothManager.Open()
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Bluetooth manager open failed: 0x" + Res.ToString("X8"))
        Else
            Dim Radio As wclBluetoothRadio = Nothing
            Res = wclBluetoothManager.GetClassicRadio(Radio)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                MessageBox.Show("Get working radio failed: 0x" + Res.ToString("X8"))
            Else
                ' Switch radio in discoverable and connectable mode.
                ' Ignore any errors. In real application you need to check the result.
                Radio.SetDiscoverable(True)
                Radio.SetConnectable(True)

                ' As wrote above it is better to start listening in AfterOpen event
                ' handler. But again, for demo app its OK to do it here.
                Res = wclRfCommServer.Listen(Radio)
                If Res <> wclErrors.WCL_E_SUCCESS Then
                    MessageBox.Show("Failed to start listening: 0x" + Res.ToString("X8"))
                End If
            End If

            If Res <> wclErrors.WCL_E_SUCCESS Then
                ' We must close Bluetooth Manager here!
                wclBluetoothManager.Close()
            End If
        End If
    End Sub

    Private Sub btClose_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btClose.Click
        Dim Res As Int32 = wclRfCommServer.Close()
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Close failed; 0x" + Res.ToString("X8"))
        Else
            wclBluetoothManager.Close()
        End If
    End Sub

    Private Sub btDisconnect_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btDisconnect.Click
        If lvClients.SelectedItems Is Nothing Then
            MessageBox.Show("Select client")
        Else
            Dim Proc As wclObexOppServer = CType(lvClients.SelectedItems(0).Tag, wclObexOppServer)
            Dim Res As Int32 = Proc.Connection.Disconnect()
            If Res <> wclErrors.WCL_E_SUCCESS Then
                MessageBox.Show("Disconnect failed: 0x" + Res.ToString("X8"))
            End If
        End If
    End Sub
End Class
