Public Class fmMain
    Private WithEvents Manager As wclBluetoothManager
    Private WithEvents Client As wclRfCommClient

    Private Sub fmMain_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load
        Manager = New wclBluetoothManager()
        Client = New wclRfCommClient()

        ' In real application you should always analize the result code.
        ' In this demo we assume that all is always OK.
        Manager.Open()

        cbAuthentication.Checked = Client.Authentication
        cbEncryption.Checked = Client.Encryption
    End Sub

    Private Sub Client_OnConnect(ByVal Sender As Object, ByVal [Error] As Integer) Handles Client.OnConnect
        If [Error] = wclErrors.WCL_E_SUCCESS Then
            lbEvents.Items.Add("Connected")

            GetBuffers()
        Else
            lbEvents.Items.Add("Connect error: 0x" + [Error].ToString("X8"))
        End If
    End Sub

    Private Sub Client_OnData(ByVal Sender As Object, ByVal Data() As Byte) Handles Client.OnData
        If Data IsNot Nothing AndAlso Data.Length > 0 Then
            Dim Str As String = System.Text.Encoding.ASCII.GetString(Data)

            lbEvents.Items.Add("Received: " + Str)
        End If
    End Sub

    Private Sub Client_OnDisconnect(ByVal Sender As Object, ByVal Reason As Integer) Handles Client.OnDisconnect
        lbEvents.Items.Add("Disconnected: 0x" + Reason.ToString("X8"))

        edReadBuffer.Text = "0"
        edWriteBuffer.Text = "0"
    End Sub

    Private Sub Manager_OnAuthenticationCompleted(Sender As Object, Radio As wclBluetooth.wclBluetoothRadio, Address As Long, [Error] As Integer) Handles Manager.OnAuthenticationCompleted
        lbEvents.Items.Add("Authentication completed: " + [Error].ToString("X8"))
    End Sub

    Private Sub Manager_OnConfirm(Sender As Object, Radio As wclBluetooth.wclBluetoothRadio, Address As Long, ByRef Confirm As Boolean) Handles Manager.OnConfirm
        Confirm = True
        lbEvents.Items.Add("Just works pairing")
    End Sub

    Private Sub Manager_OnDeviceFound(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal Address As Long) Handles Manager.OnDeviceFound
        Dim DevType As wclBluetoothDeviceType = wclBluetoothDeviceType.dtMixed
        Dim Res As Int32 = Radio.GetRemoteDeviceType(Address, DevType)

        Dim Item As ListViewItem = lvDevices.Items.Add(Address.ToString("X12"))
        Item.SubItems.Add("") ' We can not read a device's name here.
        Item.Tag = Radio ' To use it later.
        If Res <> wclErrors.WCL_E_SUCCESS Then
            Item.SubItems.Add("Error: 0x" + Res.ToString("X8"))
        Else
            Select Case DevType
                Case wclBluetoothDeviceType.dtClassic
                    Item.SubItems.Add("Classic")
                Case wclBluetoothDeviceType.dtBle
                    Item.SubItems.Add("BLE")
                Case wclBluetoothDeviceType.dtMixed
                    Item.SubItems.Add("Mixed")
                Case Else
                    Item.SubItems.Add("Unknown")
            End Select
        End If

        lbEvents.Items.Add("Device found: " + Address.ToString("X12"))
    End Sub

    Private Sub Manager_OnDiscoveringCompleted(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal [Error] As Integer) Handles Manager.OnDiscoveringCompleted
        If lvDevices.Items.Count = 0 Then
            MessageBox.Show("No devices were found.", "Information", MessageBoxButtons.OK, MessageBoxIcon.Information)
        Else
            ' Here we can update found devices names.
            For Each Item As ListViewItem In lvDevices.Items
                Dim Address As Int64 = Convert.ToInt64(Item.Text, 16)
                Dim DevName As String = ""
                Dim Res As Int32 = Radio.GetRemoteName(Address, DevName)
                If Res <> wclErrors.WCL_E_SUCCESS Then
                    Item.SubItems(1).Text = "Error: 0x" + Res.ToString("X8")
                Else
                    Item.SubItems(1).Text = DevName
                End If
            Next Item
        End If

        lbEvents.Items.Add("Discovering completed")
    End Sub

    Private Sub Manager_OnDiscoveringStarted(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio) Handles Manager.OnDiscoveringStarted
        lvDevices.Items.Clear()
        lbEvents.Items.Add("Discovering started")
    End Sub

    Private Sub Manager_OnNumericComparison(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal Address As Long, ByVal Number As UInteger, ByRef Confirm As Boolean) Handles Manager.OnNumericComparison
        ' Accept any pairing.
        Confirm = True
        lbEvents.Items.Add("Numeric comparison: " + Number.ToString())
    End Sub

    Private Sub Manager_OnPasskeyNotification(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal Address As Long, ByVal Passkey As UInteger) Handles Manager.OnPasskeyNotification
        lbEvents.Items.Add("Passkey notification: " + Passkey.ToString())
    End Sub

    Private Sub Manager_OnPasskeyRequest(Sender As Object, Radio As wclBluetooth.wclBluetoothRadio, Address As Long, ByRef Passkey As UInteger) Handles Manager.OnPasskeyRequest
        Passkey = 123456
        lbEvents.Items.Add("Passkey request: " + Passkey.ToString())
    End Sub

    Private Sub Manager_OnPinRequest(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal Address As Long, ByRef Pin As String) Handles Manager.OnPinRequest
        ' Use '0000' as PIN.
        Pin = "0000"
        lbEvents.Items.Add("Pin request: " + Pin)
    End Sub

    Private Function GetRadio() As wclBluetoothRadio
        Dim Radio As wclBluetoothRadio = Nothing
        Dim Res As Int32 = Manager.GetClassicRadio(Radio)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Get working radio failed: 0x" + Res.ToString("X8"))
            Return Nothing
        End If
        Return Radio
    End Function

    Private Sub fmMain_FormClosed(ByVal sender As System.Object, ByVal e As System.Windows.Forms.FormClosedEventArgs) Handles MyBase.FormClosed
        Client.Disconnect()
        Client = Nothing

        Manager.Close()
        Manager = Nothing
    End Sub

    Private Sub btDiscover_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btDiscover.Click
        Dim Radio As wclBluetoothRadio = GetRadio()
        If Radio IsNot Nothing Then
            If cbEnumPaired.Checked Then
                Dim Devices As Int64()
                Dim Res As Integer = Radio.EnumPairedDevices(Devices)
                If Res <> wclErrors.WCL_E_SUCCESS Then
                    MessageBox.Show("Enum paired devices failed: 0x" + Res.ToString("X8"))
                Else
                    lvDevices.Items.Clear()
                    If Devices IsNot Nothing AndAlso Devices.Length > 0 Then
                        For Each Device As Int64 In Devices
                            Manager_OnDeviceFound(Manager, Radio, Device)
                        Next
                        Manager_OnDiscoveringCompleted(Manager, Radio, wclErrors.WCL_E_SUCCESS)
                    End If
                End If
            Else
                Dim Res As Int32 = Radio.Discover(10, wclBluetoothDiscoverKind.dkClassic)
                If Res <> wclErrors.WCL_E_SUCCESS Then
                    MessageBox.Show("Error starting discovering: 0x" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
                End If
            End If
        End If
    End Sub

    Private Sub btClear_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btClear.Click
        lbEvents.Items.Clear()
    End Sub

    Private Sub btConnect_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btConnect.Click
        If lvDevices.SelectedItems.Count = 0 Then
            MessageBox.Show("Select device", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            Dim Radio As wclBluetoothRadio = GetRadio()
            If Radio IsNot Nothing Then
                Client.Address = Convert.ToInt64(lvDevices.SelectedItems(0).Text, 16)
                Client.Authentication = cbAuthentication.Checked
                Client.Encryption = cbEncryption.Checked
                Client.Timeout = Convert.ToInt32(edTimeout.Text)

                Client.Channel = 0
                Client.Service = wclUUIDs.SerialPortServiceClass_UUID

                Dim Connect As Boolean = False
                If cbServiceName.Checked Then
                    Dim Service As Guid = Client.Service
                    Dim Services As wclBluetoothService() = Nothing
                    Dim Res As Int32 = Radio.EnumRemoteServices(Client.Address, Service, Services)
                    If Res <> wclErrors.WCL_E_SUCCESS Then
                        MessageBox.Show("Service enumerating error: 0x" + Res.ToString("X8"))
                    Else
                        If Services Is Nothing OrElse Services.Length = 0 Then
                            MessageBox.Show("Services not found")
                        Else
                            Dim s As wclBluetoothService
                            For Each s In Services
                                If s.Name = edServiceName.Text Then
                                    Client.Channel = s.Channel
                                    Connect = True
                                    Exit For
                                End If
                            Next s

                            If Not Connect Then
                                MessageBox.Show("Service not found")
                            End If
                        End If
                    End If
                Else
                    Connect = True
                End If

                If Connect Then
                    Dim Res As Int32 = Client.Connect(Radio)
                    If Res <> wclErrors.WCL_E_SUCCESS Then
                        MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
                    End If
                End If
            End If
        End If
    End Sub

    Private Sub btDisconnect_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btDisconnect.Click
        Dim Res As Int32 = Client.Disconnect()
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
        End If
    End Sub

    Private Sub btSend_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btSend.Click
        Dim Ansi As Byte() = System.Text.Encoding.ASCII.GetBytes(edText.Text)
        Dim Sent As UInt32 = 0
        Dim Res As Int32 = Client.Write(Ansi, Sent)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Error: 0x" + Res.ToString("X8"))
        Else
            MessageBox.Show("Sent: " + Sent.ToString() + " from " + Ansi.Length.ToString())
        End If
    End Sub

    Private Sub btWriteBuffers_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btWriteBuffers.Click
        ' Try to set read buffer size first.
        Dim Size As UInt32 = Convert.ToUInt32(edReadBuffer.Text)
        Dim Res As Int32 = Client.SetReadBufferSize(Size)
        If Res = wclErrors.WCL_E_SUCCESS Then
            lbEvents.Items.Add("Read buffer size changed.")
        Else
            lbEvents.Items.Add("Set read buffer size error: 0x" + Res.ToString("X8"))
        End If

        ' Now write buffer.
        Size = Convert.ToUInt32(edWriteBuffer.Text)
        Res = Client.SetWriteBufferSize(Size)
        If Res = wclErrors.WCL_E_SUCCESS Then
            lbEvents.Items.Add("Write buffer size changed.")
        Else
            lbEvents.Items.Add("Set write buffer size error: 0x" + Res.ToString("X8"))
        End If
    End Sub

    Private Sub btGetBuffers_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btGetBuffers.Click
        GetBuffers()
    End Sub

    Private Sub GetBuffers()
        ' Try to get read and write buffers size.
        Dim Size As UInt32 = 0
        Dim Res As Int32 = Client.GetReadBufferSize(Size)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            lbEvents.Items.Add("Get read buffer error: 0x" + Res.ToString("X8"))
        Else
            edReadBuffer.Text = Size.ToString()
        End If

        Size = 0
        Res = Client.GetWriteBufferSize(Size)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            lbEvents.Items.Add("Get write buffer error: 0x" + Res.ToString("X8"))
        Else
            edWriteBuffer.Text = Size.ToString()
        End If
    End Sub

    Private Sub btPair_Click(sender As System.Object, e As System.EventArgs) Handles btPair.Click
        If lvDevices.SelectedItems.Count = 0 Then
            MessageBox.Show("Select device", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            Dim Radio As wclBluetoothRadio = GetRadio()
            If Radio IsNot Nothing Then
                Dim Res As Integer = Radio.RemotePair(Convert.ToInt64(lvDevices.SelectedItems(0).Text, 16), wclBluetoothPairingMethod.pmClassic)
                If Res <> wclErrors.WCL_E_SUCCESS Then
                    MessageBox.Show("Pair failed; 0x" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Warning)
                End If
            End If
        End If
    End Sub

    Private Sub btUnpair_Click(sender As System.Object, e As System.EventArgs) Handles btUnpair.Click
        If lvDevices.SelectedItems.Count = 0 Then
            MessageBox.Show("Select device", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            Dim Radio As wclBluetoothRadio = GetRadio()
            If Radio IsNot Nothing Then
                Dim Res As Int32 = Radio.RemoteUnpair(Convert.ToInt64(lvDevices.SelectedItems(0).Text, 16), wclBluetoothPairingMethod.pmClassic)
                If Res <> wclErrors.WCL_E_SUCCESS Then
                    MessageBox.Show("Unpaired failed: 0x" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Warning)
                End If
            End If
        End If
    End Sub
End Class
