Public Class fmMain
    Private WithEvents Server As wclRfCommServer
    Private WithEvents Manager As wclBluetoothManager

    Private Sub btClear_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btClear.Click
        lbEvents.Items.Clear()
    End Sub

    Private Sub fmMain_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load
        Manager = New wclBluetoothManager()
        Server = New wclRfCommServer()

        ' In real application you should always analize the result code.
        ' In this demo we assume that all is always OK.
        Manager.Open()

        cbAuthentication.Checked = Server.Authentication
        cbEncryption.Checked = Server.Encryption
    End Sub

    Private Sub Server_OnClosed(ByVal Sender As Object, ByVal Reason As Integer) Handles Server.OnClosed
        lbEvents.Items.Add("Server closed on radio " + CType(Sender, wclRfCommServer).Radio.ApiName + "with reason: 0x" + Reason.ToString("X8"))

        lvClients.Items.Clear()

        edReadBuffer.Text = ""
        edWriteBuffer.Text = ""
    End Sub

    Private Sub Server_OnConnect(ByVal Sender As Object, ByVal Client As wclBluetooth.wclRfCommServerClientConnection, ByVal [Error] As Integer) Handles Server.OnConnect
        Dim Server As wclRfCommServerConnection = CType(Client.Server, wclRfCommServerConnection)

        lbEvents.Items.Add("Client connected: " + Client.Address.ToString("X12") + " on Radio " + Server.Radio.ApiName + " error: 0x" + [Error].ToString("X8"))

        Dim Item As ListViewItem = lvClients.Items.Add(Client.Address.ToString("X12"))
        Item.SubItems.Add(GetDeviceName(Client.Address))
        Item.Tag = Client
    End Sub

    Private Sub Server_OnData(ByVal Sender As Object, ByVal Client As wclBluetooth.wclRfCommServerClientConnection, ByVal Data() As Byte) Handles Server.OnData
        If Data IsNot Nothing AndAlso Data.Length > 0 Then
            Dim Str As String = Encoding.ASCII.GetString(Data)

            Dim Server As wclRfCommServerConnection = CType(Client.Server, wclRfCommServerConnection)
            lbEvents.Items.Add("Received: " + Str + " from " + Client.Address.ToString("X12") + " on radio " + Server.Radio.ApiName)
        End If
    End Sub

    Private Sub Server_OnDisconnect(ByVal Sender As Object, ByVal Client As wclBluetooth.wclRfCommServerClientConnection, ByVal Reason As Integer) Handles Server.OnDisconnect
        Dim Server As wclRfCommServerConnection = CType(Client.Server, wclRfCommServerConnection)

        lbEvents.Items.Add("Client disconnected: " + Client.Address.ToString("X12") + " on radio " + Server.Radio.ApiName + " reason: 0x" + Reason.ToString("X8"))

        For Each Item As ListViewItem In lvClients.Items
            If Item.Tag.Equals(Client) Then
                lvClients.Items.Remove(Item)
                Exit For
            End If
        Next Item
    End Sub

    Private Sub Server_OnListen(ByVal sender As Object, ByVal e As System.EventArgs) Handles Server.OnListen
        Dim Server As wclRfCommServer = CType(sender, wclRfCommServer)
        lbEvents.Items.Add("Server listening on radio " + Server.Radio.ApiName + " on channel: " + Server.AssignedChannel.ToString())
    End Sub

    Private Sub Manager_OnNumericComparison(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal Address As Long, ByVal Number As UInteger, ByRef Confirm As Boolean) Handles Manager.OnNumericComparison
        ' Accept any pairing.
        Confirm = True
        lbEvents.Items.Add("Numeric comparison: " + Number.ToString())
    End Sub

    Private Sub Manager_OnPasskeyNotification(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal Address As Long, ByVal Passkey As UInteger) Handles Manager.OnPasskeyNotification
        lbEvents.Items.Add("Passkey notification: " + Passkey.ToString())
    End Sub

    Private Sub Manager_OnPasskeyRequest(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal Address As Long, ByRef Passkey As UInteger) Handles Manager.OnPasskeyRequest
        ' Use 123456 as passkey.
        Passkey = 123456
        lbEvents.Items.Add("Passkey request: " + Passkey.ToString())
    End Sub

    Private Sub Manager_OnPinRequest(ByVal Sender As Object, ByVal Radio As wclBluetooth.wclBluetoothRadio, ByVal Address As Long, ByRef Pin As String) Handles Manager.OnPinRequest
        ' Use '0000' as PIN.
        Pin = "0000"
        lbEvents.Items.Add("Pin request: " + Pin)
    End Sub

    Private Sub fmMain_FormClosed(ByVal sender As System.Object, ByVal e As System.Windows.Forms.FormClosedEventArgs) Handles MyBase.FormClosed
        Server.Close()
        Server = Nothing

        Manager.Close()
        Manager = Nothing
    End Sub

    Private Function GetRadio() As wclBluetoothRadio
        Dim Radio As wclBluetoothRadio = Nothing
        Dim Res As Int32 = Manager.GetClassicRadio(Radio)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Get working radio failed: 0x" + Res.ToString("X8"))
            Return Nothing
        End If
        Return Radio
    End Function

    Private Sub btListen_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btListen.Click
        Dim Radio As wclBluetoothRadio = GetRadio()
        If Radio IsNot Nothing Then
            ' Switch radio in discoverable and connectable mode.
            ' Ignore any errors. In real application you need to check the result.
            Radio.SetDiscoverable(True)
            Radio.SetConnectable(True)

            Server.Authentication = cbAuthentication.Checked
            Server.Encryption = cbEncryption.Checked

            Server.Service = wclUUIDs.SerialPortServiceClass_UUID
            Server.ServiceName = edServiceName.Text

            Dim Res As Int32 = Server.Listen(Radio)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
            End If
        End If
    End Sub

    Private Sub btClose_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btClose.Click
        Dim Res As Int32 = Server.Close()
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
        End If
    End Sub

    Private Sub btDisconnect_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btDisconnect.Click
        If lvClients.SelectedItems.Count = 0 Then
            MessageBox.Show("Select client", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            Dim Res As Int32 = CType(lvClients.SelectedItems(0).Tag, wclRfCommServerClientConnection).Disconnect()
            If Res <> wclErrors.WCL_E_SUCCESS Then
                MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
            End If
        End If
    End Sub

    Private Sub btSend_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btSend.Click
        If lvClients.SelectedItems.Count = 0 Then
            MessageBox.Show("Select client", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            Dim Ansi As Byte() = Encoding.ASCII.GetBytes(edText.Text)
            Dim Sent As UInt32 = 0
            Dim Res As Int32 = CType(lvClients.SelectedItems(0).Tag, wclRfCommServerClientConnection).Write(Ansi, Sent)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                MessageBox.Show("Error: 0x" + Res.ToString("X8"))
            Else
                MessageBox.Show("Sent: " + Sent.ToString() + " from " + Ansi.Length.ToString())
            End If
        End If
    End Sub

    Private Sub btGetBuffers_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btGetBuffers.Click
        If lvClients.SelectedItems.Count = 0 Then
            MessageBox.Show("Select client", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            Dim Client As wclRfCommServerClientConnection = CType(lvClients.SelectedItems(0).Tag, wclRfCommServerClientConnection)

            ' Get read buffer.
            Dim Size As UInt32 = 0
            Dim Res As Int32 = Client.GetReadBufferSize(Size)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                lbEvents.Items.Add("Get read buffer error: 0x" + Res.ToString("X8"))
            Else
                edReadBuffer.Text = Size.ToString()
            End If

            ' Get write buffer.
            Size = 0
            Res = Client.GetWriteBufferSize(Size)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                lbEvents.Items.Add("Get write buffer error: 0x" + Res.ToString("X8"))
            Else
                edWriteBuffer.Text = Size.ToString()
            End If
        End If
    End Sub

    Private Sub btSetBuffers_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btSetBuffers.Click
        If lvClients.SelectedItems.Count = 0 Then
            MessageBox.Show("Select client", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
        Else
            Dim Client As wclRfCommServerClientConnection = CType(lvClients.SelectedItems(0).Tag, wclRfCommServerClientConnection)

            ' Set read buffer.
            Dim Size As UInt32 = Convert.ToUInt32(edReadBuffer.Text)
            Dim Res As Int32 = Client.SetReadBufferSize(Size)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                lbEvents.Items.Add("Set read buffer error: 0x" + Res.ToString("X8"))
            End If

            ' Set write buffer.
            Size = Convert.ToUInt32(edWriteBuffer.Text)
            Res = Client.SetWriteBufferSize(Size)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                lbEvents.Items.Add("Set write buffer error: 0x" + Res.ToString("X8"))
            End If
        End If
    End Sub

    Private Function GetDeviceName(ByVal Address As Int64) As String
        Dim Radio As wclBluetoothRadio = GetRadio()
        If Radio IsNot Nothing Then
            Dim DevName As String = ""
            Dim Res As Int32 = Radio.GetRemoteName(Address, DevName)
            If Res = wclErrors.WCL_E_SUCCESS Then
                Return DevName
            End If

            Return "Error: 0x" + Res.ToString("X8")
        End If

        Return ""
    End Function
End Class
