﻿Public Class fmMain
    Private WithEvents FManager As wclBluetoothManager
    Private WithEvents FClient As wclWiiRemoteClient

    Private Function GetRadio() As wclBluetoothRadio
        Dim Radio As wclBluetoothRadio = Nothing
        Dim Res As Int32 = FManager.GetClassicRadio(Radio)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Get working radio failed: 0x" + Res.ToString("X8"))
            Return Nothing
        End If
        Return Radio
    End Function

    Private Sub fmMain_Load(sender As System.Object, e As System.EventArgs) Handles MyBase.Load
        FManager = New wclBluetoothManager()
        FClient = New wclWiiRemoteClient()

        FManager.Open()
        ClearAccel()
        ClearIr()
        ClearExtensions()
    End Sub

    Private Sub FClient_OnAccelChanged(Sender As Object, Accel As wclBluetooth.wclWiiRemoteAccel) Handles FClient.OnAccelChanged
        laX.Text = Accel.Values.X.ToString()
        laY.Text = Accel.Values.Y.ToString()
        laZ.Text = Accel.Values.Z.ToString()
    End Sub

    Private Sub FClient_OnBalanceBoardChanged(Sender As Object, Board As wclBluetooth.wclWiiRemoteBalanceBoard) Handles FClient.OnBalanceBoardChanged
        laBbTopRight.Text = Board.SensorsKg.TopRight.ToString("N2")
        laBbTopLeft.Text = Board.SensorsKg.TopLeft.ToString("N2")
        laBbBottomRight.Text = Board.SensorsKg.BottomRight.ToString("N2")
        laBbBottomLeft.Text = Board.SensorsKg.BottomLeft.ToString("N2")
    End Sub

    Private Sub FClient_OnButtonsChanged(Sender As Object, Buttons As wclBluetooth.wclWiiRemoteButtons) Handles FClient.OnButtonsChanged
        cbA.Checked = Buttons.A
        cbB.Checked = Buttons.B
        cbPlus.Checked = Buttons.Plus
        cbHome.Checked = Buttons.Home
        cbMinus.Checked = Buttons.Minus
        cbOne.Checked = Buttons.One
        cbTwo.Checked = Buttons.Two
        cbUp.Checked = Buttons.Up
        cbDown.Checked = Buttons.Down
        cbLeft.Checked = Buttons.Left
        cbRight.Checked = Buttons.Right
    End Sub

    Private Sub FClient_OnConnect(Sender As Object, [Error] As Integer) Handles FClient.OnConnect
        If [Error] <> wclErrors.WCL_E_SUCCESS Then
            Trace("Connection failed", [Error])
        Else
            Trace("Connected")
            SetLeds()
        End If
    End Sub

    Private Sub FClient_OnDisconnect(Sender As Object, Reason As Integer) Handles FClient.OnDisconnect
        Trace("Disconnected. Reason", Reason)

        laBattLevel.Text = "0.00"
        cbLed1.Checked = False
        cbLed2.Checked = False
        cbLed3.Checked = False
        cbLed4.Checked = False

        cbA.Checked = False
        cbB.Checked = False
        cbPlus.Checked = False
        cbHome.Checked = False
        cbMinus.Checked = False
        cbOne.Checked = False
        cbTwo.Checked = False
        cbUp.Checked = False
        cbDown.Checked = False
        cbLeft.Checked = False
        cbRight.Checked = False

        ClearAccel()
        ClearIr()
        ClearExtensions()
    End Sub

    Private Sub FClient_OnExtensionAttached(sender As Object, e As System.EventArgs) Handles FClient.OnExtensionAttached
        Dim Ext As String
        Select Case FClient.Extension
            Case wclWiiRemoteExtension.wiiNoExtension
                Ext = "wiiNoExtension"
            Case wclWiiRemoteExtension.wiiNunchuk
                Ext = "wiiNunchuk"
                laNunchukState.Text = "Attached"
            Case wclWiiRemoteExtension.wiiClassicController
                Ext = "wiiClassicController"
            Case wclWiiRemoteExtension.wiiGuitar
                Ext = "wiiGuitar"
            Case wclWiiRemoteExtension.wiiDrums
                Ext = "wiiDrums"
            Case wclWiiRemoteExtension.wiiBalanceBoard
                Ext = "wiiBalanceBoard"
                laBalanceBoardState.Text = "Attached"
            Case Else
                Ext = "Unknonw"
        End Select

        Trace("Extension attached: " + Ext)
    End Sub

    Private Sub FClient_OnExtensionDetached(sender As Object, e As System.EventArgs) Handles FClient.OnExtensionDetached
        Trace("Extension detached")
        ClearExtensions()
    End Sub

    Private Sub FClient_OnIrChanged(Sender As Object, Ir As wclBluetooth.wclWiiRemoteIrSensors) Handles FClient.OnIrChanged
        Select Case Ir.Mode
            Case wclWiiRemoteIrMode.wiiIrOff
                ClearIr()
            Case wclWiiRemoteIrMode.wiiIrBasicMode
                laIrMode.Text = "Basic"
            Case wclWiiRemoteIrMode.wiiIrExtendedMode
                laIrMode.Text = "Extended"
            Case wclWiiRemoteIrMode.wiiIrFullMode
                laIrMode.Text = "Full"
            Case Else
                laIrMode.Text = "Unknown"
        End Select

        If Ir.Mode <> wclWiiRemoteIrMode.wiiIrOff Then
            Select Case Ir.Sensitivity
                Case wclWiiRemoteIrSensitivity.wiiIrLevelOff
                    laIrSensitivity.Text = "Off"
                Case wclWiiRemoteIrSensitivity.wiiIrLevel1
                    laIrSensitivity.Text = "Level 1"
                Case wclWiiRemoteIrSensitivity.wiiIrLevel2
                    laIrSensitivity.Text = "Level 2"
                Case wclWiiRemoteIrSensitivity.wiiIrLevel3
                    laIrSensitivity.Text = "Level 3"
                Case wclWiiRemoteIrSensitivity.wiiIrLevel4
                    laIrSensitivity.Text = "Level 4"
                Case wclWiiRemoteIrSensitivity.wiiIrLevel5
                    laIrSensitivity.Text = "Level 5"
                Case wclWiiRemoteIrSensitivity.wiiIrLevelMax
                    laIrSensitivity.Text = "Maximum"
                Case Else
                    laIrSensitivity.Text = "Unknown"
            End Select

            If Ir.Sensors(0).Found Then
                laSensor0Position.Text = Ir.Sensors(0).Position.X.ToString() + "; " + Ir.Sensors(0).Position.Y.ToString()
                laSensor0Size.Text = Ir.Sensors(0).Size.ToString()
                laSensor0Found.Text = "True"
            Else
                laSensor0Position.Text = "0; 0"
                laSensor0Size.Text = "0"
                laSensor0Found.Text = "False"
            End If

            If Ir.Sensors(1).Found Then
                laSensor1Position.Text = Ir.Sensors(1).Position.X.ToString() + "; " + Ir.Sensors(1).Position.Y.ToString()
                laSensor1Size.Text = Ir.Sensors(1).Size.ToString()
                laSensor1Found.Text = "True"
            Else
                laSensor1Position.Text = "0; 0"
                laSensor1Size.Text = "0"
                laSensor1Found.Text = "False"
            End If

            If Ir.Sensors(2).Found Then
                laSensor2Position.Text = Ir.Sensors(2).Position.X.ToString() + "; " + Ir.Sensors(2).Position.Y.ToString()
                laSensor2Size.Text = Ir.Sensors(2).Size.ToString()
                laSensor2Found.Text = "True"
            Else
                laSensor2Position.Text = "0; 0"
                laSensor2Size.Text = "0"
                laSensor2Found.Text = "False"
            End If

            If Ir.Sensors(3).Found Then
                laSensor3Position.Text = Ir.Sensors(3).Position.X.ToString() + "; " + Ir.Sensors(3).Position.Y.ToString()
                laSensor3Size.Text = Ir.Sensors(3).Size.ToString()
                laSensor3Found.Text = "True"
            Else
                laSensor3Position.Text = "0; 0"
                laSensor3Size.Text = "0"
                laSensor3Found.Text = "False"
            End If
        End If
    End Sub

    Private Sub ClearAccel()
        laX.Text = "0"
        laY.Text = "0"
        laZ.Text = "0"
    End Sub

    Private Sub ClearIr()
        laIrMode.Text = "Off"
        laIrSensitivity.Text = "Off"

        laSensor0Position.Text = "0; 0"
        laSensor0Size.Text = "0"
        laSensor0Found.Text = "False"

        laSensor1Position.Text = "0; 0"
        laSensor1Size.Text = "0"
        laSensor1Found.Text = "False"

        laSensor2Position.Text = "0; 0"
        laSensor2Size.Text = "0"
        laSensor2Found.Text = "False"

        laSensor3Position.Text = "0; 0"
        laSensor3Size.Text = "0"
        laSensor3Found.Text = "False"
    End Sub

    Private Sub FClient_OnNunchukChanged(Sender As Object, Nunchuk As wclBluetooth.wclWiiRemoteNunchuk) Handles FClient.OnNunchukChanged
        laNunchukX.Text = Nunchuk.Accel.Values.X.ToString()
        laNunchukY.Text = Nunchuk.Accel.Values.Y.ToString()
        laNunchukZ.Text = Nunchuk.Accel.Values.Z.ToString()

        laNunchukJoyX.Text = Nunchuk.Joystick.X.ToString()
        laNunchukJoyY.Text = Nunchuk.Joystick.Y.ToString()

        cbNunchukC.Checked = Nunchuk.C
        cbNunchukZ.Checked = Nunchuk.Z
    End Sub

    Private Sub ClearNunchuk()
        laNunchukState.Text = "Detached"

        laNunchukX.Text = "0"
        laNunchukY.Text = "0"
        laNunchukZ.Text = "0"

        laNunchukJoyX.Text = "0"
        laNunchukJoyY.Text = "0"

        cbNunchukC.Checked = False
        cbNunchukZ.Checked = False
    End Sub

    Private Sub ClearExtensions()
        ClearNunchuk()
        ClearBalanceBoard()
    End Sub

    Private Sub ClearBalanceBoard()
        laBalanceBoardState.Text = "Detached"

        laBbTopRight.Text = "0"
        laBbTopLeft.Text = "0"
        laBbBottomRight.Text = "0"
        laBbBottomLeft.Text = "0"
    End Sub

    Private Sub FClient_OnStatusChanged(Sender As Object, Batt As Double, Leds As wclBluetooth.wclWiiRemoteLeds) Handles FClient.OnStatusChanged
        Trace("Status changed")

        laBattLevel.Text = Batt.ToString("N2")

        cbLed1.Checked = Leds.Led1
        cbLed2.Checked = Leds.Led2
        cbLed3.Checked = Leds.Led3
        cbLed4.Checked = Leds.Led4
    End Sub

    Private Sub FManager_OnDeviceFound(Sender As Object, Radio As wclBluetooth.wclBluetoothRadio, Address As Long) Handles FManager.OnDeviceFound
        Dim Item As ListViewItem = lvDevices.Items.Add(Address.ToString("X12"))
        Item.SubItems.Add("")
    End Sub

    Private Sub FManager_OnDiscoveringCompleted(Sender As Object, Radio As wclBluetooth.wclBluetoothRadio, [Error] As Integer) Handles FManager.OnDiscoveringCompleted
        For i = 0 To lvDevices.Items.Count - 1
            Dim Item As ListViewItem = lvDevices.Items(i)
            Dim Addr As Int64 = Convert.ToInt64(Item.Text, 16)

            Dim Name As String = ""
            Dim Res As Int32 = Radio.GetRemoteName(Addr, Name)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                Name = "Error: 0x" + Res.ToString("X9")
                Item.SubItems(1).Text = Name
            End If
        Next i

        Trace("Discovering completed with result", [Error])
    End Sub

    Private Sub FManager_OnDiscoveringStarted(Sender As Object, Radio As wclBluetooth.wclBluetoothRadio) Handles FManager.OnDiscoveringStarted
        lvDevices.Items.Clear()
        Trace("Discovering started")
    End Sub

    Private Sub fmMain_FormClosed(sender As System.Object, e As System.Windows.Forms.FormClosedEventArgs) Handles MyBase.FormClosed
        FClient.Disconnect()
        FManager.Close()
    End Sub

    Private Sub btDiscover_Click(sender As System.Object, e As System.EventArgs) Handles btDiscover.Click
        Dim Radio As wclBluetoothRadio = GetRadio()
        If Not Radio Is Nothing Then
            Dim Res As Int32 = Radio.Discover(10, wclBluetoothDiscoverKind.dkClassic)
            If Res <> wclErrors.WCL_E_SUCCESS Then
                MessageBox.Show("Discover failed; 0x" + Res.ToString("X8"))
            End If
        End If
    End Sub

    Private Sub btConnect_Click(sender As System.Object, e As System.EventArgs) Handles btConnect.Click
        If FClient.State <> wclClientState.csDisconnected Then
            MessageBox.Show("Wii Remote is connected")
        Else
            Dim Radio As wclBluetoothRadio = GetRadio()
            If Not Radio Is Nothing Then
                If lvDevices.SelectedItems.Count = 0 Then
                    MessageBox.Show("Select device")
                Else
                    FClient.Address = Convert.ToInt64(lvDevices.SelectedItems(0).Text, 16)
                    Dim Res As Int32 = FClient.Connect(Radio)
                    If Res <> wclErrors.WCL_E_SUCCESS Then
                        MessageBox.Show("Unable connect: 0x" + Res.ToString("X8"))
                    End If
                End If
            End If
        End If
    End Sub

    Private Sub btDisconnect_Click(sender As System.Object, e As System.EventArgs) Handles btDisconnect.Click
        Dim Res As Int32 = FClient.Disconnect()
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Unable disconnect: 0x" + Res.ToString("X8"))
        End If
    End Sub

    Private Sub btClear_Click(sender As System.Object, e As System.EventArgs) Handles btClear.Click
        lbLog.Items.Clear()
    End Sub

    Private Sub Trace(Msg As String)
        lbLog.Items.Add(Msg)
    End Sub

    Private Sub Trace(Msg As String, Res As Int32)
        Trace(Msg + ": 0x" + Res.ToString("X8"))
    End Sub

    Private Sub btGetStatus_Click(sender As System.Object, e As System.EventArgs) Handles btGetStatus.Click
        Dim Res As Int32 = FClient.GetStatus()
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Get status failed; 0x" + Res.ToString("X8"))
        End If
    End Sub

    Private Sub btSetLeds_Click(sender As System.Object, e As System.EventArgs) Handles btSetLeds.Click
        SetLeds()
    End Sub

    Private Sub btGetLeds_Click(sender As System.Object, e As System.EventArgs) Handles btGetLeds.Click
        If FClient.State <> wclClientState.csConnected Then
            MessageBox.Show("Not connected")
        Else
            cbLed1.Checked = FClient.Leds.Led1
            cbLed2.Checked = FClient.Leds.Led2
            cbLed3.Checked = FClient.Leds.Led3
            cbLed4.Checked = FClient.Leds.Led4
        End If
    End Sub

    Private Sub SetLeds()
        Dim Res As Int32 = FClient.SetLeds(cbLed1.Checked, cbLed2.Checked, cbLed3.Checked, cbLed4.Checked)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Set LEDs failed: 0x" + Res.ToString("X8"))
        End If
    End Sub

    Private Sub btEnableAccel_Click(sender As System.Object, e As System.EventArgs) Handles btEnableAccel.Click
        Dim Res As Int32 = FClient.EnableAccel()
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Enable accelerometer failed: 0x" + Res.ToString("X8"))
        End If
    End Sub

    Private Sub btDisableAccel_Click(sender As System.Object, e As System.EventArgs) Handles btDisableAccel.Click
        Dim Res As Int32 = FClient.DisableAccel()
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Disable accelerometer failed: 0x" + Res.ToString("X8"))
        Else
            ClearAccel()
        End If
    End Sub

    Private Sub btEnableIr_Click(sender As System.Object, e As System.EventArgs) Handles btEnableIr.Click
        Dim Res As Int32 = FClient.SetIrSensitivity(wclWiiRemoteIrSensitivity.wiiIrLevelMax)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Enable IR failed: 0x" + Res.ToString("X8"))
        End If
    End Sub

    Private Sub btDisableIr_Click(sender As System.Object, e As System.EventArgs) Handles btDisableIr.Click
        Dim Res As Int32 = FClient.SetIrSensitivity(wclWiiRemoteIrSensitivity.wiiIrLevelOff)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Disable IR failed: 0x" + Res.ToString("X8"))
        Else
            ClearIr()
        End If
    End Sub

    Private Sub btRumbleOn_Click(sender As System.Object, e As System.EventArgs) Handles btRumbleOn.Click
        Dim Res As Int32 = FClient.SetRumble(True)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Set Rumble ON failed; 0x" + Res.ToString("X8"))
        End If
    End Sub

    Private Sub btRumbleOff_Click(sender As System.Object, e As System.EventArgs) Handles btRumbleOff.Click
        Dim Res As Int32 = FClient.SetRumble(False)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            MessageBox.Show("Set Rumble OFF failed; 0x" + Res.ToString("X8"))
        End If
    End Sub
End Class
